/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.datamodels;

/**
 * Encapsulation of material recyclability and rarity factors.
 */
public class MaterialFactors {
	/**
	 * Material mass in grams.
	 */
	private final double mass_in_grams;

	/**
	 * Rarity is a measure of the physical value of raw materials as a function of
	 * their relative scarcity in the crust and the energy costs required to extract
	 * and refine the materials.
	 */
	private final double rarity;

	/**
	 * According to RECYCLIA and INDUMETAL propositions, one of the main aims of the
	 * methodology when the materials are considered (in terms of providing
	 * incentives) should be the recyclability of the materials. To take this into
	 * account, for each raw material included in a device, its rarity needs to be
	 * multiplied by a factor that confirms its recyclability, a i , which will be
	 * equal to its recyclability ratio whenever it can be recycled (a value that
	 * will vary from 0 to 1).
	 */
	private final double recyclability;

	/**
	 * No default constructor, it does not make sense. All parameters should be
	 * provided by user.
	 */
	@SuppressWarnings("unused")
	private MaterialFactors() {
		mass_in_grams = rarity = recyclability = -1.0;
	}

	/**
	 * Constructor.
	 *
	 * @param mass_in_grams material mass in grams.
	 * @param rarity        the value of the material rarity factor.
	 * @param recyclability the value of the material recyclability factor.
	 *
	 * @throws IllegalArgumentException if the values provided are out of range.
	 */
	public MaterialFactors(double mass_in_grams, double rarity, double recyclability) throws IllegalArgumentException {
		if (mass_in_grams >= 0.0) {
			this.mass_in_grams = mass_in_grams;
		} else {
			throw new IllegalArgumentException("Invalid value for mass_in_grams: " + mass_in_grams);
		}

		if (rarity >= 0.0) {
			this.rarity = rarity;
		} else {
			throw new IllegalArgumentException("Invalid value for rarity: " + rarity);
		}

		if (recyclability >= 0.0 && recyclability <= 1.0) {
			this.recyclability = recyclability;
		} else {
			throw new IllegalArgumentException("Invalid value for recyclability: " + recyclability);
		}
	}

	/**
	 * @return the mass in grams of this material.
	 */
	public double getMass_in_grams() {
		return mass_in_grams;
	}

	/**
	 * @return the value of the material recyclability factor.
	 */
	public double getRecyclability() {
		return recyclability;
	}

	/**
	 * @return the value of the material rarity factor.
	 */
	public double getRarity() {
		return rarity;
	}

	@Override
	public boolean equals(Object obj) {
		try {
			final MaterialFactors o = (MaterialFactors) obj;
			// Equality means per member equality
			return mass_in_grams == o.mass_in_grams && recyclability == o.recyclability && rarity == o.rarity;
		} catch (Exception e) {
			return false;
		}
	}
}
