/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.client;

import java.util.logging.Logger;

import javax.ws.rs.client.Client;
import javax.ws.rs.ext.ContextResolver;

import org.glassfish.jersey.client.ClientConfig;
import org.glassfish.jersey.client.HttpUrlConnectorProvider;
import org.glassfish.jersey.client.JerseyClientBuilder;
import org.glassfish.jersey.jackson.JacksonFeature;
import org.glassfish.jersey.media.multipart.MultiPartFeature;

import com.fasterxml.jackson.databind.ObjectMapper;

import gr.iccs.isense.circ4life.logging.Log;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager.AccessTokens;

/**
 * Common REST Web Service client functionality.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public abstract class RESTWebServiceClient {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(RESTWebServiceClient.class);

	/**
	 * The Username required for protected web service resources.
	 */
	private String m_username = null;

	/**
	 * The Password required for protected web service resources.
	 */
	private String m_password = null;

	/**
	 * Access Control Manager session access tokens.
	 */
	private AccessTokens m_accesstokens = null;

	/**
	 * @return the Access Control Manager registered Client ID this client should
	 *         use.
	 */
	public abstract String getClientID();

	/**
	 * Set HTTP header for all subsequent requests (e.g. Authorization).
	 * 
	 * @param name  the name of the HTTP header.
	 * @param value the value of the HTTP header.
	 */
	public abstract void setHTTPHeader(final String name, final String value);

	/**
	 * @return the username to use for the authorized HTTP session.
	 */
	public String getUsername() {
		return m_username;
	}

	/**
	 * Set the username to use for the authorized HTTP session.
	 * 
	 * @param username the username to use for the authorized HTTP session.
	 */
	public void setUsername(String username) {
		this.m_username = username;
	}

	/**
	 * @return the password to use for the authorized HTTP session.
	 */
	public String getPassword() {
		return m_password;
	}

	/**
	 * Set the password to use for the authorized HTTP session.
	 * 
	 * @param password the password to use for the authorized HTTP session.
	 */
	public void setPassword(String password) {
		this.m_password = password;
	}

	/**
	 * Start a user session by contacting the Access Control Manager for
	 * authentication using the credentials set with
	 * {@link RESTWebServiceClient#setUsername(String)} and
	 * {@link RESTWebServiceClient#setPassword(String)}. Once this method returns
	 * <code>true</code> all requests to the web service will be performed with
	 * proper authorization and protected resources will be available.
	 * 
	 * @return <code>true</code> if user session is authenticated,
	 *         <code>false</code> otherwise.
	 */
	public boolean startSession() {
		try {
			if (null == m_username) {
				logger.severe("Username is not set!");

				return false;
			}

			if (null == m_password) {
				logger.severe("Password is not set!");

				return false;
			}

			// If the tokens are already set, try to end their session
			if (null != m_accesstokens) {
				endSession();
			}

			m_accesstokens = AccessControlManager.getAccessTokens(getClientID(), m_username, m_password);
		} catch (Exception e) {
			return false;
		}

		if (null != m_accesstokens) {
			// Unfortunately the OpenAPI generator does not add the proper Authentications
			// in the org.openapitools.client.ApiClient class. The member
			// protected Map<String, Authentication> authentications
			// is empty.
			// getApiClient().setBearerToken(m_accesstokens.getAccessToken());
			setHTTPHeader("Authorization", "Bearer " + m_accesstokens.getAccessToken());

			return true;
		}

		return false;
	}

	/**
	 * End a user session by contacting the Access Control Manager using the current
	 * authorization session tokens. Once this method returns <code>true</code> the
	 * user session is finished and all requests to the web service will only
	 * succeed for unprotected resources and fail for protected ones.
	 * 
	 * @return <code>true</code> if user session is properly ended,
	 *         <code>false</code> otherwise.
	 */
	public boolean endSession() {
		if (AccessControlManager.endSession(getClientID(), m_accesstokens)) {
			m_accesstokens = null;
			return true;
		}

		return false;
	}

	/**
	 * Refresh a user session to avoid expiration of access.
	 * 
	 * @return <code>true</code> if session access is refreshed successfully,
	 *         <code>false</code> otherwise.
	 */
	public boolean refreshSession() {
		try {
			m_accesstokens = AccessControlManager.refreshAccessTokens(getClientID(), m_accesstokens, true);
		} catch (Exception e) {
			return false;
		}

		if (null != m_accesstokens) {
			// Unfortunately the OpenAPI generator does not add the proper Authentications
			// in the org.openapitools.client.ApiClient class. The member
			// protected Map<String, Authentication> authentications
			// is empty.
			// getApiClient().setBearerToken(m_accesstokens.getAccessToken());
			setHTTPHeader("Authorization", "Bearer " + m_accesstokens.getAccessToken());

			return true;
		}

		return false;
	}

	/**
	 * Set the access tokens of this client to the specified ones. End the current
	 * session if it is active. If the provided object is <code>null</code>, nothing
	 * is done.
	 * 
	 * @param accesstokens valid access tokens.
	 */
	protected void setSession(final AccessTokens accesstokens) {
		if (null == accesstokens) {
			logger.warning("Tried to set null access tokens!");

			return;
		}

		// If the tokens are already set, try to end their session
		if (null != m_accesstokens) {
			endSession();
		}

		m_accesstokens = accesstokens;

		// Unfortunately the OpenAPI generator does not add the proper Authentications
		// in the org.openapitools.client.ApiClient class. The member
		// protected Map<String, Authentication> authentications
		// is empty.
		// getApiClient().setBearerToken(m_accesstokens.getAccessToken());
		setHTTPHeader("Authorization", "Bearer " + m_accesstokens.getAccessToken());
	}

	/**
	 * Build a JAX-RS Client from Jersey Builder instead of automatic service
	 * provider client.
	 * 
	 * @param jsoncxtresolver the JSON object mapper context resolver.
	 * @return a new {@link Client} object from the {@link JerseyClientBuilder}.
	 */
	protected Client buildJerseyClient(final ContextResolver<ObjectMapper> jsoncxtresolver) {
		final JerseyClientBuilder builder = new JerseyClientBuilder();

		final ClientConfig clientConfig = new ClientConfig();
		clientConfig.register(MultiPartFeature.class);
		clientConfig.register(jsoncxtresolver);
		clientConfig.register(JacksonFeature.class);
		clientConfig.property(HttpUrlConnectorProvider.SET_METHOD_WORKAROUND, true);

		builder.withConfig(clientConfig);

		return builder.build();
	}
}
