/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.config;

import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;
import java.util.logging.Level;

import org.glassfish.jersey.logging.LoggingFeature.Verbosity;

/**
 * Common configuration properties for all REST services.
 */
public class ConfigProperties {
	/**
	 * Configuration properties file in classpath.
	 */
	private static final String CFG_PROPERTIES = "config.properties";

	/**
	 * Logger level, possible values in {@link Level}.
	 */
	public static final String PROPERTY_LOGLEVEL = "logging.level";

	/**
	 * Logger level, possible values in {@link Verbosity}.
	 */
	public static final String PROPERTY_LOGVERBOSITY = "logging.verbosity";

	/**
	 * Property that holds whether the Access Control Manager is enabled,
	 * <code>true</code> or <code>false</code>.
	 */
	public static final String PROPERTY_ACCESSCONTROLMANAGER = "accesscontrolmanager.enabled";

	/**
	 * Property that holds the Access Control Manager URL.
	 */
	public static final String PROPERTY_ACCESSCONTROLMANAGER_URL = "accesscontrolmanager.url";

	/**
	 * Property that holds the Access Control Manager Realm.
	 */
	public static final String PROPERTY_ACCESSCONTROLMANAGER_REALM = "accesscontrolmanager.realm";

	/**
	 * Property that holds the user name for accessing the Access Control Manager
	 * Admin REST API.
	 */
	public static final String PROPERTY_ACCESSCONTROLMANAGER_ADMINUSER = "accesscontrolmanager.admin.username";

	/**
	 * Property that holds the password for accessing the Access Control Manager
	 * Admin REST API.
	 */
	public static final String PROPERTY_ACCESSCONTROLMANAGER_ADMINPASS = "accesscontrolmanager.admin.password";

	/**
	 * Property that holds the Client ID for accessing the Access Control Manager
	 * Admin REST API.
	 */
	public static final String PROPERTY_ACCESSCONTROLMANAGER_ADMINCLIENT = "accesscontrolmanager.admin.clientid";

	/**
	 * Property that holds the database server host.
	 */
	public static final String PROPERTY_DATABASE_HOST = "database.host";

	/**
	 * Property that holds the database server port.
	 */
	public static final String PROPERTY_DATABASE_PORT = "database.port";

	/**
	 * Property that holds the configured file uploads directory.
	 */
	private static final String PROPERTY_UPLOADS_DIR = "webservice.fileuploadsdir";

	/**
	 * Configuration overrides. This object holds overrides for configuration
	 * properties in CFG_PROPERTIES.
	 */
	private static final Properties Overrides = new Properties();

	/**
	 * Get configuration property from a resource in the class path of a specified
	 * class. If the property is set in overrides with
	 * {@link ConfigProperties#setOverride(String, String)} then this value is
	 * returned.
	 * 
	 * @param <T>      the object type to use its {@link Class} for accessing the
	 *                 class loader.
	 * @param clazz    the object type {@link Class} for class loader resource
	 *                 access.
	 * @param property the property name to get the value from.
	 * @return the property's value if found in overrides or class path resource
	 *         file, <code>null</code> otherwise.
	 */
	public static <T> String getConfigProperty(final Class<T> clazz, final String property) {
		// First check the overrides
		final String override = Overrides.getProperty(property);
		if (null != override) {
			return override;
		}

		// Then read the configuration file
		final InputStream istream = clazz.getClassLoader().getResourceAsStream(CFG_PROPERTIES);
		if (null == istream) {
			return null;
		}
		final Properties cfgprops = new Properties();
		try {
			cfgprops.load(istream);
			istream.close();
		} catch (IOException e) {
			return null;
		}

		return cfgprops.getProperty(property);
	}

	/**
	 * Override a configuration property value.
	 * 
	 * @param property the property name to set the override value.
	 * @param value    the override value.
	 */
	public static void setOverride(final String property, final String value) {
		Overrides.setProperty(property, value);
	}

	/**
	 * Check if the Access Control Manager property is enabled in the resource
	 * configuration of the specified {@link Class}.
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return <code>true</code> if the Access Control Manager property is found and
	 *         set to true, <code>false</code> otherwise.
	 */
	public static <T> boolean isAccessControlManagerEnabled(final Class<T> clazz) {
		final String value = getConfigProperty(clazz, PROPERTY_ACCESSCONTROLMANAGER);
		if (null != value && value.equalsIgnoreCase("true")) {
			return true;
		}

		return false;
	}

	/**
	 * Get the Access Control Manager URL.
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return the Access Control Manager URL if found in configuration,
	 *         <code>null</code> otherwise.
	 */
	public static <T> String getAccessControlManagerURL(final Class<T> clazz) {
		return getConfigProperty(clazz, PROPERTY_ACCESSCONTROLMANAGER_URL);
	}

	/**
	 * Get the Access Control Manager Realm.
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return the Access Control Manager Realm if found in configuration,
	 *         <code>null</code> otherwise.
	 */
	public static <T> String getAccessControlManagerRealm(final Class<T> clazz) {
		return getConfigProperty(clazz, PROPERTY_ACCESSCONTROLMANAGER_REALM);
	}

	/**
	 * Get the property the holds the user name for accessing the Access Control
	 * Manager Admin REST API.
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return the Access Control Manager Admin REST API access username if found in
	 *         configuration, <code>null</code> otherwise.
	 */
	public static <T> String getAccessControlManagerAdminAPIUser(final Class<T> clazz) {
		return getConfigProperty(clazz, PROPERTY_ACCESSCONTROLMANAGER_ADMINUSER);
	}

	/**
	 * Get the property the holds the password for accessing the Access Control
	 * Manager Admin REST API.
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return the Access Control Manager Admin REST API access password if found in
	 *         configuration, <code>null</code> otherwise.
	 */
	public static <T> String getAccessControlManagerAdminAPIPass(final Class<T> clazz) {
		return getConfigProperty(clazz, PROPERTY_ACCESSCONTROLMANAGER_ADMINPASS);
	}

	/**
	 * Get the property the holds the Client ID for accessing the Access Control
	 * Manager Admin REST API.
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return the Access Control Manager Admin REST API Client ID if found in
	 *         configuration, <code>null</code> otherwise.
	 */
	public static <T> String getAccessControlManagerAdminAPIClient(final Class<T> clazz) {
		return getConfigProperty(clazz, PROPERTY_ACCESSCONTROLMANAGER_ADMINCLIENT);
	}

	/**
	 * Get the property that holds the database host.
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return the database host if found in configuration, <code>null</code>
	 *         otherwise.
	 */
	public static <T> String getDatabaseHost(final Class<T> clazz) {
		return getConfigProperty(clazz, PROPERTY_DATABASE_HOST);
	}

	/**
	 * Get the property that holds the database port.
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return the database port if found in configuration and can be parsed to an
	 *         {@link Integer}, <code>null</code> otherwise.
	 */
	public static <T> Integer getDatabasePort(final Class<T> clazz) {
		try {
			final String dbportstr = getConfigProperty(clazz, PROPERTY_DATABASE_PORT);
			if (null == dbportstr) {
				return null;
			}

			return Integer.parseInt(dbportstr);
		} catch (Exception e) {
			return null;
		}

	}

	/**
	 * Get logging level from configuration file.
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return the logging level if it is found and supported, {@link Level#INFO}
	 *         otherwise.
	 */
	public static <T> Level getLoggingLevel(final Class<T> clazz) {
		try {
			return Level.parse(getConfigProperty(clazz, PROPERTY_LOGLEVEL));
		} catch (NullPointerException | IllegalArgumentException e) {
			return Level.INFO;
		}
	}

	/**
	 * Get logging verbosity from configuration file.
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return the logging verbosity level if it is found and supported,
	 *         {@link Verbosity#PAYLOAD_TEXT} otherwise.
	 */
	public static <T> Verbosity getLoggingVerbosity(final Class<T> clazz) {
		try {
			return fromString(getConfigProperty(clazz, PROPERTY_LOGVERBOSITY));
		} catch (NullPointerException | IllegalArgumentException e) {
			return Verbosity.PAYLOAD_TEXT;
		}
	}

	/**
	 * Get the file path of the file uploads directory. This is used by Web Services
	 * that support file uploads or file creation on server-side.
	 * 
	 * <strong>CAUTION:The returned file path {@link String} is not guaranteed to
	 * exist! The application should make sure to create the directory in case it is
	 * not existent in the local host.</strong>
	 * 
	 * @param <T>   the object type to use its {@link Class} for accessing the class
	 *              loader.
	 * @param clazz the object type {@link Class} for class loader resource access.
	 * @return a {@link String} representing the configured file uploads directory
	 *         if found in configuration, <code>null</code> otherwise.
	 */
	public static <T> String getFileUploadsDir(final Class<T> clazz) {
		return getConfigProperty(clazz, PROPERTY_UPLOADS_DIR);
	}

	/**
	 * Convert {@link String} verbosity level to {@link Verbosity}.
	 * 
	 * @param verbosity the {@link String} representation of verbosity.
	 * @return the corresponding {@link Verbosity} if a match is found.
	 * @throws NullPointerException     if the specified {@link String} is
	 *                                  <code>null</code>.
	 * @throws IllegalArgumentException if the specified {@link String} does not
	 *                                  match any {@link Verbosity} value.
	 */
	private static Verbosity fromString(final String verbosity) throws NullPointerException, IllegalArgumentException {
		if (null == verbosity) {
			throw new NullPointerException("null String verbosity");
		}

		if (verbosity.equalsIgnoreCase("HEADERS_ONLY")) {
			return Verbosity.HEADERS_ONLY;
		} else if (verbosity.equalsIgnoreCase("PAYLOAD_TEXT")) {
			return Verbosity.PAYLOAD_TEXT;
		} else if (verbosity.equalsIgnoreCase("PAYLOAD_ANY")) {
			return Verbosity.PAYLOAD_ANY;
		} else {
			throw new IllegalArgumentException("Unknown String verbosity: " + verbosity);
		}
	}
}
