/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.security.Principal;
import java.util.ArrayList;
import java.util.Base64;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Logger;
import java.util.regex.Pattern;

import javax.json.Json;
import javax.json.JsonObject;
import javax.json.JsonReader;
import javax.naming.ConfigurationException;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;

import org.apache.http.Consts;
import org.apache.http.HttpEntity;
import org.apache.http.NameValuePair;
import org.apache.http.StatusLine;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.utils.URIBuilder;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.message.BasicNameValuePair;
import org.keycloak.admin.client.Keycloak;
import org.keycloak.representations.idm.CredentialRepresentation;
import org.keycloak.representations.idm.GroupRepresentation;
import org.keycloak.representations.idm.UserRepresentation;

import gr.iccs.isense.circ4life.logging.Log;
import gr.iccs.isense.circ4life.rest.config.ConfigProperties;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager.LoginResult.ErrorCode;
import gr.iccs.isense.circ4life.servlet.AccessControlManagerFilter;
import gr.iccs.isense.circ4life.validation.EndUserRegistrationFormValidator;

/**
 * Common Access Control Manager functionality. <br>
 * <br>
 * All authentication and Authorization of the ICT Platform is implemented with
 * the help of a Keycloak (https://www.keycloak.org/) server. The server is
 * deployed at https://circ4life.iccs.gr/auth/ and automatically handles all
 * integrated applications. OpenID Connect (https://openid.net/connect/) is the
 * methodology of the authentication and authorization. <br>
 * <br>
 * OpenID Connect is an identity mechanism based on OAuth 2.0 protocol. The end
 * users of a service (humans or applications) provide their credentials and a
 * registered Client ID into a pre-defined Token endpoint in exchange of an
 * Access Token. The Client ID is something the administrator of the Keycloak
 * server has already setup and the credentials can be obtained the same way or
 * with some kind of registration. With this triplet at hand, a user or an
 * application program can gain authorization by exchanging them for an Access
 * Token that allows access to protected resources by supplying the token along
 * the resource request.
 */
public class AccessControlManager {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(AccessControlManager.class);

	/**
	 * The client ID to use in order to login and access the Eco-credit calculator
	 * REST services.
	 */
	public final static String ECOCREDITCALC_CLIENTID = "EcoCreditCalculatorWS";

	/**
	 * The client ID to use in order to login and access the EndUserModule REST
	 * services.
	 */
	public final static String ENDUSERMODULE_CLIENTID = "EndUserModuleWS";

	/**
	 * The client ID to use in order to login and access the RecycleModule REST
	 * services.
	 */
	public final static String RECYCLEMODULE_CLIENTID = "RecycleModuleWS";

	/**
	 * The client ID to use in order to login and access the RetailerModule REST
	 * services.
	 */
	public final static String RETAILERMODULE_CLIENTID = "RetailerModuleWS";

	/**
	 * Keycloak default UUID length (as of version 4.8.3.Final).
	 */
	public final static int UUIDLENGTH = 36;

	/**
	 * Keycloak UUID pattern regular expression (as of version 4.8.3.Final).
	 * Keycloak UUIDs are of format: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
	 * 
	 * E.g. db624777-c948-4544-938f-0c5749396fce
	 */
	private final static String UUIDPATTERNREGEX = "\\w{8}-\\w{4}-\\w{4}-\\w{4}-\\w{12}";

	/**
	 * The group name of the Eco-account enabled users.
	 */
	public final static String ECOACCOUNTS_GROUP = "EcoAccounts";

	/**
	 * Encapsulation of Access and Refresh tokens.
	 */
	public static final class AccessTokens {
		/**
		 * Minimum accepted value in seconds between the local time and the access token
		 * times to consider the token not expired.
		 */
		private static final int EXPIRATION_MINIMUM_VALID = 5;
		/**
		 * Access Control Manager access token.
		 */
		private String accessToken;

		/**
		 * Access Control Manager refresh token.
		 */
		private String refreshToken;

		/**
		 * The time in seconds since January 1, 1970, 00:00:00 GMT, when this object was
		 * created.
		 */
		private final int timeLocal;

		/**
		 * The current time MUST be before the time represented by the exp Claim.
		 */
		private final int expclaim;

		/**
		 * The iat Claim can be used to reject tokens that were issued too far away from
		 * the current time, limiting the amount of time that nonces need to be stored
		 * to prevent attacks. The acceptable range is Client specific.
		 */
		private final int iatclaim;

		/**
		 * The time difference between the Access Control Manager server and client.
		 */
		private final int timeSkew;

		/**
		 * Constructor.
		 * 
		 * @param accessToken  Access Control Manager access token.
		 * @param refreshToken Access Control Manager refresh token.
		 * @throws NullPointerException     if the accessToken is malformed.
		 * @throws IllegalArgumentException if the accessToken contains expired data.
		 */
		public AccessTokens(final String accessToken, final String refreshToken)
				throws NullPointerException, IllegalArgumentException {
			super();
			this.accessToken = accessToken;
			this.refreshToken = refreshToken;
			timeLocal = ((int) (new Date().getTime() / 1000));
			logger.info("Received token at time (Epoch secs): " + timeLocal);

			// Decode the accessToken to get 'exp' and 'iat' properties
			final JsonObject jsontoken = decodeToken(accessToken);
			expclaim = jsontoken.getInt("exp");
			iatclaim = jsontoken.getInt("iat");

			logger.info("Access token exp Claim (Epoch secs): " + expclaim);
			logger.info("Access token iat Claim (Epoch secs): " + iatclaim);

			// Calculate approximate server/client difference.
			timeSkew = timeLocal - iatclaim;
			logger.info("Time skew: " + timeSkew);

			if (timeLocal >= expclaim) {
				throw new IllegalArgumentException("Local time is later than exp Claim!");
			}

			if (isTokenExpired()) {
				throw new IllegalArgumentException("Access token already expired!");
			}
		}

		/**
		 * Decode token {@link String} to a JSON object. Taken from Keycloak Javascript
		 * adapter:
		 * 
		 * <pre>
		 * function decodeToken(str) {
		 *     str = str.split('.')[1];
		 *     str = str.replace('/-/g', '+');
		 *     str = str.replace('/_/g', '/');
		 *     switch(str.length % 4) {
		 *         case 0: break;
		 *         case 2: str += '=='; break;
		 *         case 3: str += '='; break;
		 *         default: throw 'Invalid token';
		 *     }
		 *     str = (str + '===').slice(0, str.length + (str.length % 4));
		 *     str = str.replace(/-/g, '+').replace(/_/g, '/');
		 *     str = decodeURIComponent(escape(atob(str)));
		 *     str = JSON.parse(str);
		 *     return str;
		 * }
		 * </pre>
		 * 
		 * @param token the token {@link String} to decode.
		 * @return the decoded JSON object, or <code>null</code> on failure.
		 */
		public JsonObject decodeToken(String token) {
			final String[] parts = token.split("\\.");
			String encoded = parts[1];
			// Replace dashes and underscores
			encoded = encoded.replaceAll("-", "+");
			encoded = encoded.replaceAll("_", "/");
			// Put some padding in order to make the length a multiple of 4
			switch (encoded.length() % 4) {
			case 0: {
				// nothing
			}
				break;
			case 2: {
				encoded = encoded.concat("==");
			}
				break;
			case 3: {
				encoded = encoded.concat("=");
			}
				break;
			default: {
				throw new IllegalArgumentException("Invalid token length!");
			}
			}
			final int length = encoded.length(), modulo4 = length % 4;
			encoded = encoded.concat("===");
			encoded = encoded.substring(0, length + modulo4);
			encoded = encoded.replaceAll("-", "+");
			encoded = encoded.replaceAll("_", "/");
			final byte[] data = Base64.getDecoder().decode(encoded);
			final JsonReader reader = Json.createReader(new ByteArrayInputStream(data));

			return reader.readObject();
		}

		/**
		 * @return the Access Control Manager access token.
		 */
		public String getAccessToken() {
			return accessToken;
		}

		/**
		 * @return the Access Control Manager refresh token.
		 */
		public String getRefreshToken() {
			return refreshToken;
		}

		/**
		 * This implementation is taken from the Javascript Keycloak client adapter,
		 * function isTokenExpired. The main logic is the following: When the access
		 * token is received after a successful authentication, save the current local
		 * time. Calculate time difference between server and client from the local time
		 * and server's response JSON property 'iat'. To check the validity of tokens at
		 * any time, use the current time and compare it with the value of the first
		 * step. Take into consideration time difference between server and client, plus
		 * an optional small minimum validity value.
		 * 
		 * From the OpenID Connect Specification:
		 * https://openid.net/specs/openid-connect-core-1_0.html
		 * 
		 * 1) The current time MUST be before the time represented by the exp Claim.
		 * 
		 * 2) The time skew calculated from iat Claim should be small.
		 * 
		 * @return <code>true</code> if token has expired, false otherwise.
		 */
		public boolean isTokenExpired() {
			final int expireTime = expclaim + timeSkew - ((int) (new Date().getTime() / 1000));

			return expireTime - EXPIRATION_MINIMUM_VALID < 0;
		}

		/**
		 * @return <code>true</code> if tokens can be successfully decoded,
		 *         <code>false</code> otherwise.
		 */
		public boolean isTokenValid() {
			try {
				return (null != decodeToken(accessToken)) && null != decodeToken(refreshToken);
			} catch (Exception e) {
				return false;
			}
		}

		/**
		 * Clear the tokens in this object. This should be performed after an error
		 * response of the Access Control Manager or a refresh.
		 */
		public void clearTokens() {
			accessToken = refreshToken = "";
		}
	}

	/**
	 * Encapsulation of login result.
	 */
	public static final class LoginResult {
		public enum ErrorCode {
			OK, WRONG_CREDENTIALS, EMAIL_NOT_VERIFIED
		};

		/**
		 * The error code to return to the client.
		 */
		public ErrorCode error_code;

		/**
		 * The access tokens.
		 */
		public AccessTokens tokens;
	}

	/**
	 * @return <code>true</code> if Access Control Manager is enabled,
	 *         <code>false</code> otherwise.
	 */
	public static boolean isEnabled() {
		return AccessControlManagerFilter.isAccessControlManagerEnabled();
	}

	/**
	 * Check if the specified {@link SecurityContext} contains an authenticated
	 * user.
	 * 
	 * @param securityContext the {@link SecurityContext} to check for user
	 *                        authentication.
	 * @return <code>true</code> if the user is authenticated and user name can be
	 *         retrieved, <code>false</code> otherwise.
	 */
	public static boolean isUserAuthenticated(final SecurityContext securityContext) {
		return (null != getPrincipalName(securityContext));
	}

	/**
	 * Check if the specified user has verified his email address.
	 * 
	 * @param useruuid the user's Universally Unique ID.
	 * @return <code>true</code> if email address is verified, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isUserEmailVerified(final String useruuid) {
		final UserRepresentation rep = getUser(useruuid);
		if (null == rep) {
			logger.severe("Could not find user " + useruuid + " for check of email verification!");

			return false;
		}

		return rep.isEmailVerified();
	}

	/**
	 * Set the email verification status of the specified user.
	 * 
	 * @param useruuid      the user's Universally Unique ID.
	 * @param emailVerified flag to denote if the email should be set as verified or
	 *                      not.
	 * @return <code>true</code> if set properly, <code>false</code> otherwise.
	 */
	public static boolean setUserEmailVerified(final String useruuid, final boolean emailVerified) {
		final UserRepresentation rep = getUser(useruuid);

		if (null == rep) {
			logger.severe("Could not find user " + useruuid + " for setting email verification!");

			return false;
		}

		// Set only if we need it verified and is not yet
		if (emailVerified && !rep.isEmailVerified()) {
			logger.info("Setting email verification status of user " + useruuid);

			final UserRepresentation newrep = new UserRepresentation();
			newrep.setEmailVerified(emailVerified);

			return updateUser(useruuid, newrep);
		}

		// All OK
		return true;
	}

	/**
	 * Get the user name from the specified {@link SecurityContext}.
	 * 
	 * @param securityContext the {@link SecurityContext}.
	 * @return the user name as {@link String} if the {@link SecurityContext} has an
	 *         authenticated user, <code>null</code> otherwise.
	 */
	public static String getPrincipalName(final SecurityContext securityContext) {
		// Get the user from the session
		final Principal principal = securityContext.getUserPrincipal();
		if (null == principal) {
			return null;
		}

		// Get user name from the user session
		final String username = principal.getName();
		if (null == username) {
			return null;
		}

		return username;
	}

	/**
	 * Get the username's Universally Unique ID from the Access Control Manager.
	 * 
	 * @param username the registered user's Access Control Manager username.
	 * @return the username's Universally Unique ID if username is valid,
	 *         <code>null</code> otherwise.
	 */
	public static String getUsernameUUID(final String username) {
		try {
			final List<UserRepresentation> user_reps = getAdminAPIClient().realm(getRealm()).users().search(username);

			if (null == user_reps) {
				logger.warning("Username " + username + " does not match any existing users.");

				return null;
			}

			// We found exactly one
			if (1 == user_reps.size()) {
				return user_reps.get(0).getId();
			} else {
				for (final UserRepresentation user_rep : user_reps) {
					if (username.contentEquals(user_rep.getUsername())) {
						return user_rep.getId();
					}
				}

				logger.warning("Username " + username + " does not match any existing users.");

				return null;
			}
		} catch (ConfigurationException e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get the email's Universally Unique ID from the Access Control Manager.
	 * 
	 * @param email the registered user's Access Control Manager email.
	 * @return the email's Universally Unique ID if email is valid,
	 *         <code>null</code> otherwise.
	 */
	public static String getEmailUUID(final String email) {
		try {
			final List<UserRepresentation> user_reps = getAdminAPIClient().realm(getRealm()).users().search(email, 0,
					5);

			if (null == user_reps) {
				logger.warning("Email " + email + " does not match any existing users.");

				return null;
			}

			// We found exactly one
			if (1 == user_reps.size()) {
				return user_reps.get(0).getId();
			} else {
				for (final UserRepresentation user_rep : user_reps) {
					if (email.contentEquals(user_rep.getEmail())) {
						return user_rep.getId();
					}
				}

				logger.warning("Email " + email + " does not match any existing users.");

				return null;
			}
		} catch (ConfigurationException e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get the group's Universally Unique ID from the Access Control Manager.
	 * 
	 * @param group the Access Control Manager group name.
	 * @return the group's Universally Unique ID if group is a valid name,
	 *         <code>null</code> otherwise.
	 */
	public static String getGroupUUID(final String group) {
		try {
			final List<GroupRepresentation> group_reps = getAdminAPIClient().realm(getRealm()).groups().groups();

			String groupuuid = null;
			for (final GroupRepresentation g : group_reps) {
				// Save the groud id the first time we encounter the group name
				if (g.getName().equalsIgnoreCase(group)) {
					if (null == groupuuid) {
						groupuuid = g.getId();
					} else {
						// What? We already saved the group id and the same name came up again?
						// Disaster...
						logger.warning(
								"Group " + group + " has multiple representation in the Access Control Manager!");

						return null;
					}
				}
			}

			if (null == groupuuid) {
				logger.warning("Group " + group + " is non-existent!");

				return null;
			}

			return groupuuid;
		} catch (ConfigurationException e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Check if the specified user Universally Unique ID exists in the Access
	 * Control Manager.
	 * 
	 * @param useruuid the Universally Unique ID.
	 * @return <code>true</code> if exists, <code>false</code> otherwise.
	 */
	public static boolean existsUserUUID(final String useruuid) {
		if (null == getUser(useruuid)) {
			return false;
		}

		return true;
	}

	/**
	 * Check if the specified group Universally Unique ID exists in the Access
	 * Control Manager.
	 * 
	 * @param groupuuid the Universally Unique ID.
	 * @return <code>true</code> if exists, <code>false</code> otherwise.
	 */
	public static boolean existsGroupUUID(final String groupuuid) {
		if (null == getGroup(groupuuid)) {
			return false;
		}

		return true;
	}

	/**
	 * Retrieve a user's email address from Universally Unique ID.
	 * 
	 * @param useruuid the Universally Unique ID.
	 * @return the email address as {@link String} if user exists and email can be
	 *         retrieved, <code>null</code> otherwise.
	 */
	public static String getEmail(final String useruuid) {
		final UserRepresentation user = getUser(useruuid);
		if (null == user) {
			return null;
		}

		return user.getEmail();
	}

	/**
	 * Retrieve a user's username from Universally Unique ID.
	 * 
	 * @param useruuid the Universally Unique ID.
	 * @return the username as {@link String} if user exists and username can be
	 *         retrieved, <code>null</code> otherwise.
	 */
	public static String getUsername(final String useruuid) {
		final UserRepresentation user = getUser(useruuid);
		if (null == user) {
			return null;
		}

		return user.getUsername();
	}

	/**
	 * Check the format and NOT THE EXISTENCE of the specified {@link String}
	 * against the specification of a user's UUID. This can be used for quick
	 * testing without contacting the Access Control Manager, of whether the
	 * specified {@link String} could represent a user's UUID. For UUID existence,
	 * use {@link AccessControlManager#existsUserUUID(String)}.
	 * 
	 * @param useruuid the Universally Unique ID.
	 * @return <code>true</code> if the format is as expected, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isProperFormatUUID(final String useruuid) {
		// Keycloak UUIDs are of format: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
		// E.g. db624777-c948-4544-938f-0c5749396fce
		// So they have always a length of 36
		// Update AccessControlManager.UUIDLENGTH if something changes in the future
		if (null == useruuid || AccessControlManager.UUIDLENGTH != useruuid.length()
				|| !Pattern.matches(UUIDPATTERNREGEX, useruuid)) {
			logger.warning("Illegal user UUID identifier: " + useruuid);

			return false;
		}

		return true;
	}

	/**
	 * Reverse proxy of
	 * {@link AccessControlManager#getAccessTokens(String, String, String)} that
	 * first checks if the user belongs to the Eco-accounts ACM group and has
	 * verified the registration email address.
	 * 
	 * @param ClientID registered client ID on Access Control Manager.
	 * @param Username registered user name on Access Control Manager.
	 * @param Password registered user's password on Access Control Manager.
	 * @return an {@link AccessTokens} object filled with valid token from the
	 *         Access Control Manager, or <code>null</code> on failure.
	 * @throws IllegalArgumentException if the provided ClientID or credentials are
	 *                                  not recognized by the Access Control
	 *                                  Manager.
	 */
	public static LoginResult ecoAccountLogin(final String ClientID, final String Username, final String Password)
			throws IllegalArgumentException {
		try {
			final LoginResult result = new LoginResult();

			String useruuid = null;
			// Check first if the username is an email address or username
			if (null != getEmailUUID(Username)) {
				useruuid = getEmailUUID(Username);
			} else if (null != getUsernameUUID(Username)) {
				useruuid = getUsernameUUID(Username);
			}

			if (null == useruuid) {
				logger.severe("Username " + Username + " does not exist!");

				result.error_code = ErrorCode.WRONG_CREDENTIALS;
				result.tokens = null;

				return result;
			}
			logger.info("Trying to login [" + useruuid + "]");

			// Check that the credentials belong to an Eco-account user
			boolean isEcoAccount = false;
			for (final String group : getUserGroups(useruuid)) {
				if (ECOACCOUNTS_GROUP.contentEquals(group)) {
					isEcoAccount = true;
				}
			}
			if (!isEcoAccount) {
				logger.severe("Tried to login user " + useruuid + " not in group " + ECOACCOUNTS_GROUP);

				result.error_code = ErrorCode.WRONG_CREDENTIALS;
				result.tokens = null;

				return result;
			}

			// Check if the user has verified the email address
			if (!isUserEmailVerified(useruuid)) {
				logger.severe("Tried to login user " + useruuid + " that has not yet verified email address!");

				result.error_code = ErrorCode.EMAIL_NOT_VERIFIED;
				result.tokens = null;

				return result;
			}

			// All OK provide the access tokens if credentials are OK
			result.tokens = getAccessTokens(ClientID, Username, Password);
			if (null != result.tokens) {
				result.error_code = ErrorCode.OK;
			} else {
				result.error_code = ErrorCode.WRONG_CREDENTIALS;
			}

			return result;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Perform a session login on the Access Control Manager and get the
	 * access/refresh tokens.
	 * 
	 * @param ClientID registered client ID on Access Control Manager.
	 * @param Username registered user name on Access Control Manager.
	 * @param Password registered user's password on Access Control Manager.
	 * @return an {@link AccessTokens} object filled with valid token from the
	 *         Access Control Manager, or <code>null</code> on failure.
	 * @throws IllegalArgumentException if the provided ClientID or credentials are
	 *                                  not recognized by the Access Control
	 *                                  Manager.
	 */
	public static AccessTokens getAccessTokens(final String ClientID, final String Username, final String Password)
			throws IllegalArgumentException {
		try {
			// Get the token endpoint from the OpenID Connect discovery URL
			final URI tokenendpoint = getOpenIDConnectTokenEndpoint();

			final Map<String, String> params = new HashMap<String, String>();
			params.put("grant_type", "password");
			params.put("client_id", ClientID);
			params.put("username", Username);
			params.put("password", Password);

			final JsonObject ResponseJSON = httpPostURLEncodedParameters(tokenendpoint, null, params);
			if (null == ResponseJSON) {
				return null;
			}

			final String accessToken = ResponseJSON.getString("access_token"),
					refreshToken = ResponseJSON.getString("refresh_token");

			// We managed to get the tokens
			return new AccessTokens(accessToken, refreshToken);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Perform a refresh of the access tokens, optionally check the expiration.
	 * 
	 * @param ClientID         registered client ID on Access Control Manager.
	 * @param accessTokens     an {@link AccessTokens} object filled with valid
	 *                         token from the Access Control Manager that correspond
	 *                         to the session asked to end.
	 * @param check_expiration if <code>true</code> then check the expiration of the
	 *                         specified access tokens and only perform the refresh
	 *                         if they have expired.
	 * @return the rereshed access tokens on success, <code>null</code> otherwise.
	 */
	public static AccessTokens refreshAccessTokens(final String ClientID, final AccessTokens accessTokens,
			boolean check_expiration) {
		if (check_expiration && accessTokens.isTokenValid() && !accessTokens.isTokenExpired()) {
			logger.info("Valid not expired tokens, not need to refresh.");

			// return the same, they are not expired
			return accessTokens;
		}

		// Cannot refresh invalid tokens
		if (!accessTokens.isTokenValid()) {
			logger.severe("Cannot refresh invalid tokens!");

			return null;
		}

		try {
			// Get the token endpoint from the OpenID Connect discovery URL
			final URI tokenendpoint = getOpenIDConnectTokenEndpoint();

			final Map<String, String> params = new HashMap<String, String>();
			params.put("grant_type", "refresh_token");
			params.put("client_id", ClientID);
			params.put("refresh_token", accessTokens.refreshToken);

			final JsonObject ResponseJSON = httpPostURLEncodedParameters(tokenendpoint, null, params);

			if (null == ResponseJSON) {
				return null;
			}

			final String accessToken = ResponseJSON.getString("access_token"),
					refreshToken = ResponseJSON.getString("refresh_token");

			// We managed to get the new refreshed tokens
			accessTokens.clearTokens(); // invalidate the previous
			return new AccessTokens(accessToken, refreshToken);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Perform a session logout on the Access Control Manager, using the
	 * corresponding Client ID and access/refresh tokens. <br>
	 * <strong>CAUTION</strong>: This method performs a REQUEST to end the session,
	 * does not end the session even if <code>true</code> is returned. The
	 * responsible authority for this is the Access Control Manager, where all
	 * session handling is done. An HTTP 200 status code in the response indicates
	 * the proper acceptance of the request and if the Access Control Manager has no
	 * issue, the user session is then ended.
	 * 
	 * @param ClientID     registered client ID on Access Control Manager.
	 * @param accessTokens an {@link AccessTokens} object filled with valid token
	 *                     from the Access Control Manager that correspond to the
	 *                     session asked to end.
	 * @return <code>true</code> if the session end request is performed without
	 *         errors, <code>false</code> otherwise.
	 */
	public static boolean endSession(final String ClientID, final AccessTokens accessTokens) {
		try {
			// Get the end session endpoint from the OpenID Connect discovery URL
			final URI endsessionendpoint = getOpenIDConnectEndSessionEndpoint();

			final Map<String, String> headers = new HashMap<String, String>();
			headers.put("Authorization", "Bearer" + accessTokens.accessToken);

			final Map<String, String> params = new HashMap<String, String>();
			params.put("client_id", ClientID);
			params.put("refresh_token", accessTokens.refreshToken);

			final JsonObject ResponseJSON = httpPostURLEncodedParameters(endsessionendpoint, headers, params);

			// All OK?
			return (null != ResponseJSON);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Close the Access Control Manager Admin API client, logging out the user also.
	 */
	public static void closeAdminAPIClient() {
		if (null != AdminAPIClient && !AdminAPIClient.isClosed()) {
			try {
				final String clientId = ConfigProperties
						.getAccessControlManagerAdminAPIClient(AccessControlManager.class);
				if (null == clientId) {
					logger.severe("Configuration file does not contain Access Control Manager Admin API client ID!");
					throw new ConfigurationException(
							"Configuration file does not contain Access Control Manager Admin API client ID!");
				}
				logger.finest("Access Control Manager Admin API client ID: " + clientId);

				final String accessToken = AdminAPIClient.tokenManager().getAccessTokenString(),
						refreshToken = AdminAPIClient.tokenManager().getAccessToken().getRefreshToken();
				final AccessTokens tokens = new AccessTokens(accessToken, refreshToken);

				if (!endSession(clientId, tokens)) {
					logger.warning("Could not logout user from Admin REST API client");
				}
			} catch (Exception e) {
				logger.warning("Exception thrown while trying to logout user from Admin REST API client");
			}

			AdminAPIClient.close();
		}
	}

	/**
	 * Create an Eco-account enabled user in the Access Control Manager. A
	 * verification email should be sent in order to fully enable the account. The
	 * account is by default enabled but not email verified after creation.
	 * 
	 * @param username  the username.
	 * @param password  the password.
	 * @param email     the email.
	 * @param firstname the user's real world first name.
	 * @param lastname  the user's real world family name.
	 * @param locale    ISO 639 alpha-2 language code for this locale. Supported
	 *                  values are configured in the ACM.
	 * @return <code>true</code> if the user account was created successfully,
	 *         <code>false</code> otherwise.
	 */
	public static boolean createEcoAccountUser(final String username, final String password, final String email,
			final String firstname, final String lastname, final String locale) {
		try {
			// First check if the current configured Access Control Manager has the group or
			// not. If the group is non-existent, try to create it as it is a requirement to
			// group the Eco-account enabled user under the same group.
			String ecoaccgroupid = getGroupUUID(ECOACCOUNTS_GROUP);
			if (null == ecoaccgroupid) {
				if (!createEcoAccountGroup()) {
					// Cannot create the group? Sorry no user creation for you
					throw new Exception("Cannot create Eco-accounts group " + ECOACCOUNTS_GROUP);
				}

				ecoaccgroupid = getGroupUUID(ECOACCOUNTS_GROUP);
			}

			// validate input parameters
			EndUserRegistrationFormValidator.validate(username, password, email, firstname, lastname);

			// Create the user belonging to ECOACCOUNTS_GROUP
			final UserRepresentation rep = createEcoAccountRepresentation(username, password, email, firstname,
					lastname, locale);
			if (null == rep) {
				throw new Exception(
						"Cannot create user representation for Access Control Manager for " + username + "/" + email);
			}

			if (!createUser(rep)) {
				throw new Exception("User creation request failed!");
			}

			// Add to group
			if (!addUserToGroup(getUsernameUUID(username), ecoaccgroupid)) {
				// We need to remove the user as the whole transaction was not completed
				if (!removeEcoAccountUser(getUsernameUUID(username))) {
					logger.severe(
							"Could not remove user " + username + " after failing to join group " + ECOACCOUNTS_GROUP);
				}
				throw new Exception("User group join failed!");
			}

			// All OK
			return true;
		} catch (Exception e) {
			logger.severe("Cannot create Eco-account enabled user " + username + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Send the verification email to a registered but not yet enabled user, in
	 * order to verify his email address and enable the account.
	 * 
	 * @param useruuid the user's Universally Unique ID.
	 * @return <code>true</code> if the user was found, he is not yet enabled and
	 *         the email was sent successfully, <code>false</code> otherwise.
	 */
	public static boolean sendVerificationEmail(final String useruuid) {
		try {
			if (isUserEmailVerified(useruuid)) {
				logger.warning("User " + useruuid + " has already verified the email, no need to send another one.");

				return false;
			}

			getAdminAPIClient().realm(getRealm()).users().get(useruuid).sendVerifyEmail();

			return true;
		} catch (Exception e) {
			logger.severe("Could not send verification email to user " + useruuid + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Send the reset password email to a registered user, in order to change his
	 * password in case he forgot it.
	 * 
	 * @param useruuid the user's Universally Unique ID.
	 * @return <code>true</code> if the user was found and the email was sent
	 *         successfully, <code>false</code> otherwise.
	 */
	public static boolean sendResetPasswordEmail(final String useruuid) {
		try {
			final UserRepresentation rep = getUser(useruuid);
			if (null == rep) {
				logger.severe("Could not find user " + useruuid + " for sending reset password email!");

				return false;
			}

			final List<String> actions = new ArrayList<String>();
			actions.add("UPDATE_PASSWORD");
			getAdminAPIClient().realm(getRealm()).users().get(useruuid).executeActionsEmail(actions);

			return true;
		} catch (Exception e) {
			logger.severe("Could not send password reset email to user " + useruuid + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Email notification types enumeration.
	 */
	public static enum NotificationEmailType {
		ACM_UPDATED("ACM_UPDATED"), PROFILE_UPDATED("PROFILE_UPDATED"), ACCOUNT_REMOVAL("ACCOUNT_REMOVAL");

		private String m_name;

		private NotificationEmailType(String m_name) {
			this.m_name = m_name;
		}

		@Override
		public String toString() {
			return m_name;
		}
	}

	/**
	 * Send notification email to a registered user.
	 * 
	 * @param useruuid  the user's Universally Unique ID.
	 * @param emailtype email notification type.
	 * @return <code>true</code> if the user was found and the email was sent
	 *         successfully, <code>false</code> otherwise.
	 */
	public static boolean sendNotificationEmail(final String useruuid, final NotificationEmailType emailtype) {
		try {
			final UserRepresentation rep = getUser(useruuid);
			if (null == rep) {
				logger.severe("Could not find user " + useruuid + " for sending reset password email!");

				return false;
			}

			final List<String> actions = new ArrayList<String>();
			actions.add(emailtype.toString());
			getAdminAPIClient().realm(getRealm()).users().get(useruuid).executeActionsEmail(actions);

			return true;
		} catch (Exception e) {
			logger.severe("Could not send notification email to user " + useruuid + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Remove the Eco-account enabled user from the Access Control Manager.
	 * 
	 * @param useruuid the user's Universally Unique ID.
	 * @return <code>true</code> if the account removal was successful,
	 *         <code>false</code> otherwise.
	 */
	public static boolean removeEcoAccountUser(final String useruuid) {
		try {
			final Response result = getAdminAPIClient().realm(getRealm()).users().delete(useruuid);

			// All OK, group should be created
			if (299 > result.getStatus()) {
				logger.info("Successfully removed user " + useruuid);

				return true;
			}

			// Something went wrong, throw exception
			throw new Exception(
					"Status code: " + result.getStatus() + ", details: " + result.getStatusInfo().getReasonPhrase());
		} catch (Exception e) {
			logger.severe("Could not remove user " + useruuid + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Update the user representation in the Access Control Manager. Only the input
	 * parameters that are not null are updated.
	 * 
	 * @param useruuid  the user's Universally Unique ID.
	 * @param password  the updated password.
	 * @param email     the updated email.
	 * @param firstname the updated user's real world first name.
	 * @param lastname  the updated user's real world family name.
	 * @param locale    ISO 639 alpha-2 language code for this locale. Supported
	 *                  values are configured in the ACM.
	 * @return <code>true</code> if the user account was updated successfully or no
	 *         field was provided (no change required), <code>false</code>
	 *         otherwise.
	 */
	public static boolean updateEcoAccountUser(final String useruuid, final String password, final String email,
			final String firstname, final String lastname, final String locale) {
		try {
			// Perform the update if and only if there is a field that is not null
			if (null == password && null == email && null == firstname && null == lastname && null == locale) {
				// Nothing to do, bye bye
				return true;
			}

			// Perform the update if and only if the user's UUID belongs to EcoAccounts
			// group
			logger.info("Trying to update [" + useruuid + "]");
			boolean isEcoAccount = false;
			for (final String group : getUserGroups(useruuid)) {
				if (ECOACCOUNTS_GROUP.contentEquals(group)) {
					isEcoAccount = true;
				}
			}
			if (!isEcoAccount) {
				logger.severe("Tried to update user " + useruuid + " not in group " + ECOACCOUNTS_GROUP);

				return false;
			}

			// Create an updated representation from the fields that are not null
			final UserRepresentation newrep = updateEcoAccountRepresentation(password, email, firstname, lastname,
					locale);
			if (null == newrep) {
				throw new Exception("Cannot create updated user representation for Access Control Manager for "
						+ useruuid + "/" + email);
			}

			if (!updateUser(useruuid, newrep)) {
				throw new Exception("User update request failed!");
			}

			if (!sendNotificationEmail(useruuid, NotificationEmailType.ACM_UPDATED)) {
				// email sent failure is not critical to fail the whole process
				logger.severe("ACM account update email notification was not sent!");
			}

			// All OK
			return true;
		} catch (Exception e) {
			logger.severe("Could not update user " + useruuid + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * The Access Control Manager Admin API client.
	 */
	private static Keycloak AdminAPIClient = null;

	/**
	 * The Access Control Manager Realm.
	 */
	private static String Realm = null;

	/**
	 * Get Access Control Manager Realm from configuration file properties.
	 * 
	 * @return the Access Control Manager Realm.
	 * @throws ConfigurationException if a configuration error occurs.
	 */
	private static String getRealm() throws ConfigurationException {
		if (null == Realm) {
			Realm = ConfigProperties.getAccessControlManagerRealm(AccessControlManager.class);
			if (null == Realm) {
				logger.severe("Configuration file does not contain Access Control Manager Realm!");
				throw new ConfigurationException("Configuration file does not contain Access Control Manager Realm!");
			}
		}

		logger.finest("Access Control Manager Realm: " + Realm);

		return Realm;
	}

	/**
	 * Construct the Access Control Manager Admin API client from configuration file
	 * properties.
	 * 
	 * @return the Access Control Manager Admin API client.
	 * @throws ConfigurationException if a configuration error occurs.
	 */
	private static Keycloak getAdminAPIClient() throws ConfigurationException {
		// In case the Admin REST API client is not ready, read the configuration and
		// initialize it
		if (null == AdminAPIClient || AdminAPIClient.isClosed()) {
			// Get part <scheme>://<host>:<port>/auth from configuration file
			// e.g. https://circ4life.iccs.gr/auth
			final String serverUrl = ConfigProperties.getAccessControlManagerURL(AccessControlManager.class);
			if (null == serverUrl) {
				logger.severe("Configuration file does not contain Access Control Manager URL!");
				throw new ConfigurationException("Configuration file does not contain Access Control Manager URL!");
			}
			logger.finest("Access Control Manager URL: " + serverUrl);

			final String username = ConfigProperties.getAccessControlManagerAdminAPIUser(AccessControlManager.class);
			if (null == username) {
				logger.severe("Configuration file does not contain Access Control Manager Admin API username!");
				throw new ConfigurationException(
						"Configuration file does not contain Access Control Manager Admin API username!");
			}
			logger.finest("Access Control Manager Admin API username: " + username);

			final String password = ConfigProperties.getAccessControlManagerAdminAPIPass(AccessControlManager.class);
			if (null == password) {
				logger.severe("Configuration file does not contain Access Control Manager Admin API password!");
				throw new ConfigurationException(
						"Configuration file does not contain Access Control Manager Admin API password!");
			}
			logger.finest("Access Control Manager Admin API password: " + password);

			final String clientId = ConfigProperties.getAccessControlManagerAdminAPIClient(AccessControlManager.class);
			if (null == clientId) {
				logger.severe("Configuration file does not contain Access Control Manager Admin API client ID!");
				throw new ConfigurationException(
						"Configuration file does not contain Access Control Manager Admin API client ID!");
			}
			logger.finest("Access Control Manager Admin API client ID: " + clientId);

			AdminAPIClient = Keycloak.getInstance(serverUrl, getRealm(), username, password, clientId);
		}

		return AdminAPIClient;
	}

	/**
	 * Construct the OpenID Connect Discovery URL from configuration file
	 * properties.
	 * 
	 * @return the OpenID Connect Discovery URL or <code>null</code> in case of
	 *         error.
	 */
	private static String getOpenIDConnectURL() {
		// Get part <scheme>://<host>:<port>/auth from configuration file
		// e.g. https://circ4life.iccs.gr/auth
		final String serverUrl = ConfigProperties.getAccessControlManagerURL(AccessControlManager.class);
		if (null == serverUrl) {
			logger.severe("Configuration file does not contain Access Control Manager URL!");

			return null;
		}

		try {
			// We need to build a proper URL with the realm and relative OpenID Connect path
			// accctrlmngrurl + realms/<realm>/.well-known/uma2-configuration
			final URIBuilder uribuilder = new URIBuilder(serverUrl);
			uribuilder.setPath("/auth/realms/" + getRealm() + "/.well-known/uma2-configuration");
			final String openidurl = uribuilder.build().toString();
			logger.finest("OpenID Connect Discovery URL: " + openidurl);

			return openidurl;
		} catch (URISyntaxException e) {
			logger.severe("Malformed OpenID Connect Discovery URL, details: " + e.getMessage());

			return null;
		} catch (ConfigurationException e) {
			logger.severe("Malformed cinfiguration, details: " + e.getMessage());

			return null;
		}
	}

	/**
	 * Get the specified endpoint from the Access Control Manager OpenID Connect
	 * Discovery URL.
	 * 
	 * @param endpoint the endpoint property name in the JSON response.
	 * @return a {@link URI} object representing the specified endpoint, or
	 *         <code>null</code> in case of any error.
	 * @throws URISyntaxException      in case of malformed endpoint URI or
	 *                                 parameters.
	 * @throws IOException             in case of a connection I/O error.
	 * @throws ClientProtocolException in case of an HTTP protocol error.
	 */
	private static URI getOpenIDConnectEndpoint(final String endpoint)
			throws URISyntaxException, ClientProtocolException, IOException {
		final JsonObject OIDC_JSON = httpGetRequest(new URI(getOpenIDConnectURL()), null, null);

		final String endpointURI = OIDC_JSON.getJsonString(endpoint).getString();
		logger.finest("OpenID Connect Endpoint " + endpoint + ": " + endpointURI);

		return new URI(endpointURI);
	}

	/**
	 * Get the token endpoint from the Access Control Manager OpenID Connect
	 * Discovery URL.
	 * 
	 * @return a {@link URI} object representing the token endpoint.
	 * @throws URISyntaxException      in case of malformed endpoint URI or
	 *                                 parameters.
	 * @throws IOException             in case of a connection I/O error.
	 * @throws ClientProtocolException in case of an HTTP protocol error.
	 */
	private static URI getOpenIDConnectTokenEndpoint() throws URISyntaxException, ClientProtocolException, IOException {
		return getOpenIDConnectEndpoint("token_endpoint");
	}

	/**
	 * Get the end session endpoint from the Access Control Manager OpenID Connect
	 * Discovery URL.
	 * 
	 * @return a {@link URI} object representing the end session endpoint.
	 * @throws URISyntaxException      in case of malformed endpoint URI or
	 *                                 parameters.
	 * @throws IOException             in case of a connection I/O error.
	 * @throws ClientProtocolException in case of an HTTP protocol error.
	 */
	private static URI getOpenIDConnectEndSessionEndpoint()
			throws URISyntaxException, ClientProtocolException, IOException {
		return getOpenIDConnectEndpoint("end_session_endpoint");
	}

	/**
	 * Read the JSON object from the specified response if possible and release all
	 * resources bu calling the close() method of the response.
	 * 
	 * @param response the response to read the JSON object from.
	 * @return a {@link JsonObject} containing the response on success,
	 *         <code>null</code> on failure.
	 * @throws IOException if an I/O error occurs while reading or closing the
	 *                     response.
	 */
	private static JsonObject readJSONResponse(final CloseableHttpResponse response) throws IOException {
		final StatusLine statusline = response.getStatusLine();
		if (null == statusline) {
			logger.severe("HTTP Response with no status line set!");

			return null;
		}

		final int status = statusline.getStatusCode();
		// HTTP not 2XX, something happened...
		if (status > 299) {
			logger.severe("HTTP Response status error, code = " + status);

			return null;
		}

		// In case of HTTP 204 No Content, return an empty JSON object
		if (204 == status) {
			return Json.createObjectBuilder().build();
		}

		JsonObject ResponseJSON = null;
		try {
			final HttpEntity entity = response.getEntity();
			if (null != entity) {
				final InputStream is = entity.getContent();
				final JsonReader reader = Json.createReader(is);
				ResponseJSON = reader.readObject();
				is.close();
			}
		} finally {
			response.close();
		}

		return ResponseJSON;
	}

	/**
	 * Use HTTP GET and perform a request on the specified {@link URI}, adding
	 * optional HTTP headers and URL form encoded parameters.
	 * 
	 * @param endpoint   the URI to GET.
	 * @param headers    optional HTTP headers key/value map, can be
	 *                   <code>null</code> to skip.
	 * @param parameters optional URL encoded parameters key/value map, can be
	 *                   <code>null</code> to skip.
	 * @return
	 * @throws ClientProtocolException in case of an HTTP protocol error.
	 * @throws IOException             in case of a connection I/O error.
	 * @throws URISyntaxException      in case of malformed endpoint URI or
	 *                                 parameters.
	 */
	private static JsonObject httpGetRequest(final URI endpoint, final Map<String, String> headers,
			final Map<String, String> parameters) throws ClientProtocolException, IOException, URISyntaxException {
		final CloseableHttpClient client = HttpClients.createDefault();
		final HttpGet GET = new HttpGet();

		// HTTP headers
		if (null != headers) {
			for (final Entry<String, String> header : headers.entrySet()) {
				GET.setHeader(header.getKey(), header.getValue());
			}
		}

		// HTTP parameters
		final URIBuilder uribuilder = new URIBuilder(endpoint);
		if (null != parameters) {
			for (final Entry<String, String> param : parameters.entrySet()) {
				uribuilder.setParameter(URLEncoder.encode(param.getKey(), Consts.UTF_8.displayName()),
						URLEncoder.encode(param.getValue(), Consts.UTF_8.displayName()));
			}
		}
		GET.setURI(uribuilder.build());

		// Perform the request and read the JSON response
		return readJSONResponse(client.execute(GET));
	}

	/**
	 * Use HTTP POST and perform a request on the specified {@link URI}, adding
	 * optional HTTP headers and URL form encoded parameters.
	 * 
	 * @param endpoint   the URI to POST the request to.
	 * @param headers    optional HTTP headers key/value map, can be
	 *                   <code>null</code> to skip.
	 * @param parameters optional URL encoded parameters key/value map, can be
	 *                   <code>null</code> to skip.
	 * @return a {@link JsonObject} containing the response on success,
	 *         <code>null</code> on failure.
	 * @throws ClientProtocolException in case of an HTTP protocol error.
	 * @throws IOException             in case of a connection I/O error.
	 */
	private static JsonObject httpPostURLEncodedParameters(final URI endpoint, final Map<String, String> headers,
			final Map<String, String> parameters) throws ClientProtocolException, IOException {
		final CloseableHttpClient client = HttpClients.createDefault();
		final HttpPost POST = new HttpPost(endpoint);

		// HTTP headers
		if (null != headers) {
			for (final Entry<String, String> header : headers.entrySet()) {
				POST.setHeader(header.getKey(), header.getValue());
			}
		}

		// HTTP parameters
		if (null != parameters) {
			final List<NameValuePair> params = new ArrayList<NameValuePair>();
			for (final Entry<String, String> param : parameters.entrySet()) {
				params.add(new BasicNameValuePair(param.getKey(), param.getValue()));
			}
			POST.setEntity(new UrlEncodedFormEntity(params, Consts.UTF_8));
		}

		// Perform the request and read the JSON response
		return readJSONResponse(client.execute(POST));
	}

	/**
	 * Get a user representation object in order to retrieve details from it.
	 * 
	 * @param useruuid the Universally Unique ID.
	 * @return the user representation if the user UUID exists and no error happens,
	 *         <code>null</code> otherwise.
	 */
	private static UserRepresentation getUser(final String useruuid) {
		try {
			if (!isProperFormatUUID(useruuid)) {
				return null;
			}

			// Invoke the REST API resource for user representation
			return getAdminAPIClient().realm(getRealm()).users().get(useruuid).toRepresentation();
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get the specified user's group list.
	 * 
	 * @param useruuid the Universally Unique ID.
	 * @return the user group list if the user UUID exists and no error happens,
	 *         <code>null</code> otherwise.
	 */
	private static List<String> getUserGroups(final String useruuid) {
		try {
			if (!isProperFormatUUID(useruuid)) {
				return null;
			}

			final ArrayList<String> groupnames = new ArrayList<String>();
			final List<GroupRepresentation> groups = getAdminAPIClient().realm(getRealm()).users().get(useruuid)
					.groups();
			if (null != groups) {
				for (final GroupRepresentation groupRepresentation : groups) {
					groupnames.add(groupRepresentation.getName());
				}
			}

			// Invoke the REST API resource for user representation
			return groupnames;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get a group representation object in order to retrieve details from it.
	 * 
	 * @param groupuuid the Universally Unique ID.
	 * @return the group representation if the user UUID exists and no error
	 *         happens, <code>null</code> otherwise.
	 */
	private static GroupRepresentation getGroup(final String groupuuid) {
		try {
			// Invoke the REST API resource for group representation
			return getAdminAPIClient().realm(getRealm()).groups().group(groupuuid).toRepresentation();
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Create the Eco-account enabled users's group.
	 * 
	 * @return <code>true</code> if the group is created, <code>false</code>
	 *         otherwise.
	 */
	private static boolean createEcoAccountGroup() {
		final GroupRepresentation rep = new GroupRepresentation();
		rep.setName(ECOACCOUNTS_GROUP);

		// TODO: Perform group setup with specific properties: access, attributes,
		// clientRoles, id, path, realmRoles, subGroups

		return createGroup(rep);
	}

	/**
	 * Create a top level group under the currently configured Access Control
	 * Manager realm.
	 * 
	 * @param rep the group representation.
	 * @return <code>true</code> if the group was created successfully,
	 *         <code>false</code> otherwise.
	 */
	private static boolean createGroup(final GroupRepresentation rep) {
		try {
			final Response result = getAdminAPIClient().realm(getRealm()).groups().add(rep);

			// All OK, group should be created
			if (299 > result.getStatus()) {
				return true;
			}

			// Something went wrong, throw exception
			throw new Exception(
					"Status code: " + result.getStatus() + ", details: " + result.getStatusInfo().getReasonPhrase());
		} catch (Exception e) {
			logger.severe("Could not create group " + rep.getName() + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Create the Eco-account user representation from the basic properties.
	 * 
	 * @param username  the username.
	 * @param password  the password.
	 * @param email     the email.
	 * @param firstname the user's real world first name.
	 * @param lastname  the user's real world family name.
	 * @param locale    ISO 639 alpha-2 language code for this locale. Supported
	 *                  values are configured in the ACM. If this is malformed or
	 *                  not supported the default "en" is used.
	 * @return the user's representation.
	 */
	private static UserRepresentation createEcoAccountRepresentation(final String username, final String password,
			final String email, final String firstname, final String lastname, final String locale) {
		final UserRepresentation rep = new UserRepresentation();

		// Username
		rep.setUsername(username);

		// Password
		final ArrayList<CredentialRepresentation> credentials = new ArrayList<CredentialRepresentation>();
		final CredentialRepresentation credrep = new CredentialRepresentation();
		credrep.setType(CredentialRepresentation.PASSWORD);
		credrep.setValue(password);
		credentials.add(credrep);
		rep.setCredentials(credentials);

		// Email, Real world names
		rep.setEmail(email);
		rep.setFirstName(firstname);
		rep.setLastName(lastname);

		rep.setEnabled(true);
		rep.setEmailVerified(false);

		// Locale
		rep.singleAttribute("locale", convertLocale(locale));

		// TODO: Perform user setup with specific properties, roles, access control,
		// etc.

		return rep;
	}

	/**
	 * Update the Eco-account user representation from the basic properties.
	 * 
	 * @param username  the username.
	 * @param password  the password.
	 * @param email     the email.
	 * @param firstname the user's real world first name.
	 * @param lastname  the user's real world family name.
	 * @return the user's representation.
	 */
	private static UserRepresentation updateEcoAccountRepresentation(final String password, final String email,
			final String firstname, final String lastname, final String locale) {
		final UserRepresentation rep = new UserRepresentation();

		// Password
		if (null != password) {
			final ArrayList<CredentialRepresentation> credentials = new ArrayList<CredentialRepresentation>();
			final CredentialRepresentation credrep = new CredentialRepresentation();
			credrep.setType(CredentialRepresentation.PASSWORD);
			credrep.setValue(password);
			credentials.add(credrep);
			rep.setCredentials(credentials);
		}

		// Email
		if (null != email) {
			rep.setEmail(email);
		}

		if (null != firstname) {
			rep.setFirstName(firstname);
		}

		if (null != lastname) {
			rep.setLastName(lastname);
		}

		if (null != locale) {
			// Locale
			rep.singleAttribute("locale", convertLocale(locale));
		}

		return rep;
	}

	private static String convertLocale(final String locale) {
		String replocale = "en";
		if (null != locale) {
			switch (locale.toLowerCase()) {
			case "de": {
				replocale = "de";
			}
				break;
			case "el": {
				replocale = "el";
			}
				break;
			case "es": {
				replocale = "es";
			}
				break;
			case "fi": {
				replocale = "fi";
			}
				break;
			case "fr": {
				replocale = "fr-BE";
			}
				break;
			case "nl": {
				replocale = "nl-BE";
			}
				break;
			case "pl": {
				replocale = "pl";
			}
				break;
			case "sv": {
				replocale = "sv";
			}
				break;
			default: {
				replocale = "en";
			}
				break;
			}
		}

		return replocale;
	}

	/**
	 * Create a new user in the Access Control Manager with the specified
	 * representation.
	 * 
	 * @param rep the user representation.
	 * @return <code>true</code> if the user was created successfully,
	 *         <code>false</code> otherwise.
	 */
	private static boolean createUser(final UserRepresentation rep) {
		try {
			final Response result = getAdminAPIClient().realm(getRealm()).users().create(rep);

			// All OK, group should be created
			if (299 > result.getStatus()) {
				logger.info(
						"Created user " + rep.getUsername() + ", response details: " + result.getLocation().getPath());

				return true;
			}

			// Something went wrong, throw exception
			throw new Exception(
					"Status code: " + result.getStatus() + ", details: " + result.getStatusInfo().getReasonPhrase());
		} catch (Exception e) {
			logger.severe("Could not create user " + rep.getUsername() + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Add user to the specified group.
	 * 
	 * @param useruuid  the user's Universally Unique ID.
	 * @param groupuuid the group's Universally Unique ID.
	 * @return <code>true</code> if the group join was successful,
	 *         <code>false</code> otherwise.
	 */
	private static boolean addUserToGroup(final String useruuid, final String groupuuid) {
		try {
			getAdminAPIClient().realm(getRealm()).users().get(useruuid).joinGroup(groupuuid);

			return true;
		} catch (Exception e) {
			logger.severe("Could not add user " + useruuid + " to group " + groupuuid + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Update a user's representation.
	 * 
	 * @param useruuid the user's Universally Unique ID.
	 * @param rep      the user representation update.
	 * @return <code>true</code> if the update was performed successfully,
	 *         <code>false</code> otherwise.
	 */
	private static boolean updateUser(final String useruuid, final UserRepresentation rep) {
		try {
			getAdminAPIClient().realm(getRealm()).users().get(useruuid).update(rep);

			return true;
		} catch (Exception e) {
			logger.severe("Could not update user " + rep.getUsername() + ", details: " + e.getMessage());

			return false;
		}
	}
}
