/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.openapi;

import java.time.LocalDate;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.Date;

import gr.iccs.isense.circ4life.datamodels.WEEEsEndOfLifeState;

/**
 * Common OpenAPI to internal structures conversion routines and objects.
 */
public class DataConversion {
	/**
	 * Convert End of Life state string to enumeration.
	 * 
	 * @param eolstate the {@link String} describing the End of Life state.
	 * @return the enumeration if the specified {@link String} is a valid one,
	 *         <code>null</code> otherwise.
	 */
	public static WEEEsEndOfLifeState convertEoLState(final String eolstate) {
		if (eolstate.equalsIgnoreCase("working")) {
			return WEEEsEndOfLifeState.WORKING;
		} else if (eolstate.equalsIgnoreCase("repairable")) {
			return WEEEsEndOfLifeState.REUSABLE;
		} else if (eolstate.equalsIgnoreCase("broken")) {
			return WEEEsEndOfLifeState.BROKEN;
		}

		// Should never reach here
		return null;
	}

	/**
	 * Format GTIN code number to GTIN {@link String} with 14 characters length.
	 * 
	 * @param gtin The GTIN code as long integer.
	 * @return GTIN {@link String} with 14 characters length.
	 */
	public static String formatGTIN14(final long gtin) {
		return String.format("%1$014d", gtin);
	}

	/**
	 * Format date as ISO 8601 YYYY-MM-DD, e.g. 1985-01-14
	 * 
	 * @param date {@link Date} object to format as {@link String}.
	 * @return a {@link String} representing the date formatted as ISO 8601
	 *         YYYY-MM-DD.
	 */
	public static String formatDate(final Date date) {
		final LocalDate localdate = date.toInstant().atZone(ZoneId.systemDefault()).toLocalDate();
		return DateTimeFormatter.ISO_DATE.format(localdate);
	}

	/**
	 * Read date from {@link String} expecting a ISO 8601 YYYY-MM-DD format.
	 * 
	 * @param datestr the {@link String} date formatted as ISO 8601 YYYY-MM-DD.
	 * @return the {@link Date} object representing the same date if no parsing
	 *         error happens.
	 * @throws IllegalArgumentException if any error occurs due to malformed datestr
	 *                                  value.
	 */
	public static Date readDate(final String datestr) throws IllegalArgumentException {
		final LocalDate localdate = DateTimeFormatter.ISO_DATE.parse(datestr, LocalDate::from);
		return Date.from(localdate.atStartOfDay().atZone(ZoneId.systemDefault()).toInstant());
	}
}
