/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.utils;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.jar.Manifest;

/**
 * Utility class for accessing various application MANIFEST information.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class ManifestAccess {
	/**
	 * Retrieve the application title from the packaged Manifest file for a specific
	 * class.
	 *
	 * @param appclass The class to get the application title for.
	 * @return {@link String} representation of this project's title as set in Maven
	 *         POM properties in Manifest file or <code>null</code> if not found.
	 */
	public static String getAppTitleFromManifest(final Class<?> appclass) {
		return getMainManifestAttribute(appclass, "Implementation-Title");
	}

	/**
	 * Retrieve the application version from the packaged Manifest file for a
	 * specific class.
	 *
	 * @param appclass The class to get the version for.
	 *
	 * @return {@link String} representation of this project's version as set in
	 *         Maven POM properties in Manifest file or <code>null</code> if not
	 *         found.
	 */
	public static String getVersionFromManifest(final Class<?> appclass) {
		return getMainManifestAttribute(appclass, "Implementation-Version");
	}

	/**
	 * Retrieve an attribute's value from the main application MANIFEST of a
	 * specified application class.
	 *
	 * @param appclass The class to get the attribute for.
	 * @param attrname The attribute name to get the value for.
	 * @return {@link String} representation of the specified attribute's value.
	 */
	private static String getMainManifestAttribute(final Class<?> appclass, final String attrname) {
		final String manifestpath = getMainManifestPath(appclass);

		if (manifestpath == null) {
			return "Unknown (Class not found in CLASSPATH)";
		}

		InputStream is = null;
		try {
			is = new URL(manifestpath).openStream();
		} catch (MalformedURLException e) {
			return "Unknown (Malformed MANIFEST location)";
		} catch (IOException e) {
			return "Unknown (Cannot open MANIFEST location)";
		}

		Manifest m = null;
		try {
			m = new Manifest(is);
		} catch (IOException e) {
			return "Unknown (Cannot create MANIFEST object)";
		}

		final String value = m.getMainAttributes().getValue(attrname);

		return (value == null) ? "Unknown (Attribute not found)" : value;
	}

	/**
	 * Return the path of the specified application's class main MANIFEST file. The
	 * returned location is the EXPECTED location, the MANIFEST file may not be
	 * there due to build system error or local debugging (e.g. Eclipse).
	 *
	 * @param appclass The application's class we wish to get the main MANIFEST file
	 *                 for.
	 * @return A {@link String} representation of the path on success,
	 *         <code>null</code> otherwise.
	 */
	private static String getMainManifestPath(final Class<?> appclass) {
		// Classpath URL of this class, this can be a local file or a JAR file
		final URL classpathURL = appclass.getResource(appclass.getSimpleName() + ".class");
		if (classpathURL == null) {
			return null;
		}
		final String classpath = classpathURL.toString();

		// Check if we are inside a JAR or unpacked class in filesystem
		String manifest = null;
		if (classpath.startsWith("jar")) {
			// You will find the main Manifest after the ! character in this classpath
			manifest = classpath.substring(0, classpath.lastIndexOf("!") + 1);
			manifest += "/META-INF/MANIFEST.MF";
		} else if (classpath.contains("/WEB-INF/classes/")) {
			// This is a web application running from a container,
			// possibly packaged as WAR. So the location of META-INF/MANIFEST.MF
			// is on the extracted directory inside the server container.
			// <app folder>/WEB-INF/classes/your/package/relative/path/<class name>.class
			// <app folder>/META-INF/MANIFEST.MF
			final String relpath = "/WEB-INF/classes/" + appclass.getName().replace('.', '/') + ".class";
			// Remove the relative path from the complete classpath URL to get root folder
			manifest = classpath.substring(0, classpath.length() - relpath.length());
			manifest += "/META-INF/MANIFEST.MF";
		} else {
			// You will find the main Manifest in the same folder as your class
			// relative folder starts. This root folder contains the main Manifest
			// in relative path META-INF/MANIFEST.MF:
			// <root folder>/your/package/relative/path/<class name>.class
			// <root folder>/META-INF/MANIFEST.MF
			final String relpath = appclass.getName().replace('.', '/') + ".class";
			// Remove the relative path from the complete classpath URL to get root folder
			manifest = classpath.substring(0, classpath.length() - relpath.length() - 1);
			manifest += "/META-INF/MANIFEST.MF";
		}

		return manifest;
	}
}
