/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.utils;

import java.io.PrintStream;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Objects;

/**
 * Common interface and functionality of program options.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public abstract class ProgramOptions {
	/**
	 * Organization under which this program is developed.
	 */
	protected static final String ORGANIZATION = "ICCS";

	/**
	 * Organization team that develops this program.
	 */
	protected static final String ORGANIZATION_TEAM = "i-sense group";

	/**
	 * Encapsulation of an option name and type.
	 * 
	 * @param <T> The {@link Object} type of the option value.
	 */
	protected final static class Option<T> {

		/**
		 * Option name.
		 */
		private final String m_name;

		/**
		 * Option value type.
		 */
		private final Class<T> m_type;

		/**
		 * Option usage message, used for help usage construction.
		 */
		private final String m_usage;

		/**
		 * Constructor.
		 * 
		 * @param name  Option name.
		 * @param type  Option value type.
		 * @param usage Option usage message.
		 */
		public Option(final String name, final Class<T> type, final String usage) {
			m_name = name;
			m_type = type;
			m_usage = usage;
		}

		/**
		 * @return The option's name.
		 */
		public final String getName() {
			return m_name;
		}

		/**
		 * @return the options's {@link Object} type.
		 */
		Class<T> getType() {
			return m_type;
		}

		/**
		 * @return The option's usage message.
		 */
		public final String getUsage() {
			return m_usage;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#hashCode()
		 */
		@Override
		public int hashCode() {
			// Hash code from name and type not usage, to avoid passing usage on value
			// retrieval. If someone passes a name and type already present in the map and
			// changes the usage, the first found will be returned.
			return Objects.hash(m_name, m_type);
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		@Override
		public boolean equals(Object obj) {
			return this.hashCode() == obj.hashCode();
		}

	}

	/**
	 * Option name to option value map. Respects order of insertion while
	 * enumerating key/value pairs.
	 */
	private final LinkedHashMap<Option<?>, Object> name_value_map = new LinkedHashMap<Option<?>, Object>();

	/**
	 * Initialize an option.
	 *
	 * @param <T>         the type of the option value.
	 * @param optionname  The option name to set the value for.
	 * @param optionvalue The option value to set.
	 * @param usage       the option usage text.
	 * @throws IllegalArgumentException If any of the arguments is
	 *                                  <code>null</code>.
	 */
	public <T> void initValue(final String optionname, final T optionvalue, final String usage)
			throws IllegalArgumentException {
		if (optionname == null || optionvalue == null || usage == null) {
			throw new IllegalArgumentException("Null object used to initialize option!");
		}

		name_value_map.put(createKey(optionname, optionvalue.getClass(), usage), optionvalue);
	}

	/**
	 * Set a value based only on option name and type.
	 * 
	 * @param <T>         the type of the option value.
	 * @param optionname  The option name.
	 * @param optionvalue The option value.
	 * @throws IllegalArgumentException If any of the arguments is
	 *                                  <code>null</code>.
	 */
	public <T> void setValue(final String optionname, final T optionvalue) throws IllegalArgumentException {
		if (optionname == null || optionvalue == null) {
			throw new IllegalArgumentException("Null object used to initialize option!");
		}

		// The null in the usage is set freely, as it does not play part in the object's
		// hash code creation.
		final Option<?> option = createKey(optionname, optionvalue.getClass(), null);

		name_value_map.put(option, optionvalue);
	}

	/**
	 * Get the value for an option. You must specify the type of the value. Uses the
	 * hash code to retrieve from the Hash Map directly.
	 * 
	 * @param <T>        the type of the class type.
	 * @param optionname The option name to get the value of.
	 * @param optiontype The class type of the specified option.
	 * @return The option value.
	 */
	public <T> T getValue(final String optionname, final Class<T> optiontype) {
		// The null in the usage is set freely, as it does not play part in the object's
		// hash code creation.
		final Option<T> option = createKey(optionname, optiontype, null);

		return option.getType().cast(name_value_map.get(option));
	}

	/**
	 * @return The options name to value map.
	 */
	public final Map<Option<?>, Object> getOptions() {
		return name_value_map;
	}

	/**
	 * Helper creator of Option Hash map key objects.
	 * 
	 * @param <T>   the option type class type.
	 * @param name  Option name.
	 * @param type  Option type.
	 * @param usage Option usage, <code>null</code> if you want just a key for
	 *              retrieval.
	 * @return a newly created option object.
	 */
	public static <T> Option<T> createKey(final String name, final Class<T> type, final String usage) {
		return new Option<>(name, type, usage);
	}

	/**
	 * Prints program's usage help message to the specified output stream.
	 * 
	 * @param os the output stream to direct messages to.
	 */
	public abstract void printUsage(final PrintStream os);

	/**
	 * Parse the program's command line options.
	 * 
	 * @param arguments Array of {@link String} containing the command line
	 *                  arguments (as passed to main function of main application
	 *                  class).
	 * @return <code>true</code> on successful parsing, <code>false</code>
	 *         otherwise.
	 */
	public abstract boolean parseCommandLine(final String[] arguments);

	/**
	 * Removes single or double quotes from the start and end of an input
	 * {@link String}. Unfortunately the Java VM passes quoted command line
	 * arguments to the application.
	 * 
	 * @param str Input {@link String} which may be wrapped with single or double
	 *            quotes.
	 * @return The str {@link String} without single or double quotes.
	 */
	protected static String removeQuotes(final String str) {
		final String unquoted = str.replaceAll("^[\"']|[\"']$", "");

		return unquoted;
	}

	/**
	 * Check if the argument in the specified index is an assignable, non option
	 * value.
	 * 
	 * @param args  Arguments {@link String} array.
	 * @param index Index of argument to check.
	 * @return <code>true</code> if the value in the args_index position is OK,
	 *         <code>false</code> otherwise.
	 */
	protected static boolean isOptionValue(final String[] args, final int index) {
		if (index < args.length && !args[index].startsWith("--")) {
			return true;
		}

		return false;
	}

	/**
	 * This is the program options validator and should be implemented by all
	 * subclasses.
	 * 
	 * @return <code>true</code> if the options set via
	 *         {@link ProgramOptions#setValue(String, Object)} are valid,
	 *         <code>false</code> otherwise.
	 */
	protected abstract boolean validateOptions();
}
