/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.validation;

import java.util.logging.Logger;
import java.util.regex.Pattern;

import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;

import gr.iccs.isense.circ4life.logging.Log;

/**
 * Email address validation class.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EmailAddressValidator {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(EmailAddressValidator.class);

	/**
	 * Validate the specified e-mail address. First validation on rules of RFC822
	 * are performed and then extra ICT platform specific rules. This method
	 * validates ONLY the email address itself, NOT THE EXISTENCE of the mailbox it
	 * targets or the existence of the remote mail server.
	 * 
	 * @param email the email address to validate.
	 * @return <code>true</code> if the email address is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final String email) {
		try {
			// Unfortunatelly some addresses that are invalid for RFC822 implementation
			// of javax.mail.internet.InternetAddress are considered valid by the official
			// regular expression of RFC5322F. That is why we apply them in this order:
			// RFC822 -> RFC5322 -> Our ICT Platform rules

			// RFC822 strict validation by javax.mail.internet.InternetAddress
			validateRFC822(email);

			// RFC 5322 strict validation
			validateRFC5322(email);

			// Our own sanity regular expression.
			validateICTPlatform(email);

			// All OK
			return true;
		} catch (AddressException e) {
			logger.severe("Invalid e-mail address " + email + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * If the specified email address is not compatible with the RFC822 rules an
	 * {@link AddressException} is thrown.
	 * 
	 * @param email the email address to validate.
	 * @throws AddressException thrown on validation error.
	 */
	public static void validateRFC822(final String email) throws AddressException {
		// RFC822 strict validation by javax.mail.internet.InternetAddress
		final InternetAddress emailaddress = new InternetAddress(email, true);
		emailaddress.validate();
	}

	/**
	 * If the specified email address is not compatible with the RFC5322 regular
	 * expression an {@link AddressException} is thrown.
	 * 
	 * @param email the email address to validate.
	 * @throws AddressException thrown on validation error.
	 */
	public static void validateRFC5322(final String email) throws AddressException {
		// This regular expression is the official recommendation from RFC5322
		// https://emailregex.com/
		final String regex = "(?:[a-z0-9!#$%&'*+/=?^_`{|}~-]+" + "(?:\\.[a-z0-9!#$%&'*+/=?^_`{|}~-]+)*"
				+ "|\"(?:[\\x01-\\x08\\x0b\\x0c\\x0e-\\x1f\\x21\\x23-\\x5b\\x5d-\\x7f]|"
				+ "\\\\[\\x01-\\x09\\x0b\\x0c\\x0e-\\x7f])*\")" + "@(?:(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?\\.)"
				+ "+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?" + "|\\[(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.)"
				+ "{3}(?:25[0-5]|2[0-4][0-9]|[01]?" + "[0-9][0-9]?|[a-z0-9-]*"
				+ "[a-z0-9]:(?:[\\x01-\\x08\\x0b\\x0c\\x0e-\\x1f\\x21-\\x5a\\x53-\\x7f]"
				+ "|\\\\[\\x01-\\x09\\x0b\\x0c\\x0e-\\x7f])+)\\])";
		final Pattern pattern = Pattern.compile(regex, Pattern.CASE_INSENSITIVE);

		if (!pattern.matcher(email).matches()) {
			throw new AddressException("Email address incompatible with RFC5322: " + email);
		}
	}

	/**
	 * If the specified email address is not compatible with the rules below an
	 * {@link AddressException} is thrown.
	 * 
	 * <b>CAUTION: Do not use this validation function as a standalone validation!
	 * It must always validate addresses that have already passed through some
	 * standard validator like an RFC one.</b>
	 * 
	 * <pre>
	 * It invalidates the extra following types of addresses:
	 * 
	 * Local domains not allowed, e.g. user@localhost, user@localdomain, user@somehost.localdomain
	 * 
	 * IP addresses not allowed, e.g. user@127.0.0.1, user@10.0.1.1, user@8.8.8.8, user@147.102.5.97, user@192.168.1.245
	 * 
	 * Number of characters in top level domain must not less than 2 or more than 6,
	 * e.g. username@yahoo.c, username@yahoo.corporate
	 * </pre>
	 * 
	 * 
	 * @param email the email address to validate.
	 * @throws AddressException thrown on validation error.
	 */
	public static void validateICTPlatform(final String email) throws AddressException {
		{
			// Local domains not allowed, e.g. user@localhost, user@localdomain,
			// user@somehost.localdomain
			final String regex = "^.*@localhost|.*\\.localdomain$";
			final Pattern pattern = Pattern.compile(regex, Pattern.CASE_INSENSITIVE);

			if (pattern.matcher(email).matches()) {
				throw new AddressException("Email address has local host or domain: " + email);
			}
		}

		{
			// IP addresses not allowed, e.g. user@127.0.0.1, user@10.0.1.1, user@8.8.8.8,
			// user@147.102.5.97, user@192.168.1.245
			final String ipv4regex = "^.*@(([01]?\\d\\d?|2[0-4]\\d|25[0-5])\\.){3}([01]?\\d\\d?|2[0-4]\\d|25[0-5])$";
			final Pattern ipv4pattern = Pattern.compile(ipv4regex, Pattern.CASE_INSENSITIVE);
			final String ipv6regex = "^.*@([0-9a-f]{1,4}:){7}([0-9a-f]){1,4}$";
			final Pattern ipv6pattern = Pattern.compile(ipv6regex, Pattern.CASE_INSENSITIVE);

			if (ipv4pattern.matcher(email).matches() || ipv6pattern.matcher(email).matches()) {
				throw new AddressException("Email address has IP address instead of domain name: " + email);
			}
		}

		{
			// Number of characters in top level domain must not less than 2 or more than 6,
			// e.g. username@yahoo.c, username@yahoo.corporate
			final String regex = "^.*@(?:[a-zA-Z0-9-]+\\.)+[a-zA-Z]{2,6}$";
			final Pattern pattern = Pattern.compile(regex, Pattern.CASE_INSENSITIVE);

			if (!pattern.matcher(email).matches()) {
				throw new AddressException(
						"Email address has top level domain name with less than 2 or more than 6 characters: " + email);
			}
		}
	}
}
