/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.validation;

import java.net.MalformedURLException;
import java.net.URL;
import java.time.LocalDate;
import java.time.ZoneId;
import java.util.Date;
import java.util.logging.Logger;
import java.util.regex.Pattern;

import gr.iccs.isense.circ4life.logging.Log;

/**
 * User registration form validator.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EndUserRegistrationFormValidator {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(EndUserRegistrationFormValidator.class);

	/**
	 * Allowed usernames regular expression, allowed only A-Z, a-z,0-9.
	 */
	protected final static String USERNAMEREGEX = "^[A-Za-z0-9]+";

	/**
	 * Allowed first, last name regular expression, At least 2 and not more than 256
	 * characters of length, allows any character except numbers(0-9) and these
	 * {@code (<>, {}, [], !, @, #, $, %, ^, &, *, +, \, |, /, ?, `, ~)}
	 */
	private final static String REALNAMEREGEX = "^[^0-9\\<\\>\\{\\}\\[\\]\\!\\@\\#\\$\\%\\^\\&\\*\\+\\|\\/\\?\\`\\~]+";

	/**
	 * Allowed URL protocol regular expression, only http/https allowed.
	 */
	private final static String HTTPPATTERNREGEX = "^(http|https)";

	/**
	 * Allowed telephone number regular expression, allowed only + as first
	 * character, 0-9.
	 */
	private final static String TELEPHONEPATTERNREGEX = "^\\+?[0-9]+";

	/**
	 * Maximum length of free text property fields.
	 */
	private final static int MAXFREETEXTLENGTH = 256;

	/**
	 * Check the validity of required parameters.
	 * 
	 * @param username   user name. <strong>Required</strong>. At least 8 characters
	 *                   of length, allowed only A-Z, a-z, 0-9.
	 * @param password   password. <strong>Required</strong>. At least 8 characters
	 *                   of length, not equal to username.
	 * @param email      e-mail. <strong>Required</strong>. Accepts only valid email
	 *                   addresses according to RFC822, RFC5322. Extra conditions
	 *                   that need to be met for ICT Platform: Local domains not
	 *                   allowed, e.g. user@localhost, user@localhost.localdomain,
	 *                   user@somehost.localdomain, IP addresses not allowed, e.g.
	 *                   user@127.0.0.1, user@10.0.1.1, user@8.8.8.8,
	 *                   user@147.102.5.97, user@192.168.1.245, Number of characters
	 *                   in top level domain must not less than 2 or more than 6,
	 *                   e.g. username@yahoo.c, username@yahoo.corporate.
	 * @param givenName  first name. <strong>Required</strong>. At least 2
	 *                   characters of length, allowed first, last name regular
	 *                   expression, At least 2 and not more than 256 characters of
	 *                   length, allows any character except numbers(0-9) and these
	 *                   {@code (<>, {}, [], !, @, #, $, %, ^, &, *, +, \, |, /, ?,
	 *                   `, ~)}.
	 * @param familyName last name. <strong>Required</strong>. At least 2 characters
	 *                   of length, allowed first, last name regular expression, At
	 *                   least 2 and not more than 256 characters of length, allows
	 *                   any character except numbers(0-9) and these {@code (<>, {},
	 *                   [], !, @, #, $, %, ^, &, *, +, \, |, /, ?, `, ~)}.
	 * @return <code>true</code> if all parameters are valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final String username, // REQUIRED -> ACM
			final String password, // REQUIRED -> ACM
			final String email, // REQUIRED -> ACM,DB
			final String givenName, // REQUIRED -> ACM,DB
			final String familyName // REQUIRED -> ACM,DB
	) {
		try {
			// validate and throw on error
			validate(username, password, email, givenName, familyName);
		} catch (Exception e) {
			logger.severe("Invalid registration form, details: " + e.getMessage());

			return false;
		}

		// All OK
		return true;
	}

	/**
	 * Check the validity of all possible parameters.
	 * 
	 * @param username         user name. <strong>Required</strong>. At least 8
	 *                         characters of length, allowed only A-Z, a-z,0-9.
	 * @param password         password. <strong>Required</strong>. At least 8
	 *                         characters of length, not equal to username.
	 * @param email            e-mail. <strong>Required</strong>. Accepts only valid
	 *                         email addresses according to RFC822, RFC5322. Extra
	 *                         conditions that need to be met for ICT Platform:
	 *                         Local domains not allowed, e.g. user@localhost,
	 *                         user@localhost.localdomain,
	 *                         user@somehost.localdomain, IP addresses not allowed,
	 *                         e.g. user@127.0.0.1, user@10.0.1.1, user@8.8.8.8,
	 *                         user@147.102.5.97, user@192.168.1.245, Number of
	 *                         characters in top level domain must not less than 2
	 *                         or more than 6, e.g. username@yahoo.c,
	 *                         username@yahoo.corporate.
	 * @param givenName        first name. <strong>Required</strong>. At least 2
	 *                         characters of length, allowed first, last name
	 *                         regular expression, At least 2 and not more than 256
	 *                         characters of length, allows any character except
	 *                         numbers(0-9) and these {@code (<>, {}, [], !, @, #, $, %, ^,
	 *                         &, *, +, \, |, /, ?, `, ~)}.
	 * @param familyName       last name. <strong>Required</strong>. At least 2
	 *                         characters of length, allowed first, last name
	 *                         regular expression, At least 2 and not more than 256
	 *                         characters of length, allows any character except
	 *                         numbers(0-9) and these {@code (<>, {}, [], !, @, #, $, %, ^,
	 *                         &, *, +, \, |, /, ?, `, ~)}.
	 * @param birthDate        Date of birth. <strong>Optional</strong>. Must
	 *                         represent a Date of the past, at least 5 years old.
	 * @param addressCountry   Country code. <strong>Optional</strong>. Country 2
	 *                         letter ISO 639 alpha-2 code
	 * @param postalCode       Postal code of address area.
	 *                         <strong>Optional</strong>. Cannot be an empty string.
	 * @param streetAddress    Street address (street name, number, city, area).
	 *                         <strong>Optional</strong>. Cannot be an empty string.
	 * @param addressLocality  The locality/city in which the street address is
	 *                         located. <strong>Optional</strong>. Cannot be an
	 *                         empty string.
	 * @param addressRegion    The region in which the city belongs.
	 *                         <strong>Optional</strong>. Cannot be an empty string.
	 * @param telephone        Telephone number. <strong>Optional</strong>. At least
	 *                         4 characters of length, allowed only + as first
	 *                         character,0-9.
	 * @param userURL          User's profile URL. <strong>Optional</strong>. Valid
	 *                         HTTP, HTTPs URL.
	 * @param gender           Gender. <strong>Optional</strong>. Free text.
	 * @param organization     A user's affiliated organization, e.g. Company,
	 *                         University, Club, Team, etc.
	 *                         <strong>Optional</strong>. Free Text.
	 * @param organizationURL  User's organization URL. <strong>Optional</strong>.
	 *                         Valid HTTP, HTTPs URL.
	 * @param recyclebinuserid Intelligent Bin User code as an 11 digit integer.
	 *                         <strong>Optional</strong>. ZIP code is first 5 digits
	 *                         (e.g. 16004), 6 last digits are sequential (e.g.
	 *                         000001). Minimum = 0, Maximum = 99999999999.
	 * @return <code>true</code> if all parameters are valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final String username, // REQUIRED -> ACM
			final String password, // REQUIRED -> ACM
			final String email, // REQUIRED -> ACM,DB
			final String givenName, // REQUIRED -> ACM,DB
			final String familyName, // REQUIRED -> ACM,DB
			final Date birthDate, // OPTIONAL -> DB
			final String addressCountry, // OPTIONAL -> DB
			final String postalCode, // OPTIONAL -> DB
			final String streetAddress, // OPTIONAL -> DB
			final String addressLocality, // OPTIONAL -> DB
			final String addressRegion, // OPTIONAL -> DB
			final String telephone, // OPTIONAL -> DB
			final String userURL, // OPTIONAL -> DB
			final String gender, // OPTIONAL -> DB
			final String organization, // OPTIONAL -> DB
			final String organizationURL, // OPTIONAL -> DB
			final Long recyclebinuserid // OPTIONAL -> DB
	) {
		try {
			// validate REQUIRED and throw on error
			validate(username, password, email, givenName, familyName);

			// validate OPTIONAL and throw on error
			validate(birthDate, addressCountry, postalCode, streetAddress, addressLocality, addressRegion, telephone,
					userURL, gender, organization, organizationURL, recyclebinuserid);
		} catch (Exception e) {
			logger.severe("Invalid registration form, details: " + e.getMessage());

			return false;
		}

		// All OK
		return true;
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.
	 * 
	 * @param username   user name. At least 8 characters of length, allowed only
	 *                   A-Z, a-z, 0-9.
	 * @param password   password. At least 8 characters of length, not equal to
	 *                   username.
	 * @param email      e-mail.
	 * @param givenName  first name. Allowed first, last name regular expression, At
	 *                   least 2 and not more than 256 characters of length, allows
	 *                   any character except numbers(0-9) and these {@code (<>, {}, [],
	 *                   !, @, #, $, %, ^, &, *, +, \, |, /, ?, `, ~)}.
	 * @param familyName last name. Allowed first, last name regular expression, At
	 *                   least 2 and not more than 256 characters of length, allows
	 *                   any character except numbers(0-9) and these {@code (<>, {}, [],
	 *                   !, @, #, $, %, ^, &, *, +, \, |, /, ?, `, ~)}.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	public static void validate(final String username, final String password, final String email,
			final String givenName, final String familyName) throws IllegalArgumentException {
		// Credentials
		validateCredentials(username, password);

		// Email validation
		validateEmail(email);

		// User's real world name
		validateUserRealName(givenName, familyName);
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.
	 * 
	 * @param birthDate        birthday date.
	 * @param addressCountry   2-letter country code, e.g. GR.
	 * @param postalCode       Postal code number.
	 * @param streetAddress    Full street address, name, number, town.
	 * @param addressLocality  City name.
	 * @param addressRegion    Address region.
	 * @param telephone        telephone number.
	 * @param userURL          user profile URL.
	 * @param gender           gender.
	 * @param organization     affiliated organization name.
	 * @param organizationURL  affiliated organization URL.
	 * @param recyclebinuserid recycle bin system user ID.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	public static void validate(final Date birthDate, final String addressCountry, final String postalCode,
			final String streetAddress, final String addressLocality, final String addressRegion,
			final String telephone, final String userURL, final String gender, final String organization,
			final String organizationURL, final Long recyclebinuserid) throws IllegalArgumentException {
		// Birthday
		validateBirthday(birthDate);

		// Address
		validateAddress(addressCountry, postalCode, streetAddress, addressLocality, addressRegion);

		// Telephone
		validateTelephoneNumber(telephone);

		// User's profile URL
		if (null != userURL) {
			try {
				validateHTTPURL(userURL);
			} catch (MalformedURLException e) {
				throw new IllegalArgumentException("Illegal user profile URL: " + userURL);
			}
		}

		// Validate Gender as free text
		validateFreeText(gender);

		// Validate organization name as free text
		validateFreeText(organization);

		// Organization URL
		if (null != organizationURL) {
			try {
				validateHTTPURL(organizationURL);
			} catch (MalformedURLException e) {
				throw new IllegalArgumentException("Illegal user organization URL: " + userURL);
			}
		}

		// Recycle Bin User ID
		validateRecycleBinUserID(recyclebinuserid);
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.
	 * 
	 * @param email           e-mail.
	 * @param givenName       first name.
	 * @param familyName      last name.
	 * @param birthDate       birthday date.
	 * @param addressCountry  2-letter country code, e.g. GR.
	 * @param postalCode      Postal code number.
	 * @param streetAddress   Full street address, name, number, town.
	 * @param addressLocality City name.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	public static void validate(final String email, final String givenName, final String familyName,
			final Date birthDate, final String addressCountry, final String postalCode, final String streetAddress,
			final String addressLocality) throws IllegalArgumentException {
		// Email validation
		validateEmail(email);

		// User's real world name
		validateUserRealName(givenName, familyName);

		// Birthday
		validateBirthday(birthDate);

		// Address
		validateAddress(addressCountry, postalCode, streetAddress, addressLocality, null);
	}

	/**
	 * Validate the parameters passed and throw {@link MalformedURLException} in
	 * case of any error.
	 * 
	 * @param url HTTP(s) URL. <strong>MUST NOT BE <code>null</code></strong>
	 * @throws MalformedURLException thrown if any of the required input parameters
	 *                               is not valid.
	 */
	public static void validateHTTPURL(final String url) throws MalformedURLException {
		final URL Url = new URL(url);
		if (!Pattern.matches(HTTPPATTERNREGEX, Url.getProtocol())) {
			throw new MalformedURLException("Invalid protocol in URL: " + Url.getProtocol());
		}
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.<br>
	 * 
	 * <strong>CAUTION: You cannot validate password without username, as one of the
	 * requirements is for the password to be different from the username.</strong>
	 * 
	 * @param username user name. At least 8 characters of length, allowed only A-Z,
	 *                 a-z, 0-9.
	 * @param password password. At least 8 characters of length, not equal to
	 *                 username.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	public static void validateCredentials(final String username, final String password)
			throws IllegalArgumentException {
		// General text validation
		validateFreeText(username);
		// At least 8 characters of length, allowed only A-Z, a-z, 0-9
		if (null == username || 8 > username.length() || !Pattern.matches(USERNAMEREGEX, username)) {
			throw new IllegalArgumentException("Illegal user name: " + username);
		}

		// General text validation
		validateFreeText(password);
		// At least 8 characters of length, not equal to username
		if (null == password || 8 > password.length() || password.equalsIgnoreCase(username)) {
			throw new IllegalArgumentException("Illegal user password: " + password);
		}
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.
	 * 
	 * @param email Email address.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	public static void validateEmail(final String email) throws IllegalArgumentException {
		// General text validation
		validateFreeText(email);

		if (null == email || !EmailAddressValidator.isValid(email)) {
			throw new IllegalArgumentException("Illegal user e-mail: " + email);
		}
	}

	/**
	 * Validate user's first name. Allowed first, last name regular expression, At
	 * least 2 and not more than 256 characters of length, allows any character
	 * except numbers(0-9) and these {@code (<>, {}, [], !, @, #, $, %, ^, &, *, +, \, |,
	 * /, ?, `, ~)}.
	 * 
	 * @param givenName first name.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	public static void validateGivenName(final String givenName) throws IllegalArgumentException {
		// General text validation
		validateFreeText(givenName);
		// At least 2 characters of length, allowed only A-Z, a-z
		if (null == givenName || 2 > givenName.length() || !Pattern.matches(REALNAMEREGEX, givenName)) {
			throw new IllegalArgumentException("Illegal user first name: " + givenName);
		}
	}

	/**
	 * Validate user's last name. Allowed first, last name regular expression, At
	 * least 2 and not more than 256 characters of length, allows any character
	 * except numbers(0-9) and these {@code (<>, {}, [], !, @, #, $, %, ^, &, *, +, \, |,
	 * /, ?, `, ~)}.
	 * 
	 * @param familyName last name.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	public static void validateFamilyName(final String familyName) throws IllegalArgumentException {
		// General text validation
		validateFreeText(familyName);
		// At least 2 characters of length, allowed only A-Z, a-z
		if (null == familyName || 2 > familyName.length() || !Pattern.matches(REALNAMEREGEX, familyName)) {
			throw new IllegalArgumentException("Illegal user family name: " + familyName);
		}
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.<br>
	 * 
	 * <strong>CAUTION: You cannot validate first name or last name alone during
	 * registration, as both are required. This function fits this purpose.</strong>
	 * 
	 * @param givenName  first name.
	 * @param familyName last name.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	private static void validateUserRealName(final String givenName, final String familyName)
			throws IllegalArgumentException {
		validateGivenName(givenName);
		validateFamilyName(familyName);
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.
	 * 
	 * @param birthDate User's birthday.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	private static void validateBirthday(final Date birthDate) throws IllegalArgumentException {
		// Must represent a Date of the past, at least 5 years old
		if (null != birthDate) {
			final LocalDate today = LocalDate.now();
			final LocalDate fiveyearsago = LocalDate.of(today.getYear() - 5, today.getMonth(), today.getDayOfMonth());
			final Date tocheck = Date.from(fiveyearsago.atStartOfDay().atZone(ZoneId.systemDefault()).toInstant());
			if (!birthDate.before(tocheck)) {
				throw new IllegalArgumentException("Illegal user birthday: " + birthDate);
			}
		}
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.
	 * 
	 * @param addressCountry  User's Country of residence.
	 * @param postalCode      Address postal code.
	 * @param streetAddress   Street address.
	 * @param addressLocality City/Locality.
	 * @param addressRegion   Address region.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	private static void validateAddress(final String addressCountry, final String postalCode,
			final String streetAddress, final String addressLocality, final String addressRegion)
			throws IllegalArgumentException {
		// General text validation
		validateFreeText(addressCountry);
		validateFreeText(postalCode);
		validateFreeText(streetAddress);
		validateFreeText(addressLocality);
		validateFreeText(addressRegion);

		// 2 characters length only, Country name or 2 letter ISO 639 alpha-2 code
		if (null != addressCountry && 2 != addressCountry.length()) {
			throw new IllegalArgumentException("Illegal user address country code: " + addressCountry);
		}

		// Cannot be an empty string
		if (null != postalCode && 0 == postalCode.length()) {
			throw new IllegalArgumentException("Illegal user postal code: " + postalCode);
		}

		// Cannot be an empty string
		if (null != streetAddress && 0 == streetAddress.length()) {
			throw new IllegalArgumentException("Illegal user street address: " + streetAddress);
		}

		// Cannot be an empty string
		if (null != addressLocality && 0 == addressLocality.length()) {
			throw new IllegalArgumentException("Illegal user address locality/city: " + addressLocality);
		}

		// Cannot be an empty string
		if (null != addressRegion && 0 == addressRegion.length()) {
			throw new IllegalArgumentException("Illegal user address region: " + addressRegion);
		}
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.
	 * 
	 * @param telephone Telephone number.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	private static void validateTelephoneNumber(final String telephone) throws IllegalArgumentException {
		// General text validation
		validateFreeText(telephone);

		if (null != telephone && !Pattern.matches(TELEPHONEPATTERNREGEX, telephone)) {
			throw new IllegalArgumentException("Illegal telephone: " + telephone);
		}
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.
	 * 
	 * @param freetext Free text.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	private static void validateFreeText(final String freetext) throws IllegalArgumentException {
		if (null != freetext && MAXFREETEXTLENGTH < freetext.length()) {
			throw new IllegalArgumentException(
					"Free text exceeds maximum number of characters: " + freetext.substring(0, 30) + "....");
		}
	}

	/**
	 * Validate the parameters passed and throw {@link IllegalArgumentException} in
	 * case of any error.
	 * 
	 * @param recyclebinuserid Recycle Bin User ID.
	 * @throws IllegalArgumentException thrown if any of the required input
	 *                                  parameters is not valid.
	 */
	private static void validateRecycleBinUserID(final Long recyclebinuserid) throws IllegalArgumentException {
		// minimum: 0, maximum: 99999999999
		if (null != recyclebinuserid
				&& (0 > recyclebinuserid.doubleValue() || 99999999999L < recyclebinuserid.doubleValue())) {
			throw new IllegalArgumentException("Illegal recycle bin user ID value: " + recyclebinuserid);
		}
	}
}
