/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.validation;

import java.util.logging.Logger;
import java.util.regex.Pattern;

import gr.iccs.isense.circ4life.logging.Log;

/**
 * User login/logout forms validator.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EndUserSessionFormValidator {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(EndUserSessionFormValidator.class);

	/**
	 * Check the validity of parameters.
	 * 
	 * @param username user name.
	 * @param password password.
	 * @return <code>true</code> if all parameters are valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final String username, final String password) {
		try {
			// Check if the user provided email or username in order to login
			// Keycloak does not care at all, they are equivalent
			if (null != username && EmailAddressValidator.isValid(username)) {
				// Email used as username
				logger.info("Login form uses email:" + username);
			} else if (null == username || 8 > username.length()
					|| !Pattern.matches(EndUserRegistrationFormValidator.USERNAMEREGEX, username)) {
				// At least 8 characters of length, allowed only A-Z, a-z, 0-9
				throw new IllegalArgumentException("Illegal user name: " + username);
			}

			// At least 8 characters of length, not equal to username
			if (null == password || 8 > password.length() || password.equalsIgnoreCase(username)) {
				throw new IllegalArgumentException("Illegal user password: " + password);
			}
		} catch (Exception e) {
			logger.severe("Invalid login form, details: " + e.getMessage());

			return false;
		}

		// All OK
		return true;
	}

	/**
	 * Check the validity of parameters.
	 * 
	 * @param refreshtoken the last, not expired refresh token retrieved from the
	 *                     Access Control Manager.
	 * @return <code>true</code> if all parameters are valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final String refreshtoken) {
		try {
			if (null == refreshtoken || 0 == refreshtoken.length()) {
				throw new IllegalArgumentException("Illegal refresh token: " + refreshtoken);
			}
		} catch (Exception e) {
			logger.severe("Invalid logout form, details: " + e.getMessage());

			return false;
		}

		// All OK
		return true;
	}
}
