/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.test;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import org.junit.Assert;
import org.junit.Test;

import gr.iccs.isense.circ4life.validation.EmailAddressValidator;

/**
 * Tests on email validation.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EmailAddressValidatorTests {
	private final static String VALIDSUFFIX = ".valid.emails.txt";
	private final static String INVALIDSUFFIX = ".invalid.emails.txt";

	@Test
	public void testEmailAddresses() throws IOException {
		// ---------------------------------------------------------------------
		// RFC822 standard, most lenient
		{
			// VALID
			final BufferedReader reader = new BufferedReader(new InputStreamReader(
					EmailAddressValidatorTests.class.getResourceAsStream("/RFC822" + VALIDSUFFIX), "UTF-8"));

			String email;
			while (null != (email = reader.readLine())) {
				try {
					EmailAddressValidator.validateRFC822(email);
					// Should reach here
					Assert.assertTrue(true);
				} catch (Exception e) {
					System.err.println(e.getMessage());
					// Should NOT reach here
					Assert.assertTrue(false);
				}
			}

			reader.close();
		}

		{
			// INVALID
			final BufferedReader reader = new BufferedReader(new InputStreamReader(
					EmailAddressValidatorTests.class.getResourceAsStream("/RFC822" + INVALIDSUFFIX), "UTF-8"));

			String email;
			while (null != (email = reader.readLine())) {
				try {
					EmailAddressValidator.validateRFC822(email);
					// Should NOT reach here
					Assert.assertTrue(false);
				} catch (Exception e) {
					System.err.println(e.getMessage());
					// Should reach here
					Assert.assertTrue(true);
				}
			}

			reader.close();
		}
		// ---------------------------------------------------------------------

		// ---------------------------------------------------------------------
		// RFC822 standard, updated and stricter than RFC822
		{
			// VALID
			final BufferedReader reader = new BufferedReader(new InputStreamReader(
					EmailAddressValidatorTests.class.getResourceAsStream("/RFC5322" + VALIDSUFFIX), "UTF-8"));

			String email;
			while (null != (email = reader.readLine())) {
				try {
					EmailAddressValidator.validateRFC5322(email);
					// Should reach here
					Assert.assertTrue(true);
				} catch (Exception e) {
					System.err.println(e.getMessage());
					// Should NOT reach here
					Assert.assertTrue(false);
				}
			}

			reader.close();
		}

		{
			// INVALID
			final BufferedReader reader = new BufferedReader(new InputStreamReader(
					EmailAddressValidatorTests.class.getResourceAsStream("/RFC5322" + INVALIDSUFFIX), "UTF-8"));

			String email;
			while (null != (email = reader.readLine())) {
				try {
					EmailAddressValidator.validateRFC5322(email);
					// Should NOT reach here
					Assert.assertTrue(false);
				} catch (Exception e) {
					System.err.println(e.getMessage());
					// Should reach here
					Assert.assertTrue(true);
				}
			}

			reader.close();
		}
		// ---------------------------------------------------------------------

		// ---------------------------------------------------------------------
		// Our ICT Platform rules
		{
			// VALID
			final BufferedReader reader = new BufferedReader(new InputStreamReader(
					EmailAddressValidatorTests.class.getResourceAsStream("/ICTRules" + VALIDSUFFIX), "UTF-8"));

			String email;
			while (null != (email = reader.readLine())) {
				try {
					EmailAddressValidator.validateICTPlatform(email);
					// Should reach here
					Assert.assertTrue(true);
				} catch (Exception e) {
					System.err.println(e.getMessage());
					// Should NOT reach here
					Assert.assertTrue(false);
				}
			}

			reader.close();
		}

		{
			// INVALID
			final BufferedReader reader = new BufferedReader(new InputStreamReader(
					EmailAddressValidatorTests.class.getResourceAsStream("/ICTRules" + INVALIDSUFFIX), "UTF-8"));

			String email;
			while (null != (email = reader.readLine())) {
				try {
					EmailAddressValidator.validateICTPlatform(email);
					// Should NOT reach here
					Assert.assertTrue(false);
				} catch (Exception e) {
					System.err.println(e.getMessage());
					// Should reach here
					Assert.assertTrue(true);
				}
			}

			reader.close();
		}
		// ---------------------------------------------------------------------

		// ---------------------------------------------------------------------
		// All above chained as RFC822 -> RFC5322 -> ICT rules
		{
			// VALID
			final BufferedReader reader = new BufferedReader(new InputStreamReader(
					EmailAddressValidatorTests.class.getResourceAsStream("/ALL" + VALIDSUFFIX), "UTF-8"));

			String email;
			while (null != (email = reader.readLine())) {
				Assert.assertTrue(EmailAddressValidator.isValid(email));
			}

			reader.close();
		}

		{
			// INVALID
			final BufferedReader reader = new BufferedReader(new InputStreamReader(
					EmailAddressValidatorTests.class.getResourceAsStream("/ALL" + INVALIDSUFFIX), "UTF-8"));

			String email;
			while (null != (email = reader.readLine())) {
				Assert.assertFalse(EmailAddressValidator.isValid(email));
			}

			reader.close();
		}
		// ---------------------------------------------------------------------

		// ---------------------------------------------------------------------
		// Try also our list of partners
		{
			// VALID
			final BufferedReader reader = new BufferedReader(new InputStreamReader(
					EmailAddressValidatorTests.class.getResourceAsStream("/Partners.emails.txt"), "UTF-8"));

			String email;
			while (null != (email = reader.readLine())) {
				Assert.assertTrue(EmailAddressValidator.isValid(email));
			}

			reader.close();
		}
		// ---------------------------------------------------------------------
	}
}
