/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.integration.test;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;

import javax.naming.ConfigurationException;

import org.glassfish.jersey.logging.LoggingFeature.Verbosity;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import gr.iccs.isense.circ4life.rest.config.ConfigProperties;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager.AccessTokens;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager.LoginResult;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager.LoginResult.ErrorCode;
import gr.iccs.isense.circ4life.test.IntegrationTestsCommon;
import gr.iccs.isense.circ4life.test.IntegrationTestsCommon.TestUser;

/**
 * Integration tests for the Access Control Manager.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class AccessControlManagerIT {
	private Level logginglevel = null;
	private Verbosity loggingverbosity = null;
	private String acmurl = null;
	private String acmrealm = null;
	private String adminuser = null;
	private String adminpass = null;
	private String adminclient = null;

	@Test
	public void itLoginLogout() {
		// Login
		System.out.println("Testing user login...");
		final LoginResult result = AccessControlManager.ecoAccountLogin(AccessControlManager.ENDUSERMODULE_CLIENTID,
				TestUser.Username, TestUser.Password);
		Assert.assertNotNull(result);
		System.out.println("Login result [" + result.error_code + "]");
		Assert.assertTrue(result.error_code == ErrorCode.OK);
		Assert.assertNotNull(result.tokens);

		// Logout
		System.out.println("Testing user logout...");
		final boolean loggedout = AccessControlManager.endSession(AccessControlManager.ENDUSERMODULE_CLIENTID,
				result.tokens);
		Assert.assertTrue(loggedout);
	}

	@Test
	public void itRefreshToken() {
		// Login
		System.out.println("Testing user login...");
		final LoginResult result = AccessControlManager.ecoAccountLogin(AccessControlManager.ENDUSERMODULE_CLIENTID,
				TestUser.Username, TestUser.Password);
		Assert.assertNotNull(result);
		System.out.println("Login result [" + result.error_code + "]");
		Assert.assertTrue(result.error_code == ErrorCode.OK);
		Assert.assertNotNull(result.tokens);

		// Refresh
		System.out.println("Testing token refresh...");
		AccessTokens refreshedtokens = AccessControlManager
				.refreshAccessTokens(AccessControlManager.ENDUSERMODULE_CLIENTID, result.tokens, false);
		Assert.assertNotNull(refreshedtokens);
		Assert.assertTrue("Access token not refreshed!",
				!result.tokens.getAccessToken().contentEquals(refreshedtokens.getAccessToken()));
		Assert.assertTrue("Refresh token not refreshed!",
				!result.tokens.getRefreshToken().contentEquals(refreshedtokens.getRefreshToken()));

		// Ensure that previous tokens are invalid
		Assert.assertTrue(!result.tokens.isTokenValid());

		// Not expired tokens, should be the same
		System.out.println("Testing token refresh...");
		AccessTokens notrefreshedtokens = AccessControlManager
				.refreshAccessTokens(AccessControlManager.ENDUSERMODULE_CLIENTID, refreshedtokens, true);
		Assert.assertNotNull(notrefreshedtokens);
		Assert.assertTrue("Access token refreshed!",
				refreshedtokens.getAccessToken().contentEquals(notrefreshedtokens.getAccessToken()));
		Assert.assertTrue("Refresh token refreshed!",
				refreshedtokens.getRefreshToken().contentEquals(notrefreshedtokens.getRefreshToken()));

		// Logout
		System.out.println("Testing user logout...");
		final boolean loggedout = AccessControlManager.endSession(AccessControlManager.ENDUSERMODULE_CLIENTID,
				notrefreshedtokens);
		Assert.assertTrue(loggedout);
	}

	@Test
	public void itAdminRESTAPI() {
		System.out.println("Testing Admin REST API...");

		final List<String> users_to_test = Arrays.asList(TestUser.Username, adminuser);
		for (final String user : users_to_test) {
			System.out.println("Testing Admin REST API, User: " + user);
			final String useruuid = AccessControlManager.getUsernameUUID(user);
			Assert.assertNotNull(useruuid);
			System.out.println("Testing Admin REST API, User UUID: " + useruuid);
			Assert.assertTrue(AccessControlManager.existsUserUUID(useruuid));
			Assert.assertNotNull(AccessControlManager.getEmail(useruuid));
		}

		Assert.assertFalse(AccessControlManager.existsUserUUID("UUID_THAT_IS_NOT_VALID"));
	}

	@Test
	public void itAdminSendEmails() {
		System.out.println("Testing Admin REST API Email Sending...");

		final List<String> users_to_test = Arrays.asList(TestUser.Username);
		for (final String user : users_to_test) {
			System.out.println("Testing Admin REST API, User: " + user);
			final String useruuid = AccessControlManager.getUsernameUUID(user);
			Assert.assertNotNull(useruuid);
			System.out.println("Testing Admin REST API, User UUID: " + useruuid);
			// Already verified this should return false
			Assert.assertFalse(AccessControlManager.sendVerificationEmail(useruuid));
			// Sent all possible actions
			Assert.assertTrue(AccessControlManager.sendResetPasswordEmail(useruuid));
			// Sent notifications
			Assert.assertTrue(AccessControlManager.sendNotificationEmail(useruuid,
					AccessControlManager.NotificationEmailType.PROFILE_UPDATED));
			Assert.assertTrue(AccessControlManager.sendNotificationEmail(useruuid,
					AccessControlManager.NotificationEmailType.ACCOUNT_REMOVAL));
			Assert.assertTrue(AccessControlManager.sendNotificationEmail(useruuid,
					AccessControlManager.NotificationEmailType.ACM_UPDATED));
		}
	}

	@Before
	public void BeforeTests() throws URISyntaxException, ConfigurationException {
		System.out.println("Preparing tests...");

		logginglevel = ConfigProperties.getLoggingLevel(AccessControlManagerIT.class);
		Assert.assertNotNull(logginglevel);
		loggingverbosity = ConfigProperties.getLoggingVerbosity(AccessControlManagerIT.class);
		Assert.assertNotNull(loggingverbosity);
		acmurl = ConfigProperties.getAccessControlManagerURL(AccessControlManagerIT.class);
		Assert.assertNotNull(acmurl);
		acmrealm = ConfigProperties.getAccessControlManagerRealm(AccessControlManagerIT.class);
		Assert.assertNotNull(acmrealm);
		adminuser = ConfigProperties.getAccessControlManagerAdminAPIUser(AccessControlManagerIT.class);
		Assert.assertNotNull(adminuser);
		adminpass = ConfigProperties.getAccessControlManagerAdminAPIPass(AccessControlManagerIT.class);
		Assert.assertNotNull(adminpass);
		adminclient = ConfigProperties.getAccessControlManagerAdminAPIClient(AccessControlManagerIT.class);
		Assert.assertNotNull(adminclient);

		AccessControlManager.logger.setLevel(logginglevel);

		System.out.println("Logging Level:" + logginglevel);
		System.out.println("Logging verbosity:" + loggingverbosity);
		System.out.println("ACM URL:" + acmurl);
		System.out.println("ACM Realm:" + acmrealm);
		System.out.println("ACM Admin API User:" + adminuser);
		System.out.println("ACM Admin API Password:" + adminpass);
		System.out.println("ACM Admin API Client ID:" + adminclient);

		// Check connectivity to Access Control Manager in order to fail early and have
		// a cleaner output for the user running the integration tests
		final URI uri = new URI(acmurl);
		int port = uri.getPort();
		if (-1 == port) {
			// port not in URI, we need to get the default from the protocol
			if (uri.getScheme().equalsIgnoreCase("HTTP")) {
				port = 80;
			} else if (uri.getScheme().equalsIgnoreCase("HTTPS")) {
				port = 443;
			} else {
				throw new ConfigurationException("Cannot determine server port from URI " + uri);
			}
		}
		if (!IntegrationTestsCommon.canConnect(uri.getHost(), port)) {
			throw new ConfigurationException("Cannot connect to Access Control Manager at " + uri + "! Is it running?");
		}

		// Create the user for integration testing
		Assert.assertTrue(AccessControlManager.createEcoAccountUser(TestUser.Username, TestUser.Password,
				TestUser.Email, TestUser.Firstname, TestUser.Lastname, "en"));
		final String useruuid = AccessControlManager.getUsernameUUID(TestUser.Username);
		Assert.assertNotNull(useruuid);
		Assert.assertTrue(AccessControlManager.sendVerificationEmail(useruuid));
		// Verify manually for integration testing
		Assert.assertTrue(AccessControlManager.setUserEmailVerified(useruuid, true));
	}

	@After
	public void AfterTests() {
		System.out.println("Clean up tests...");

		// Remove the user from the Access Control Manager
		Assert.assertTrue(
				AccessControlManager.removeEcoAccountUser(AccessControlManager.getUsernameUUID(TestUser.Username)));

		// Do not leave stale session of Admin REST Client
		AccessControlManager.closeAdminAPIClient();
	}

	public static void main(String[] args) {
		final AccessControlManagerIT it = new AccessControlManagerIT();

		try {
			it.BeforeTests();
			it.itLoginLogout();
			it.itRefreshToken();
			it.itAdminRESTAPI();
			it.AfterTests();
		} catch (Exception e) {
			System.err.println(e.getMessage());
		}
	}
}
