/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.test;

import java.net.Socket;
import java.net.URI;
import java.net.URISyntaxException;

import javax.naming.ConfigurationException;

import gr.iccs.isense.circ4life.rest.config.ConfigProperties;

public class IntegrationTestsCommon {
	/**
	 * Test user details for integration testing
	 */
	public static final class TestUser {
		/**
		 * Sample UUID for testing Eco-account records without creating user in the
		 * Access Control Manager.
		 */
		public static final String SAMPLEUUID = "db624777-c948-4544-938f-0c5749396fce";

		/**
		 * Access Control Manager user name for integration testing.
		 */
		public static final String Username = "IntegrationTester";

		/**
		 * Access Control Manager user password for integration testing.
		 */
		public static final String Password = "iccs.isense.circ4life.it.2019";

		/**
		 * Access Control Manager user email for integration testing.
		 */
		public static final String Email = "integration_tester@iccs.gr";

		/**
		 * Access Control Manager user first name for integration testing.
		 */
		public static final String Firstname = "Integration";

		/**
		 * Access Control Manager user last name for integration testing.
		 */
		public static final String Lastname = "Tester";

		// Eco-account details
		public static final String birthday = "2010-05-01";
		public static final String country = "GR";
		public static final String postalcode = "12345678";
		public static final String street = "Foo Street 40";
		public static final String city = "Foo city";
		public static final String region = "Foo region";
		public static final String telephone = "1234567890";
		public static final String image = "https://example.com/profile.png";
		public static final String url = "https://example.com/profile.htm";
		public static final String gender = "Male";
		public static final String organizationname = "Organization";
		public static final String organizationURL = "https://example.com";
		public static final long recyclebinuserid = 16004000001L;
	}

	/**
	 * Start local application server instance for integration testing.
	 * 
	 * @param DISPLAYNAME              The Web Application context display name
	 *                                 (web.xml display-name tag equivalent).
	 * @param APPCLASSNAME             The full application class name, e.g.
	 *                                 gr.iccs.isense.circ4life.rest.EcoCreditCalculatorModule.
	 * @param LOCALHOST_URI            The locahost URI to listen to with pattern:
	 *                                 http://127.0.0.1:PORT/CONTEXTPATH/RESOURCEPATH.
	 * @param ACCESSCONTROLSKIPPATTERN URI path pattern to skip access control for.
	 */
	public static void startLocalAppServer(final String DISPLAYNAME, final String APPCLASSNAME,
			final String LOCALHOST_URI, final String ACCESSCONTROLSKIPPATTERN) {
		final String args[] = { DISPLAYNAME, APPCLASSNAME, LOCALHOST_URI, ACCESSCONTROLSKIPPATTERN };
		new Thread(new Runnable() {
			public void run() {
				ApplicationServer.main(args);
			}
		}).start();
	}

	/**
	 * @return <code>true</code> if the local application server instance has
	 *         started, <code>false</code> otherwise.
	 */
	public static boolean isLocalAppServerReady() {
		return (null != ApplicationServer.ServerInstance) && ApplicationServer.ServerInstance.isStarted();
	}

	/**
	 * Stop the local application server.
	 */
	public static void stopLocalAppServer() {
		while (isLocalAppServerReady()) {
			ApplicationServer.ServerInstance.shutdownNow();
		}
	}

	/**
	 * Get the module directory of the specified class.
	 * 
	 * @param <T>   the type of the class.
	 * @param clazz the class.
	 * @return the module directory fo the class.
	 */
	public static <T> String getModuleDir(final Class<T> clazz) {
		return clazz.getClassLoader().getResource(".").getPath();
	}

	/**
	 * Check if we can make a TCP connection to the specified host and port.
	 * 
	 * @param host     the hostname or IP address.
	 * @param tcp_port the TCP port to connect to.
	 * @return <code>true</code> if a TCP connection is successful,
	 *         <code>false</code> otherwise.
	 */
	public static boolean canConnect(final String host, final int tcp_port) {
		try (final Socket sock = new Socket(host, tcp_port)) {
			sock.close();

			// All OK
			return true;
		} catch (Exception e) {
			return false;
		}
	}

	/**
	 * Check if the configured Access Control Manager is accessible.
	 * 
	 * @throws ConfigurationException if the current configuration is not valid.
	 * @throws URISyntaxException     if the conffigured Access Control Manager URL
	 *                                is malformed.
	 */
	public static void checkAccessControlManagerConnection() throws ConfigurationException, URISyntaxException {
		final String url = ConfigProperties.getAccessControlManagerURL(IntegrationTestsCommon.class);
		if (null == url) {
			throw new ConfigurationException(
					"Configuration is missing required property for Access Control Manager URL!");
		}

		final URI acm_uri = new URI(url);
		final String host = acm_uri.getHost();
		int port = acm_uri.getPort();
		if (-1 == port) {
			// Post not explicitly defined in URL, try to detect it from scheme.
			final String scheme = acm_uri.getScheme();
			if (scheme.equalsIgnoreCase("http")) {
				port = 80;
			} else if (scheme.equalsIgnoreCase("https")) {
				port = 443;
			} else {
				throw new ConfigurationException("Cannot detect port number from URL " + url);
			}
		}

		if (!canConnect(host, port)) {
			throw new ConfigurationException(
					"Cannot connect to Access Control Manager server at " + host + ":" + port + "! Is it running?");
		}
	}

	/**
	 * Check if the configured database server is accessible.
	 * 
	 * @throws ConfigurationException if the current configuration is not valid.
	 */
	public static void checkDatabaseServerConnection() throws ConfigurationException {
		final String dbhost = ConfigProperties.getDatabaseHost(IntegrationTestsCommon.class);
		if (null == dbhost) {
			throw new ConfigurationException("Configuration is missing required property for database host!");
		}

		final Integer dbport = ConfigProperties.getDatabasePort(IntegrationTestsCommon.class);
		if (null == dbport) {
			throw new ConfigurationException("Configuration is missing required property for database port!");
		}

		if (!canConnect(dbhost, dbport)) {
			throw new ConfigurationException(
					"Cannot connect to database server at " + dbhost + ":" + dbport + "! Is it running?");
		}
	}

	/**
	 * Get the integration testing database import directory.
	 * 
	 * @param <T>   the type of the class.
	 * @param clazz the class.
	 * @return the path to the integration testing database import directory.
	 */
	public static <T> String getDatabaseImportDirectory(Class<T> clazz) {
		return getModuleDir(clazz) + "../../../DataAccessObject/src/it/resources/databases";
	}
}
