/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.integration.test;

import java.util.ArrayList;
import java.util.Collection;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import gr.iccs.isense.circ4life.EcoCreditCalculator;
import gr.iccs.isense.circ4life.FoodParameters;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductType;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO.WEEEsLifespanDistributionParameters;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO.WEEEsWeights;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO.WasteType;
import gr.iccs.isense.circ4life.datamodels.MaterialFactors;
import gr.iccs.isense.circ4life.datamodels.WEEEsEndOfLifeState;
import gr.iccs.isense.circ4life.test.TestsCommon;
import gr.iccs.isense.circ4life.test.TestsCommon.EcoCreditCalculatorBioWasteTest;

/**
 * Integration tests for the Data Access Object (DAO) for the Eco-credits
 * calculator.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EcoCreditCalculatorDAOIT extends PersistenceLayerDAOIT {
	private static EcoCreditCalculatorDAO dao = null;
	private static EcoCreditCalculator calculator = null;

	private final class ProductTypeTest {
		private final ProductType type;
		private final WasteType wastetype;

		public ProductTypeTest(ProductType type, WasteType wastetype) {
			this.type = type;
			this.wastetype = wastetype;
		}
	}

	private final class EcoCreditsParametersTest {
		private final Integer hskey;
		private final Double lifespanAlpha;
		private final Double lifespanBeta;
		private final Double WA;
		private final Double WB;
		private final Double WC;

		public EcoCreditsParametersTest(Integer hskey, Double lifespanAlpha, Double lifespanBeta, Double WA, Double WB,
				Double WC) {
			this.hskey = hskey;
			this.lifespanAlpha = lifespanAlpha;
			this.lifespanBeta = lifespanBeta;
			this.WA = WA;
			this.WB = WB;
			this.WC = WC;
		}
	}

	@Before
	public void createStaticObjects() {
		if (null == dao) {
			dao = new EcoCreditCalculatorDAO();
		}

		if (null == calculator) {
			calculator = new EcoCreditCalculator();
		}
	}

	@Test
	public void itProductTypes() {
		Assert.assertNotNull(dao);

		final ProductTypeTest[] valid = { new ProductTypeTest(new ProductType(23, null, null), WasteType.FOOD),
				new ProductTypeTest(new ProductType(847130, 303, null), WasteType.WEEES),
				new ProductTypeTest(new ProductType(950410, 702, null), WasteType.WEEES) };

		for (final ProductTypeTest t : valid) {
			Assert.assertTrue(t.wastetype == dao.getProductWasteType(t.type.HSkey));
			if (null != t.type.UNUkey) {
				Assert.assertTrue(t.type.HSkey.equals(dao.getHSKeyFromUNUkey(t.type.UNUkey)));
				Assert.assertTrue(t.type.UNUkey.equals(dao.getUNUKeyFromHSkey(t.type.HSkey)));
			} else {
				Assert.assertNull(dao.getHSKeyFromUNUkey(t.type.UNUkey));
				Assert.assertNull(dao.getUNUKeyFromHSkey(t.type.HSkey));
			}
		}
	}

	@Test
	public void itProductTypesEcoCreditsParameters() {
		Assert.assertNotNull(dao);

		final EcoCreditsParametersTest[] defaults = { // The expected default values
				new EcoCreditsParametersTest(23, null, null, 0.0, 0.0, 0.0), // Food waste
				new EcoCreditsParametersTest(847130, 1.5, 5.2, 0.00006224, 4.706, 15.0), // Laptops/tablets
				new EcoCreditsParametersTest(950410, 1.5, 5.2, 0.00006224, 4.706, 15.0) // Game consoles
		};

		for (final EcoCreditsParametersTest t : defaults) {
			if (null != t.lifespanAlpha && null != t.lifespanBeta) {
				final Integer unukey = dao.getUNUKeyFromHSkey(t.hskey);
				Assert.assertNotNull(unukey);
				final WEEEsLifespanDistributionParameters lfsp = dao.getWEEEsLifespanDistributionParameters(unukey);

				Assert.assertNotNull(lfsp);
				Assert.assertTrue(t.lifespanAlpha.doubleValue() == lfsp.A.doubleValue()
						&& t.lifespanBeta.doubleValue() == lfsp.B.doubleValue());
				final WEEEsWeights weights = dao.getWEEEsWeights(unukey);
				Assert.assertNotNull(weights);
				Assert.assertTrue(
						t.WA.doubleValue() == weights.A.doubleValue() && t.WB.doubleValue() == weights.B.doubleValue()
								&& t.WC.doubleValue() == weights.C.doubleValue());

				final Collection<MaterialFactors> working = dao.getMaterialFactors(unukey, WEEEsEndOfLifeState.WORKING);
				final Collection<MaterialFactors> broken = dao.getMaterialFactors(unukey, WEEEsEndOfLifeState.BROKEN);

				// Sort now the golden standard
				final ArrayList<MaterialFactors> expectedworking = new ArrayList<MaterialFactors>();
				for (final MaterialFactors e : TestsCommon.working_tablet_material_factors) {
					expectedworking.add(e);
				}
				final ArrayList<MaterialFactors> expectedbroken = new ArrayList<MaterialFactors>();
				for (final MaterialFactors e : TestsCommon.broken_tablet_material_factors) {
					expectedbroken.add(e);
				}

				// Compare the collections
				Assert.assertTrue(expectedworking.size() == working.size());
				Assert.assertTrue(expectedbroken.size() == broken.size());
				for (final MaterialFactors e : expectedworking) {
					Assert.assertTrue(working.contains(e));
				}
				for (final MaterialFactors e : expectedbroken) {
					Assert.assertTrue(broken.contains(e));
				}
			} else {
				final Integer unukey = dao.getUNUKeyFromHSkey(t.hskey);
				Assert.assertNull(unukey);
				Assert.assertNull(dao.getWEEEsLifespanDistributionParameters(unukey));
				Assert.assertNull(dao.getWEEEsWeights(unukey));
				Assert.assertNull(dao.getMaterialFactors(unukey, WEEEsEndOfLifeState.WORKING));
				Assert.assertNull(dao.getMaterialFactors(unukey, WEEEsEndOfLifeState.BROKEN));
			}
		}
	}

	@Test
	public void itBioWasteEcoCreditsFromMass() {
		Assert.assertNotNull(dao);
		Assert.assertNotNull(calculator);

		final EcoCreditCalculatorBioWasteTest tests[] = { // Eco-credits corresponding to waste mass
				new EcoCreditCalculatorBioWasteTest(1.0, 1000), // 1 Kg
				new EcoCreditCalculatorBioWasteTest(2.01, 2000), // 2 Kg
				new EcoCreditCalculatorBioWasteTest(3.02, 3000), // 3 Kg
				new EcoCreditCalculatorBioWasteTest(4.03, 4000), // 4 Kg
				new EcoCreditCalculatorBioWasteTest(7.55, 5000), // 5 Kg
				new EcoCreditCalculatorBioWasteTest(9.06, 6000), // 6 Kg
				new EcoCreditCalculatorBioWasteTest(10.57, 7000), // 7 Kg
				new EcoCreditCalculatorBioWasteTest(12.08, 8000), // 8 Kg
				new EcoCreditCalculatorBioWasteTest(17.99, 9000), // 9 Kg
				new EcoCreditCalculatorBioWasteTest(19.99, 10000), // 10 Kg
				new EcoCreditCalculatorBioWasteTest(21.99, 11000), // 11 Kg
				new EcoCreditCalculatorBioWasteTest(23.99, 12000), // 12 Kg
				new EcoCreditCalculatorBioWasteTest(25.99, 13000), // 13 Kg
				new EcoCreditCalculatorBioWasteTest(27.99, 14000), // 14 Kg
				new EcoCreditCalculatorBioWasteTest(29.99, 15000), // 15 Kg
				new EcoCreditCalculatorBioWasteTest(31.99, 16000) // 16 Kg
		};

		for (final EcoCreditCalculatorBioWasteTest t : tests) {
			final FoodParameters params = dao.getFoodParameters(t.waste_mass);
			final double calculatedresult = calculator.calculate(params);
			Assert.assertEquals(t.expectedresult, calculatedresult, TestsCommon.fractional_max_accepted_delta);
		}
	}

	public static void main(String[] args) {
		final EcoCreditCalculatorDAOIT it = new EcoCreditCalculatorDAOIT();

		try {
			it.itInitializeDatabase();
			it.itProductTypes();
			it.itProductTypesEcoCreditsParameters();
			it.itBioWasteEcoCreditsFromMass();
		} catch (Exception e) {
			System.err.println(e.getMessage());
		}
		it.itCleanupDatabase();
	}
}
