/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.integration.test;

import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map.Entry;

import javax.naming.ConfigurationException;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import gr.iccs.isense.circ4life.dao.EcoAccountDAO;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerAdminDAO;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAO;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAOFactory;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;
import gr.iccs.isense.circ4life.test.IntegrationTestsCommon;
import gr.iccs.isense.circ4life.test.IntegrationTestsCommon.TestUser;

/**
 * Integration tests for the Data Access Object (DAO) interface for persistence
 * layer.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class PersistenceLayerDAOIT {
	protected static final String ImportDirectory = "src/it/resources/databases";
	protected static EcoAccountDAO ecoaccountdao = null;
	protected PersistenceLayerDAO dbdao = null;
	private PersistenceLayerAdminDAO admindbdao = null;
	private static String useruuid = null;

	@Before
	public void itInitializeDatabase()
			throws ConfigurationException, IllegalArgumentException, MalformedURLException, URISyntaxException {
		// Check that we can connect to the Access Control Manager server and the
		// database server in order to quit early and inform user running the
		// integration tests
		IntegrationTestsCommon.checkAccessControlManagerConnection();
		IntegrationTestsCommon.checkDatabaseServerConnection();

		if (null == dbdao) {
			dbdao = PersistenceLayerDAOFactory.getDefaultInstance();
		}
		Assert.assertNotNull(dbdao);

		if (null == admindbdao) {
			admindbdao = PersistenceLayerDAOFactory.getDefaultAdminInstance();
		}
		Assert.assertNotNull(admindbdao);

		// Import everything from the src/it/resources/databases directory
		// TODO: Start a fresh instance of a MongoDB server for integration testing that
		// does not have any of our application databases. The keepdata = true here
		// means that all imported data are imported without looking for existing
		// application databases.
		Assert.assertTrue(admindbdao.importDatabases(ImportDirectory, true));

		// Create the user with Eco-account
		if (null == ecoaccountdao) {
			ecoaccountdao = new EcoAccountDAO();
		}

		useruuid = ecoaccountdao.createEcoAccount(TestUser.Username, TestUser.Password, TestUser.Email,
				TestUser.Firstname, TestUser.Lastname, DataConversion.readDate(TestUser.birthday), TestUser.country,
				TestUser.postalcode, TestUser.street, TestUser.city, TestUser.region, TestUser.telephone,
				new URL(TestUser.image), new URL(TestUser.url), TestUser.gender, TestUser.organizationname,
				new URL(TestUser.organizationURL), "en");
		Assert.assertNotNull(useruuid);
		Assert.assertTrue(ecoaccountdao.updateEcoAccountWithRecycleBinUserID(useruuid, TestUser.recyclebinuserid));
	}

	@Test
	public void itCheckDatabases() {
		final String[] dbnames = { PersistenceLayerDAO.EcoAccounts.DBNAME,
				PersistenceLayerDAO.MasterProductData.DBNAME };

		for (final String dbname : dbnames) {
			System.out.println("Checking existence of database: " + dbname);
			Assert.assertTrue(dbdao.databaseExists(dbname));
		}
	}

	@Test
	public void itCheckCollections() {
		// TODO: find a way to import and check Live Product Data also
		final HashMap<String, String[]> dbnames2collections = new HashMap<String, String[]>();
		dbnames2collections.put(PersistenceLayerDAO.EcoAccounts.DBNAME,
				new String[] { PersistenceLayerDAO.EcoAccounts.USERSCOLLECTION });
		dbnames2collections.put(PersistenceLayerDAO.MasterProductData.DBNAME,
				new String[] { PersistenceLayerDAO.MasterProductData.PRODUCTSCOLLECTION,
						PersistenceLayerDAO.MasterProductData.PRODUCTTYPESCOLLECTION });

		for (final Entry<String, String[]> e : dbnames2collections.entrySet()) {
			for (final String collection : e.getValue()) {
				System.out.println("Checking existence of collection: " + e.getKey() + "." + collection);
				Assert.assertTrue(dbdao.collectionExists(e.getKey(), collection));
			}
		}
	}

	@After
	public void itCleanupDatabase() {
		final String uid = AccessControlManager.getUsernameUUID(TestUser.Username);
		Assert.assertNotNull(uid);
		Assert.assertTrue(uid.contentEquals(useruuid));
		Assert.assertTrue(ecoaccountdao.removeEcoAccount(uid));
		Assert.assertNull(ecoaccountdao.getEcoAccount(uid));

		// If we reach here, then a successful import is performed during this
		// integration test and we should have all application databases setup properly.
		if (null != admindbdao) {
			Assert.assertTrue(admindbdao.dropApplicationDatabases(true));
		}

		// Do not leave stale session of Admin REST Client
		AccessControlManager.closeAdminAPIClient();
	}

	public static void main(String[] args) {
		final PersistenceLayerDAOIT it = new PersistenceLayerDAOIT();

		try {
			it.itInitializeDatabase();
			it.itCheckDatabases();
			it.itCheckCollections();
		} catch (Exception e) {
			System.err.println(e.getMessage());
		}

		it.itCleanupDatabase();
	}
}
