/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.integration.test;

import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;

import javax.json.Json;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import gr.iccs.isense.circ4life.dao.CommonDAO.ProductID;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductType;
import gr.iccs.isense.circ4life.dao.ProductMasterDataDAO;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAO;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAO.MasterProductData;
import gr.iccs.isense.circ4life.json.JacksonJSONSerialization;

/**
 * Integration tests for the Data Access Object (DAO) for product Master Data.
 */
public class ProductMasterDataDAOIT extends PersistenceLayerDAOIT {
	private final class ProductTest {
		public final ProductRecord expected;
		public final String productid;

		public ProductTest(ProductRecord expected, String productid) {
			this.expected = expected;
			this.productid = productid;
		}
	}

	private final class BarcodeTest {
		public final Long expectedGTIN;
		public final String expectedSerialNumber;
		public final String Barcode;

		public BarcodeTest(Long expectedGTIN, String expectedSerialNumber, String Barcode) {
			this.expectedGTIN = expectedGTIN;
			this.expectedSerialNumber = expectedSerialNumber;
			this.Barcode = Barcode;
		}
	}

	private static ProductMasterDataDAO dao = null;

	@Before
	public void createDAO() {
		if (null == dao) {
			dao = new ProductMasterDataDAO();
		}
	}

	@Test
	public void itProductTypes() throws IOException {
		Assert.assertNotNull(dao);

		// Check that we return exactly the same number of records as found in the
		// import directory.
		final File importfile = new File(ImportDirectory + File.separator + MasterProductData.DBNAME + File.separator
				+ MasterProductData.PRODUCTTYPESCOLLECTION + File.separator + "ProductTypes.en.json");
		final InputStream is = new FileInputStream(importfile);
		final String prodtypeslistcollection = Json.createReader(is).readObject()
				.getString(PersistenceLayerDAO.LOCALE_COLLECTION_PROPERTY);
		final File importdir = new File(
				ImportDirectory + File.separator + MasterProductData.DBNAME + File.separator + prodtypeslistcollection);
		is.close();
		Assert.assertTrue(importdir.isDirectory());
		final FilenameFilter jsonfilesfilter = new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				return dir.getAbsolutePath().equals(importdir.getAbsolutePath()) && name.endsWith(".json");
			}
		};
		final int prodtypesimportcount = importdir.listFiles(jsonfilesfilter).length;

		{
			// Known and default locale
			final Collection<ProductType> prodtypes = dao.getSupportedProductTypes("en");
			Assert.assertNotNull(prodtypes);
			Assert.assertTrue(prodtypesimportcount == prodtypes.size());
		}

		{
			// Unknown locale, should return the default of 'en'
			final Collection<ProductType> prodtypes = dao.getSupportedProductTypes("foobar");
			Assert.assertNotNull(prodtypes);
			Assert.assertTrue(prodtypesimportcount == prodtypes.size());
		}

		{
			// Locale found but missing product types list collection
			final Collection<ProductType> prodtypes = dao.getSupportedProductTypes("missing");
			Assert.assertNotNull(prodtypes);
			Assert.assertTrue(0 == prodtypes.size());
		}

		try {
			// Locale found but missing product types list collection
			dbdao.getProductTypes("missing");
			// Should not reach here
			Assert.assertTrue(false);
		} catch (Exception e) {
			// Should reach here
			Assert.assertTrue(true);
		}

		final String productid = "12345678901234";
		{
			// Known product ID
			final ProductType prodtype = dao.getProductType(productid);
			Assert.assertNotNull(prodtype);
		}

		{
			// Unknown product ID
			final ProductType prodtype = dao.getProductType("FooBar");
			Assert.assertNull(prodtype);
		}

		{
			final Collection<ProductType> prodtypes = dao.getSupportedProductTypes("en");
			for (final ProductType productType : prodtypes) {
				Assert.assertNotNull(dbdao.getProductTypeFromHSkey(productType.HSkey));
				if (null != productType.UNUkey) {
					Assert.assertNotNull(dbdao.getProductTypeFromUNUkey(productType.UNUkey));
				}
			}
		}
	}

	@Test
	public void itProductRetrieval() throws IOException {
		Assert.assertNotNull(dao);

		// List of test to perform, filled dynamically below
		final ArrayList<ProductTest> producttests = new ArrayList<ProductTest>();

		// Read the JSON product files into the test structures
		final File importfile = new File(ImportDirectory + File.separator + MasterProductData.DBNAME + File.separator
				+ MasterProductData.PRODUCTSCOLLECTION + File.separator + "Products.en.json");
		final InputStream is = new FileInputStream(importfile);
		final String productslistcollection = Json.createReader(is).readObject()
				.getString(PersistenceLayerDAO.LOCALE_COLLECTION_PROPERTY);
		final File importdir = new File(
				ImportDirectory + File.separator + MasterProductData.DBNAME + File.separator + productslistcollection);
		is.close();
		Assert.assertTrue(importdir.isDirectory());
		final FilenameFilter jsonfilesfilter = new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				return dir.getAbsolutePath().equals(importdir.getAbsolutePath()) && name.endsWith(".json");
			}
		};
		for (final File prodjsonfile : importdir.listFiles(jsonfilesfilter)) {
			final InputStream prodis = new FileInputStream(prodjsonfile);

			final ProductRecord product = JacksonJSONSerialization
					.StringtoJSON(Json.createReader(prodis).readObject().toString(), ProductRecord.class);

			producttests.add(new ProductTest(product, product.getMasterData().getProductData().getGtin14()));

			prodis.close();
		}

		// Add some malformed data to check for null
		producttests.add(new ProductTest(null, "Foobar"));

		// Perform the testing
		for (final ProductTest test : producttests) {
			{
				// Known product ID and default locale
				final ProductRecord product = dao.getProductRecord(test.productid, "en");
				if (null != test.expected) {
					Assert.assertTrue(test.expected.getMasterData().getProductData()
							.equals(product.getMasterData().getProductData()));
				} else {
					Assert.assertNull(product);
				}
			}

			{
				// Known product ID and unknown locale, should return the default of 'en'
				final ProductRecord product = dao.getProductRecord(test.productid, "foobar");
				if (null != test.expected) {
					Assert.assertTrue(test.expected.getMasterData().getProductData()
							.equals(product.getMasterData().getProductData()));
				} else {
					Assert.assertNull(product);
				}
			}

			{
				// Known product ID and Locale found but missing products list collection
				final ProductRecord product = dao.getProductRecord(test.productid, "missing");
				Assert.assertNull(product);
			}
		}
	}

	@Test
	public void itProductBarcodes() {
		Assert.assertNotNull(dao);

		// Create tests for barcode conversions to Product IDs.
		final BarcodeTest[] validbarcodetests = {
				new BarcodeTest(80614141123458L, "6789", "(01)80614141123458(21)6789"),
				new BarcodeTest(4047111123453L, "123456789", "(01)04047111123453(21)123456789") };

		for (final BarcodeTest test : validbarcodetests) {
			final ProductID prodid = dao.getProductIDFromBarcode(test.Barcode);
			Assert.assertNotNull(prodid);
			Assert.assertEquals(test.expectedGTIN, prodid.GTIN14);
			Assert.assertEquals(test.expectedSerialNumber, prodid.SerialNumber);
		}

		final BarcodeTest[] invalidbarcodetests = { new BarcodeTest(0L, "TODO", "TODO") };

		for (final BarcodeTest test : invalidbarcodetests) {
			final ProductID prodid = dao.getProductIDFromBarcode(test.Barcode);
			Assert.assertNull(prodid);
		}
	}

	public static void main(String[] args) {
		final ProductMasterDataDAOIT it = new ProductMasterDataDAOIT();

		try {
			it.itInitializeDatabase();
			it.itProductTypes();
			it.itProductRetrieval();
			it.itProductBarcodes();
		} catch (Exception e) {
			System.err.println(e.getMessage());
		}
		it.itCleanupDatabase();
	}
}
