/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.dao;

import java.net.MalformedURLException;
import java.util.Arrays;
import java.util.Date;
import java.util.regex.Pattern;

import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;

import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAO;
import gr.iccs.isense.circ4life.dao.schema.Address;
import gr.iccs.isense.circ4life.dao.schema.Affiliation;
import gr.iccs.isense.circ4life.dao.schema.Audience;
import gr.iccs.isense.circ4life.dao.schema.Brand;
import gr.iccs.isense.circ4life.dao.schema.CustomEnum;
import gr.iccs.isense.circ4life.dao.schema.EcoAccount;
import gr.iccs.isense.circ4life.dao.schema.EcoAccounting;
import gr.iccs.isense.circ4life.dao.schema.EcoBalance;
import gr.iccs.isense.circ4life.dao.schema.EcoCredit;
import gr.iccs.isense.circ4life.dao.schema.EcoCredits;
import gr.iccs.isense.circ4life.dao.schema.EcoDebit;
import gr.iccs.isense.circ4life.dao.schema.EcoPoints;
import gr.iccs.isense.circ4life.dao.schema.EoLState;
import gr.iccs.isense.circ4life.dao.schema.HSkey;
import gr.iccs.isense.circ4life.dao.schema.LifeCycle;
import gr.iccs.isense.circ4life.dao.schema.Lifetime;
import gr.iccs.isense.circ4life.dao.schema.Manufacturer;
import gr.iccs.isense.circ4life.dao.schema.MasterData;
import gr.iccs.isense.circ4life.dao.schema.Months;
import gr.iccs.isense.circ4life.dao.schema.OrganicChemicalExergy;
import gr.iccs.isense.circ4life.dao.schema.OrganicMass;
import gr.iccs.isense.circ4life.dao.schema.Parameters;
import gr.iccs.isense.circ4life.dao.schema.Product;
import gr.iccs.isense.circ4life.dao.schema.Purchased;
import gr.iccs.isense.circ4life.dao.schema.Recyclability;
import gr.iccs.isense.circ4life.dao.schema.Recycled;
import gr.iccs.isense.circ4life.dao.schema.Resources;
import gr.iccs.isense.circ4life.dao.schema.UNUkey;
import gr.iccs.isense.circ4life.dao.schema.WeightOfEoLState;
import gr.iccs.isense.circ4life.dao.schema.WeightOfExergy;
import gr.iccs.isense.circ4life.dao.schema.WeightOfLifetime;
import gr.iccs.isense.circ4life.dao.schema.Years;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;
import gr.iccs.isense.circ4life.validation.EndUserRegistrationFormValidator;

/**
 * Data Access Object (DAO) common data models and methods.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class CommonDAO {
	/**
	 * Encapsulation of all used Schema.org types. -
	 */
	private static final class SchemaOrgTypes {
		public final static String Product = "Product";
		public final static String IndividualProduct = "IndividualProduct";
		public final static String Brand = "Brand";
		public final static String Organization = "Organization";
	}

	/**
	 * Encapsulation of product type standardized codes.
	 */
	public static class ProductType {
		/**
		 * Product Harmonized System 6 digit code as described in
		 * http://www.wcoomd.org/en/topics/nomenclature/overview/what-is-the-harmonized-system.aspx.
		 * Required for all products.
		 */
		@JsonProperty(PersistenceLayerDAO.MasterProductData.PRODUCTTYPESHSKEY)
		public final Integer HSkey;

		/**
		 * Electronic product UNU-KEY code as described in
		 * https://unu.edu/projects/e-waste-quantification.html#outputs. Applies only in
		 * electronics and related products.
		 */
		@JsonProperty(PersistenceLayerDAO.MasterProductData.PRODUCTTYPESUNUKEY)
		public final Integer UNUkey;

		/**
		 * Text description.
		 */
		@JsonProperty(PersistenceLayerDAO.MasterProductData.PRODUCTTYPESDESCR)
		public final String Description;

		/**
		 * Constructor.
		 * 
		 * @param hskey       Product Harmonized System 6 digit code.
		 * @param unukey      Electronic product UNU-KEY code
		 * @param description Text description.
		 * @throws IllegalArgumentException if the hskey is invalid.
		 */
		public ProductType(Integer hskey, Integer unukey, String description) throws IllegalArgumentException {
			if (null == hskey) {
				throw new IllegalArgumentException("Invalid HSkey value: " + hskey);
			}
			HSkey = hskey;
			UNUkey = unukey;
			Description = description;
		}

		/**
		 * Constructor.
		 * 
		 * @param hskey Product Harmonized System 6 digit code.
		 */
		public ProductType(Integer hskey) {
			this(hskey, null, null);
		}

		@Override
		public String toString() {
			return "[" + PersistenceLayerDAO.MasterProductData.PRODUCTTYPESHSKEY + " = " + HSkey + ", "
					+ PersistenceLayerDAO.MasterProductData.PRODUCTTYPESUNUKEY + " = " + UNUkey + "]";
		}
	}

	/**
	 * Encapsulation of product ID: GTIN14 + Optional Serial Number.
	 */
	public static class ProductID {
		/**
		 * Product GTIN code.
		 */
		public Long GTIN14;

		/**
		 * Product serial number.
		 */
		public String SerialNumber;

		/**
		 * Constructor for JSON serialization.
		 */
		public ProductID() {
			GTIN14 = null;
			SerialNumber = null;
		}

		/**
		 * Constructor.
		 * 
		 * @param gtin14       Product GTIN code.
		 * @param serialnumber Product serial number.
		 */
		public ProductID(Long gtin14, String serialnumber) {
			GTIN14 = gtin14;
			SerialNumber = serialnumber;
		}

		/**
		 * Constructor.
		 * 
		 * @param gtin14 Product GTIN code.
		 */
		public ProductID(Long gtin14) {
			this(gtin14, null);
		}

		/**
		 * Constructor.
		 * 
		 * @param productid Product ID string. Format used is GTIN14.SerialNumber or
		 *                  GTIN14.examples: 04047111123453.012345678901,
		 *                  04047111123453.
		 * @throws IllegalArgumentException if the productid has malformed format.
		 * @throws NumberFormatException    if the GTIN14 part of the productid cannot
		 *                                  be parse to a {@link Long}.
		 */
		public ProductID(final String productid) throws IllegalArgumentException, NumberFormatException {
			if (null == productid || productid.length() < 8
					|| !Pattern.matches("\\d{8,14}|\\d{8,14}\\..+", productid)) {
				throw new IllegalArgumentException("Invalid product ID string: " + productid
						+ ", expected GTIN.SerialNumber or GTIN, where GTIN can be 8 to 14 digits long");
			}

			if (Pattern.matches("\\d{8,14}\\..+", productid)) {
				final String tokens[] = productid.split("\\.");
				final int gtindigits = tokens[0].length();
				if (9 <= gtindigits && 11 >= gtindigits) {
					throw new IllegalArgumentException(
							"Unsupported GTIN format, expected GTIN-8,12,13,14, got " + gtindigits + " digits");
				}
				GTIN14 = Long.valueOf(tokens[0]);
				SerialNumber = tokens[1];
			} else {
				final int gtindigits = productid.length();
				if (9 <= gtindigits && 11 >= gtindigits) {
					throw new IllegalArgumentException(
							"Unsupported GTIN format, expected GTIN-8,12,13,14, got " + gtindigits + " digits");
				}
				GTIN14 = Long.valueOf(productid);
				SerialNumber = null;
			}
		}

		@Override
		public String toString() {
			final String gtin = DataConversion.formatGTIN14(GTIN14);

			return null == SerialNumber ? gtin : gtin + "." + SerialNumber;
		}

		/**
		 * @return the GTIN14 code.
		 */
		@JsonProperty("GTIN14")
		public Long getGTIN14() {
			return GTIN14;
		}

		/**
		 * Set the GTIN14 code.
		 * 
		 * @param gtin14 the code to set.
		 */
		public void setGTIN14(Long gtin14) {
			GTIN14 = gtin14;
		}

		/**
		 * @return the serial number.
		 */
		@JsonProperty("SerialNumber")
		public String getSerialNumber() {
			return SerialNumber;
		}

		/**
		 * Set the serial number.
		 * 
		 * @param serialnumber the serial number to set.
		 */
		public void setSerialNumber(String serialnumber) {
			SerialNumber = serialnumber;
		}
	}

	/**
	 * Encapsulation of product Master Data record.
	 */
	public static class ProductRecord extends gr.iccs.isense.circ4life.dao.schema.ProductMasterData {
		/**
		 * Required for the class serialization support.
		 */
		private static final long serialVersionUID = 4506229815528855899L;

		/**
		 * Constructor, required for the class serialization support.
		 */
		public ProductRecord() {
			super();
		}

		/**
		 * Constructor.
		 *
		 * @param type      product type code keys.
		 * @param gtin14    GTIN14 code.
		 * @param productID Unique Identification String for this product model.
		 */
		public ProductRecord(final ProductType type, final String gtin14, final String productID) {
			super();
			if (null == type) {
				throw new RuntimeException(
						ProductRecord.class.getName() + ": construction with null " + ProductType.class.getName());
			}
			withMasterData(initMasterData(type, gtin14, productID)).withEcoAccounting(initEcoAccounting());
		}

		/**
		 * @return {@link ProductType} from Recources information.
		 */
		@JsonIgnore
		public ProductType getProductType() {
			final Resources resources = getMasterData().getLifeCycle().getResources();
			final Integer hskey = resources.getHSkey().getValue(),
					unukey = (null == resources.getUNUkey()) ? null : resources.getUNUkey().getValue();
			final ProductType productType = new ProductType(hskey, unukey, null);

			return productType;
		}

		/**
		 * Create the absolute minimal {@link MasterData} object, initialized with the
		 * required properties.
		 * 
		 * @param type      product type code keys.
		 * @param gtin14    GTIN14 code.
		 * @param productID Unique Identification String for this product model.
		 * @return the created and initialized object.
		 */
		private MasterData initMasterData(final ProductType type, final String gtin14, final String productID) {
			final MasterData masterData = new MasterData();

			masterData.withProductData(initProductData(gtin14, productID));
			masterData.withLifeCycle(initLifeCycle(type));

			return masterData;
		}

		/**
		 * Create the absolute minimal {@link EcoAccounting} object, initialized with
		 * the required properties.
		 * 
		 * @return the created and initialized object.
		 */
		private EcoAccounting initEcoAccounting() {
			final EcoAccounting ecoAccounting = new EcoAccounting();

			ecoAccounting.withEcoPoints(new EcoPoints());

			return ecoAccounting;
		}

		/**
		 * Create the absolute minimal {@link Product} object, initialized with the
		 * required properties.
		 * 
		 * @param gtin14    GTIN14 code.
		 * @param productID Unique Identification String for this product model.
		 * @return the created and initialized object.
		 */
		private Product initProductData(final String gtin14, final String productID) {
			final Product productData = new Product();

			productData.setType(SchemaOrgTypes.Product);
			productData.withGtin14(gtin14).withProductID(productID);

			return productData;
		}

		/**
		 * Create the absolute minimal {@link LifeCycle} object, initialized with the
		 * required properties.
		 * 
		 * @param type product type.
		 * @return the created and initialized object.
		 */
		public LifeCycle initLifeCycle(final ProductType type) {
			final LifeCycle lifeCycle = new LifeCycle();

			lifeCycle.withResources(initResources(type));

			return lifeCycle;
		}

		/**
		 * Create the absolute minimal {@link Resources} object, initialized with the
		 * required properties.
		 * 
		 * @param type product type code keys.
		 * @return the created and initialized object.
		 */
		private Resources initResources(final ProductType type) {
			final Resources resources = new Resources();

			resources.withHSkey(new HSkey().withValue(type.HSkey)).withUNUkey(new UNUkey().withValue(type.UNUkey))
					.withRecyclability(initRecyclability());

			return resources;
		}

		/**
		 * Create the absolute minimal {@link Recyclability} object, initialized with
		 * the required properties.
		 * 
		 * @return the created and initialized object.
		 */
		public static Recyclability initRecyclability() {
			final Recyclability recyclability = new Recyclability();

			recyclability.withLifetime(new Lifetime().withYears(new Years()).withMonths(new Months()))
					.withParameters(initParameters()).withEoLState(initEoLState());

			return recyclability;
		}

		private static Parameters initParameters() {
			final Parameters parameters = new Parameters();

			parameters.withOrganicChemicalExergy(new OrganicChemicalExergy());
			parameters.withOrganicMass(new OrganicMass());
			parameters.withWeightOfEoLState(new WeightOfEoLState());
			parameters.withWeightOfExergy(new WeightOfExergy());
			parameters.withWeightOfLifetime(new WeightOfLifetime());

			return parameters;
		}

		private static EoLState initEoLState() {
			final EoLState eolstate = new EoLState();

			eolstate.getCustomEnum().add(new CustomEnum().withName("Working").withValue(1.0));
			eolstate.getCustomEnum().add(new CustomEnum().withName("Re-usable").withValue(0.5));
			eolstate.getCustomEnum().add(new CustomEnum().withName("Broken").withValue(0.0));

			return eolstate;
		}
	}

	/**
	 * Encapsulation of product Master Data record for an individual instance of a
	 * product.
	 */
	public static class IndividualProductData extends gr.iccs.isense.circ4life.dao.schema.IndividualProduct {
		/**
		 * Required for the class serialization support.
		 */
		private static final long serialVersionUID = 6356644774551533975L;

		/**
		 * Product type codes.
		 */
		private final ProductType product_type;

		/**
		 * Constructor, required for the class serialization support.
		 */
		public IndividualProductData() {
			super();
			setType(SchemaOrgTypes.IndividualProduct);
			product_type = new ProductType(-1);
		}

		/**
		 * Constructor.
		 * 
		 * @param type         product type code keys.
		 * @param gtin14       GTIN14 code.
		 * @param productID    Unique Identification String for this product model.
		 * @param serialNumber the individual product serial number code.
		 * @param name         the product name.
		 * @param purchaseDate the date of purchase.
		 */
		public IndividualProductData(final ProductType type, final String gtin14, final String productID,
				final String serialNumber, final String name, final Date purchaseDate) {
			super();
			setType(SchemaOrgTypes.IndividualProduct);
			setId(gtin14 + "." + serialNumber);
			if (null == type) {
				throw new RuntimeException(
						ProductRecord.class.getName() + ": construction with null " + ProductType.class.getName());
			}
			product_type = type;
			// Product
			withGtin14(gtin14).withProductID(productID).withName(name)
					.withBrand(new Brand().withType(Arrays.asList(SchemaOrgTypes.Brand)))
					.withDescription("Not available").withImage(null).withAudience(new Audience())
					.withManufacturer(new Manufacturer().withType(SchemaOrgTypes.Organization))
					.withCategory("Not available").withMaterial("Not available")
					// IndividualProduct
					.withSerialNumber(serialNumber).withPrice("Not available")
					.withPurchaseDate(DataConversion.formatDate(purchaseDate)).withRecycleDate("Not available")
					.withEcoCredits(new EcoCredits().withValue(0.0));
		}

		/**
		 * Constructor from superclass schema object.
		 * 
		 * @param type product type code keys.
		 * @param obj  the superclass object to copy propertied from.
		 * @throws Exception if the obj contains malformed properties.
		 */
		public IndividualProductData(final ProductType type,
				final gr.iccs.isense.circ4life.dao.schema.IndividualProduct obj) throws Exception {
			this(type, obj.getGtin14(), obj.getProductID(), obj.getSerialNumber(), obj.getName(),
					DataConversion.readDate(obj.getPurchaseDate()));

			// Product
			withBrand(obj.getBrand()).withDescription(obj.getDescription()).withImage(obj.getImage())
					.withAudience(obj.getAudience()).withManufacturer(obj.getManufacturer())
					.withCategory(obj.getCategory()).withMaterial(obj.getMaterial())
					// IndividualProduct
					.withPrice(obj.getPrice()).withPurchaseDate(obj.getPurchaseDate())
					.withRecycleDate(obj.getRecycleDate()).getEcoCredits().withValue((obj.getEcoCredits().getValue()));
		}

		/**
		 * Constructor from product master data record and serial number.
		 * 
		 * @param productRecord product master data record.
		 * @param serialNumber  the individual product serial number code.
		 * @param purchaseDate  the date of purchase.
		 */
		public IndividualProductData(final ProductRecord productRecord, final String serialNumber,
				final Date purchaseDate) {
			this(productRecord.getProductType(), productRecord.getMasterData().getProductData().getGtin14(),
					productRecord.getMasterData().getProductData().getProductID(), serialNumber,
					productRecord.getMasterData().getProductData().getName(), purchaseDate);
			final Product productdata = productRecord.getMasterData().getProductData();

			// Product
			withBrand(productdata.getBrand()).withDescription(productdata.getDescription())
					.withImage(productdata.getImage()).withAudience(productdata.getAudience())
					.withManufacturer(productdata.getManufacturer()).withCategory(productdata.getCategory())
					.withMaterial(productdata.getMaterial())
					// IndividualProduct
					.withPrice("Not Available").withPurchaseDate("Not available").withRecycleDate("Not available")
					.getEcoCredits().withValue(0.0);
		}

		/**
		 * Get the product type.
		 * 
		 * @return the product type.
		 */
		@JsonIgnore
		public ProductType getProductType() {
			return product_type;
		}

		/**
		 * Get the purchase date string as a {@link Date} object.
		 * 
		 * @return the {@link Date} object representing the purchase date.
		 * @throws IllegalArgumentException if
		 *                                  {@link gr.iccs.isense.circ4life.dao.schema.IndividualProduct#getPurchaseDate()}
		 *                                  returns a malformed {@link String} for date
		 *                                  representation.
		 */
		@JsonIgnore
		public Date getPurchaseDateObject() throws IllegalArgumentException {
			return DataConversion.readDate(this.getPurchaseDate());
		}
	}

	/**
	 * Encapsulation of Eco-account record.
	 */
	public static class EcoAccountRecord extends gr.iccs.isense.circ4life.dao.schema.EcoAccountRecord {
		/**
		 * Required for the class serialization support.
		 */
		private static final long serialVersionUID = -2666865553407765917L;

		/**
		 * Constructor, required for the class serialization support.
		 */
		public EcoAccountRecord() {
			super();
		}

		/**
		 * Constructor with minimum required parameters.
		 * 
		 * @param identifier      Universally Unique ID.
		 * @param email           e-mail.
		 * @param givenName       first name.
		 * @param familyName      last name.
		 * @param birthDate       birthday date.
		 * @param addressCountry  2-letter country code, e.g. GR.
		 * @param postalCode      Postal code number.
		 * @param streetAddress   Full street address, name, number, town.
		 * @param addressLocality City name.
		 * @throws IllegalArgumentException thrown if any of the required input
		 *                                  parameters is not valid.
		 */
		public EcoAccountRecord(final String identifier, final String email, final String givenName,
				final String familyName, final Date birthDate, final String addressCountry, final String postalCode,
				final String streetAddress, final String addressLocality) throws IllegalArgumentException {
			super();
			try {
				if (!AccessControlManager.isProperFormatUUID(identifier)) {
					throw new IllegalArgumentException("Malformed Access Control Manager user UUID: " + identifier);
				}

				// Check the rest of the properties and throw accordingly on validation error
				EndUserRegistrationFormValidator.validate(email, givenName, familyName, birthDate, addressCountry,
						postalCode, streetAddress, addressLocality);
			} catch (Exception e) {
				throw new IllegalArgumentException(
						"Invalid arguments during creation of Eco-account record object! Details: " + e.getMessage());
			}

			withEcoAccount(initEcoAccount(identifier, email, givenName, familyName, birthDate, addressCountry,
					postalCode, streetAddress, addressLocality)).withEcoBalance(initEcoBalance());
		}

		/**
		 * Constructor will all possible arguments.
		 * 
		 * @param identifier       Universally Unique ID. REQUIRED.
		 * @param email            e-mail. REQUIRED.
		 * @param givenName        first name. REQUIRED.
		 * @param familyName       last name. REQUIRED.
		 * @param birthDate        birthday date. REQUIRED.
		 * @param addressCountry   2-letter country code, e.g. GR. REQUIRED.
		 * @param postalCode       Postal code number. REQUIRED.
		 * @param streetAddress    Full street address, name, number, town. REQUIRED.
		 * @param addressLocality  City name. REQUIRED.
		 * @param addressRegion    Region name. OPTIONAL.
		 * @param telephone        Telephone number. OPTIONAL.
		 * @param imageurl         URL of profile image. OPTIONAL.
		 * @param userurl          URL of profile. OPTIONAL.
		 * @param gender           Gender, can be male, female or free text. OPTIONAL.
		 * @param organizationname Affiliated organization name. OPTIONAL.
		 * @param organizationurl  Affiliated organization URL. OPTIONAL.
		 * @throws IllegalArgumentException thrown if any of the required input
		 *                                  parameters is not valid.
		 */
		public EcoAccountRecord(final String identifier, final String email, final String givenName,
				final String familyName, final Date birthDate, final String addressCountry, final String postalCode,
				final String streetAddress, final String addressLocality, final String addressRegion,
				final String telephone, final String imageurl, final String userurl, final String gender,
				final String organizationname, final String organizationurl) throws IllegalArgumentException {
			this(identifier, email, givenName, familyName, birthDate, addressCountry, postalCode, streetAddress,
					addressLocality);

			// Check URLs
			try {
				if (null != imageurl) {
					EndUserRegistrationFormValidator.validateHTTPURL(imageurl);
				}

				if (null != userurl) {
					EndUserRegistrationFormValidator.validateHTTPURL(userurl);
				}

				if (null != organizationurl) {
					EndUserRegistrationFormValidator.validateHTTPURL(organizationurl);
				}
			} catch (MalformedURLException e) {
				throw new IllegalArgumentException("Malformed URL, details: " + e.getMessage());
			}

			// Fill in the rest of the provided parameters
			final Affiliation organization = new Affiliation().withName(organizationname).withUrl(organizationurl);
			this.getEcoAccount().withTelephone(telephone).withImage(imageurl).withUrl(userurl).withGender(gender)
					.withAffiliation(organization);
			this.getEcoAccount().getAddress().setAddressRegion(addressRegion);
		}

		/**
		 * Create the absolute minimal {@link EcoAccount} object, initialized with the
		 * required properties.
		 * 
		 * @param identifier      Universally Unique ID.
		 * @param email           e-mail.
		 * @param givenName       first name.
		 * @param familyName      last name.
		 * @param birthDate       birthday date.
		 * @param addressCountry  2-letter country code, e.g. GR.
		 * @param postalCode      Postal code number.
		 * @param streetAddress   Full street address, name, number, town.
		 * @param addressLocality City name.
		 * @return the created and initialized object.
		 */
		private EcoAccount initEcoAccount(final String identifier, final String email, final String givenName,
				final String familyName, final Date birthDate, final String addressCountry, final String postalCode,
				final String streetAddress, final String addressLocality) {
			final EcoAccount ecoAccount = new EcoAccount();

			final String birthday = (null == birthDate) ? null : DataConversion.formatDate(birthDate);
			ecoAccount.withIdentifier(identifier).withName(givenName + " " + familyName)
					.withAddress(initAddress(addressCountry, postalCode, streetAddress, addressLocality))
					.withBirthDate(birthday).withEmail(email).withFamilyName(familyName).withGivenName(givenName);

			return ecoAccount;
		}

		/**
		 * Create the absolute minimal {@link Address} object, initialized with the
		 * required properties.
		 * 
		 * @param addressCountry
		 * @param postalCode
		 * @param streetAddress
		 * @param addressLocality
		 * @return the created and initialized object.
		 */
		private Address initAddress(final String addressCountry, final String postalCode, final String streetAddress,
				final String addressLocality) {
			final Address address = new Address();

			address.withAddressCountry(addressCountry).withPostalCode(postalCode).withStreetAddress(streetAddress)
					.withAddressLocality(addressLocality);

			return address;
		}

		/**
		 * Create the absolute minimal {@link EcoBalance} object, initialized with the
		 * required properties.
		 * 
		 * @return the created and initialized object.
		 */
		private EcoBalance initEcoBalance() {
			final EcoBalance ecoBalance = new EcoBalance();

			ecoBalance.withRecycleBinUserID(null).withEcoCredit(initEcoCredit()).withEcoDebit(initEcoDebit())
					.withPurchased(new Purchased()).withRecycled(new Recycled());

			return ecoBalance;
		}

		/**
		 * Create the absolute minimal {@link EcoCredit} object, initialized with the
		 * required properties.
		 * 
		 * @return the created and initialized object.
		 */
		private EcoCredit initEcoCredit() {
			final EcoCredit ecoCredit = new EcoCredit();

			ecoCredit.withMaxValue(1000000.0).withMinValue(0.0).withUnitCode(null).withValue(0.0);

			return ecoCredit;
		}

		/**
		 * Create the absolute minimal {@link EcoDebit} object, initialized with the
		 * required properties.
		 * 
		 * @return the created and initialized object.
		 */
		private EcoDebit initEcoDebit() {
			final EcoDebit ecoDebit = new EcoDebit();

			ecoDebit.withMaxValue(1000000.0).withMinValue(0.0).withUnitCode(null).withValue(0.0);

			return ecoDebit;
		}
	}

	public abstract static class FeedBack {
		/**
		 * The Eco-account user Universally Unique ID that submitted the feedback.
		 */
		private String useruuid;

		/**
		 * Feedback score in 5 star rating scale (1=bad, 5=Good).
		 */
		private Integer userscore;

		/**
		 * Feedback free text comments.
		 */
		private String freetext;

		/**
		 * @return the Eco-account user Universally Unique ID.
		 */
		@JsonProperty("UserID")
		public String getUseruuid() throws IllegalArgumentException {
			return useruuid;
		}

		/**
		 * @return the Feedback score in 5 star rating scale (1=bad, 5=Good).
		 */
		@JsonProperty("Score")
		public Integer getUserscore() {
			return userscore;
		}

		/**
		 * @return the Feedback free text comments.
		 */
		@JsonProperty("Comment")
		public String getFreetext() {
			return freetext;
		}

		/**
		 * Set the Eco-account user Universally Unique ID.
		 * 
		 * @param useruuid the Eco-account user Universally Unique ID.
		 * @throws IllegalArgumentException if the provided user ID is not as expected
		 *                                  from the Access Control Manager.
		 */
		public void setUseruuid(String useruuid) throws IllegalArgumentException {
			if (!AccessControlManager.isProperFormatUUID(useruuid)) {
				throw new IllegalArgumentException("Malformed User ID: " + useruuid);
			}
			this.useruuid = useruuid;
		}

		/**
		 * Set the Feedback score.
		 * 
		 * @param userscore the Feedback score in 5 star rating scale (1=bad, 5=Good).
		 * @throws IllegalArgumentException if the provided score is outside of range
		 *                                  [1,5].
		 */
		public void setUserscore(Integer userscore) throws IllegalArgumentException {
			if (1 > userscore || 5 < userscore) {
				throw new IllegalArgumentException("Producers feedback score out of range [1,5], score = " + userscore);
			}
			this.userscore = userscore;
		}

		/**
		 * Set the Feedback free text comments.
		 * 
		 * @param freetext the Feedback free text comments.
		 */
		public void setFreetext(String freetext) {
			this.freetext = freetext;
		}

		/**
		 * Set the Eco-account user Universally Unique ID.
		 * 
		 * @param useruuid the Eco-account user Universally Unique ID.
		 * @return this object.
		 * @throws IllegalArgumentException if the provided user ID is not as expected
		 *                                  from the Access Control Manager.
		 */
		public FeedBack withUseruuid(String useruuid) throws IllegalArgumentException {
			setUseruuid(useruuid);
			return this;
		}

		/**
		 * Set the Feedback score.
		 * 
		 * @param userscore the Feedback score in 5 star rating scale (1=bad, 5=Good).
		 * @return this object.
		 * @throws IllegalArgumentException if the provided score is outside of range
		 *                                  [1,5].
		 */
		public FeedBack withUserscore(Integer userscore) throws IllegalArgumentException {
			setUserscore(userscore);
			return this;
		}

		/**
		 * Set the Feedback free text comments.
		 * 
		 * @param freetext the Feedback free text comments.
		 * @return this object.
		 */
		public FeedBack withFreetext(String freetext) {
			setFreetext(freetext);
			return this;
		}
	}

	/**
	 * User's product feedback score and comment encapsulation.
	 */
	public static class UserProductFeedback extends FeedBack {
		/**
		 * Product identification (SGTIN = GTIN14 + Serial Number).
		 */
		private ProductID productid;

		/**
		 * @return the Product identification.
		 */
		@JsonProperty("ProductID")
		public ProductID getProductid() {
			return productid;
		}

		/**
		 * Set the Product identification.
		 * 
		 * @param productid the Product identification.
		 */
		public void setProductid(ProductID productid) {
			this.productid = productid;
		}

		/**
		 * Set the Product identification.
		 * 
		 * @param productid the Product identification.
		 * @return this object.
		 */
		public UserProductFeedback withProductid(ProductID productid) {
			setProductid(productid);
			return this;
		}

		@Override
		public UserProductFeedback withUseruuid(String useruuid) throws IllegalArgumentException {
			return (UserProductFeedback) super.withUseruuid(useruuid);
		}

		@Override
		public UserProductFeedback withUserscore(Integer userscore) throws IllegalArgumentException {
			return (UserProductFeedback) super.withUserscore(userscore);
		}

		@Override
		public UserProductFeedback withFreetext(String freetext) {
			return (UserProductFeedback) super.withFreetext(freetext);
		}

		@Override
		public String toString() {
			return "UserProductFeedback [getProductid()=" + getProductid() + ", getUseruuid()=" + getUseruuid()
					+ ", getUserscore()=" + getUserscore() + ", getFreetext()=" + getFreetext() + ", getClass()="
					+ getClass() + ", hashCode()=" + hashCode() + ", toString()=" + super.toString() + "]";
		}

	}

	/**
	 * User's system feedback score and comment encapsulation.
	 */
	public static class SystemFeedback extends FeedBack {
		@Override
		public SystemFeedback withUseruuid(String useruuid) throws IllegalArgumentException {
			return (SystemFeedback) super.withUseruuid(useruuid);
		}

		@Override
		public SystemFeedback withUserscore(Integer userscore) throws IllegalArgumentException {
			return (SystemFeedback) super.withUserscore(userscore);
		}

		@Override
		public SystemFeedback withFreetext(String freetext) {
			return (SystemFeedback) super.withFreetext(freetext);
		}

		@Override
		public String toString() {
			return "SystemFeedback [getUseruuid()=" + getUseruuid() + ", getUserscore()=" + getUserscore()
					+ ", getFreetext()=" + getFreetext() + ", getClass()=" + getClass() + ", hashCode()=" + hashCode()
					+ ", toString()=" + super.toString() + "]";
		}
	}

	/**
	 * Parse the specified barcode in order to retrieve the product ID (GTIN14 +
	 * Serial Number). The supported barcode formats are:
	 * 
	 * <br>
	 * GS1-128 barcode with AI of 01 (GTIN) and/or 21 (Serial Number). Example:
	 * (01)80614141123458(21)6789
	 * 
	 * @param barcode the barcode to parse.
	 * @return a {@link ProductID} object representing the GTIN14 and optional
	 *         Serial Number on successful parsing, <code>null</code> otherwise.
	 */
	public static ProductID readProductBarcode(final String barcode) {
		try {
			final String REGEXGTIN14 = "^\\(01\\)\\d{14}$";
			final String REGEXGTIN14SERIALNUM = "^\\(01\\)\\d{14}\\(21\\)\\w{1,20}$";

			if (null == barcode || barcode.length() < 18 || barcode.length() > 42
					|| !Pattern.matches(REGEXGTIN14 + "|" + REGEXGTIN14SERIALNUM, barcode)) {
				throw new IllegalArgumentException("Invalid Barcode string: " + barcode
						+ ", expected (01)<GTIN14>(21)<SerialNumber> or (01)<GTIN14>");
			}

			final String productid = barcode.replaceAll("\\(01\\)", "").replaceAll("\\(21\\)", ".");
			final ProductID prodid = new ProductID(productid);

			return prodid;
		} catch (Exception e) {
			return null;
		}
	}
}
