/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.dao;

import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Base64;
import java.util.Collection;
import java.util.Date;
import java.util.logging.Logger;

import gr.iccs.isense.circ4life.EcoCreditCalculator;
import gr.iccs.isense.circ4life.EcoCreditParameters;
import gr.iccs.isense.circ4life.dao.CommonDAO.EcoAccountRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.IndividualProductData;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductType;
import gr.iccs.isense.circ4life.dao.CommonDAO.SystemFeedback;
import gr.iccs.isense.circ4life.dao.CommonDAO.UserProductFeedback;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO.WasteType;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerAdminDAO;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAO;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAOFactory;
import gr.iccs.isense.circ4life.dao.schema.Audience;
import gr.iccs.isense.circ4life.dao.schema.Brand;
import gr.iccs.isense.circ4life.dao.schema.IndividualProduct;
import gr.iccs.isense.circ4life.dao.schema.Manufacturer;
import gr.iccs.isense.circ4life.dao.schema.Product;
import gr.iccs.isense.circ4life.datamodels.WEEEsEndOfLifeState;
import gr.iccs.isense.circ4life.logging.Log;
import gr.iccs.isense.circ4life.rest.config.ConfigProperties;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager.NotificationEmailType;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;

/**
 * Data Access Object (DAO) for Eco-account data retrieval.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EcoAccountDAO {
	/**
	 * Logger.
	 */
	private final static Logger logger = Log.getConfiguredLogger(EcoAccountDAO.class);

	/**
	 * Placholder image URL for Eco-accounts.
	 */
	private static final String PLACEHOLDER_IMG = "https://circ4life.iccs.gr/images/placeholders/ecoaccount.png";

	/**
	 * Administration Persistence layer Data Access Object.
	 */
	private static final PersistenceLayerAdminDAO admindbdao = PersistenceLayerDAOFactory.getDefaultAdminInstance();

	/**
	 * Persistence layer Data Access Object.
	 */
	private static final PersistenceLayerDAO dbdao = PersistenceLayerDAOFactory.getDefaultInstance();

	/**
	 * Product Master Data Access Object.
	 */
	private static final ProductMasterDataDAO productmasterdatadao = new ProductMasterDataDAO();

	/**
	 * Eco-credits calculator Data Access Object.
	 */
	private static final EcoCreditCalculatorDAO calculatordao = new EcoCreditCalculatorDAO();

	/**
	 * Eco-credits calculator.
	 */
	private static final EcoCreditCalculator calculator = new EcoCreditCalculator();

	/**
	 * Constructor.
	 * 
	 * @throws IllegalStateException if the object cannot be initialized properly.
	 *                               Check exception detail message for error
	 *                               information.
	 */
	public EcoAccountDAO() throws IllegalStateException {
		logger.setLevel(ConfigProperties.getLoggingLevel(EcoAccountDAO.class));

		if (null == admindbdao) {
			throw new IllegalStateException("Could not initialize administration persistence layer!");
		}

		if (null == dbdao) {
			throw new IllegalStateException("Could not initialize persistence layer!");
		}
	}

	/**
	 * Create a new Eco-account after a valid user registration request. The user is
	 * created on both the Access Control Manager and persistent database. A
	 * verification email should be sent in order to fully enable the account. The
	 * account is by default enabled but not email verified after creation.
	 * 
	 * <br>
	 * <b>CAUTION:</b>The input parameters are NOT validated in this function. The
	 * validation should have occurred in an earlier stage before we reach the call
	 * of this method.
	 * 
	 * @param username   the user name. REQUIRED. Saved in Access Control Manager.
	 * @param password   the user's password. REQUIRED. Saved in Access Control
	 *                   Manager.
	 * @param email      the user's email address. REQUIRED. Saved in Access Control
	 *                   Manager and Eco-account database record.
	 * @param firstname  the user's real world first name. REQUIRED. Saved in Access
	 *                   Control Manager and Eco-account database record.
	 * @param lastname   the user's real world last name. REQUIRED. Saved in Access
	 *                   Control Manager and Eco-account database record.
	 * @param birthday   the user's birthday. REQUIRED. Saved in Eco-account
	 *                   database record.
	 * @param country    the user's 2-letter country code, e.g. GR. REQUIRED. Saved
	 *                   in Eco-account database record.
	 * @param postalcode the user's postal code. REQUIRED. Saved in Eco-account
	 *                   database record.
	 * @param street     the user's street address. REQUIRED. Saved in Eco-account
	 *                   database record.
	 * @param city       the user's location/city. REQUIRED. Saved in Eco-account
	 *                   database record.
	 * @param region     the user's region. OPTIONAL. Saved in Eco-account database
	 *                   record.
	 * @param telephone  the user's telephone. OPTIONAL. Saved in Eco-account
	 *                   database record.
	 * @param image      the user's avatar image URL. OPTIONAL. Saved in Eco-account
	 *                   database record.
	 * @param userurl    the user's Web homepage. OPTIONAL. Saved in Eco-account
	 *                   database record.
	 * @param gender     the user's gender. OPTIONAL. Saved in Eco-account database
	 *                   record.
	 * @param orgname    the user's organization name. OPTIONAL. Saved in
	 *                   Eco-account database record.
	 * @param orgurl     the user's organization Web page. OPTIONAL. Saved in
	 *                   Eco-account database record.
	 * @param locale     ISO 639 alpha-2 language code for this locale. Supported
	 *                   values are configured in the ACM.
	 * @return the user's UUID if the Eco-account was created successfully,
	 *         <code>null</code> otherwise.
	 */
	public String createEcoAccount(final String username, final String password, final String email,
			final String firstname, final String lastname, final Date birthday, final String country,
			final String postalcode, final String street, final String city, final String region,
			final String telephone, final URL image, final URL userurl, final String gender, final String orgname,
			final URL orgurl, final String locale) {
		try {
			// 1. Create the user in the Access Control Manager.
			if (!AccessControlManager.createEcoAccountUser(username, password, email, firstname, lastname, locale)) {
				logger.severe("Could not create Eco-account user in Access Control Manager!");

				return null;
			}

			// 2. Create the Eco-account in the persistent database.
			final String useruuid = AccessControlManager.getUsernameUUID(username);
			if (null == useruuid) {
				logger.severe("Could not get Eco-account user UUID from Access Control Manager!");
				// If this happens, it means that we may have created the user in step above but
				// there is trouble getting details with his username. This is fatal in our case
				// as no further step can be performed. The user cannot be removed without his
				// UUID, so we may end up with stale users.

				return null;
			}
			final String imageurlstr = null == image ? null : image.toString(),
					userurlstr = null == userurl ? null : userurl.toString(),
					orgurlstr = null == orgurl ? null : orgurl.toString();
			final EcoAccountRecord ecoaccount = new EcoAccountRecord(useruuid, email, firstname, lastname, birthday,
					country, postalcode, street, city, region, telephone, imageurlstr, userurlstr, gender, orgname,
					orgurlstr);
			if (!admindbdao.createEcoAccount(ecoaccount)) {
				// Could not create the Eco-account in the database, we need to revert
				// everything we have done.
				logger.severe("Could not create Eco-account user record in database!");

				if (!AccessControlManager.removeEcoAccountUser(useruuid)) {
					logger.severe("Could not remove Eco-account user from Access Control Manager!");
				}

				return null;
			}

			// 3. Create a user validation record in the persistent database, pending until
			// the confirmation email is opened by the user and the registration validation
			// link is visited.
			if (!AccessControlManager.sendVerificationEmail(useruuid)) {
				// Could not send the verification email, we need to revert
				// everything we have done.
				logger.severe("Could not send Eco-account verification email!");

				if (!AccessControlManager.removeEcoAccountUser(useruuid)) {
					logger.severe("Could not remove Eco-account user from Access Control Manager!");
				}

				if (!admindbdao.removeEcoAccount(useruuid)) {
					logger.severe("Could not remove Eco-account user record from database!");
				}

				return null;
			}

			// 4. Everything was OK, return the user UUID.
			return useruuid;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			// Try to clean up
			final String useruuid = AccessControlManager.getUsernameUUID(username);
			AccessControlManager.removeEcoAccountUser(useruuid);
			admindbdao.removeEcoAccount(useruuid);

			return null;
		}
	}

	/**
	 * Remove an Eco-account from the Access Control Manager and the persistent
	 * database.
	 * 
	 * @param useruuid the user's Universally Unique ID.
	 * @return <code>true</code> if the account was successfully removed,
	 *         <code>false</code> otherwise.
	 */
	public boolean removeEcoAccount(final String useruuid) {
		try {
			// 1. Notify the user, now that his email is still in the ACM
			if (!AccessControlManager.sendNotificationEmail(useruuid, NotificationEmailType.ACCOUNT_REMOVAL)) {
				logger.severe("Could not notify via email Eco-account user " + useruuid + " for his account removal!");
				// We need to go on, email problems cannot fail this process.
			}

			// 2. Remove the user from the Access Control Manager.
			if (!AccessControlManager.removeEcoAccountUser(useruuid)) {
				logger.severe("Could not remove Eco-account user " + useruuid + " from the Access Control Manager!");

				return false;
			}

			// 3. Remove record from the persistent database.
			if (!admindbdao.removeEcoAccount(useruuid)) {
				// TODO: What should we do if the account was removed from Access Control
				// Manager but not from database? This is a stale record.
				logger.severe("Could not remove Eco-account record " + useruuid + " from the database!");

				return false;
			}

			return true;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Update Eco-account with the provided paramaters, if and only if the parameter
	 * is not null.
	 * 
	 * <br>
	 * <b>CAUTION:</b>The input parameters are NOT validated in this function. The
	 * validation should have occurred in an earlier stage before we reach the call
	 * of this method.
	 * 
	 * @param useruuid   the user's Universally Unique ID.
	 * @param password   the user's password. REQUIRED. Saved in Access Control
	 *                   Manager.
	 * @param email      the user's email address. REQUIRED. Saved in Access Control
	 *                   Manager and Eco-account database record.
	 * @param firstname  the user's real world first name. REQUIRED. Saved in Access
	 *                   Control Manager and Eco-account database record.
	 * @param lastname   the user's real world last name. REQUIRED. Saved in Access
	 *                   Control Manager and Eco-account database record.
	 * @param birthday   the user's birthday. REQUIRED. Saved in Eco-account
	 *                   database record.
	 * @param country    the user's 2-letter country code, e.g. GR. REQUIRED. Saved
	 *                   in Eco-account database record.
	 * @param postalcode the user's postal code. REQUIRED. Saved in Eco-account
	 *                   database record.
	 * @param street     the user's street address. REQUIRED. Saved in Eco-account
	 *                   database record.
	 * @param city       the user's location/city. REQUIRED. Saved in Eco-account
	 *                   database record.
	 * @param region     the user's region. OPTIONAL. Saved in Eco-account database
	 *                   record.
	 * @param telephone  the user's telephone. OPTIONAL. Saved in Eco-account
	 *                   database record.
	 * @param image      the user's avatar image URL. OPTIONAL. Saved in Eco-account
	 *                   database record.
	 * @param userurl    the user's Web homepage. OPTIONAL. Saved in Eco-account
	 *                   database record.
	 * @param gender     the user's gender. OPTIONAL. Saved in Eco-account database
	 *                   record.
	 * @param orgname    the user's organization name. OPTIONAL. Saved in
	 *                   Eco-account database record.
	 * @param orgurl     the user's organization Web page. OPTIONAL. Saved in
	 *                   Eco-account database record.
	 * @param locale     ISO 639 alpha-2 language code for this locale. Supported
	 *                   values are configured in the ACM.
	 * @return <code>true</code> if the account was successfully updated,
	 *         <code>false</code> otherwise.
	 */
	public boolean updateEcoAccount(final String useruuid, final String password, final String email,
			final String firstname, final String lastname, final Date birthday, final String country,
			final String postalcode, final String street, final String city, final String region,
			final String telephone, final URL image, final URL userurl, final String gender, final String orgname,
			final URL orgurl, final String locale) {
		try {
			// Update the ACM fields first
			if (!AccessControlManager.updateEcoAccountUser(useruuid, password, email, firstname, lastname, locale)) {
				logger.severe("Could not update Eco-account user in Access Control Manager!");

				return false;
			}

			// Flag for DB profile update
			boolean profile_updated = false;
			// Update the persistent layer Eco-account record if there is a change in value
			if (null != email) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.email", email)) {
					logger.warning("ACM Eco-account updated but no update on DB email!");
				} else {
					profile_updated = true;
				}
			}

			if (null != firstname) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.givenName", firstname)) {
					logger.severe("ACM Eco-account updated but no update on DB first name!");
				} else {
					profile_updated = true;
				}
			}

			if (null != lastname) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.familyName", lastname)) {
					logger.severe("ACM Eco-account updated but no update on DB last name!");
				} else {
					profile_updated = true;
				}
			}

			// Update merged name property
			if (null != firstname || null != lastname) {
				final EcoAccountRecord account = getEcoAccount(useruuid);
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.name",
						account.getEcoAccount().getGivenName() + " " + account.getEcoAccount().getFamilyName())) {
					logger.severe("ACM Eco-account updated but no update on DB name!");
				} else {
					profile_updated = true;
				}
			}

			if (null != birthday) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.birthDate",
						DataConversion.formatDate(birthday))) {
					logger.severe("ACM Eco-account updated but no update on DB birth day!");
				} else {
					profile_updated = true;
				}
			}

			if (null != country) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.address.addressCountry", country)) {
					logger.severe("ACM Eco-account updated but no update on DB address country!");
				} else {
					profile_updated = true;
				}
			}

			if (null != postalcode) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.address.postalCode", postalcode)) {
					logger.severe("ACM Eco-account updated but no update on DB address postal code!");
				} else {
					profile_updated = true;
				}
			}

			if (null != street) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.address.streetAddress", street)) {
					logger.severe("ACM Eco-account updated but no update on DB address street!");
				} else {
					profile_updated = true;
				}
			}

			if (null != city) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.address.addressLocality", city)) {
					logger.severe("ACM Eco-account updated but no update on DB address city!");
				} else {
					profile_updated = true;
				}
			}

			if (null != region) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.address.addressRegion", region)) {
					logger.severe("ACM Eco-account updated but no update on DB address region!");
				} else {
					profile_updated = true;
				}
			}

			if (null != telephone) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.telephone", telephone)) {
					logger.severe("ACM Eco-account updated but no update on DB telephone!");
				} else {
					profile_updated = true;
				}
			}

			if (null != image) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.image", image.toString())) {
					logger.severe("ACM Eco-account updated but no update on DB image URL!");
				} else {
					profile_updated = true;
				}
			}

			if (null != userurl) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.url", userurl.toString())) {
					logger.severe("ACM Eco-account updated but no update on DB user URL!");
				} else {
					profile_updated = true;
				}
			}

			if (null != gender) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.gender", gender)) {
					logger.severe("ACM Eco-account updated but no update on DB gender!");
				} else {
					profile_updated = true;
				}
			}

			if (null != orgname) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.affiliation.name", orgname)) {
					logger.severe("ACM Eco-account updated but no update on DB Organization name!");
				} else {
					profile_updated = true;
				}
			}

			if (null != orgurl) {
				if (!updateEcoAccountTextProperty(useruuid, "EcoAccount.affiliation.url", orgurl.toString())) {
					logger.severe("ACM Eco-account updated but no update on DB Organization URL!");
				} else {
					profile_updated = true;
				}
			}

			if (profile_updated
					&& !AccessControlManager.sendNotificationEmail(useruuid, NotificationEmailType.PROFILE_UPDATED)) {
				// Could not send the profile update email
				logger.severe("Could not send Eco-account profile update email!");
			}

			return true;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Get the Eco-account UUID from the user name and email address.
	 * 
	 * @param username the user name.
	 * @param email    the user email address.
	 * @return the the user's Universally Unique ID if found,
	 *         <code>null</code>otherwise.
	 */
	public String getEcoAccountID(final String username, final String email) {
		try {
			final String useruuid = AccessControlManager.getUsernameUUID(username);
			if (null == useruuid) {
				logger.info("Could not find Eco-account with username " + username);

				return null;
			}
			final String usermail = AccessControlManager.getEmail(useruuid);
			if (!email.equalsIgnoreCase(usermail)) {
				logger.warning("User account " + useruuid + " does not have email address equal to " + email);

				return null;
			}

			return useruuid;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Generate an anonymous user ID from the corresponding Eco-account.
	 * 
	 * @param principalname the username of the Eco-account as represented in the
	 *                      user session security context.
	 * @return a generated anonymous user ID calculated from his Eco-account
	 *         non-volatile properties, <code>null</code> if any error occurs.
	 */
	public String getEcoAccountAnonymousID(final String principalname) {
		try {
			// Does the UUID exists first and for all in the Access Control Manager?
			if (!AccessControlManager.existsUserUUID(principalname)) {
				return null;
			}

			return encodeUserID(principalname);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			// Unhandled error, nothing we can do
			return null;
		}
	}

	/**
	 * Get the recycle bin user ID from an Eco-account.
	 * 
	 * @param principalname the username of the Eco-account as represented in the
	 *                      user session security context.
	 * @return the recycle bin system user ID on success, -1 on failure or if the
	 *         Eco-account has not mapped one yet.
	 */
	public long getEcoAccountRecycleBinUserID(final String principalname) {
		final EcoAccountRecord ecoaccount = getEcoAccount(principalname);
		if (null == ecoaccount) {
			logger.warning("No eco-account record for UUID " + principalname);

			return -1;
		}

		try {
			final Long recyclebinuserid = ecoaccount.getEcoBalance().getRecycleBinUserID();

			return recyclebinuserid.longValue();
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return -1;
		}
	}

	/**
	 * Get the complete EcoAccountRecord of the specified username.
	 * 
	 * @param principalname the username of the Eco-account as represented in the
	 *                      user session security context.
	 * @return the corresponding {@link EcoAccountRecord}, or <code>null</code> on
	 *         error.
	 */
	public EcoAccountRecord getEcoAccount(final String principalname) {
		try {
			// Does the UUID exists first and for all in the Access Control Manager?
			if (!AccessControlManager.existsUserUUID(principalname)) {
				return null;
			}

			final EcoAccountRecord ecoaccount = dbdao.getEcoAccountRecord(principalname);

			// Set placeholder picture if the user has not yet updated his profile picture
			// in the database
			final String profileimgURL = ecoaccount.getEcoAccount().getImage();
			if (null == profileimgURL || profileimgURL.isEmpty()) {
				ecoaccount.getEcoAccount().setImage(PLACEHOLDER_IMG);
			}

			return ecoaccount;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			// Unhandled error, nothing we can do
			return null;
		}
	}

	/**
	 * Get the Eco-account record ID for a user from the Recycle Bin specific
	 * consumer ID.
	 * 
	 * @param recycleBinUserID The consumer's ID as specified by the Recycle Bin
	 *                         system.
	 * @return The Eco-account Id that corresponds to the consumerid, or
	 *         <code>null</code> if there is no mapping.
	 */
	public String getEcoAccountIDFromRecycleBinUserID(final Long recycleBinUserID) {
		try {
			final String useruuid = dbdao.getEcoAccountRecord(recycleBinUserID).getEcoAccount().getIdentifier();

			// Does the UUID exists first and for all in the Access Control Manager?
			if (!AccessControlManager.existsUserUUID(useruuid)) {
				return null;
			}

			return useruuid;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			// Unhandled error, nothing we can do
			return null;
		}
	}

	/**
	 * Get list of purchased products of specified Eco-account.
	 * 
	 * @param principalname the username of the Eco-account as represented in the
	 *                      user session security context.
	 * @param hskey         if not <code>null</code> this parameter specifies the
	 *                      type of products to retrieve.
	 * @param locale        text properties locale as ISO 639 alpha-2 language code,
	 *                      e.g. 'en', 'el', 'es', etc.
	 * @return a {@link Collection} of {@link IndividualProductData} objects filled
	 *         with the essential properties.
	 * @throws Exception if there are malformed properties in the recycle products
	 *                   list.
	 */
	public Collection<IndividualProductData> getEcoAccountPurchasedList(final String principalname, final Integer hskey,
			final String locale) throws Exception {
		final ArrayList<IndividualProductData> result = new ArrayList<IndividualProductData>();

		// Does the UUID exists first and for all in the Access Control Manager?
		if (!AccessControlManager.existsUserUUID(principalname)) {
			return result;
		}

		final EcoAccountRecord ecoaccount = dbdao.getEcoAccountRecord(principalname);
		if (null == ecoaccount || null == ecoaccount.getEcoBalance()
				|| null == ecoaccount.getEcoBalance().getPurchased()
				|| null == ecoaccount.getEcoBalance().getPurchased().getItemListElement()) {
			logger.warning("Missing Eco-balance properties from user ID: " + principalname);

			return result;
		}

		for (final IndividualProduct e : ecoaccount.getEcoBalance().getPurchased().getItemListElement()) {
			final String productid = e.getGtin14() + "." + e.getSerialNumber();
			final ProductType type = productmasterdatadao.getProductType(productid);
			if (null != hskey && null != type && 0 != hskey.compareTo(type.HSkey)) {
				// if the type is determined and it is different than the one provided, skip
				// this product
				continue;
			}
			if (null != hskey && null == type) {
				// if the type cannot be determined and type is provided, skip this product
				continue;
			} else {
				// Enrich product with any extra information in the Master Product Database if
				// possible and add it to the returned list.
				result.add(enrichProduct(e, locale));
			}
		}

		return result;
	}

	/**
	 * Get list of recycled products of specified Eco-account.
	 * 
	 * @param principalname the username of the Eco-account as represented in the
	 *                      user session security context.
	 * @param locale        text properties locale as ISO 639 alpha-2 language code,
	 *                      e.g. 'en', 'el', 'es', etc.
	 * @return a {@link Collection} of {@link IndividualProductData} objects filled
	 *         with the essential properties.
	 * @throws Exception if there are malformed properties in the recycle products
	 *                   list.
	 */
	public Collection<IndividualProductData> getEcoAccountRecycledList(final String principalname, final String locale)
			throws Exception {
		final ArrayList<IndividualProductData> result = new ArrayList<IndividualProductData>();

		// Does the UUID exists first and for all in the Access Control Manager?
		if (!AccessControlManager.existsUserUUID(principalname)) {
			return result;
		}

		final EcoAccountRecord ecoaccount = dbdao.getEcoAccountRecord(principalname);
		if (null == ecoaccount || null == ecoaccount.getEcoBalance() || null == ecoaccount.getEcoBalance().getRecycled()
				|| null == ecoaccount.getEcoBalance().getRecycled().getItemListElement()) {
			logger.warning("Missing Eco-balance properties from user ID: " + principalname);

			return result;
		}

		for (final IndividualProduct e : ecoaccount.getEcoBalance().getRecycled().getItemListElement()) {
			// Enrich product with any extra information in the Master Product Database if
			// possible and add it to the returned list.
			result.add(enrichProduct(e, locale));
		}

		return result;
	}

	/**
	 * Get the user Universally Unique ID from the corresponding anonymous ID.
	 * 
	 * @param anonymousid the anonymous ID.
	 * @return the user Universally Unique ID if there is a user with the specified
	 *         anonymous ID, <code>null</code> otherwise.
	 */
	public String getEcoAccountIDFromAnonymousID(final String anonymousid) {
		final String useruuid = decodeUserID(anonymousid);

		// Does the UUID exists first and for all in the Access Control Manager?
		if (!AccessControlManager.existsUserUUID(useruuid)) {
			return null;
		}

		// Does it have an Eco-account?
		try {
			if (null == dbdao.getEcoAccountRecord(useruuid)) {
				return null;
			}
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}

		// All OK
		return useruuid;
	}

	/**
	 * Update the specified Eco-account with the specified WEEE recycled item and
	 * parameters. This method will effectively add a new item in the Eco-accounts
	 * recycled items list and calculate the Eco-credits of the specific item,
	 * adding them to the account's corresponding Eco-balance.
	 * 
	 * @param useruuid    the Eco-account user Universally Unique ID.
	 * @param recycledate the date of the recycling.
	 * @param years       years of product's use.
	 * @param extramonths extra months added to years.
	 * @param eolstate    End of Life state, applies only to WEEEs products.
	 * @param item        item product details, used to gather Eco-credits
	 *                    calculation parameters.
	 * @return <code>true</code> if calculations were performed and the transaction
	 *         was submitted to the Eco-account successfully, <code>false</code>
	 *         otherwise.
	 */
	public boolean updateEcoAccountWithRecycledWEEEItem(final String useruuid, final Date recycledate,
			final Integer years, final Integer extramonths, final WEEEsEndOfLifeState eolstate,
			final IndividualProductData item) {
		try {
			// Check if the product item is in our Product Master Data
			final ProductType prodtype = productmasterdatadao.getProductType(item.getGtin14());
			if (null == prodtype) {
				// the recycled product is not in our database, we cannot retrieve its
				// eco-credit parameters. We need to give something to the user though so we
				// retrieve the product type from user input and trust it. The application that
				// can send the request is a trusted back-end application, not a consumer form.
				return updateEcoAccountWithRecycledWEEEItem(useruuid, recycledate, years, extramonths, eolstate,
						item.getProductType());
			}

			// Get the details of the product if found in our Product Master Data
			final IndividualProductData enricheditem = enrichProduct(prodtype, item, PersistenceLayerDAO.DEFAULTLOCALE);

			// Calculate the Eco-credits of the specific product in its current state and
			// retrieve the required parameters from the database. These may be product
			// specific, if the product record is rich with details, or product type general
			// values, which brings us to the result of the logic above. Nothing we can do,
			// it is a best effort service.
			final EcoCreditParameters params = calculatordao
					.getProductEcoCreditParameters(enricheditem.getGtin14().toString(), years, extramonths, eolstate);
			final Double ecocredits = calculator.calculate(params);
			enricheditem.getEcoCredits().setValue(ecocredits);
			logger.fine("Eco-credits for recycled item " + enricheditem.getGtin14() + " = " + ecocredits);

			// OK the Eco-credits are calculated, let's update the user's Eco-account. We
			// need to add the item and also to increase the Eco-credits he owns.
			return dbdao.updateEcoAccountWithRecycledItem(useruuid, recycledate, enricheditem);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Same base functionality as
	 * {@link EcoAccountDAO#updateEcoAccountWithRecycledWEEEItem(String, Date, Integer, Integer, WEEEsEndOfLifeState, IndividualProductData)}
	 * but the Eco-credits calculations are performed with the default parameters of
	 * the product's type.
	 * 
	 * @param useruuid    the Eco-account user Universally Unique ID.
	 * @param recycledate the date of the recycling.
	 * @param years       years of product's use.
	 * @param extramonths extra months added to years.
	 * @param eolstate    End of Life state, applies only to WEEEs products.
	 * @param itemtype    item product type, used to gather Eco-credits calculation
	 *                    parameters.
	 * @return <code>true</code> if calculations were performed and the transaction
	 *         was submitted to the Eco-account successfully, <code>false</code>
	 *         otherwise.
	 */
	public boolean updateEcoAccountWithRecycledWEEEItem(final String useruuid, final Date recycledate,
			final Integer years, final Integer extramonths, final WEEEsEndOfLifeState eolstate,
			final ProductType itemtype) {
		try {
			// Get the default Eco-credits parameters for this product type
			final WasteType wastetype = calculatordao.getProductWasteType(itemtype.HSkey);

			// Calculate the Eco-credits of the specific product type
			EcoCreditParameters params = null;
			switch (wastetype) {
			case WEEES: {
				params = calculatordao.getWEEEsParameters(itemtype, years, extramonths, eolstate);
			}
				break;
			case FOOD:
			case UNKNOWN:
			default: {
				logger.severe("Waste type " + wastetype + " is not valid for WEEEs products!");

				return false;
			}
			}
			final Double ecocredits = calculator.calculate(params);
			logger.fine("Eco-credits for recycled item type " + itemtype.HSkey + " = " + ecocredits);

			// Create a dummy item, since we do not have details on the recycled one
			final IndividualProductData item = new IndividualProductData(itemtype, "Not Available", "Not Available",
					"Not Available", "Not Available", new Date());
			item.getEcoCredits().setValue(ecocredits);
			logger.fine("Eco-credits for recycled item " + itemtype + " = " + ecocredits);

			// OK the Eco-credits are calculated, let's update the user's Eco-account. We
			// need to add the item and also to increase the Eco-credits he owns.
			return dbdao.updateEcoAccountWithRecycledItem(useruuid, recycledate, item);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Update the specified Eco-account with the specified Food and Urban Organic
	 * waste recycled item and parameters. This method will effectively add a new
	 * item in the Eco-accounts recycled items list and calculate the Eco-credits of
	 * the specific item, adding them to the account's corresponding Eco-balance.
	 * 
	 * @param useruuid      the Eco-account user Universally Unique ID.
	 * @param recycledate   the date of the recycling.
	 * @param mass_in_grams the mass in grams of the recycled item.
	 * @param itemtype      item product type, used to gather Eco-credits
	 *                      calculation parameters.
	 * @return <code>true</code> if calculations were performed and the transaction
	 *         was submitted to the Eco-account successfully, <code>false</code>
	 *         otherwise.
	 */
	public boolean updateEcoAccountWithRecycledBioWasteItem(final String useruuid, final Date recycledate,
			final Double mass_in_grams, final ProductType itemtype) {
		try {
			// Get the default Eco-credits parameters for this product type
			final WasteType wastetype = calculatordao.getProductWasteType(itemtype.HSkey);

			// Calculate the Eco-credits of the specific product type
			EcoCreditParameters params = null;
			switch (wastetype) {
			case FOOD: {
				params = calculatordao.getFoodParameters(mass_in_grams);
			}
				break;
			case WEEES:
			case UNKNOWN:
			default: {
				logger.severe("Waste type " + wastetype + " is not valid for Bio-waste products!");

				return false;
			}
			}
			final Double ecocredits = calculator.calculate(params);
			logger.fine("Eco-credits for recycled item type " + itemtype.HSkey + " = " + ecocredits);

			// Create a dummy item, since we do not have details on the recycled one
			final IndividualProductData item = new IndividualProductData(itemtype, "Not Available", "Not Available",
					"Not Available", "Not Available", new Date());

			item.getEcoCredits().setValue(ecocredits);
			logger.fine("Eco-credits for recycled item " + itemtype + " = " + ecocredits);

			// OK the Eco-credits are calculated, let's update the user's Eco-account. We
			// need to add the item and also to increase the Eco-credits he owns.
			return dbdao.updateEcoAccountWithRecycledItem(useruuid, recycledate, item);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Update the specified Eco-account with the specified Eco-debits and their
	 * corresponding purchase list.
	 * 
	 * @param useruuid   the Eco-account user Universally Unique ID.
	 * @param purchases  the purchased product list.
	 * @param ecodebits  the Eco-debits to charge the user by the retailer, based on
	 *                   the purchased list.
	 * @param ecocredits the Eco-credits to remove from the user Eco-account for the
	 *                   purchased list.
	 * @return <code>true</code> if the transaction was submitted to the Eco-account
	 *         successfully, <code>false</code> otherwise.
	 */
	public boolean updateEcoAccountWithPurchases(final String useruuid,
			final Collection<IndividualProductData> purchases, final double ecodebits, final double ecocredits) {
		try {
			return dbdao.updateEcoAccountWithPurchases(useruuid, purchases, ecodebits, ecocredits);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Update Eco-account with mapped Recycle Bin system user ID.
	 * 
	 * @param useruuid         the Eco-account user Universally Unique ID.
	 * @param RecycleBinUserID the Recycle Bin system user ID.
	 * @return <code>true</code> if the transaction is successful,
	 *         <code>false</code> otherwise.
	 */
	public boolean updateEcoAccountWithRecycleBinUserID(final String useruuid, long RecycleBinUserID) {
		try {
			return dbdao.updateEcoAccountWithRecycleBinUserID(useruuid, RecycleBinUserID);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Update a text field property of the Eco-account record of a specified user.
	 * 
	 * @param useruuid the Eco-account user Universally Unique ID.
	 * @param property the field property name. in JSON dotted notation (e.g.
	 *                 EcoAccount.image, EcoAccount.gender).
	 * @param value    the {@link String} value of the field property to set.
	 * @return <code>true</code> if the transaction is successful,
	 *         <code>false</code> otherwise.
	 */
	public boolean updateEcoAccountTextProperty(final String useruuid, final String property, final String value) {
		try {
			return dbdao.updateEcoAccountTextProperty(useruuid, property, value);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Update Eco-account's balance by adding the specified Eco-debits and
	 * subtracting the Eco-credits.
	 * 
	 * @param useruuid      the Eco-account user Universally Unique ID.
	 * @param addecodebits  add this positive number to account's Eco-debits.
	 * @param subecocredits subtract this positive number from the account's
	 *                      Eco-credits.
	 * @return <code>true</code> if the transaction is successful,
	 *         <code>false</code> otherwise.
	 */
	public boolean updateEcoAccountBalance(final String useruuid, final double addecodebits,
			final double subecocredits) {
		try {
			return dbdao.updateEcoAccountBalance(useruuid, addecodebits, subecocredits);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Update the products user feedback in the LiveProductData database.
	 * 
	 * @param feedback the list of updated product feedback objects.
	 * @return true if the feedback was applied successfully for all items in the
	 *         provided list, <code>false</code> otherwise.
	 */
	public boolean updateProductsFeedback(final Collection<UserProductFeedback> feedback) {
		try {
			return dbdao.updateProductsFeedback(feedback);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Update the system user feedback in the LiveProductData database.
	 * 
	 * @param feedback the system feedback to update.
	 * @return true if the feedback was applied successfully, <code>false</code>
	 *         otherwise.
	 */
	public boolean updateSystemFeedback(final SystemFeedback feedback) {
		try {
			return dbdao.updateSystemFeedback(feedback);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	/**
	 * Encode Eco-account Universally Unique ID.
	 * 
	 * @param useruuid the Eco-account user Universally Unique ID.
	 * @return a encoded Eco-account Universally Unique ID.
	 */
	private String encodeUserID(final String useruuid) {
		try {
			final String encoded = Base64.getEncoder().encodeToString(useruuid.getBytes());

			return "UUID" + encoded;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Decode Eco-account Universally Unique ID.
	 * 
	 * @param encodeduseruuid the encoded Eco-account user Universally Unique ID.
	 * @return the decoded Eco-account Universally Unique ID.
	 */
	private String decodeUserID(final String encodeduseruuid) {
		try {
			final byte[] decoded = Base64.getDecoder().decode(encodeduseruuid.replaceFirst("^UUID", ""));

			return new String(decoded, StandardCharsets.UTF_8);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Try to find product properties in the Master Product Database in order to
	 * enrich the input individual product object.
	 * 
	 * @param indivproduct the individual product object to enrich.
	 * @param locale       the locale to use for text properties.
	 * @return the enriched object.
	 * @throws Exception in case of any error during the process.
	 */
	private IndividualProductData enrichProduct(final IndividualProduct indivproduct, final String locale)
			throws Exception {
		final String gtin = indivproduct.getGtin14();
		ProductType type = productmasterdatadao.getProductType(gtin);
		if (null == type) {
			logger.warning("Cannot determine product type codes for product with GTIN " + gtin + "! Using dummy.");
			type = new ProductType(-1);
		}

		final ProductRecord productrecord = productmasterdatadao.getProductRecord(gtin, locale);
		if (null != productrecord) {
			// Master Product record found, use it to fill all possible properties.
			final Product productdata = productrecord.getMasterData().getProductData();
			indivproduct.withProductID(productdata.getProductID()).withManufacturer(productdata.getManufacturer());
		}

		return new IndividualProductData(type, indivproduct);
	}

	/**
	 * Try to find product properties in the Master Product Database in order to
	 * enrich the input individual product object.
	 * 
	 * @param type         pre-defined product type.
	 * @param itemtoenrich the individual product object to enrich.
	 * @param locale       the locale to use for text properties.
	 * @return the enriched object.
	 * @throws Exception in case of any error during the process.
	 */
	private IndividualProductData enrichProduct(final ProductType type, final IndividualProductData itemtoenrich,
			final String locale) throws Exception {
		// Cannot accept null
		if (null == type || null == itemtoenrich) {
			logger.severe("Null arguments passed to enrich product process!");

			return null;
		}

		final IndividualProductData result = new IndividualProductData(type, itemtoenrich);
		final ProductRecord productrecord = productmasterdatadao.getProductRecord(itemtoenrich.getGtin14(), locale);
		if (null != productrecord) {
			final Audience audience = productrecord.getMasterData().getProductData().getAudience();
			if (null != audience) {
				result.setAudience(audience);
			}
			final Brand brand = productrecord.getMasterData().getProductData().getBrand();
			if (null != brand) {
				result.setBrand(brand);
			}
			final String category = productrecord.getMasterData().getProductData().getCategory();
			if (null != category) {
				result.setCategory(category);
			}
			final String description = productrecord.getMasterData().getProductData().getDescription();
			if (null != description) {
				result.setDescription(description);
			}
			final String image = productrecord.getMasterData().getProductData().getImage();
			if (null != image) {
				result.setImage(image);
			}
			final Manufacturer manufacturer = productrecord.getMasterData().getProductData().getManufacturer();
			if (null != manufacturer) {
				result.setManufacturer(manufacturer);
			}
			final String material = productrecord.getMasterData().getProductData().getMaterial();
			if (null != material) {
				result.setMaterial(material);
			}
			final String name = productrecord.getMasterData().getProductData().getName();
			if (null != name) {
				result.setName(name);
			}
			final String productID = productrecord.getMasterData().getProductData().getProductID();
			if (null != productID) {
				result.setProductID(productID);
			}
		} else {
			logger.info("GTIN14: " + itemtoenrich.getGtin14()
					+ " not found in Master Product Data, cannot enrich product item.");
		}

		return result;
	}
}
