/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.dao;

import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Logger;

import gr.iccs.isense.circ4life.EcoCreditParameters;
import gr.iccs.isense.circ4life.FoodParameters;
import gr.iccs.isense.circ4life.WEEEsLifetimeFactorCalculator;
import gr.iccs.isense.circ4life.WEEEsMaterialThermodynamicRarityCalculator;
import gr.iccs.isense.circ4life.WEEEsParameters;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductType;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAO;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAOFactory;
import gr.iccs.isense.circ4life.dao.schema.LifespanDistributionParameters;
import gr.iccs.isense.circ4life.dao.schema.Material;
import gr.iccs.isense.circ4life.dao.schema.Parameters;
import gr.iccs.isense.circ4life.dao.schema.Recyclability;
import gr.iccs.isense.circ4life.dao.schema.Resources;
import gr.iccs.isense.circ4life.datamodels.MaterialFactors;
import gr.iccs.isense.circ4life.datamodels.WEEEsEndOfLifeState;
import gr.iccs.isense.circ4life.logging.Log;
import gr.iccs.isense.circ4life.rest.config.ConfigProperties;

/**
 * Data Access Object (DAO) for Eco-credits calculation data and parameters.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EcoCreditCalculatorDAO {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(EcoCreditCalculatorDAO.class);

	/**
	 * Persistence layer Data Access Object.
	 */
	private static final PersistenceLayerDAO persistencelayerdao = PersistenceLayerDAOFactory.getDefaultInstance();

	/**
	 * Product Master Data Access Object.
	 */
	private static final ProductMasterDataDAO productmasterdatadao = new ProductMasterDataDAO();

	/**
	 * Constant weights for Waste Electrical and Electronic Equipment.
	 */
	public class WEEEsWeights {
		public final Double A;
		public final Double B;
		public final Double C;

		public WEEEsWeights(Double a, Double b, Double c) {
			A = a;
			B = b;
			C = c;
		}
	}

	/**
	 * Lifespan Weibull Distribution parameters for Waste Electrical and Electronic
	 * Equipment.
	 */
	public class WEEEsLifespanDistributionParameters {
		public final Double A;
		public final Double B;

		public WEEEsLifespanDistributionParameters(Double a, Double b) {
			A = a;
			B = b;
		}
	}

	/**
	 * Product waste type enumeration.
	 */
	public enum WasteType {
		UNKNOWN, FOOD, WEEES;
	}

	/**
	 * Constructor.
	 * 
	 * @throws IllegalStateException if the object cannot be initialized properly.
	 *                               Check exception detail message for error
	 *                               information.
	 */
	public EcoCreditCalculatorDAO() throws IllegalStateException {
		logger.setLevel(ConfigProperties.getLoggingLevel(EcoCreditCalculatorDAO.class));

		if (null == persistencelayerdao) {
			throw new IllegalStateException("Could not initialize persistence layer!");
		}
	}

	/**
	 * Determine product waste type from the Harmonized System code key.
	 * 
	 * @param hskey the Harmonized System code key.
	 * @return the waste type, or {@link WasteType} enumeration for unknown.
	 */
	public WasteType getProductWasteType(final Integer hskey) {
		// In the Harmonized System the WEEEs fall under the Tier 1 category 84
		// and Tier 2 8471. The food waste is under Tier 1 category 23 in
		// general
		final String hskeystr = hskey.toString();
		if (hskeystr.startsWith("23")) {
			return WasteType.FOOD;
		} else if (hskeystr.startsWith("8471") || hskeystr.startsWith("9504")) {
			return WasteType.WEEES;
		}

		return WasteType.UNKNOWN;
	}

	/**
	 * Get the corresponding Harmonized System code key from a UNU key.
	 * 
	 * @param unukey the UNU key.
	 * @return the corresponding Harmonized System code key, or <code>null</code> if
	 *         a mapping does not exist.
	 */
	public Integer getHSKeyFromUNUkey(final Integer unukey) {
		try {
			return persistencelayerdao.getProductTypeFromUNUkey(unukey).HSkey;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get the corresponding UNU key from Harmonized System code key.
	 * 
	 * @param hskey the Harmonized System code key.
	 * @return the corresponding UNU key, or <code>null</code> if a mapping does not
	 *         exist.
	 */
	public Integer getUNUKeyFromHSkey(final Integer hskey) {
		try {
			return persistencelayerdao.getProductTypeFromHSkey(hskey).UNUkey;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get the constant parameters Alpha and Beta of Lifespan Weibull Distribution
	 * parameters for Waste Electrical and Electronic Equipment from UNU code key.
	 * 
	 * @param unukey the UNU key.
	 * @return the parameters of the lifespan distribution, or <code>null</code> if
	 *         there are no parameters corresponding to the specified UNU key.
	 */
	public WEEEsLifespanDistributionParameters getWEEEsLifespanDistributionParameters(final Integer unukey) {
		final Integer hskey = getHSKeyFromUNUkey(unukey);
		if (null == hskey) {
			logger.warning("No HS key corresponding to the specified UNU key: " + unukey);

			return null;
		}

		try {
			final LifespanDistributionParameters params = persistencelayerdao
					.getDefaultLifespanDistributionParameters(hskey);
			if (null == params) {
				logger.severe("No default Lifespan Distribution parameters for the specified UNU key: " + unukey);

				return null;
			}

			final Double alpha = params.getAlpha().getValue(), beta = params.getBeta().getValue();

			return new WEEEsLifespanDistributionParameters(alpha, beta);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get weight constants that correspond to the specified UNU key.
	 * 
	 * @param unukey the UNU key of the product type.
	 * @return the weights A,B,C, or <code>null</code> if there are no weights
	 *         corresponding to the specified UNU key.
	 */

	public WEEEsWeights getWEEEsWeights(final Integer unukey) {
		if (null == unukey) {
			return null;
		}

		if (isWEEEsUNUKey(unukey)) {
			try {
				final Parameters parameters = persistencelayerdao.getDefaultParameters(getHSKeyFromUNUkey(unukey));

				return new WEEEsWeights(parameters.getWeightOfExergy().getValue(),
						parameters.getWeightOfEoLState().getValue(), parameters.getWeightOfLifetime().getValue());
			} catch (Exception e) {
				logger.severe(e.getMessage());

				return null;
			}
		}

		// The WEEEs weight constants apply only to the UNU keys of WEEEs products.
		return null;
	}

	/**
	 * Get the default materials list from the HS key product type. The returned
	 * list is not for a specific product, rather a list of reference materials list
	 * values for every product that belongs in the category specified by the HS
	 * key.
	 * 
	 * @param hskey the Harmonized System code key.
	 * @return a list of default materials, or <code>null</code> if there is no
	 *         corresponding list for this HS key.
	 */
	public Collection<Material> getDefaultMaterials(final Integer hskey) {
		try {
			return persistencelayerdao.getDefaultMaterialFactors(hskey);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get the material factors list from the UNU key and the product's End of Life
	 * state. The returned list is not for a specific product, rather a list of
	 * reference material factor values for every product that belongs in the
	 * category specified by the UNU key.
	 * 
	 * @param unukey   the UNU key of the product type.
	 * @param eolstate the End of Life state.
	 * @return a list of material factors, or <code>null</code> if there is no
	 *         corresponding list for this UNU key.
	 */
	public Collection<MaterialFactors> getMaterialFactors(final Integer unukey, final WEEEsEndOfLifeState eolstate) {
		final Integer hskey = getHSKeyFromUNUkey(unukey);
		if (null == hskey) {
			logger.warning("No HS key corresponding to the specified UNU key: " + unukey);

			return null;
		}

		try {
			final Collection<Material> materials = getDefaultMaterials(hskey);
			if (null == materials) {
				logger.warning("No default materials list for the specified HS key: " + hskey);

				return null;
			}

			final ArrayList<MaterialFactors> result = new ArrayList<MaterialFactors>();

			for (final Material m : materials) {
				final double mass_in_grams = m.getFactors().getMass().getValue(),
						rarity = m.getFactors().getRarity().getValue(),
						recyclability = m.getFactors().getRecyclabilityfactor().getValue();
				// Only End of Life state of Broken has specific values, coming from the
				// database.
				// In the working/reusable cases the recyclability is equal to 1.0.
				if (eolstate == WEEEsEndOfLifeState.BROKEN) {
					result.add(new MaterialFactors(mass_in_grams, rarity, recyclability));
				} else {
					result.add(new MaterialFactors(mass_in_grams, rarity, 1.0));
				}
			}

			return result;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get the Food waste parameters from the product type specified with the
	 * Harmonized System code key. The returned value is not for a specific product,
	 * rather reference values for the product type.
	 * 
	 * @param hskey the Harmonized System code key.
	 * @return the food parameters required for Eco-credit calculation, or
	 *         <code>null</code> if the Harmonized System code key cannot be used.
	 */
	public FoodParameters getFoodParameters(final Integer hskey) {
		try {
			final Parameters parameters = persistencelayerdao.getDefaultParameters(hskey);
			if (null == parameters) {
				return null;
			}

			return new FoodParameters(parameters.getWeightOfExergy().getValue(),
					parameters.getOrganicChemicalExergy().getValue(), parameters.getOrganicMass().getValue());
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Calculate the Food waste parameters of the specified product type dynamically
	 * in order to return the parameters with a modified coefficient D. The
	 * coefficient D will be responsible to convert the energy contained in the
	 * wastes into the incentive given to the final user and it can modified and
	 * adjusted to different incentive programs. The parameter D will be applied in
	 * the expression in order to balance the expected incentives associated to
	 * Eco-credits to the final incentive provided.
	 * 
	 * @param mass_in_grams Organic waste mass in grams.
	 * @return the food parameters required for Eco-credit calculation, or
	 *         <code>null</code> in case of error.
	 */
	public FoodParameters getFoodParameters(final Double mass_in_grams) {
		try {
			// We only keep the Chemical exergy and calculate dynamically the rest according
			// to D2.5: Incentive schemes for collaborative reuse/recyclingof products
			// The unfortunate thing is that there is no proper formula in D2.5, so we have
			// to just comply with the mass to eco-credits mapping of Table 6.
			// We use polynomial regression on value ranges since it seems the table in the
			// deliverable does not contain values comming from a single formula.
			double coeffD = 1.0;
			if (mass_in_grams <= 4000) { // 4Kg
				coeffD = (-5.801557144 * Math.pow(10.0, -21.0) * Math.pow(mass_in_grams, 2.0) + 0.00101 * mass_in_grams
						- 0.01) / mass_in_grams;
			} else if (mass_in_grams <= 8000) { // 5 - 8 Kg
				coeffD = (-8.875759458 * Math.pow(10.0, -20.0) * Math.pow(mass_in_grams, 2.0) + 0.00151 * mass_in_grams
						- 3.637978807 * Math.pow(10.0, -12.0)) / mass_in_grams;
			} else { // 9Kg or more
				// 4.820758691·10-20 x2 + 0.002 x - 0.01
				coeffD = (4.820758691 * Math.pow(10.0, -20.0) * Math.pow(mass_in_grams, 2.0) + 0.002 * mass_in_grams
						- 0.01) / mass_in_grams;
			}

			// Set to 1.0 so that Eco-credits formula Eco-credits = D * b_chemical_exergy *
			// m_organic_waste does not break down later
			final double exergy = 1.0;

			return new FoodParameters(coeffD, exergy, mass_in_grams);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Create the Waste Electrical and Electronic Equipment Eco-credits formula
	 * parameters from the product, lifetime and state.
	 * 
	 * @param prodtype    The product type codes.
	 * @param years       years of product usage.
	 * @param extramonths extra months of usage added to years.
	 * @param eolstate    the End of Life state.
	 * @return the WEEEs parameters required for Eco-credit calculation, or
	 *         <code>null</code> if the product details cannot be used.
	 */
	public WEEEsParameters getWEEEsParameters(final ProductType prodtype, final Integer years,
			final Integer extramonths, final WEEEsEndOfLifeState eolstate) {
		Integer unukey = null;
		if (null == prodtype.UNUkey) {
			unukey = getUNUKeyFromHSkey(prodtype.HSkey);
		} else {
			unukey = prodtype.UNUkey;
		}

		final WEEEsWeights weeesABC = getWEEEsWeights(unukey);
		if (null == weeesABC) {
			return null;
		}
		final WEEEsLifespanDistributionParameters weeeslfspanAB = getWEEEsLifespanDistributionParameters(unukey);
		if (null == weeeslfspanAB) {
			return null;
		}
		final Collection<MaterialFactors> matfactors = getMaterialFactors(unukey, eolstate);

		final WEEEsMaterialThermodynamicRarityCalculator materialcalc = new WEEEsMaterialThermodynamicRarityCalculator(
				matfactors);
		final WEEEsLifetimeFactorCalculator ltcalc = new WEEEsLifetimeFactorCalculator(years, extramonths,
				weeeslfspanAB.A, weeeslfspanAB.B);

		return new WEEEsParameters(weeesABC.A, materialcalc, weeesABC.B, eolstate, weeesABC.C, ltcalc);
	}

	/**
	 * Check if the specified UNU key corresponds to WEEEs waste.
	 * 
	 * @param unukey UNU key to check.
	 * @return <code>true</code> if the specified UNU key corresponds to WEEEs
	 *         waste, <code>false</code> otherwise.
	 */
	private boolean isWEEEsUNUKey(Integer unukey) {
		final Integer hskey = getHSKeyFromUNUkey(unukey);
		if (null == hskey) {
			// Unknown UNU key for our database, we cannot determine if is WEEEs
			return false;
		}

		switch (getProductWasteType(hskey)) {
		case WEEES: {
			return true;
		}
		case FOOD:
		case UNKNOWN:
		default: {
			return false;
		}
		}
	}

	/**
	 * Get a product's parameters required for Eco-credits calculation.
	 * 
	 * @param productid   the product ID code key.
	 * @param years       years of product usage.
	 * @param extramonths extra months of usage added to years.
	 * @param eolstate    the End of Life state.
	 * @return The parameters required for Eco-credits calculation, or
	 *         <code>null</code> if the product or all required parameters do not
	 *         exist in the persistent data.
	 */
	public EcoCreditParameters getProductEcoCreditParameters(final String productid, final Integer years,
			final Integer extramonths, final WEEEsEndOfLifeState eolstate) {
		final ProductType prodtype = productmasterdatadao.getProductType(productid);
		if (null == prodtype) {
			return null;
		}
		final WasteType wastetype = getProductWasteType(prodtype.HSkey);
		final ProductRecord product = productmasterdatadao.getProductRecord(productid,
				PersistenceLayerDAO.DEFAULTLOCALE);

		try {
			// First try to get the parameters from the product record resources...
			final Resources resources = product.getMasterData().getLifeCycle().getResources();

			final Recyclability recyclability = resources.getRecyclability();
			if (null == recyclability) {
				throw new Exception("Product with ID " + productid + " is missing recyclability properties.");
			}

			switch (wastetype) {
			case FOOD: {
				final Double D = recyclability.getParameters().getWeightOfExergy().getValue(),
						chemical_exergy = recyclability.getParameters().getOrganicChemicalExergy().getValue(),
						mass_organic_waste = recyclability.getParameters().getOrganicMass().getValue();

				return new FoodParameters(D, chemical_exergy, mass_organic_waste);
			}
			case WEEES: {
				final Double A = recyclability.getParameters().getWeightOfExergy().getValue(),
						B = recyclability.getParameters().getWeightOfEoLState().getValue(),
						C = recyclability.getParameters().getWeightOfLifetime().getValue();

				final Collection<MaterialFactors> material_factors = new ArrayList<MaterialFactors>();
				for (final Material m : resources.getMaterialList()) {
					final double mass = m.getFactors().getMass().getValue(),
							rarity = m.getFactors().getRarity().getValue(),
							recycle = m.getFactors().getRecyclabilityfactor().getValue();
					// Only End of Life state of Broken has specific values, coming from the
					// database.
					// In the working/reusable cases the recyclability is equal to 1.0.
					if (eolstate == WEEEsEndOfLifeState.BROKEN) {
						material_factors.add(new MaterialFactors(mass, rarity, recycle));
					} else {
						material_factors.add(new MaterialFactors(mass, rarity, 1.0));
					}
				}

				final WEEEsMaterialThermodynamicRarityCalculator materialcalc = new WEEEsMaterialThermodynamicRarityCalculator(
						material_factors);
				final WEEEsLifespanDistributionParameters weeeslfspanAB = getWEEEsLifespanDistributionParameters(
						prodtype.UNUkey);
				final WEEEsLifetimeFactorCalculator ltcalc = new WEEEsLifetimeFactorCalculator(years, extramonths,
						weeeslfspanAB.A, weeeslfspanAB.B);

				return new WEEEsParameters(A, materialcalc, B, eolstate, C, ltcalc);
			}
			case UNKNOWN:
			default: {
				throw new Exception("Product with ID " + productid + " has unknown recycle waste type.");
			}
			}
		} catch (Exception e) {
			logger.fine("Eco-credits parameters not found in product record, details: " + e.getMessage());
		}

		// If the product record does not have the parameters, then grab the product
		// type and get the default parameters for this product type.
		switch (wastetype) {
		case FOOD: {
			return getFoodParameters(prodtype.HSkey);
		}
		case WEEES: {
			return getWEEEsParameters(prodtype, years, extramonths, eolstate);
		}
		case UNKNOWN:
		default: {
			return null;
		}
		}
	}
}
