/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.dao;

import java.util.ArrayList;
import java.util.Collection;
import java.util.logging.Logger;

import gr.iccs.isense.circ4life.FoodParameters;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductID;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductType;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO.WEEEsWeights;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO.WasteType;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAO;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAOFactory;
import gr.iccs.isense.circ4life.dao.schema.Material;
import gr.iccs.isense.circ4life.dao.schema.OrganicChemicalExergy;
import gr.iccs.isense.circ4life.dao.schema.OrganicMass;
import gr.iccs.isense.circ4life.dao.schema.Parameters;
import gr.iccs.isense.circ4life.dao.schema.Resources;
import gr.iccs.isense.circ4life.dao.schema.WeightOfEoLState;
import gr.iccs.isense.circ4life.dao.schema.WeightOfExergy;
import gr.iccs.isense.circ4life.dao.schema.WeightOfLifetime;
import gr.iccs.isense.circ4life.logging.Log;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;

/**
 * Data Access Object (DAO) for product Master data retrieval.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class ProductMasterDataDAO {
	/**
	 * Logger.
	 */
	private static final Logger logger = Log.getConfiguredLogger(ProductMasterDataDAO.class);

	/**
	 * Placholder image URL for Eco-accounts.
	 */
	private static final String PLACEHOLDER_IMG = "https://circ4life.iccs.gr/images/placeholders/product.png";

	/**
	 * Persistence layer Data Access Object.
	 */
	private static final PersistenceLayerDAO persistencelayerdao = PersistenceLayerDAOFactory.getDefaultInstance();

	/**
	 * Constructor.
	 */
	public ProductMasterDataDAO() {
		if (null == persistencelayerdao) {
			throw new IllegalStateException("Could not initialize persistence layer!");
		}
	}

	/**
	 * Get list of supported product types defined by standard code keys and
	 * localized text description.
	 * 
	 * @param locale text properties locale as ISO 639 alpha-2 language code, e.g.
	 *               'en', 'el', 'es', etc.
	 * @return a {@link Collection} of {@link ProductType} objects. The
	 *         {@link Collection} may be empty in case of error, but it will be not
	 *         <code>null</code>.
	 */
	public Collection<ProductType> getSupportedProductTypes(final String locale) {
		final Collection<ProductType> product_types = new ArrayList<ProductType>();

		try {
			final Collection<ProductType> db_prod_types = persistencelayerdao.getProductTypes(locale);
			if (null == db_prod_types || 0 == db_prod_types.size()) {
				logger.severe("Could not retrieve product types from database!");

				return product_types;
			}

			product_types.addAll(db_prod_types);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			// Unhandled error, nothing we can do
			product_types.clear();
		}

		return product_types;
	}

	/**
	 * Get the Product type codes from the product ID.
	 * 
	 * @param productid Product ID string. Format used is GTIN14.SerialNumber or
	 *                  GTIN14.examples: 04047111123453.012345678901,
	 *                  04047111123453.
	 * @return the Product type codes or <code>null</code> in case of malformed
	 *         product ID or product not found.
	 */
	public ProductType getProductType(final String productid) {
		try {
			final ProductID prodid = getProductID(productid);
			if (null == prodid) {
				logger.warning("Malformed product ID: " + productid);

				return null;
			}

			// Lookup in the database
			return persistencelayerdao.getProductType(DataConversion.formatGTIN14(prodid.GTIN14));
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Create product ID object for easier codes access.
	 * 
	 * @param productid Product ID string. Format used is GTIN14.SerialNumber or
	 *                  GTIN14.examples: 04047111123453.012345678901,
	 *                  04047111123453.
	 * @return a {@link ProductID} object representing the specified SGTIN on
	 *         successful conversion, <code>null</code> otherwise.
	 */
	public ProductID getProductID(final String productid) {
		try {
			return new ProductID(productid);
		} catch (IllegalArgumentException e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Retrieve the Product Master Data record from a product ID (SGTIN).
	 * 
	 * @param productid product ID (SGTIN).
	 * @param locale    text properties locale as ISO 639 alpha-2 language code,
	 *                  e.g. 'en', 'el', 'es', etc.
	 * @return the Product Master Data record if found, <code>null</code> otherwise.
	 */
	public ProductRecord getProductRecord(final String productid, final String locale) {
		try {

			final ProductRecord productrec = persistencelayerdao.getProductRecord(productid, locale);

			// Set placeholder picture if the product has no image URL in the database
			final String productimgURL = productrec.getMasterData().getProductData().getImage();
			if (null == productimgURL || productimgURL.isEmpty()) {
				productrec.getMasterData().getProductData().setImage(PLACEHOLDER_IMG);
			}

			// Fill in missing Resources
			final ProductType type = getProductType(productid);
			// Type could be determined, let's get the Recyclability.Parameters of this type
			final EcoCreditCalculatorDAO ecocreddao = new EcoCreditCalculatorDAO();
			final WasteType wastetype = ecocreddao.getProductWasteType(type.HSkey);

			// Get default parameters for this product type
			final Parameters defaultparameters = new Parameters();
			switch (wastetype) {
			case FOOD: {
				final FoodParameters foodparams = ecocreddao.getFoodParameters(type.HSkey);

				if (null != foodparams) {
					defaultparameters.setWeightOfExergy(new WeightOfExergy().withValue(foodparams.getWeightOfExergy()));
					defaultparameters.setOrganicChemicalExergy(
							new OrganicChemicalExergy().withValue(foodparams.getChemicalExergy()));
					defaultparameters.setOrganicMass(new OrganicMass().withValue(foodparams.getOrganicWasteMass()));
				}

				defaultparameters.setWeightOfEoLState(null);
				defaultparameters.setWeightOfLifetime(null);
			}
				break;
			case WEEES: {
				final WEEEsWeights weeesweights = ecocreddao.getWEEEsWeights(type.UNUkey);

				if (null != weeesweights) {
					defaultparameters.setWeightOfExergy(new WeightOfExergy().withValue(weeesweights.A));
					defaultparameters.setWeightOfEoLState(new WeightOfEoLState().withValue(weeesweights.B));
					defaultparameters.setWeightOfLifetime(new WeightOfLifetime().withValue(weeesweights.C));
				}

				defaultparameters.setOrganicChemicalExergy(null);
				defaultparameters.setOrganicMass(null);
			}
				break;
			default: {
				logger.warning("Could not determine WasteType for product: " + productid);
			}
				break;

			}

			final Resources resources = productrec.getMasterData().getLifeCycle().getResources();

			// Create the Recyclability if it does not exist
			if (null == resources.getRecyclability()) {
				resources.withRecyclability(ProductRecord.initRecyclability());

				// Clean up fields that do not apply to waste type
				if (WasteType.FOOD == wastetype) {
					resources.getRecyclability().setEoLState(null);
					resources.getRecyclability().setLifetime(null);
				}

				logger.info("Setting default Recyclability Parameters for product " + productid);
				resources.getRecyclability().withParameters(defaultparameters);
			}

			// Set default product type parameters
			if (null == resources.getRecyclability().getParameters()) {
				logger.info("Setting default Recyclability Parameters for product " + productid);

				resources.getRecyclability().withParameters(defaultparameters);
			}

			// Set default materials list
			if (WasteType.WEEES == wastetype && null != type.UNUkey
					&& (null == resources.getMaterialList() || 0 == resources.getMaterialList().size())) {
				logger.info("Setting default material list for product " + productid);

				resources.setMaterialList(new ArrayList<Material>(ecocreddao.getDefaultMaterials(type.HSkey)));
			}

			return productrec;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Retrieve the Product Eco-points from a product ID (SGTIN).
	 * 
	 * @param productid product ID (SGTIN).
	 * @return the Eco-points value if product record found, <code>null</code>
	 *         otherwise.
	 */
	public Double getProductEcoPoints(final String productid) {
		try {
			return persistencelayerdao.getProductRecord(productid, PersistenceLayerDAO.DEFAULTLOCALE).getEcoAccounting()
					.getEcoPoints().getValue().doubleValue();
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Parse the {@link String} representation of supported barcode and return the
	 * corresponding product ID.
	 * 
	 * @param barcode the {@link String} representation of a barcode.
	 * @return the product ID that corresponds to the barcode, <code>null</code> if
	 *         no correspondence found or barcode is not supported.
	 */
	public ProductID getProductIDFromBarcode(final String barcode) {
		try {
			return CommonDAO.readProductBarcode(barcode);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}
}
