/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.dao.persistence;

import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import javax.json.Json;
import javax.json.JsonObject;

import org.bson.Document;
import org.bson.conversions.Bson;

import com.mongodb.client.model.Filters;
import com.mongodb.client.model.IndexOptions;
import com.mongodb.client.model.ReplaceOptions;
import com.mongodb.client.result.DeleteResult;
import com.mongodb.client.result.UpdateResult;

import gr.iccs.isense.circ4life.dao.CommonDAO.EcoAccountRecord;
import gr.iccs.isense.circ4life.json.JacksonJSONSerialization;
import gr.iccs.isense.circ4life.logging.Log;

/**
 * Data Access Object (DAO) for persistence layer data administration of MongoDB
 * databases.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class MongoDBPersistenceLayerAdminDAO extends MongoDBPersistenceLayerDAO implements PersistenceLayerAdminDAO {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(MongoDBPersistenceLayerAdminDAO.class);

	/**
	 * Filename found under a collection directory, containing the collection
	 * documents index properties JSON object. It's existence is optional, only
	 * collections with indexed documents should use it.
	 */
	public final static String CollectionIndexFilename = "_collection_.index";

	/**
	 * Constructor.
	 * 
	 * @param dbhost The database server hostname or IP address.
	 * @param dbport The database server listening port.
	 * @throws IllegalArgumentException if any of the expected databases/collections
	 *                                  is not present of the database server host.
	 */
	public MongoDBPersistenceLayerAdminDAO(final String dbhost, final Integer dbport) throws IllegalArgumentException {
		super(dbhost, dbport);
	}

	@Override
	public boolean dropDatabase(String databasename) {
		try {
			mongoClient.getDatabase(databasename).drop();
			logger.info("Dropped database " + databasename);

			return true;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	@Override
	public boolean dropCollection(String databasename, String collectionname) {
		try {
			mongoClient.getDatabase(databasename).getCollection(collectionname).drop();
			logger.info("Dropped collection " + collectionname + " of database " + databasename);

			return true;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	@Override
	public boolean importDatabases(final String importdirectory) {
		return importDatabases(importdirectory, true);
	}

	@Override
	public boolean importDatabases(final String importdirectory, boolean keepdata) {
		// If data are not kept, drop everything up front. If the databases are not
		// found or not dropped due to error, the whole import process stops here.
		if (!keepdata && !dropApplicationDatabases(false)) {
			logger.severe("Could not find or drop the application known databases, "
					+ "aborting to avoid import on unknown application...");

			return false;
		}

		// We import data by creating records and updating current ones as they are
		// found.
		try {
			final File importdirfile = new File(importdirectory);
			// First level of hierarchy: directories are considered database names
			for (final File dbdir : importdirfile.listFiles()) {
				if (dbdir.isDirectory()) {
					// Database directory found.
					final String dbname = dbdir.getName();
					if (databaseExists(dbname)) {
						logger.info("Database " + dbname + " found. Updating Collections with data in "
								+ dbdir.getAbsolutePath());
					} else {
						logger.info("Database " + dbname + " not found, creating...");
					}

					// Second level of hierarchy: directories are considered database collection
					// names
					for (final File collectiondir : dbdir.listFiles()) {
						if (collectiondir.isDirectory()) {
							// Collection directory found
							final String collectionname = collectiondir.getName();
							if (collectionExists(dbname, collectionname)) {
								logger.info("Collection " + collectionname + " found in database " + dbname
										+ ", updating...");
							} else {
								logger.info("Collection " + collectionname + " not found in database " + dbname
										+ ", creating...");
							}

							// third level of hierarchy: JSON files are considered JSON record documents to
							// add to current collection
							final FilenameFilter jsonfilesfilter = new FilenameFilter() {
								@Override
								public boolean accept(File dir, String name) {
									return dir.getAbsolutePath().equals(collectiondir.getAbsolutePath())
											&& name.endsWith(".json");
								}
							};

							final ArrayList<Document> toimport = new ArrayList<Document>();
							for (final File jsondoc : collectiondir.listFiles(jsonfilesfilter)) {
								// JSON Document file
								logger.info("Found JSON Document file for import " + jsondoc.getAbsolutePath());

								final InputStream is = new FileInputStream(jsondoc);
								final Document document = Document.parse(Json.createReader(is).readObject().toString());

								// Add to the list of documents to insert
								toimport.add(document);

								is.close();
							}

							// Create index
							collectionCreateIndex(dbname, collectiondir);

							// We detected a list of document to add
							if (0 < toimport.size()) {
								if (!keepdata || !databaseExists(dbname)) {
									// Insert documents to the collection
									mongoClient.getDatabase(dbname).getCollection(collectionname).insertMany(toimport);
								} else {
									// Replace options with upcert = true in order to create documents not already
									// present
									final ReplaceOptions replaceOptions = new ReplaceOptions().upsert(true);

									// Get index property names
									final JsonObject indexjson = collectionGetIndex(collectiondir);
									if (null != indexjson) {
										final Set<String> indexkeys = Document
												.parse(indexjson.getJsonObject("keys").toString()).keySet();

										// Update current documents, create new only if missing
										for (final Document d : toimport) {
											// Create a filter from the index properties and the values found in the
											// imported document.
											final Document filter = new Document();
											for (final String indexkey : indexkeys) {
												filter.append(indexkey, getValueFromPropertyPath(d, indexkey));
											}

											// Replace the document that matches the filter
											final UpdateResult replaceresult = mongoClient.getDatabase(dbname)
													.getCollection(collectionname)
													.replaceOne(filter, d, replaceOptions);

											// Remove the updated document from the list of imported documents
											if (replaceresult.wasAcknowledged()
													&& 1 == replaceresult.getModifiedCount()) {
												logger.info("Successfully updated document with: " + filter.toJson());
											}
										}
									} else {
										// Insert all documents, no index properties in this collection
										mongoClient.getDatabase(dbname).getCollection(collectionname)
												.insertMany(toimport);
									}
								}
							}
						}
					}
				}
			}
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}

		// All OK
		return true;
	}

	@Override
	public boolean importDatabases(final URL importresourceURL, boolean keepdata) {
		// TODO: Implement
		logger.warning("Unimplented method " + Thread.currentThread().getStackTrace()[1].getMethodName());

		return false;
	}

	@Override
	public boolean dropApplicationDatabases(final boolean force) {
		final String[] dbnames = { PersistenceLayerDAO.EcoAccounts.DBNAME, PersistenceLayerDAO.LiveProductData.DBNAME,
				PersistenceLayerDAO.MasterProductData.DBNAME };

		// Check for known database existence first, we do not want to start importing
		// stuff to a database server for an alien application!
		for (final String dbname : dbnames) {
			if (!force && !databaseExists(dbname)) {
				logger.severe("Could not find database " + dbname);

				return false;
			}
		}

		// Drop databases, all are found or we are forced to do it
		for (final String dbname : dbnames) {
			if (!force && !dropDatabase(dbname)) {
				logger.severe("Could not drop database " + dbname);

				return false;
			} else if (databaseExists(dbname)) {
				// drop no matter what
				dropDatabase(dbname);
			}
		}

		return true;
	}

	@Override
	public boolean createEcoAccount(final EcoAccountRecord ecoaccount) {
		try {
			// 1. First check that the email does not already exist.
			final Bson filter = Filters.eq(EcoAccounts.USEREMAILPROPERTY, ecoaccount.getEcoAccount().getEmail());
			for (final Document d : EcoAccountsDB.getCollection(EcoAccounts.USERSCOLLECTION).find(filter)) {
				// If we find any, we cannot create a new one.
				logger.severe("Cannot create Eco-account from object: " + ecoaccount
						+ ", email already exists in database! Account with same email: "
						+ d.getString(EcoAccounts.USERUUIDPROPERTY));

				return false;
			}

			// 2. Create the account
			// TODO: Escape the text properties before insert to the database?
			final String jsonstr = JacksonJSONSerialization.JSONtoString(ecoaccount);
			final Document document = Document.parse(jsonstr);
			EcoAccountsDB.getCollection(EcoAccounts.USERSCOLLECTION).insertOne(document);

			return true;
		} catch (Exception e) {
			logger.severe("Cannot create Eco-account from object: " + ecoaccount);

			return false;
		}
	}

	@Override
	public boolean removeEcoAccount(final String useruuid) {
		try {
			// Find the record, this function should only return a valid, not duplicate
			// result
			final EcoAccountRecord record = getEcoAccountRecord(useruuid);
			if (null == record) {
				logger.severe("User Eco-account not found for " + useruuid);

				return false;
			}

			final Bson filter = Filters.eq(EcoAccounts.USERUUIDPROPERTY, useruuid);
			final DeleteResult result = EcoAccountsDB.getCollection(EcoAccounts.USERSCOLLECTION).deleteOne(filter);

			return result.wasAcknowledged() && 1 == result.getDeletedCount();
		} catch (Exception e) {
			logger.severe("Cannot remove Eco-account of user: " + useruuid);

			return false;
		}
	}

	/**
	 * Read the collection index file in a JSON object.
	 * 
	 * @param collectiondir the collection import directory.
	 * @return the JSON object read in success, <code>null</code> otherwise.
	 * @throws IOException thrown in index file read error.
	 */
	private JsonObject collectionGetIndex(final File collectiondir) throws IOException {
		// Read the indices file of this collection. This will create defined indices or
		// update documents based on the index property value. If a document with the
		// same index property value is found in the database, it is updated with the
		// data read from import directory. If not, then a new record is created.
		final File collectionindexfile = new File(
				collectiondir.getAbsolutePath() + File.separator + CollectionIndexFilename);
		if (collectionindexfile.exists()) {
			// collection directory has index file, let's read it
			logger.info("Found Collection index file for import " + collectionindexfile.getAbsolutePath());

			final InputStream is = new FileInputStream(collectionindexfile);
			final JsonObject indexjson = Json.createReader(is).readObject();
			is.close();

			final Document indexkeys = Document.parse(indexjson.getJsonObject("keys").toString());
			if (null == indexkeys) {
				logger.severe(collectionindexfile.getAbsolutePath() + " does not contain index keys property!");
				throw new NullPointerException(
						collectionindexfile.getAbsolutePath() + " does not contain index keys property!");
			}

			return indexjson;
		} else {
			// No index file
			return null;
		}
	}

	/**
	 * Create the specified collection's index.
	 * 
	 * @param dbname        the database name containing the collection.
	 * @param collectiondir the import directory of the collection data.
	 * @throws IOException if the collection index file cannot be read.
	 */
	private void collectionCreateIndex(final String dbname, final File collectiondir) throws IOException {
		final JsonObject indexjson = collectionGetIndex(collectiondir);
		if (null != indexjson) {
			final Document indexkeys = Document.parse(indexjson.getJsonObject("keys").toString());
			final Document indexoptions = Document.parse(indexjson.getJsonObject("options").toString());

			logger.info(dbname + "." + collectiondir.getName() + " index: " + indexkeys.toJson());

			// Create the indices
			if (null == indexoptions) {
				mongoClient.getDatabase(dbname).getCollection(collectiondir.getName()).createIndex(indexkeys);
			} else {
				final IndexOptions options = new IndexOptions();
				final Boolean unique = indexoptions.getBoolean("unique");
				if (null != unique) {
					options.unique(unique);
				}
				logger.info(dbname + "." + collectiondir.getName() + " index options: " + options.toString());
				mongoClient.getDatabase(dbname).getCollection(collectiondir.getName()).createIndex(indexkeys, options);
			}
		} else {
			logger.info(dbname + "." + collectiondir.getName() + " no index file found for , skipping index creation.");
		}
	}

	/**
	 * Get the value of a property dotted key path.
	 * 
	 * @param doc          the document containing the property embedded in the
	 *                     specified path.
	 * @param propertypath the property path to follow in order to reach the value.
	 * @return the value of the property if found in the specified document, using
	 *         the specified path, <code>null</code> otherwise.
	 */
	private Object getValueFromPropertyPath(final Document doc, final String propertypath) {
		final List<String> keys = (propertypath.contains(".")) ? Arrays.asList(propertypath.split("\\."))
				: Arrays.asList(propertypath);

		return doc.getEmbedded(keys, Object.class);
	}
}
