/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.dao.persistence;

import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.logging.Logger;

import javax.json.Json;
import javax.json.JsonArray;
import javax.json.JsonObject;

import org.bson.Document;
import org.bson.conversions.Bson;
import org.bson.json.JsonMode;
import org.bson.json.JsonWriterSettings;

import com.mongodb.client.MongoClient;
import com.mongodb.client.MongoClients;
import com.mongodb.client.MongoCollection;
import com.mongodb.client.MongoDatabase;
import com.mongodb.client.model.Filters;
import com.mongodb.client.model.ReplaceOptions;
import com.mongodb.client.model.Updates;
import com.mongodb.client.result.UpdateResult;

import gr.iccs.isense.circ4life.dao.CommonDAO;
import gr.iccs.isense.circ4life.dao.CommonDAO.EcoAccountRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.IndividualProductData;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductID;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductType;
import gr.iccs.isense.circ4life.dao.CommonDAO.SystemFeedback;
import gr.iccs.isense.circ4life.dao.CommonDAO.UserProductFeedback;
import gr.iccs.isense.circ4life.dao.schema.LifespanDistributionParameters;
import gr.iccs.isense.circ4life.dao.schema.Material;
import gr.iccs.isense.circ4life.dao.schema.Parameters;
import gr.iccs.isense.circ4life.dao.schema.Recyclability;
import gr.iccs.isense.circ4life.json.JacksonJSONSerialization;
import gr.iccs.isense.circ4life.logging.Log;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;

/**
 * Data Access Object (DAO) for persistence layer data retrieval from MongoDB
 * database.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class MongoDBPersistenceLayerDAO implements PersistenceLayerDAO {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(MongoDBPersistenceLayerDAO.class);

	/**
	 * MongoDB record ID property field that should be removed before deserialized
	 * retrieved BSON documents to POJOs.
	 */
	private final static String MongoDBIDField = "_id";

	/**
	 * MongoDB driver access client.
	 */
	protected final MongoClient mongoClient;

	/**
	 * Access handle to 1.1.1.1 Master Product Data database.
	 */
	private final MongoDatabase MasterProductDataDB;

	/**
	 * Access handle to 1.1.1.2 User Eco-accounts database.
	 */
	protected final MongoDatabase EcoAccountsDB;

	/**
	 * Access handle to 1.1.1.3 Live Product Data database.
	 */
	private final MongoDatabase LiveProductDataDB;

	/**
	 * JSON output settings for conversion of MongoDB BSON documents to plain JSON
	 * that Jackson can deserialize.
	 */
	private final JsonWriterSettings JSONOutSettings;

	/**
	 * Constructor.
	 * 
	 * @param dbhost The database server hostname or IP address.
	 * @param dbport The database server listening port.
	 * @throws IllegalArgumentException if any of the expected databases/collections
	 *                                  is not present of the database server host.
	 */
	public MongoDBPersistenceLayerDAO(final String dbhost, final Integer dbport) throws IllegalArgumentException {
		// Initialize MongoDB driver connection
		logger.info("Connecting to MongoDB server at " + dbhost + ":" + dbport);
		mongoClient = MongoClients.create("mongodb://" + dbhost + ":" + dbport);

		// Get database handles
		MasterProductDataDB = mongoClient.getDatabase(MasterProductData.DBNAME);
		printCollectionCounts(MasterProductData.DBNAME, MasterProductData.PRODUCTSCOLLECTION);
		printCollectionCounts(MasterProductData.DBNAME, MasterProductData.PRODUCTTYPESCOLLECTION);

		EcoAccountsDB = mongoClient.getDatabase(EcoAccounts.DBNAME);
		printCollectionCounts(EcoAccounts.DBNAME, EcoAccounts.USERSCOLLECTION);

		LiveProductDataDB = mongoClient.getDatabase(LiveProductData.DBNAME);
		printCollectionCounts(LiveProductData.DBNAME, LiveProductData.EVENTSCOLLECTION);
		printCollectionCounts(LiveProductData.DBNAME, LiveProductData.PRODUCTSFEEDBACKCOLLECTION);
		printCollectionCounts(LiveProductData.DBNAME, LiveProductData.SYSTEMFEEDBACKCOLLECTION);

		JSONOutSettings = JsonWriterSettings.builder().outputMode(JsonMode.RELAXED).build();
	}

	@Override
	public boolean databaseExists(final String dbname) {
		try {
			for (final String name : mongoClient.listDatabaseNames()) {
				if (name.equals(dbname)) {
					return true;
				}
			}

			return false;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	@Override
	public boolean collectionExists(final String dbname, final String collectionname) {
		try {
			for (final String name : mongoClient.getDatabase(dbname).listCollectionNames()) {
				if (name.equals(collectionname)) {
					return true;
				}
			}

			return false;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return false;
		}
	}

	@Override
	public EcoAccountRecord getEcoAccountRecord(final String userid) throws IllegalStateException {
		Document ecoaccountdoc = null;

		// We need to find ONLY ONE Eco-account with the specified userid. If we find
		// more than one we need to throw an exception.
		final Bson filter = Filters.eq(EcoAccounts.USERUUIDPROPERTY, userid);
		for (final Document d : EcoAccountsDB.getCollection(EcoAccounts.USERSCOLLECTION).find(filter)) {
			if (null == ecoaccountdoc) {
				// Initial loop, save the document that matched
				ecoaccountdoc = d;
			} else {
				// We have saved the first match and found another one, this is an exception.
				throw new IllegalStateException(
						"Duplicate Eco-account records with " + EcoAccounts.USERUUIDPROPERTY + ": " + userid);
			}
		}

		// We haven't found this userid in our userbase
		if (null == ecoaccountdoc) {
			logger.fine("Eco-account record document not found with " + EcoAccounts.USERUUIDPROPERTY + ": " + userid);

			return null;
		}

		// Convert the found document record to our object.
		try {
			// Remove the MongoDB special property '_id'
			ecoaccountdoc.remove(MongoDBIDField);

			return JacksonJSONSerialization.StringtoJSON(ecoaccountdoc.toJson(JSONOutSettings), EcoAccountRecord.class);
		} catch (IOException e) {
			throw new IllegalStateException(
					"Cannot read Eco-account record with " + EcoAccounts.USERUUIDPROPERTY + ": " + userid);
		}
	}

	@Override
	public EcoAccountRecord getEcoAccountRecord(final Long recycleBinUserID) throws IllegalStateException {
		Document ecoaccountdoc = null;

		// We need to find ONLY ONE Eco-account with the specified Recycle Bin User ID.
		// If we find more than one we need to throw an exception.
		final Bson filter = Filters.eq(EcoAccounts.RECYCLEBINUSERID, recycleBinUserID.longValue());
		for (final Document d : EcoAccountsDB.getCollection(EcoAccounts.USERSCOLLECTION).find(filter)) {
			if (null == ecoaccountdoc) {
				// Initial loop, save the document that matched
				ecoaccountdoc = d;
			} else {
				// We have saved the first match and found another one, this is an exception.
				throw new IllegalStateException(
						"Duplicate Eco-account records with " + EcoAccounts.RECYCLEBINUSERID + ": " + recycleBinUserID);
			}
		}

		// We haven't found this recycleBinUserID in our userbase
		if (null == ecoaccountdoc) {
			logger.fine("Eco-account record document not found with " + EcoAccounts.RECYCLEBINUSERID + ": "
					+ recycleBinUserID);

			return null;
		}

		// Convert the found document record to our object.
		try {
			// Remove the MongoDB special property '_id'
			ecoaccountdoc.remove(MongoDBIDField);

			return JacksonJSONSerialization.StringtoJSON(ecoaccountdoc.toJson(JSONOutSettings), EcoAccountRecord.class);
		} catch (IOException e) {
			throw new IllegalStateException(
					"Cannot read Eco-account record with " + EcoAccounts.RECYCLEBINUSERID + ": " + recycleBinUserID);
		}
	}

	@Override
	public Collection<ProductType> getProductTypes(String locale) throws IllegalStateException {
		try {
			final MongoCollection<Document> documents = MasterProductDataDB
					.getCollection(MasterProductData.PRODUCTTYPESCOLLECTION);

			if (null == documents || 0L == documents.countDocuments()) {
				throw new IllegalStateException(
						MasterProductData.PRODUCTTYPESCOLLECTION + " returned a NULL or empty collection!");
			}

			// Get the document that matches our document or English 'en' if it does not
			// exist
			Document localized_doc = documents.find(Filters.eq(LOCALE_JSON_PROPERTY, locale)).first();
			if (null == localized_doc) {
				// No database record for the specified locale, let's try the default locale
				// 'en'
				localized_doc = documents.find(Filters.eq(LOCALE_JSON_PROPERTY, DEFAULTLOCALE)).first();
				if (null == localized_doc) {
					// No record at all, bye bye
					throw new IllegalStateException(
							MasterProductData.PRODUCTTYPESCOLLECTION + " does not contain any record, for any locale!");
				}
			}

			// Now that we found the localized document let's get the list of product types.
			final String prodtypeslistcollection = localized_doc.getString(LOCALE_COLLECTION_PROPERTY);
			final MongoCollection<Document> prodtypes = MasterProductDataDB.getCollection(prodtypeslistcollection);
			if (null == prodtypes || 0L == prodtypes.countDocuments()) {
				throw new IllegalStateException(prodtypeslistcollection + " returned a NULL or empty collection!");
			}

			final Collection<ProductType> result = new ArrayList<CommonDAO.ProductType>();

			for (final Document d : prodtypes.find()) {
				// Read the product types one by one and fill up the result
				try {
					final Integer hskey = d.getInteger(MasterProductData.PRODUCTTYPESHSKEY),
							unukey = d.getInteger(MasterProductData.PRODUCTTYPESUNUKEY);
					final String description = d.getString(MasterProductData.PRODUCTTYPESDESCR);
					result.add(new ProductType(hskey, unukey, description));
				} catch (Exception e) {
					logger.warning("Product Type JSON document error: " + e.getMessage());

					continue;
				}
			}

			return result;
		} catch (IllegalArgumentException e) {
			logger.severe(MasterProductData.PRODUCTTYPESCOLLECTION + " is invalid! Details: " + e.getMessage());

			return null;
		}
	}

	@Override
	public ProductType getProductType(final String GTIN14) throws IllegalStateException {
		try {
			// Convert to ProductID for validation
			new ProductID(GTIN14);
		} catch (IllegalArgumentException e) {
			logger.severe("Invalid GTIN14 code: " + GTIN14 + ", details: " + e.getMessage());

			return null;
		}

		// GTIN is OK, let's try to find the product
		try {
			final ProductRecord product = getProductRecord(GTIN14, DEFAULTLOCALE);

			// Get the HSKey and retrieve the full product type from the product types
			// collection.
			return product.getProductType();
		} catch (IllegalArgumentException e) {
			logger.severe(MasterProductData.PRODUCTSCOLLECTION + " is invalid! Details: " + e.getMessage());

			return null;
		}
	}

	@Override
	public ProductType getProductTypeFromHSkey(Integer hskey) throws IllegalStateException {
		if (null == hskey) {
			logger.warning("Tried to get product type with null HS key!");

			return null;
		}

		try {
			final MongoCollection<Document> documents = MasterProductDataDB
					.getCollection(MasterProductData.PRODUCTTYPESCOLLECTION);

			if (null == documents || 0L == documents.countDocuments()) {
				throw new IllegalStateException(
						MasterProductData.PRODUCTTYPESCOLLECTION + " returned a NULL or empty collection!");
			}

			final Document localized_doc = documents.find(Filters.eq(LOCALE_JSON_PROPERTY, DEFAULTLOCALE)).first();
			if (null == localized_doc) {
				// No record at all, bye bye
				throw new IllegalStateException(
						MasterProductData.PRODUCTTYPESCOLLECTION + " does not contain any record, for any locale!");
			}

			// Now that we found the localized document let's get the list of product types.
			final String prodtypeslistcollection = localized_doc.getString(LOCALE_COLLECTION_PROPERTY);

			Document result = null;
			final Bson filter = Filters.eq(MasterProductData.PRODUCTTYPESHSKEY, hskey);
			for (final Document d : MasterProductDataDB.getCollection(prodtypeslistcollection).find(filter)) {
				if (null == result) {
					// Initial loop, save the document that matched
					result = d;
				} else {
					// We have saved the first match and found another one, this is an exception.
					throw new IllegalStateException("Duplicate product type records with "
							+ MasterProductData.PRODUCTTYPESHSKEY + ": " + hskey);
				}
			}

			if (null == result) {
				logger.fine("Not found product type with " + MasterProductData.PRODUCTTYPESHSKEY + ": " + hskey);

				return null;
			}

			final Integer HSkey = result.getInteger(MasterProductData.PRODUCTTYPESHSKEY),
					UNUkey = result.getInteger(MasterProductData.PRODUCTTYPESUNUKEY);
			final String description = result.getString(MasterProductData.PRODUCTTYPESDESCR);

			return new ProductType(HSkey, UNUkey, description);
		} catch (IllegalArgumentException e) {
			logger.severe(MasterProductData.PRODUCTTYPESCOLLECTION + " is invalid! Details: " + e.getMessage());

			return null;
		}
	}

	@Override
	public ProductType getProductTypeFromUNUkey(Integer unukey) throws IllegalStateException {
		if (null == unukey) {
			logger.warning("Tried to get product type with null UNU key!");

			return null;
		}

		try {
			final MongoCollection<Document> documents = MasterProductDataDB
					.getCollection(MasterProductData.PRODUCTTYPESCOLLECTION);

			if (null == documents || 0L == documents.countDocuments()) {
				throw new IllegalStateException(
						MasterProductData.PRODUCTTYPESCOLLECTION + " returned a NULL or empty collection!");
			}

			final Document localized_doc = documents.find(Filters.eq(LOCALE_JSON_PROPERTY, DEFAULTLOCALE)).first();
			if (null == localized_doc) {
				// No record at all, bye bye
				throw new IllegalStateException(
						MasterProductData.PRODUCTTYPESCOLLECTION + " does not contain any record, for any locale!");
			}

			// Now that we found the localized document let's get the list of product types.
			final String prodtypeslistcollection = localized_doc.getString(LOCALE_COLLECTION_PROPERTY);

			Document result = null;
			final Bson filter = Filters.eq(MasterProductData.PRODUCTTYPESUNUKEY, unukey);
			for (final Document d : MasterProductDataDB.getCollection(prodtypeslistcollection).find(filter)) {
				if (null == result) {
					// Initial loop, save the document that matched
					result = d;
				} else {
					// We have saved the first match and found another one, this is an exception.
					throw new IllegalStateException("Duplicate product type records with "
							+ MasterProductData.PRODUCTTYPESUNUKEY + ": " + unukey);
				}
			}

			if (null == result) {
				logger.fine("Not found product type with " + MasterProductData.PRODUCTTYPESUNUKEY + ": " + unukey);

				return null;
			}

			final Integer HSkey = result.getInteger(MasterProductData.PRODUCTTYPESHSKEY),
					UNUkey = result.getInteger(MasterProductData.PRODUCTTYPESUNUKEY);
			final String description = result.getString(MasterProductData.PRODUCTTYPESDESCR);

			return new ProductType(HSkey, UNUkey, description);
		} catch (IllegalArgumentException e) {
			logger.severe(MasterProductData.PRODUCTTYPESCOLLECTION + " is invalid! Details: " + e.getMessage());

			return null;
		}
	}

	@Override
	public ProductRecord getProductRecord(String productid, String locale) throws IllegalStateException {
		ProductID pid = null;
		try {
			// Convert to ProductID for validation
			pid = new ProductID(productid);
		} catch (IllegalArgumentException e) {
			logger.severe("Invalid product ID code: " + productid + ", details: " + e.getMessage());

			return null;
		}

		// GTIN is OK, let's try to find the product
		try {
			final MongoCollection<Document> documents = MasterProductDataDB
					.getCollection(MasterProductData.PRODUCTSCOLLECTION);

			if (null == documents || 0L == documents.countDocuments()) {
				throw new IllegalStateException(
						MasterProductData.PRODUCTSCOLLECTION + " returned a NULL or empty collection!");
			}

			// Get the document that matches our document or English 'en' if it does not
			// exist
			Document localized_doc = documents.find(Filters.eq(LOCALE_JSON_PROPERTY, locale)).first();
			if (null == localized_doc) {
				// No database record for the specified locale, let's try the default locale
				// 'en'
				localized_doc = documents.find(Filters.eq(LOCALE_JSON_PROPERTY, DEFAULTLOCALE)).first();
				if (null == localized_doc) {
					// No record at all, bye bye
					throw new IllegalStateException(
							MasterProductData.PRODUCTSCOLLECTION + " does not contain any record, for any locale!");
				}
			}

			// Now that we found the localized document let's get the list of products.
			final String productscollection = localized_doc.getString(LOCALE_COLLECTION_PROPERTY);
			final MongoCollection<Document> products = MasterProductDataDB.getCollection(productscollection);
			if (null == products || 0L == products.countDocuments()) {
				throw new IllegalStateException(productscollection + " returned a NULL or empty collection!");
			}

			// Try to find the product record and check for duplication
			Document product = null;
			for (final Document d : products
					.find(Filters.eq(MasterProductData.PRODUCTGTIN, DataConversion.formatGTIN14(pid.GTIN14)))) {
				if (null == product) {
					// Initial loop, save the document that matched
					product = d;
				} else {
					// We have saved the first match and found another one, this is an exception.
					throw new IllegalStateException("Duplicate Master Product Data records with "
							+ MasterProductData.PRODUCTGTIN + ": " + pid.GTIN14.toString());
				}
			}

			// We haven't found this GTIN in our database
			if (null == product) {
				logger.fine("Master Product Data record document not found with " + MasterProductData.PRODUCTGTIN + ": "
						+ pid.GTIN14.toString());

				return null;
			}

			product.remove(MongoDBIDField);

			return JacksonJSONSerialization.StringtoJSON(product.toJson(JSONOutSettings), ProductRecord.class);
		} catch (IllegalArgumentException e) {
			logger.severe(MasterProductData.PRODUCTSCOLLECTION + " is invalid! Details: " + e.getMessage());

			return null;
		} catch (IOException e) {
			throw new IllegalStateException("Cannot read Master Product Data record with "
					+ MasterProductData.PRODUCTGTIN + ": " + pid.GTIN14.toString());
		}
	}

	@Override
	public LifespanDistributionParameters getDefaultLifespanDistributionParameters(final Integer hskey)
			throws IllegalStateException {
		final Document d = getDefaultEcoCreditParameters(hskey);
		if (null == d) {
			return null;
		}

		try {
			final JsonObject obj = Json.createReader(new StringReader(d.toJson(JSONOutSettings))).readObject();
			final String lifespandistrib = obj.get(MasterProductData.ECOCREDITPARAMSLIFESPANDISTRIB).toString();

			return JacksonJSONSerialization.StringtoJSON(lifespandistrib, LifespanDistributionParameters.class);
		} catch (Exception e) {
			logger.severe("Data conversion error during Lifespan Distribution parameters retrieval, details: "
					+ e.getMessage());

			return null;
		}
	}

	@Override
	public Parameters getDefaultParameters(Integer hskey) throws IllegalStateException {
		final Document d = getDefaultEcoCreditParameters(hskey);
		if (null == d) {
			return null;
		}

		try {
			final JsonObject obj = Json.createReader(new StringReader(d.toJson(JSONOutSettings))).readObject();
			final Recyclability recyclability = JacksonJSONSerialization.StringtoJSON(
					obj.get(MasterProductData.ECOCREDITPARAMSRECYCLABILITY).toString(), Recyclability.class);

			return recyclability.getParameters();
		} catch (Exception e) {
			logger.severe("Data conversion error during default Parameters retrieval, details: " + e.getMessage());

			return null;
		}
	}

	@Override
	public Collection<Material> getDefaultMaterialFactors(Integer hskey) throws IllegalStateException {
		final Document d = getDefaultEcoCreditParameters(hskey);
		if (null == d) {
			return null;
		}

		try {
			final ArrayList<Material> materialslist = new ArrayList<Material>();

			final JsonObject obj = Json.createReader(new StringReader(d.toJson(JSONOutSettings))).readObject();
			final JsonArray materials = obj.getJsonArray(MasterProductData.ECOCREDITPARAMSMATERIALS);
			materials.forEach(m -> {
				final JsonObject material = (JsonObject) m;
				try {
					materialslist.add(JacksonJSONSerialization.StringtoJSON(material.toString(), Material.class));
				} catch (IOException e) {
					logger.severe(e.getMessage());
				}
			});

			return materialslist;
		} catch (Exception e) {
			logger.severe("Data conversion error during Material Factors retrieval, details: " + e.getMessage());

			return null;
		}
	}

	@Override
	public boolean updateEcoAccountWithRecycledItem(final String useruuid, final Date recycledate,
			final IndividualProductData item) throws IllegalStateException {
		// Check input parameters
		if (null == useruuid || null == recycledate || null == item) {
			logger.severe("Null arguments not allowed while updating Eco-account!");

			return false;
		}

		// Set the recycle date, just in case
		item.setRecycleDate(DataConversion.formatDate(recycledate));

		Double ecocredits = 0.0;
		if (null == item.getEcoCredits() || null == item.getEcoCredits().getValue()) {
			logger.warning(
					"Missing Eco-credits from Recycled Item, we will add the item but no Eco-balance change will occur.");
		} else {
			ecocredits = item.getEcoCredits().getValue().doubleValue();
		}

		try {
			// Filter to find the Eco-account record to update.
			final Bson filter = Filters.eq(EcoAccounts.USERUUIDPROPERTY, useruuid);
			// The update documents
			final Bson itemjson = Document.parse(JacksonJSONSerialization.JSONtoString(item));
			final Bson ecocreditsupdate = Updates.inc(EcoAccounts.ECOCREDITS, ecocredits),
					recycledlistupdate = Updates.push(EcoAccounts.RECYCLEDLIST, itemjson);
			final Bson update = Updates.combine(ecocreditsupdate, recycledlistupdate);

			final UpdateResult updateresult = EcoAccountsDB.getCollection(EcoAccounts.USERSCOLLECTION).updateOne(filter,
					update);

			// All OK
			return updateresult.wasAcknowledged() && 1 == updateresult.getMatchedCount()
					&& 1 == updateresult.getModifiedCount();
		} catch (Exception e) {
			logger.severe("Could not add recycled item to Eco-account " + useruuid + ", details: " + e.getMessage());

			return false;
		}
	}

	@Override
	public boolean updateEcoAccountWithPurchases(final String useruuid, Collection<IndividualProductData> purchases,
			final double ecodebits, final double ecocredits) throws IllegalStateException {
		// Check input parameters
		if (null == useruuid || null == purchases || 0 == purchases.size()) {
			logger.severe("Null arguments or empty lists not allowed while updating Eco-account!");

			return false;
		}

		try {
			// First check if the Eco-credits that are requested to use and remove from the
			// user's Eco-account are available. If not then everything stop here and we
			// cannot process this request.
			final EcoAccountRecord record = getEcoAccountRecord(useruuid);
			if (null != record) {
				final double avail_ecocredits = record.getEcoBalance().getEcoCredit().getValue().doubleValue();

				// Not enough Eco-credits in user's account
				if (avail_ecocredits < ecocredits) {
					return false;
				}
			}

			// Filter to find the Eco-account record to update.
			final Bson filter = Filters.eq(EcoAccounts.USERUUIDPROPERTY, useruuid);
			final Bson ecodebitsupdate = Updates.inc(EcoAccounts.ECODEBITS, ecodebits);
			// MongoDB has no decrease update, so we use the increase with an amount
			// multiplied by -1
			final Bson ecocreditsupdate = Updates.inc(EcoAccounts.ECOCREDITS, -1.0 * ecocredits);
			final ArrayList<Bson> updates = new ArrayList<Bson>();
			for (final IndividualProductData purchase : purchases) {
				final Bson purchasejson = Document.parse(JacksonJSONSerialization.JSONtoString(purchase));
				updates.add(purchasejson);
			}
			final Bson update = Updates.combine(ecodebitsupdate, ecocreditsupdate,
					Updates.pushEach(EcoAccounts.PURCHASEDLIST, updates));

			final UpdateResult updateresult = EcoAccountsDB.getCollection(EcoAccounts.USERSCOLLECTION).updateOne(filter,
					update);

			// Check result
			return updateresult.wasAcknowledged() && 1 == updateresult.getMatchedCount()
					&& 1 == updateresult.getModifiedCount();
		} catch (Exception e) {
			logger.severe(
					"Could not add purchased item list to Eco-account " + useruuid + ", details: " + e.getMessage());

			return false;
		}
	}

	@Override
	public boolean updateProductsFeedback(Collection<UserProductFeedback> feedback) throws IllegalStateException {
		// Check input parameters
		if (null == feedback || 0 == feedback.size()) {
			logger.severe("Null arguments or empty list is not allowed while updating products user feedback!");

			return false;
		}

		try {
			final ReplaceOptions replaceOptions = new ReplaceOptions().upsert(true);
			for (final UserProductFeedback upf : feedback) {
				if (null == upf.getUseruuid() || null == upf.getProductid()) {
					throw new Exception("Invalid product user feedback: " + upf);
				}

				final Bson useridfilter = Filters.eq(LiveProductData.FEEDBACKUSERID, upf.getUseruuid()),
						gtinfilter = Filters.eq(LiveProductData.PRODUCTSFEEDBACKGTIN, upf.getProductid().getGTIN14());
				final Bson filter = Filters.and(useridfilter, gtinfilter);
				final Document replacement = Document.parse(JacksonJSONSerialization.JSONtoString(upf));

				final UpdateResult updateresult = LiveProductDataDB
						.getCollection(LiveProductData.PRODUCTSFEEDBACKCOLLECTION)
						.replaceOne(filter, replacement, replaceOptions);

				if (!updateresult.wasAcknowledged()) {
					throw new Exception("Could not update product user feedback, user ID = " + upf.getUseruuid()
							+ ", GTIN = " + upf.getProductid().getGTIN14());
				}
			}

			// All OK
			return true;
		} catch (Exception e) {
			logger.severe("Could not update products user feedback, details: " + e.getMessage());

			return false;
		}
	}

	@Override
	public boolean updateSystemFeedback(SystemFeedback feedback) throws IllegalStateException {
		// Check input parameters
		if (null == feedback || null == feedback.getUseruuid() || null == feedback.getUserscore()) {
			logger.severe("Null arguments not allowed while updating system user feedback!");

			return false;
		}

		try {
			final Bson filter = Filters.eq(LiveProductData.FEEDBACKUSERID, feedback.getUseruuid());
			final Document replacement = Document.parse(JacksonJSONSerialization.JSONtoString(feedback));
			final ReplaceOptions replaceOptions = new ReplaceOptions().upsert(true);

			final UpdateResult updateresult = LiveProductDataDB.getCollection(LiveProductData.SYSTEMFEEDBACKCOLLECTION)
					.replaceOne(filter, replacement, replaceOptions);

			// Check result
			return updateresult.wasAcknowledged();
		} catch (Exception e) {
			logger.severe("Could not update products user feedback, details: " + e.getMessage());

			return false;
		}
	}

	@Override
	public boolean updateEcoAccountWithRecycleBinUserID(String useruuid, long recycleBinUserID)
			throws IllegalStateException {
		// Check input parameters
		if (null == useruuid || 0 > recycleBinUserID || 99999999999L < recycleBinUserID) {
			logger.severe("Null/Invalid arguments while updating Eco-account!");

			return false;
		}

		try {
			// Filter to find the Eco-account record to update.
			final Bson filter = Filters.eq(EcoAccounts.USERUUIDPROPERTY, useruuid);
			final Bson update = Updates.set(EcoAccounts.RECYCLEBINUSERID, recycleBinUserID);

			final UpdateResult updateresult = EcoAccountsDB.getCollection(EcoAccounts.USERSCOLLECTION).updateOne(filter,
					update);

			// Check result
			return updateresult.wasAcknowledged() && 1 == updateresult.getMatchedCount()
					&& 1 == updateresult.getModifiedCount();
		} catch (Exception e) {
			logger.severe("Could not map Recycle Bin User ID " + recycleBinUserID + " to Eco-account " + useruuid
					+ ", details: " + e.getMessage());

			return false;
		}
	}

	@Override
	public boolean updateEcoAccountTextProperty(final String useruuid, final String property, final String value)
			throws IllegalStateException {
		// Check input parameters
		if (null == useruuid || null == property) {
			logger.severe("Null/Invalid arguments while updating Eco-account!");

			return false;
		}

		try {
			// Filter to find the Eco-account record to update.
			final Bson filter = Filters.eq(EcoAccounts.USERUUIDPROPERTY, useruuid);
			final Bson update = Updates.set(property, value);

			final UpdateResult updateresult = EcoAccountsDB.getCollection(EcoAccounts.USERSCOLLECTION).updateOne(filter,
					update);

			// Check result
			return updateresult.wasAcknowledged() && 1 == updateresult.getMatchedCount()
					&& 1 == updateresult.getModifiedCount();
		} catch (Exception e) {
			logger.severe("Could not update property value of " + property + " to Eco-account " + useruuid
					+ ", details: " + e.getMessage());

			return false;
		}
	}

	@Override
	public boolean updateEcoAccountBalance(String useruuid, double addecodebits, double subecocredits)
			throws IllegalStateException {
		// Check input parameters
		if (null == useruuid || 0.0 > addecodebits || 0.0 > subecocredits) {
			logger.severe("Null/Invalid arguments while updating Eco-account!");

			return false;
		}

		try {
			// Check that the Eco-account has enough Eco-credits to subtract
			final EcoAccountRecord ecoaccount = getEcoAccountRecord(useruuid);
			if (null == ecoaccount) {
				logger.fine(
						"Eco-account record document not found with " + EcoAccounts.USERUUIDPROPERTY + ": " + useruuid);

				return false;
			}

			if (subecocredits > ecoaccount.getEcoBalance().getEcoCredit().getValue().doubleValue()) {
				logger.fine("Eco-account " + useruuid + " does not have enough Eco-credits to subtract");

				return false;
			}

			// Filter to find the Eco-account record to update.
			final Bson filter = Filters.eq(EcoAccounts.USERUUIDPROPERTY, useruuid);
			final Bson ecodebitsupdate = Updates.inc(EcoAccounts.ECODEBITS, addecodebits),
					ecocreditsupdate = Updates.inc(EcoAccounts.ECOCREDITS, -1.0 * subecocredits);
			final Bson update = Updates.combine(ecodebitsupdate, ecocreditsupdate);

			final UpdateResult updateresult = EcoAccountsDB.getCollection(EcoAccounts.USERSCOLLECTION).updateOne(filter,
					update);

			// Check result
			return updateresult.wasAcknowledged() && 1 == updateresult.getMatchedCount()
					&& 1 == updateresult.getModifiedCount();
		} catch (Exception e) {
			logger.severe("Could not update balance of Eco-account " + useruuid + ", details: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Print the collection counts to log, taking into consideration localization.
	 * 
	 * @param dbname the database name.
	 * @param clname the collection name.
	 */
	private void printCollectionCounts(final String dbname, final String clname) {
		// Get the database handle
		final MongoDatabase db = mongoClient.getDatabase(dbname);

		// Get collection
		final MongoCollection<Document> collection = db.getCollection(clname);
		if (null == collection || 0L == collection.countDocuments()) {
			logger.info(dbname + "." + clname + " has no records");
			return;
		}

		for (final Document doc : collection.find()) {
			// Check if the collection is localized
			final String locale = doc.getString(LOCALE_JSON_PROPERTY);
			if (null == locale) {
				// Not localized collection, print its count and exit
				logger.info(dbname + "." + clname + " has no localization, records: " + collection.countDocuments());
				return;
			}

			// Get localized collection name
			final String localecollection = doc.getString(LOCALE_COLLECTION_PROPERTY);
			if (collectionExists(dbname, localecollection)) {
				// Print localized collection count
				logger.info(dbname + "." + clname + " has localization for " + locale + ", records: "
						+ db.getCollection(localecollection).countDocuments());
			} else {
				logger.severe("Locale " + locale + " has no existing collection " + localecollection);
			}
		}
	}

	/**
	 * Retrieve the document that contains the default Eco-credits calculation
	 * parameters for the specified product type HS key.
	 * 
	 * @param hskey the product type's HS key.
	 * @return the parameters object if parameters are found, <code>null</code>
	 *         otherwise.
	 */
	private Document getDefaultEcoCreditParameters(final Integer hskey) {
		if (null == hskey) {
			logger.warning("Tried to get default Weibull Lifespan Distribution parameters with null HS key!");

			return null;
		}

		try {
			final MongoCollection<Document> documents = MasterProductDataDB
					.getCollection(MasterProductData.ECOCREDITPARAMSCOLLECTION);

			if (null == documents || 0L == documents.countDocuments()) {
				throw new IllegalStateException(
						MasterProductData.ECOCREDITPARAMSCOLLECTION + " returned a NULL or empty collection!");
			}

			Document result = null;
			final Bson filter = Filters.eq(MasterProductData.PRODUCTTYPESHSKEY, hskey);
			for (final Document d : documents.find(filter)) {
				if (null == result) {
					// Initial loop, save the document that matched
					result = d;
				} else {
					// We have saved the first match and found another one, this is an exception.
					throw new IllegalStateException("Duplicate Eco-credits parameters records with "
							+ MasterProductData.PRODUCTTYPESHSKEY + ": " + hskey);
				}
			}

			if (null == result) {
				logger.fine(
						"Not found Eco-credits parameters with " + MasterProductData.PRODUCTTYPESHSKEY + ": " + hskey);

				return null;
			}

			return result;
		} catch (IllegalArgumentException e) {
			logger.severe(MasterProductData.ECOCREDITPARAMSCOLLECTION + " is invalid! Details: " + e.getMessage());

			return null;
		}
	}
}
