/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.dao.persistence;

import java.net.URL;

import gr.iccs.isense.circ4life.dao.CommonDAO.EcoAccountRecord;

/**
 * Data Access Object (DAO) interface for persistence layer data administration.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public interface PersistenceLayerAdminDAO extends PersistenceLayerDAO {
	/**
	 * Drop the specified database, effectively deleting all data collections it
	 * contains.
	 * 
	 * @param databasename the database to drop.
	 * @return <code>true</code> if the operation was successful, <code>false</code>
	 *         otherwise.
	 */
	public boolean dropDatabase(final String databasename);

	/**
	 * Drop all known databases of our application, as defined in
	 * {@link PersistenceLayerDAO}.
	 * 
	 * @param force if this is set to <code>true</code> the databases will be
	 *              dropped no matter if they exist or not. If set to
	 *              <code>false</code> they are dropped <strong>ONLY</strong> if
	 *              they are found.
	 * @return <code>true</code> if and only if all known databases are found and
	 *         dropped successfully, <code>false</code> otherwise.
	 */
	public boolean dropApplicationDatabases(final boolean force);

	/**
	 * Drop the specified collection from the specified database.
	 * 
	 * @param databasename   the database that contains the collection to drop.
	 * @param collectionname the collection to drop.
	 * @return <code>true</code> if the database collection was dropped
	 *         successfully, <code>false</code> otherwise.
	 */
	public boolean dropCollection(final String databasename, final String collectionname);

	/**
	 * Same functionality with
	 * {@link PersistenceLayerAdminDAO#importDatabases(String, boolean)} if
	 * <code>keepdata</code> is set to <code>true</code>. This method will not drop
	 * databases or collections, it will update the data with what is available in
	 * <code>importdirectory</code>.
	 * 
	 * @param importdirectory Directory that contains the import data in the
	 *                        hierarchy shown above.
	 * @return <code>true</code> if the import finishes successfully,
	 *         <code>false</code> otherwise.
	 */
	public boolean importDatabases(final String importdirectory);

	/**
	 * Import data into the persistence layer databases from a directory in the
	 * local host file-system. The directory should be prepared with the following
	 * structure:
	 * 
	 * <pre>
	IMPORT_DIR
	|_DatabaseName1
	    |_CollectionName1
	        |_JSONDocument1.json
	        |_...
	        |_JSONDocumentN.json
	    |_...
	    |_CollectionNameN
	|_...
	|_DatabaseNameN
	 * </pre>
	 * 
	 * The IMPORT_DIR is the <code>importdirectory</code> method parameter. The
	 * first level of directories (DatabaseName1 ... DatabaseNameN) underneath it
	 * are the database names, each one will create a new database if it does not
	 * already exist. The second level of directories (CollectionName1 ...
	 * CollectionNameN) are the collection names of each database. Finally the third
	 * level of JSON files are the Documents to import, one JSON file corresponds to
	 * one Document inserted into the corresponding collection.
	 * 
	 * You can keep the current database data or drop all databases with the
	 * parameter <code>keepdata</code>.
	 * 
	 * <strong>CAUTION:</strong> Be very careful when <code>keepdata</code> is set
	 * to <code>true</code>. The imported data may cause updates that will result in
	 * inconsistent records. Never mix different versions of the databases, the
	 * import directory should have data that is consistent with the schemas
	 * expected from the current databases.
	 * 
	 * @param importdirectory Directory that contains the import data in the
	 *                        hierarchy shown above.
	 * @param keepdata        if set to <code>false</code> all databases will be
	 *                        dropped. Set to <code>true</code> will keep all data
	 *                        and update records accordingly.
	 * @return <code>true</code> if the import finishes successfully,
	 *         <code>false</code> otherwise.
	 */
	public boolean importDatabases(final String importdirectory, boolean keepdata);

	/**
	 * Performs the same functionality as
	 * {@link PersistenceLayerAdminDAO#importDatabases(String, boolean)}, but the
	 * import data reside in the class-path as a resource.
	 * 
	 * @param importresourceURL the resource URL location that contains the import
	 *                          data in the hierarchy specified in
	 *                          {@link PersistenceLayerAdminDAO#importDatabases(String, boolean)}.
	 * @param keepdata          if set to <code>false</code> all databases will be
	 *                          dropped. Set to <code>true</code> will keep all data
	 *                          and update records accordingly.
	 * @return <code>true</code> if the import finishes successfully,
	 *         <code>false</code> otherwise.
	 */
	public boolean importDatabases(final URL importresourceURL, boolean keepdata);

	/**
	 * Create an Eco-account record.
	 * 
	 * @param ecoaccount the object containing all details of the record.
	 * @return <code>true</code> if the record was created successfully,
	 *         <code>false</code> otherwise.
	 */
	public boolean createEcoAccount(final EcoAccountRecord ecoaccount);

	/**
	 * Remove an Eco-account record.
	 * 
	 * @param useruuid the Eco-account user Universally Unique ID.
	 * @return <code>true</code> if Eco-account record removed successfully,
	 *         <code>false</code> otherwise.
	 */
	public boolean removeEcoAccount(final String useruuid);
}
