/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.dao.persistence;

import java.util.Collection;
import java.util.Date;

import gr.iccs.isense.circ4life.dao.CommonDAO.EcoAccountRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.IndividualProductData;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductType;
import gr.iccs.isense.circ4life.dao.CommonDAO.SystemFeedback;
import gr.iccs.isense.circ4life.dao.CommonDAO.UserProductFeedback;
import gr.iccs.isense.circ4life.dao.schema.LifespanDistributionParameters;
import gr.iccs.isense.circ4life.dao.schema.Material;
import gr.iccs.isense.circ4life.dao.schema.Parameters;

/**
 * Data Access Object (DAO) interface for persistence layer data retrieval.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public interface PersistenceLayerDAO {
	/**
	 * JSON property for language locale of multi-lingual records.
	 */
	static final String LOCALE_JSON_PROPERTY = "locale";

	/**
	 * JSON property for localized collection name of multi-lingual records.
	 */
	static final String LOCALE_COLLECTION_PROPERTY = "locale_collection";

	/**
	 * Default locale for multi-lingual records.
	 */
	static final String DEFAULTLOCALE = "en";

	/**
	 * 1.1.1.1 Master Product Data database details.
	 */
	static final class MasterProductData {
		public static final String DBNAME = "MasterProductData";
		public static final String PRODUCTSCOLLECTION = "Products";
		public static final String PRODUCTGTIN = "MasterData.ProductData.gtin14";
		public static final String PRODUCTHSKEY = "MasterData.LifeCycle.Resources.HSkey";
		public static final String PRODUCTTYPESCOLLECTION = "ProductTypes";
		public static final String PRODUCTTYPESHSKEY = "HSkey";
		public static final String PRODUCTTYPESUNUKEY = "UNUkey";
		public static final String PRODUCTTYPESDESCR = "Description";
		public static final String ECOCREDITPARAMSCOLLECTION = "EcoCreditParameters";
		public static final String ECOCREDITPARAMSRECYCLABILITY = "Recyclability";
		public static final String ECOCREDITPARAMSMATERIALS = "Material List";
		public static final String ECOCREDITPARAMSLIFESPANDISTRIB = "LifespanDistribution";
	}

	/**
	 * 1.1.1.2 User Eco-accounts database details.
	 */
	static final class EcoAccounts {
		public static final String DBNAME = "EcoAccounts";
		public static final String USERSCOLLECTION = "Users";
		public static final String USERUUIDPROPERTY = "EcoAccount.identifier";
		public static final String USEREMAILPROPERTY = "EcoAccount.email";
		public static final String RECYCLEBINUSERID = "EcoBalance.RecycleBinUserID";
		public static final String ECOCREDITS = "EcoBalance.EcoCredit.value";
		public static final String ECODEBITS = "EcoBalance.EcoDebit.value";
		public static final String RECYCLEDLIST = "EcoBalance.Recycled.itemListElement";
		public static final String PURCHASEDLIST = "EcoBalance.Purchased.itemListElement";
	}

	/**
	 * 1.1.1.3 Live Product Data database details. TODO: Remove Live Product Data
	 * references
	 */
	static final class LiveProductData {
		public static final String DBNAME = "LiveProductData";
		public static final String EVENTSCOLLECTION = "Events";
		public static final String PRODUCTSFEEDBACKCOLLECTION = "ProductsFeedback";
		public static final String SYSTEMFEEDBACKCOLLECTION = "SystemFeedback";
		public static final String FEEDBACKUSERID = "UserID";
		public static final String PRODUCTSFEEDBACKGTIN = "ProductID.GTIN14";
	}

	/**
	 * Check if the currently connected database server contains a database with the
	 * specified name.
	 * 
	 * @param dbname the database name to check for existence.
	 * @return <code>true</code> if the specified database exists,
	 *         <code>false</code> otherwise.
	 */
	public boolean databaseExists(final String dbname);

	/**
	 * Check if the currently connected database server contains the specified
	 * collection inside the specified database.
	 * 
	 * @param dbname         the database name to check for collection existence.
	 * @param collectionname the collection to check for existence.
	 * @return <code>true</code> if the specified collection is found inside the
	 *         specified database, <code>false</code> otherwise.
	 */
	public boolean collectionExists(final String dbname, final String collectionname);

	/**
	 * Get the Eco-account record from Eco-account user Universally Unique ID.
	 * 
	 * @param useruuid the Eco-account user Universally Unique ID.
	 * @return the {@link EcoAccountRecord} with data the correspond to
	 *         <code>useruuid</code> if found, <code>null</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. duplicate user records).
	 */
	public EcoAccountRecord getEcoAccountRecord(final String useruuid) throws IllegalStateException;

	/**
	 * Get the Eco-account record from Eco-account user Recycle Bin ID.
	 * 
	 * @param recycleBinUserID the Eco-account user Recycle Bin ID.
	 * @return the {@link EcoAccountRecord} with data the correspond to
	 *         <code>recycleBinUserID</code> if found, <code>null</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. duplicate user records).
	 */
	public EcoAccountRecord getEcoAccountRecord(final Long recycleBinUserID) throws IllegalStateException;

	/**
	 * Get the supported product types list from the Master Product Database.
	 * 
	 * @param locale text properties locale as ISO 639 alpha-2 language code, e.g.
	 *               'en', 'el', 'es', etc.
	 * @return a {@link Collection} of {@link ProductType} objects on successful
	 *         retrieval, <code>null</code> on failure.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public Collection<ProductType> getProductTypes(final String locale) throws IllegalStateException;

	/**
	 * Get the product type of a specific product in the Master Product Database.
	 * 
	 * @param GTIN14 the Global Trade Item Number of the product, in 14-digit
	 *               format.
	 * @return the product type if the product is found, <code>null</code>
	 *         otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public ProductType getProductType(final String GTIN14) throws IllegalStateException;

	/**
	 * Get the product type that corresponds to the specified HS key.
	 * 
	 * @param hskey the HS key.
	 * @return the corresponding product type if found, <code>null</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public ProductType getProductTypeFromHSkey(final Integer hskey) throws IllegalStateException;

	/**
	 * Get the product type that corresponds to the specified UNU key.
	 * 
	 * @param unukey the UNU key.
	 * @return the corresponding product type if found, <code>null</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public ProductType getProductTypeFromUNUkey(final Integer unukey) throws IllegalStateException;

	/**
	 * Get the full product record from the Master Product Database.
	 * 
	 * @param productid product ID (SGTIN).
	 * @param locale    text properties locale as ISO 639 alpha-2 language code,
	 *                  e.g. 'en', 'el', 'es', etc.
	 * @return the full product record if found, <code>null</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. duplicate product records).
	 */
	public ProductRecord getProductRecord(final String productid, final String locale) throws IllegalStateException;

	/**
	 * Get the default Weibull Lifespan Distribution parameters for the specified
	 * product type. It only applies to product type of Waste Electrical and
	 * Electronic Equipment (WEEEs).
	 * 
	 * @param hskey the product type's HS key.
	 * @return the parameters object if parameters are found, <code>null</code>
	 *         otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public LifespanDistributionParameters getDefaultLifespanDistributionParameters(final Integer hskey)
			throws IllegalStateException;

	/**
	 * Get the default Recyclability parameters.
	 * 
	 * @param hskey the product type's HS key.
	 * @return the object containing the default Recyclability parameters if found,
	 *         <code>null</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public Parameters getDefaultParameters(final Integer hskey) throws IllegalStateException;

	/**
	 * Get the default material factors for the specified product type.
	 * 
	 * @param hskey the product type's HS key.
	 * @return the list of material factors if found, or <code>null</code>
	 *         otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public Collection<Material> getDefaultMaterialFactors(final Integer hskey) throws IllegalStateException;

	/**
	 * Update the Eco-account of the specified user with a new recycled item, along
	 * with the Eco-credits gained by the recycling action. The Eco-credits are
	 * included in the item object
	 * 
	 * @param useruuid    the Eco-account user Universally Unique ID.
	 * @param recycledate the date the recycling took place, kept for reference.
	 * @param item        the individual item details.
	 * @return <code>true</code> if the transaction is successful,
	 *         <code>false</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public boolean updateEcoAccountWithRecycledItem(final String useruuid, final Date recycledate,
			final IndividualProductData item) throws IllegalStateException;

	/**
	 * Update the Eco-account of the specified user with purchase list, along with
	 * the Eco-debits of the purchase action.
	 * 
	 * @param useruuid   the Eco-account user Universally Unique ID.
	 * @param purchases  the purchased product list.
	 * @param ecodebits  the Eco-debits to charge the user by the retailer, based on
	 *                   the purchased list.
	 * @param ecocredits the Eco-credits to remove from the user Eco-account for the
	 *                   purchased list.
	 * @return <code>true</code> if the transaction is successful,
	 *         <code>false</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public boolean updateEcoAccountWithPurchases(final String useruuid,
			final Collection<IndividualProductData> purchases, final double ecodebits, final double ecocredits)
			throws IllegalStateException;

	/**
	 * Update the products user feedback in the LiveProductData database.
	 * 
	 * @param feedback the list of updated product feedback objects.
	 * @return true if the feedback was applied successfully for all items in the
	 *         provided list, <code>false</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public boolean updateProductsFeedback(final Collection<UserProductFeedback> feedback) throws IllegalStateException;

	/**
	 * Update the system user feedback in the LiveProductData database.
	 * 
	 * @param feedback the system feedback to update.
	 * @return true if the feedback was applied successfully, <code>false</code>
	 *         otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public boolean updateSystemFeedback(final SystemFeedback feedback) throws IllegalStateException;

	/**
	 * Update Eco-account with mapped Recycle Bin system user ID.
	 * 
	 * @param useruuid         the Eco-account user Universally Unique ID.
	 * @param recycleBinUserID the Recycle Bin system user ID.
	 * @return <code>true</code> if the transaction is successful,
	 *         <code>false</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public boolean updateEcoAccountWithRecycleBinUserID(final String useruuid, final long recycleBinUserID)
			throws IllegalStateException;

	/**
	 * Update a text field property of the Eco-account record of a specified user.
	 * 
	 * @param useruuid the Eco-account user Universally Unique ID.
	 * @param property the field property name. in JSON dotted notation (e.g.
	 *                 EcoAccount.image, EcoAccount.gender).
	 * @param value    the {@link String} value of the field property to set.
	 * @return <code>true</code> if the transaction is successful,
	 *         <code>false</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public boolean updateEcoAccountTextProperty(final String useruuid, final String property, final String value)
			throws IllegalStateException;

	/**
	 * Update Eco-account's balance by adding the specified Eco-debits and
	 * subtracting the Eco-credits.
	 * 
	 * @param useruuid      the Eco-account user Universally Unique ID.
	 * @param addecodebits  add this positive number to account's Eco-debits.
	 * @param subecocredits subtract this positive number from the account's
	 *                      Eco-credits.
	 * @return <code>true</code> if the transaction is successful,
	 *         <code>false</code> otherwise.
	 * @throws IllegalStateException in any case the persistence layer contains an
	 *                               illegal state (e.g. no records found).
	 */
	public boolean updateEcoAccountBalance(final String useruuid, final double addecodebits, final double subecocredits)
			throws IllegalStateException;
}
