/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.dao.persistence;

import java.util.logging.Logger;

import javax.naming.ConfigurationException;

import gr.iccs.isense.circ4life.logging.Log;
import gr.iccs.isense.circ4life.rest.config.ConfigProperties;

/**
 * Data Access Object (DAO) object factory for persistence layer data retrieval.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class PersistenceLayerDAOFactory {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(PersistenceLayerDAOFactory.class);

	/**
	 * Default singleton of persistence layer access.
	 */
	private static PersistenceLayerDAO DefaultInstancePersistenceLayerDAO = null;

	/**
	 * Default singleton of persistence layer administration.
	 */
	private static PersistenceLayerAdminDAO DefaultInstancePersistenceLayerAdminDAO = null;

	/**
	 * Get the default singleton access object of the {@link PersistenceLayerDAO}
	 * interface. Database server connection parameters are read from internal
	 * configuration file.
	 * 
	 * @return the default persistence layer singleton access object on successful
	 *         initialization, <code>null</code> otherwise.
	 */
	public static PersistenceLayerDAO getDefaultInstance() {
		try {
			// Lazy initialization
			if (null == DefaultInstancePersistenceLayerDAO) {
				// Get configuration parameters
				final String dbhost = ConfigProperties.getDatabaseHost(PersistenceLayerDAOFactory.class);
				if (null == dbhost) {
					throw new ConfigurationException("Configuration is missing required property for database host!");
				}

				final Integer dbport = ConfigProperties.getDatabasePort(PersistenceLayerDAOFactory.class);
				if (null == dbport) {
					throw new ConfigurationException("Configuration is missing required property for database port!");
				}

				// MongoDB is the default.
				DefaultInstancePersistenceLayerDAO = createDefault(dbhost, dbport);
			}

			return DefaultInstancePersistenceLayerDAO;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Create an object of the {@link PersistenceLayerDAO} interface with the
	 * default implementation.
	 * 
	 * @param dbhost The database server hostname or IP address.
	 * @param dbport The database server listening port.
	 * 
	 * @return an instance of the default persistence layer access object on
	 *         successful creation, <code>null</code> otherwise.
	 */
	private static PersistenceLayerDAO createDefault(final String dbhost, final Integer dbport) {
		try {
			// MongoDB is the default.
			return new MongoDBPersistenceLayerDAO(dbhost, dbport);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get the default singleton administration object of the
	 * {@link PersistenceLayerAdminDAO} interface. Database server connection
	 * parameters are read from internal configuration file.
	 * 
	 * @return the default persistence layer singleton administration object on
	 *         successful initialization, <code>null</code> otherwise.
	 */
	public static PersistenceLayerAdminDAO getDefaultAdminInstance() {
		try {
			// Lazy initialization
			if (null == DefaultInstancePersistenceLayerAdminDAO) {
				// Get configuration parameters
				final String dbhost = ConfigProperties.getDatabaseHost(PersistenceLayerDAOFactory.class);
				if (null == dbhost) {
					throw new ConfigurationException("Configuration is missing required property for database host!");
				}

				final Integer dbport = ConfigProperties.getDatabasePort(PersistenceLayerDAOFactory.class);
				if (null == dbport) {
					throw new ConfigurationException("Configuration is missing required property for database port!");
				}

				// MongoDB is the default.
				DefaultInstancePersistenceLayerAdminDAO = createDefaultAdmin(dbhost, dbport);
			}

			return DefaultInstancePersistenceLayerAdminDAO;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Get the default singleton administration object of the
	 * {@link PersistenceLayerAdminDAO} interface. Database server connection
	 * parameters are passed in method.
	 * 
	 * @param dbhost The database server hostname or IP address.
	 * @param dbport The database server listening port.
	 * @return the default persistence layer singleton administration object on
	 *         successful initialization, <code>null</code> otherwise.
	 */
	public static PersistenceLayerAdminDAO getDefaultAdminInstance(final String dbhost, final Integer dbport) {
		try {
			// Lazy initialization
			if (null == DefaultInstancePersistenceLayerAdminDAO) {
				// MongoDB is the default.
				DefaultInstancePersistenceLayerAdminDAO = createDefaultAdmin(dbhost, dbport);
			}

			return DefaultInstancePersistenceLayerAdminDAO;
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}

	/**
	 * Create an object of the {@link PersistenceLayerAdminDAO} interface with the
	 * default implementation.
	 * 
	 * @param dbhost The database server hostname or IP address.
	 * @param dbport The database server listening port.
	 * @return an instance of the default persistence layer administration object on
	 *         successful creation, <code>null</code> otherwise.
	 */
	private static PersistenceLayerAdminDAO createDefaultAdmin(final String dbhost, final Integer dbport) {
		try {
			// MongoDB is the default.
			return new MongoDBPersistenceLayerAdminDAO(dbhost, dbport);
		} catch (Exception e) {
			logger.severe(e.getMessage());

			return null;
		}
	}
}
