/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.json;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;

import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.ObjectMapper;

/**
 * Jackson annotated serialization utilities for JSON objects.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class JacksonJSONSerialization {
	/**
	 * Try to serialize JSON annotated object to {@link String}.
	 * 
	 * @param o the JSON annotated object.
	 * @return {@link String} representation of the object if serialization
	 *         succeeds, <code>null</code> otherwise.
	 */
	public static String JSONtoString(Object o) {
		final JsonFactory messagePackFactory = new JsonFactory();
		messagePackFactory.configure(JsonGenerator.Feature.AUTO_CLOSE_TARGET, true);
		final ObjectMapper objmapper = new ObjectMapper(messagePackFactory);
		final ByteArrayOutputStream byteos = new ByteArrayOutputStream();

		try {
			objmapper.writeValue(byteos, o);
			final String result = byteos.toString(StandardCharsets.UTF_8.toString());

			return result;
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Same as {@link JacksonJSONSerialization#JSONtoString(Object)} but the result
	 * is formatted with the pretty writer default settings.
	 * 
	 * @param o the JSON annotated object.
	 * @return {@link String} representation of the object if serialization
	 *         succeeds, <code>null</code> otherwise.
	 */
	public static String JSONtoPrettyString(Object o) {
		final JsonFactory messagePackFactory = new JsonFactory();
		messagePackFactory.configure(JsonGenerator.Feature.AUTO_CLOSE_TARGET, true);
		final ObjectMapper objmapper = new ObjectMapper(messagePackFactory);
		final ByteArrayOutputStream byteos = new ByteArrayOutputStream();

		try {
			objmapper.writerWithDefaultPrettyPrinter().writeValue(byteos, o);
			final String result = byteos.toString(StandardCharsets.UTF_8.toString());

			return result;
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Try to de-serialize a {@link String} representing a JSON document to a JSON
	 * annotated object of specified class.
	 * 
	 * @param <T>      the JSON annotated object type.
	 * @param jsondata the {@link String} containing the JSON document.
	 * @param type     the JSON annotated object class.
	 * @return the de-serialized object of specified type on success.
	 * @throws IOException if the JSON data {@link String} cannot be parsed.
	 */
	public static <T> T StringtoJSON(final String jsondata, final Class<T> type) throws IOException {
		final JsonFactory messagePackFactory = new JsonFactory();
		messagePackFactory.configure(JsonGenerator.Feature.AUTO_CLOSE_TARGET, true);
		final ObjectMapper objmapper = new ObjectMapper(messagePackFactory);

		return objmapper.readerFor(type).readValue(jsondata);
	}
}
