/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life;

/**
 * Eco-credit formula common interface for parameters.
 *
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public interface EcoCreditParameters {
	/**
	 * Recycled product types.
	 */
	public static enum ProductWasteType {
		/**
		 * Waste Electrical and Electronic Equipment (WEEEs) recycled products
		 */
		ELECTRICAL_ELECTRONICS_WASTE,
		/**
		 * Food or Organic Urban waste recycled products
		 */
		FOOD_OR_ORGANIC_WASTE;
	}

	/**
	 * Get the product's waste type, calculation formulas are different between
	 * different waste types.
	 *
	 * @return The product waste type as defined in possible values of
	 *         {@link ProductWasteType}.
	 */
	public ProductWasteType getProductWasteType();

	/**
	 * Get the value of the constant A weight parameter of Eq. 1 in Deliverable 2.4.
	 * This parameter controls the weight factor of the sum of products between
	 * Materials thermodynamic rarity and recycle-ability factors. It does not apply
	 * to waste type of {@link ProductWasteType#FOOD_OR_ORGANIC_WASTE} and is always
	 * 0 in this case.
	 *
	 * @return the value of the constant A weight parameter of Eq. 1 in Deliverable
	 *         2.4.
	 */
	public double getWeightParameterA();

	/**
	 * Get the value of the constant B weight parameter of Eq. 1 in Deliverable 2.4.
	 * This parameter controls the weight factor of the End of Life state. It does
	 * not apply to waste type of {@link ProductWasteType#FOOD_OR_ORGANIC_WASTE} and
	 * is always 0 in this case.
	 *
	 * @return the value of the constant B weight parameter of Eq. 1 in Deliverable
	 *         2.4.
	 */
	public double getWeightParameterB();

	/**
	 * Get the value of the constant C weight parameter of Eq. 1 in Deliverable 2.4.
	 * This parameter controls the weight factor of the Lifetime factor. It does not
	 * apply to waste type of {@link ProductWasteType#FOOD_OR_ORGANIC_WASTE} and is
	 * always 0 in this case.
	 *
	 * @return the value of the constant C weight parameter of Eq. 1 in Deliverable
	 *         2.4.
	 */
	public double getWeightParameterC();

	/**
	 * Get the value of the constant D weight parameter of Eq. 3 in Deliverable 2.4.
	 * This parameter controls the weight factor of the chemical exergy. It does not
	 * apply to waste type of {@link ProductWasteType#ELECTRICAL_ELECTRONICS_WASTE}
	 * and is always 0 in this case.
	 *
	 * @return the value of the constant D weight parameter of Eq. 3 in Deliverable
	 *         2.4.
	 */
	public double getWeightParameterD();

	/**
	 * Helper function for choosing between constant weight parameter A or D
	 * according to the recycled waste type this object corresponds to.
	 *
	 * @return the value of {@link EcoCreditParameters#getWeightParameterA()} in
	 *         case this object corresponds to
	 *         {@link ProductWasteType#ELECTRICAL_ELECTRONICS_WASTE} and the value
	 *         {@link EcoCreditParameters#getWeightParameterD()} in case this object
	 *         corresponds to {@link ProductWasteType#FOOD_OR_ORGANIC_WASTE}.
	 */
	public double getWeightOfExergy();

	/**
	 * Get the value of the exergy analysis, which is the Materials Thermodynamic
	 * Rarity to Recycle-ability factor product sum or the Chemical exergy of
	 * organic waste. The value depends on the type of {@link ProductWasteType} this
	 * object corresponds to. This value should be multiplied with the result of
	 * {@link EcoCreditParameters#getWeightParameterA()} or
	 * {@link EcoCreditParameters#getWeightParameterD()} accordingly.
	 *
	 * @return the value of the exergy analysis depending on the type of
	 *         {@link ProductWasteType} this object corresponds to.
	 */
	public double getExergy();

	/**
	 * Get the End of Life (EoL) state of the recycled product. This value should be
	 * multiplied with the result of
	 * {@link EcoCreditParameters#getWeightParameterB()} as defined by Eq. 1 in
	 * Deliverable 2.4.
	 *
	 * @return the value of the End of Life (EoL) state.
	 */
	public double getEndofLifeState();

	/**
	 * Get the Lifetime factor of the recycled product. This value should be
	 * multiplied with the result of
	 * {@link EcoCreditParameters#getWeightParameterC()} as defined by Eq. 1 in
	 * Deliverable 2.4.
	 *
	 * @return the value of the Lifetime factor.
	 */
	public double getLifetimeFactor();
}
