/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life;

/**
 * Class that encapsulates the parameters and factors of Food and Urban Organic
 * product waste. Objects of this class are used to calculate Eco-points with
 * the formula Eq. 3 of Deliverable 2.4:
 *
 * Eco-credits = D * b_chemical_exergy * m_organic_waste
 *
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class FoodParameters implements EcoCreditParameters {
	/**
	 * The coefficient D will be responsible to convert the energy contained in the
	 * wastes into the incentive given to the final user and it can modified and
	 * adjusted to different incentive programs. The parameter D will be applied in
	 * the expression in order to balance the expected incentives associated to
	 * Eco-credits to the final incentive provided.
	 */
	private final double weight_param_D;

	/**
	 * Organic urban waste chemical exergy for a unit of mass.
	 */
	private final double b_chemical_exergy;

	/**
	 * Organic waste mass.
	 */
	private final double m_organic_waste;

	/**
	 * No default constructor, it does not make sense. All parameters should be
	 * provided by user.
	 */
	@SuppressWarnings("unused")
	private FoodParameters() {
		weight_param_D = -1.0;
		b_chemical_exergy = -1.0;
		m_organic_waste = -1.0;
	}

	/**
	 * Constructor.
	 *
	 * @param D                  The coefficient D factor, responsible for
	 *                           converting the exergy in the wastes to user
	 *                           incentive for proper waste disposal.
	 * @param chemical_exergy    Organic urban waste chemical exergy, which is the
	 *                           energy contained in the waste.
	 * @param mass_organic_waste Mass of waste. The unit of measurement will change
	 *                           the scale of the final Eco-credit result, so in
	 *                           order to compare Eco-credits the same mass unit
	 *                           (Kg, g) should be used.
	 */
	public FoodParameters(double D, double chemical_exergy, double mass_organic_waste) {
		weight_param_D = D;
		b_chemical_exergy = chemical_exergy;
		m_organic_waste = mass_organic_waste;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getProductWasteType()
	 */
	@Override
	public ProductWasteType getProductWasteType() {
		return ProductWasteType.FOOD_OR_ORGANIC_WASTE;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getWeightParameterA()
	 */
	@Override
	public double getWeightParameterA() {
		// In food/organic waste Eq. 1 formula A factor is the same as Eq. 3 D factor.
		// We return the value of D just in case.
		return getWeightParameterD();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getWeightParameterB()
	 */
	@Override
	public double getWeightParameterB() {
		// In food/organic waste B does not apply, so it is zero.
		return 0.0;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getWeightParameterC()
	 */
	@Override
	public double getWeightParameterC() {
		// In food/organic waste C does not apply, so it is zero.
		return 0.0;
	}

	/**
	 * @return the b_chemical_exergy.
	 */
	public double getChemicalExergy() {
		return b_chemical_exergy;
	}

	/**
	 * @return the m_organic_waste.
	 */
	public double getOrganicWasteMass() {
		return m_organic_waste;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getWeightParameterD()
	 */
	@Override
	public double getWeightParameterD() {
		return weight_param_D;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getWeightOfExergy()
	 */
	@Override
	public double getWeightOfExergy() {
		// In organic waste this is equal to D (or A in our implementation).
		return getWeightParameterD();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getExergy()
	 */
	@Override
	public double getExergy() {
		return b_chemical_exergy * m_organic_waste;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getEndofLifeState()
	 */
	@Override
	public double getEndofLifeState() {
		// In food/organic waste EoL does not apply, so it is zero.
		return 0.0;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getLifetimeFactor()
	 */
	@Override
	public double getLifetimeFactor() {
		// In food/organic waste Lifetime Factor does not apply, so it is zero.
		return 0.0;
	}
}
