/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life;

/**
 * Class that encapsulates the parameters and calculation of the Lifetime Factor
 * for Waste Electrical and Electronic Equipment. Objects of this class should
 * be used as input to the {@link WEEEsParameters} constructor. All input
 * parameters are constants during the life of the object.
 *
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class WEEEsLifetimeFactorCalculator {
	/**
	 * Years of device usage thus far.
	 */
	private final int device_years_of_use;

	/**
	 * Additional months of device usage to be added on
	 * {@link WEEEsParameters#device_years_of_use}.
	 */
	private final int device_months_of_use;

	/**
	 * Weibull Lifespan distribution alpha shape parameter. Constant value per
	 * UNU-KEY Code WEEE device type.
	 */
	private final double alpha;

	/**
	 * Weibull Lifespan distribution beta scale parameter. Constant value per
	 * UNU-KEY Code WEEE device type.
	 */
	private final double beta;

	/**
	 * No default constructor, it does not make sense. All parameters should be
	 * provided by user.
	 */
	@SuppressWarnings("unused")
	private WEEEsLifetimeFactorCalculator() {
		// set invalid values
		device_years_of_use = -1;
		device_months_of_use = -1;
		alpha = -1.0;
		beta = -1.0;
	}

	/**
	 * Constructor. All input parameters are set to final member variables in order
	 * to make them constant for this object.
	 *
	 * @param device_years_of_use  Years of device usage thus far.
	 * @param device_months_of_use Additional months of device usage to add on
	 *                             device_years_of_use.
	 * @param alpha                Weibull Lifespan distribution alpha parameter for
	 *                             a specific device type.
	 * @param beta                 Weibull Lifespan distribution beta parameter for
	 *                             a specific device type.
	 */
	public WEEEsLifetimeFactorCalculator(int device_years_of_use, int device_months_of_use, double alpha, double beta)
			throws IllegalArgumentException {
		if (device_years_of_use >= 0) {
			this.device_years_of_use = device_years_of_use;
		} else {
			throw new IllegalArgumentException("Invalid value for device_years_of_use: " + device_years_of_use);
		}
		if (device_months_of_use >= 0) {
			this.device_months_of_use = device_months_of_use;
		} else {
			throw new IllegalArgumentException("Invalid value for device_months_of_use: " + device_months_of_use);
		}
		if (alpha >= 0) {
			this.alpha = alpha;
		} else {
			throw new IllegalArgumentException("Invalid value for alpha: " + alpha);
		}
		if (beta >= 0) {
			this.beta = beta;
		} else {
			throw new IllegalArgumentException("Invalid value for beta: " + beta);
		}
	}

	/**
	 * Get the time of use in months for Lifetime Factor calculation.
	 *
	 * getTimeOfUseMonths = device_years_of_use * 12 + device_months_of_use
	 *
	 * @return the WEEE product's time of use in months.
	 */
	public double getTimeOfUseMonths() {
		return device_years_of_use * 12 + device_months_of_use;
	}

	/**
	 * Get the maximum X axis time value of the Weibull Lifespan distribution in
	 * years.
	 *
	 * @return the maximum X axis time value of the Weibull Lifespan distribution in
	 *         years.
	 */
	public double getTimeOfUseMaxYears() {
		return beta * Math.pow(((alpha - 1.0) / alpha), (1.0 / alpha));
	}

	/**
	 * Get the Y axis value of the Weibull Lifespan distribution for specified X
	 * years in fractional format.
	 *
	 * @param X_years The X axis time value in fractional years.
	 * @return the Y axis value of the Weibull Lifespan distribution value.
	 */
	public double getWeibullDistributionValue(double X_years) {
		final double X_div_beta = X_years / beta;
		final double X_factor = Math.pow(X_div_beta, alpha - 1.0);
		final double Exponent_factor = Math.exp(-1.0 * Math.pow(X_div_beta, alpha));

		return (alpha / beta) * X_factor * Exponent_factor;
	}

	/**
	 * Get the Lifetime Factor of the product by specifying the time of use in
	 * months.
	 *
	 * @return the Lifetime Factor of the product.
	 * @throws IllegalArgumentException thrown if usage time is wrong.
	 */
	public double getLifetimeFactor() throws IllegalArgumentException {
		// The Lifetime Factor is not a single formula function is has the following
		// cases (X is the time_of_use_months converted to years):
		// 1. 0 <= X < Xmax
		// 2. X = Xmax
		// 3. Xmax < X < 1.25 * Xmax
		// 4. X = 1.25 * Xmax
		// 5. X > 1.25 * Xmax

		final double use_months = getTimeOfUseMonths();
		final double X = use_months / 12.0;
		final double Xmax = getTimeOfUseMaxYears();
		final double Y = getWeibullDistributionValue(X);
		final double Ymax = getWeibullDistributionValue(Xmax);

		if ((0 <= X && X < Xmax) || (X > 1.25 * Xmax)) { // case 1 or 5
			return Y / Ymax;
		} else if (X == Xmax) { // case 2
			return 1.03;
		} else if (Xmax < X && X < 1.25 * Xmax) { // case 3
			return 1.03 + (0.08 * (X - Xmax) / Xmax);
		} else if (X == 1.25 * Xmax) { // case 4
			return 1.05;
		} else {
			// Throw exception for invalid time_of_use_months value
			throw new IllegalArgumentException("Invalid parameters for Lifetime Factor calculation: "
					+ "Weibull distribution Alpha= " + alpha + ", Beta= " + beta + ", " + device_years_of_use
					+ " years, " + device_months_of_use + " months");
		}
	}
}
