/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life;

import gr.iccs.isense.circ4life.datamodels.WEEEsEndOfLifeState;

/**
 * Class that encapsulates the parameters and factors of Waste Electrical and
 * Electronic Equipment. Objects of this class are used to calculate Eco-points
 * with the formula Eq. 1 of Deliverable 2.4.
 *
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class WEEEsParameters implements EcoCreditParameters {
	/**
	 * Constant parameter weight A to be multiplied with product materials exergy.
	 */
	private final double weight_param_A;

	/**
	 * Calculator of product material exergy, which is the sum of material
	 * recyclability and rarity products.
	 */
	private final WEEEsMaterialThermodynamicRarityCalculator exergy_calculator;

	/**
	 * Constant parameter weight B to be multiplied with End of Life state factor.
	 */
	private final double weight_param_B;

	/**
	 * End of Life state factor.
	 */
	private final WEEEsEndOfLifeState endoflife_state;
	/**
	 * Constant parameter weight C to be multiplied by Lifetime Factor.
	 */
	private final double weight_param_C;

	/**
	 * Lifetime Factor calculator.
	 */
	private final WEEEsLifetimeFactorCalculator lifetime_factor;

	/**
	 * No default constructor, it does not make sense. All parameters should be
	 * provided by user.
	 */
	@SuppressWarnings("unused")
	private WEEEsParameters() {
		weight_param_A = -1.0;
		exergy_calculator = null;
		weight_param_B = -1.0;
		endoflife_state = null;
		weight_param_C = -1.0;
		lifetime_factor = null;
	}

	/**
	 * Constructor.
	 *
	 * @param A            Constant parameter weight A to be multiplied with product
	 *                     materials exergy.
	 * @param materialcalc Calculator of product material exergy.
	 * @param B            Constant parameter weight B to be multiplied with End of
	 *                     Life state factor.
	 * @param eol          End of Life state factor.
	 * @param C            Constant parameter weight C to be multiplied by Lifetime
	 *                     Factor.
	 * @param ltcalc       Lifetime Factor calculator.
	 */
	public WEEEsParameters(double A, final WEEEsMaterialThermodynamicRarityCalculator materialcalc, double B,
			final WEEEsEndOfLifeState eol, double C, final WEEEsLifetimeFactorCalculator ltcalc) {
		weight_param_A = A;
		exergy_calculator = materialcalc;
		weight_param_B = B;
		endoflife_state = eol;
		weight_param_C = C;
		lifetime_factor = ltcalc;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getProductWasteType()
	 */
	@Override
	public ProductWasteType getProductWasteType() {
		return ProductWasteType.ELECTRICAL_ELECTRONICS_WASTE;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getWeightParameterA()
	 */
	@Override
	public double getWeightParameterA() {
		return weight_param_A;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getWeightParameterB()
	 */
	@Override
	public double getWeightParameterB() {
		return weight_param_B;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getWeightParameterC()
	 */
	@Override
	public double getWeightParameterC() {
		return weight_param_C;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getWeightParameterD()
	 */
	@Override
	public double getWeightParameterD() {
		// Always 0 in WEEEs
		return 0;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getWeightOfExergy()
	 */
	@Override
	public double getWeightOfExergy() {
		return getWeightParameterA();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getExergy()
	 */
	@Override
	public double getExergy() {
		return exergy_calculator.getMaterialsExergy();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getEndofLifeState()
	 */
	@Override
	public double getEndofLifeState() {
		return endoflife_state.Value();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see gr.iccs.isense.circ4life.EcoCreditParameters#getLifetimeFactor()
	 */
	@Override
	public double getLifetimeFactor() {
		return lifetime_factor.getLifetimeFactor();
	}
}
