/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.test;

import java.util.Arrays;
import java.util.Vector;

import org.junit.Assert;
import org.junit.Test;

import gr.iccs.isense.circ4life.EcoCreditCalculator;
import gr.iccs.isense.circ4life.EcoCreditParameters;
import gr.iccs.isense.circ4life.WEEEsLifetimeFactorCalculator;
import gr.iccs.isense.circ4life.WEEEsMaterialThermodynamicRarityCalculator;
import gr.iccs.isense.circ4life.datamodels.MaterialFactors;
import gr.iccs.isense.circ4life.datamodels.WEEEsEndOfLifeState;
import gr.iccs.isense.circ4life.WEEEsParameters;

/**
 * Tests for EcoCreditCalculator class. All correct result values are provided
 * by our partners at CIRCE.
 *
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EcoCreditCalculatorTests {
	/**
	 * WEEEs product test encapsulation.
	 */
	public final class WEEEsProductCalculatorTest {
		public final double expectedresult;
		public final int years;
		public final int months;
		public final WEEEsEndOfLifeState eolstate;
		public final MaterialFactors materialfactors[];
		public final double LTA;
		public final double LTB;
		public final double WA;
		public final double WB;
		public final double WC;

		/**
		 * Constructor.
		 * 
		 * @param expectedresult  The expected Eco-credits result.
		 * @param years           years of product usage.
		 * @param months          extra months of usage added to years.
		 * @param eolstate        the End of Life state.
		 * @param materialfactors material factors.
		 * @param LTA             Lifetime distribution Alpha.
		 * @param LTB             Lifetime distribution Beta.
		 * @param WA              the weight factor of the material exergy.
		 * @param WB              the weight factor of the End of Life state.
		 * @param WC              the weight factor of the Lifetime factor.
		 */
		public WEEEsProductCalculatorTest(double expectedresult, int years, int months, WEEEsEndOfLifeState eolstate,
				MaterialFactors[] materialfactors, double LTA, double LTB, double WA, double WB, double WC) {
			this.expectedresult = expectedresult;
			this.years = years;
			this.months = months;
			this.eolstate = eolstate;
			this.materialfactors = materialfactors;
			this.LTA = LTA;
			this.LTB = LTB;
			this.WA = WA;
			this.WB = WB;
			this.WC = WC;
		}
	}

	/**
	 * Test the result of the Lifetime Factor calculation for a single tablet.
	 */
	@Test
	public void testLifetimeFactorTablet() {
		final double tablet_alpha = 1.5;
		final double tablet_beta = 5.2;

		// Case 1: 1 tablet with 2 years and 7 months = 31 months
		{
			final double correct_lifetime_factor = 1.03267;
			final int tablet_years = 2;
			final int tablet_months = 7;

			final WEEEsLifetimeFactorCalculator calc = new WEEEsLifetimeFactorCalculator(tablet_years, tablet_months,
					tablet_alpha, tablet_beta);
			final double calculated = calc.getLifetimeFactor();

			Assert.assertEquals(correct_lifetime_factor, calculated, TestsCommon.fractional_max_accepted_delta);
		}

		// Case 2: 1 tablet with 3 years and 2 months = 38 months
		{
			final double correct_lifetime_factor = 0.9766018;
			final int tablet_years = 3;
			final int tablet_months = 2;

			final WEEEsLifetimeFactorCalculator calc = new WEEEsLifetimeFactorCalculator(tablet_years, tablet_months,
					tablet_alpha, tablet_beta);
			final double calculated = calc.getLifetimeFactor();

			Assert.assertEquals(correct_lifetime_factor, calculated, TestsCommon.fractional_max_accepted_delta);
		}

		// Case 3: For 1 tablet with 8 months = 8 months
		{
			final double correct_lifetime_factor = 0.6883697;
			final int tablet_years = 0;
			final int tablet_months = 8;

			final WEEEsLifetimeFactorCalculator calc = new WEEEsLifetimeFactorCalculator(tablet_years, tablet_months,
					tablet_alpha, tablet_beta);
			final double calculated = calc.getLifetimeFactor();

			Assert.assertEquals(correct_lifetime_factor, calculated, TestsCommon.fractional_max_accepted_delta);
		}
	}

	/**
	 * Test the result of the Materials Thermodynamic Rarity calculation for a
	 * single tablet.
	 */
	@Test
	public void testMaterialExergyTablet() {
		// Case 1: Working or repairable tablet (not recycled)
		{
			final WEEEsMaterialThermodynamicRarityCalculator calc = new WEEEsMaterialThermodynamicRarityCalculator(
					new Vector<MaterialFactors>(Arrays.asList(TestsCommon.working_tablet_material_factors)));

			final double correct_materials_exergy = 61448.3;
			final double calculated = calc.getMaterialsExergy();

			Assert.assertEquals(correct_materials_exergy, calculated, TestsCommon.integer_max_accepted_delta);
		}

		// Case 2: Broken tablet (to be recycled)
		{
			final WEEEsMaterialThermodynamicRarityCalculator calc = new WEEEsMaterialThermodynamicRarityCalculator(
					new Vector<MaterialFactors>(Arrays.asList(TestsCommon.broken_tablet_material_factors)));

			final double correct_materials_exergy = 42545.4;
			final double calculated = calc.getMaterialsExergy();

			Assert.assertEquals(correct_materials_exergy, calculated, TestsCommon.integer_max_accepted_delta);
		}
	}

	/**
	 * Test the result of the Eco-credits calculation for a single tablet.
	 */
	@Test
	public void testEcoCreditsTablet() {
		final EcoCreditCalculator calculator = new EcoCreditCalculator();

		// Tablet case, UNU-Key = 0303
		// alpha and beta have different values per UNU-key or product type
		// Lifetime Weibull Distribution alpha = 1.5 and beta = 5.2;
		// Weight constants A = 0.00006224, B = 4.706, C = 15.0;
		final WEEEsProductCalculatorTest tests[] = {
				new WEEEsProductCalculatorTest(24.020598907823427, 2, 7, WEEEsEndOfLifeState.WORKING,
						TestsCommon.working_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(21.66759890782343, 2, 7, WEEEsEndOfLifeState.REUSABLE,
						TestsCommon.working_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(18.138076406015085, 2, 7, WEEEsEndOfLifeState.BROKEN,
						TestsCommon.broken_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(23.179575898058356, 3, 2, WEEEsEndOfLifeState.WORKING,
						TestsCommon.working_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(20.82657589805836, 3, 2, WEEEsEndOfLifeState.REUSABLE,
						TestsCommon.working_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(17.297053396250014, 3, 2, WEEEsEndOfLifeState.BROKEN,
						TestsCommon.broken_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(18.85609437248686, 0, 8, WEEEsEndOfLifeState.WORKING,
						TestsCommon.working_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(16.503094372486864, 0, 8, WEEEsEndOfLifeState.REUSABLE,
						TestsCommon.working_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(12.97357187067852, 0, 8, WEEEsEndOfLifeState.BROKEN,
						TestsCommon.broken_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(8.530548936388639, 0, 0, WEEEsEndOfLifeState.WORKING,
						TestsCommon.working_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(6.177548936388639, 0, 0, WEEEsEndOfLifeState.REUSABLE,
						TestsCommon.working_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0),
				new WEEEsProductCalculatorTest(2.648026434580296, 0, 0, WEEEsEndOfLifeState.BROKEN,
						TestsCommon.broken_tablet_material_factors, 1.5, 5.2, 0.00006224, 4.706, 15.0) };

		for (final WEEEsProductCalculatorTest test : tests) {
			System.out.println("---- " + test.eolstate + " tablet -----------");
			final WEEEsMaterialThermodynamicRarityCalculator materialcalc = new WEEEsMaterialThermodynamicRarityCalculator(
					new Vector<MaterialFactors>(Arrays.asList(test.materialfactors)));
			final WEEEsLifetimeFactorCalculator ltcalc = new WEEEsLifetimeFactorCalculator(test.years, test.months,
					test.LTA, test.LTB);
			final EcoCreditParameters params = new WEEEsParameters(test.WA, materialcalc, test.WB, test.eolstate,
					test.WC, ltcalc);

			// Perform the calculation
			final double calculated = calculator.calculate(params);

			System.out.println("Usage " + test.years + " years and " + test.months + " months");
			System.out.println("Weight parameters, A = " + test.WA + ", B = " + test.WB + ", C = " + test.WC);
			System.out.println("Materials exergy = " + materialcalc.getMaterialsExergy());
			System.out.println("Lifetime Factor = " + ltcalc.getLifetimeFactor());
			System.out.println("Eco-credits = " + calculated);
			System.out.println("Expected = " + test.expectedresult);

			Assert.assertEquals(test.expectedresult, calculated, TestsCommon.fractional_max_accepted_delta);

			System.out.println("---------------------------------------------");
		}
	}

	/**
	 * For debugging this test class as an application.
	 *
	 * @param args Command line arguments array.
	 */
	public static void main(String[] args) {
		final EcoCreditCalculatorTests instance = new EcoCreditCalculatorTests();

		instance.testLifetimeFactorTablet();
		instance.testMaterialExergyTablet();
		instance.testEcoCreditsTablet();
	}
}
