/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.integration.test;

import java.net.URI;
import java.net.URISyntaxException;

import javax.naming.ConfigurationException;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import gr.iccs.isense.circ4life.client.EcoCreditCalculatorClient;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerAdminDAO;
import gr.iccs.isense.circ4life.dao.persistence.PersistenceLayerDAOFactory;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;
import gr.iccs.isense.circ4life.test.IntegrationTestsCommon;
import gr.iccs.isense.circ4life.test.IntegrationTestsCommon.TestUser;
import gr.iccs.isense.circ4life.test.TestsCommon;
import gr.iccs.isense.circ4life.test.TestsCommon.EcoCreditCalculatorProductTypeTest;

/**
 * Integration tests for the Eco-credit calculation RESTful Web Service client.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EcoCreditCalculatorClientIT {
	/**
	 * REST service details.
	 */
	private static final String SERVICE_URI = "http://127.0.0.1:8081/EcoCreditCalculator/resources";
	private static final String SERVICE_DISPNAME = "EcoCreditCalculatorWS";
	private static final String SERVICE_CLASSNAME = "gr.iccs.isense.circ4life.rest.EcoCreditCalculatorModule";
	private static final String SERVICE_ACCESSCTRLSKIP = "^/resources/(application.wadl|info/.*)";

	private PersistenceLayerAdminDAO admindbdao = null;

	@Before
	public void beforeTests() throws ConfigurationException {
		// Check that we can connect to database server in order to quit early and
		// inform user running the integration tests
		IntegrationTestsCommon.checkDatabaseServerConnection();

		if (null == admindbdao) {
			admindbdao = PersistenceLayerDAOFactory.getDefaultAdminInstance();
		}
		Assert.assertNotNull(admindbdao);

		// Import everything from the src/it/resources/databases directory
		// TODO: Start a fresh instance of a MongoDB server for integration testing that
		// does not have any of our application databases. The keepdata = true here
		// means that all imported data are imported without looking for existing
		// application databases.
		Assert.assertTrue(admindbdao.importDatabases(
				IntegrationTestsCommon.getDatabaseImportDirectory(EcoCreditCalculatorClientIT.class), true));

		// Perform the integration testing by launching the REST service from current
		// source code. The following lines will launch a server using the source code
		// that is currently checkout out from the project EcoCreditCalculatorWS
		IntegrationTestsCommon.startLocalAppServer(SERVICE_DISPNAME, SERVICE_CLASSNAME, SERVICE_URI,
				SERVICE_ACCESSCTRLSKIP);

		while (!IntegrationTestsCommon.isLocalAppServerReady()) {
			System.out.println("Waiting for local application server to start...");
			try {
				Thread.sleep(2000);
			} catch (InterruptedException e) {
				e.printStackTrace();
				break;
			}
		}

		// Create the user in Access Control Manager
		Assert.assertTrue(AccessControlManager.createEcoAccountUser(TestUser.Username, TestUser.Password,
				TestUser.Email, TestUser.Firstname, TestUser.Lastname, "en"));
		final String useruuid = AccessControlManager.getUsernameUUID(TestUser.Username);
		Assert.assertNotNull(useruuid);
		// Verify manually for integration testing
		Assert.assertTrue(AccessControlManager.setUserEmailVerified(useruuid, true));
	}

	@After
	public void afterTests() {
		if (IntegrationTestsCommon.isLocalAppServerReady()) {
			IntegrationTestsCommon.stopLocalAppServer();
		}

		// If we reach here, then a successful import is performed during this
		// integration test and we should have all application databases setup properly.
		if (null != admindbdao) {
			Assert.assertTrue(admindbdao.dropApplicationDatabases(true));
		}

		final String useruuid = AccessControlManager.getUsernameUUID(TestUser.Username);
		Assert.assertNotNull(useruuid);
		Assert.assertTrue(AccessControlManager.removeEcoAccountUser(useruuid));

		// Do not leave stale session of Admin REST Client
		AccessControlManager.closeAdminAPIClient();
	}

	@Test
	public void itEcoCreditsCalculatorWS() throws URISyntaxException {
		System.out.println("Testing client against REST service " + SERVICE_URI);

		// Create the client
		final EcoCreditCalculatorClient ecocreditsclient = new EcoCreditCalculatorClient(new URI(SERVICE_URI),
				TestUser.Username, TestUser.Password);

		// Login
		Assert.assertTrue("User login failed, could not start session!", ecocreditsclient.startSession());

		// Food waste
		Assert.assertEquals(6.0, ecocreditsclient.EcoCreditsFoodGET(1.0, 2.0, 3.0),
				TestsCommon.fractional_max_accepted_delta);

		// Laptop/Tablet example
		final EcoCreditCalculatorProductTypeTest productTypeTests[] = {
				new EcoCreditCalculatorProductTypeTest(24.020598907823427, 847130, 2, 7, "working"),
				new EcoCreditCalculatorProductTypeTest(21.66759890782343, 847130, 2, 7, "repairable"),
				new EcoCreditCalculatorProductTypeTest(18.138076406015085, 847130, 2, 7, "broken"),
				new EcoCreditCalculatorProductTypeTest(23.179575898058356, 847130, 3, 2, "working"),
				new EcoCreditCalculatorProductTypeTest(20.82657589805836, 847130, 3, 2, "repairable"),
				new EcoCreditCalculatorProductTypeTest(17.297053396250014, 847130, 3, 2, "broken"),
				new EcoCreditCalculatorProductTypeTest(18.85609437248686, 847130, 0, 8, "working"),
				new EcoCreditCalculatorProductTypeTest(16.503094372486864, 847130, 0, 8, "repairable"),
				new EcoCreditCalculatorProductTypeTest(12.97357187067852, 847130, 0, 8, "broken"),
				new EcoCreditCalculatorProductTypeTest(8.530548936388639, 847130, 0, 0, "working"),
				new EcoCreditCalculatorProductTypeTest(6.177548936388639, 847130, 0, 0, "repairable"),
				new EcoCreditCalculatorProductTypeTest(2.648026434580296, 847130, 0, 0, "broken") };

		for (final EcoCreditCalculatorProductTypeTest test : productTypeTests) {
			System.out.println(
					"HS key:" + test.hskey + ", Y: " + test.years + ", M: " + test.months + ", ST: " + test.eolstate);
			Assert.assertEquals(test.expectedresult, ecocreditsclient.EcoCreditsGET(test.hskey, test.years, test.months,
					DataConversion.convertEoLState(test.eolstate)), TestsCommon.fractional_max_accepted_delta);
		}

		// Logout
		Assert.assertTrue("User logout failed, could not end session!", ecocreditsclient.endSession());
	}

	public static void main(String[] args) throws URISyntaxException {
		final EcoCreditCalculatorClientIT it = new EcoCreditCalculatorClientIT();

		try {
			it.beforeTests();
			it.itEcoCreditsCalculatorWS();
		} catch (Exception e) {
			System.err.println(e.getMessage());
		}

		it.afterTests();
	}
}
