/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.client;

import java.net.URI;
import java.util.logging.Logger;

import org.openapitools.client.ApiException;
import org.openapitools.client.api.DefaultApi;
import org.openapitools.client.model.EcoCreditParameters;

import gr.iccs.isense.circ4life.datamodels.WEEEsEndOfLifeState;
import gr.iccs.isense.circ4life.logging.Log;
import gr.iccs.isense.circ4life.rest.client.RESTWebServiceClient;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;

/**
 * Client for the Eco-credit calculation RESTful Web Service.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EcoCreditCalculatorClient extends RESTWebServiceClient {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(EcoCreditCalculatorClient.class);

	/**
	 * Default REST Web Service URI.
	 */
	public final static String DEFAULT_SERVICE_URI = "https://circ4life.iccs.gr/EcoCreditCalculator/resources";

	/**
	 * OpenAPI generated API client.
	 */
	private final DefaultApi APIClient;

	/**
	 * Constructor. Default REST Web Service URI used and no credentials.
	 */
	public EcoCreditCalculatorClient() {
		APIClient = new DefaultApi();
		// HACK: The service provider lookup constructs a RESTEasy JAX-RS client by
		// default, because it is in the classpath. This leads to an exception:
		// javax.ws.rs.ProcessingException: RESTEASY003145: Unable to find a
		// MessageBodyReader of content-type application/json and type class <HTTP
		// Response type here>
		// This happens because the OpenAPI generator code produces a client
		// configuration only for Jersey clients.
		APIClient.getApiClient().setHttpClient(buildJerseyClient(APIClient.getApiClient().getJSON()));
		APIClient.getApiClient().setUserAgent(EcoCreditCalculatorClient.class.getName());
		APIClient.getApiClient().setBasePath(DEFAULT_SERVICE_URI);
	}

	/**
	 * Constructor.
	 * 
	 * @param basePath the web service URL endpoint.
	 */
	public EcoCreditCalculatorClient(final URI basePath) {
		this();
		APIClient.getApiClient().setBasePath(basePath.toString());
	}

	/**
	 * Constructor.
	 * 
	 * @param basePath the web service URL endpoint.
	 * @param username the Username required for protected web service resources.
	 * @param password the Password required for protected web service resources.
	 */
	public EcoCreditCalculatorClient(final URI basePath, final String username, final String password) {
		this(basePath);
		setUsername(username);
		setPassword(password);
	}

	@Override
	public String getClientID() {
		return AccessControlManager.ECOCREDITCALC_CLIENTID;
	}

	@Override
	public void setHTTPHeader(String name, String value) {
		APIClient.getApiClient().addDefaultHeader(name, value);
	}

	/**
	 * Get food Eco-credits by supplying all parameters in path.
	 * 
	 * @param exergy Organic urban waste chemical exergy for a unit of mass.
	 * @param mass   Organic waste mass.
	 * @param D      Constant parameter D.
	 * @return the Eco-credits, or <code>null</code> in case of an inaccessible API.
	 */
	public Double EcoCreditsFoodGET(Double exergy, Double mass, Double D) {
		Double ecocredits = null;

		try {
			// Try to refresh...
			if (!refreshSession()) {
				// ...on failure try to login again
				endSession();
				if (!startSession()) {
					logger.severe("Could not start or refresh HTTP user session!");

					return null;
				}
			}

			ecocredits = APIClient.ecoCreditsFoodGET(exergy, mass, D);
		} catch (ApiException e) {
			logger.severe(e.getMessage());

			return null;
		}

		return ecocredits;
	}

	/**
	 * USER_R_02 Get Eco-credits from evaluated recycled product type.
	 * 
	 * @param hskey       Product Harmonized System 6 digit code.
	 * @param years       Years of product use thus far.
	 * @param extramonths Additional months of use added to years of use.
	 * @param eolstate    Product End of Life state, should be one of 'working',
	 *                    'repairable', 'broken'.
	 * @return the Eco-credits, or <code>null</code> in case of an inaccessible API.
	 */
	public Double EcoCreditsGET(Integer hskey, Integer years, Integer extramonths, WEEEsEndOfLifeState eolstate) {
		Double ecocredits = null;

		try {
			// Try to refresh...
			if (!refreshSession()) {
				// ...on failure try to login again
				endSession();
				if (!startSession()) {
					logger.severe("Could not start or refresh HTTP user session!");

					return null;
				}
			}

			ecocredits = APIClient.ecoCreditsGET(hskey, years, extramonths, eolstate.toString());
		} catch (ApiException e) {
			logger.severe(e.getMessage());

			return null;
		}

		return ecocredits;
	}

	/**
	 * Calculate the Eco-credits of a recycled product with the specified parameters
	 * according to the formula of Eq. 1 in Deliverable 2.4.
	 * 
	 * @param ecoCreditParameters All Eco-credit parameters encapsulated in an
	 *                            object.
	 * @return the Eco-credits, or <code>null</code> in case of an inaccessible API.
	 */
	public Double EcoCreditsPOST(EcoCreditParameters ecoCreditParameters) {
		Double ecocredits = null;

		try {
			// Try to refresh...
			if (!refreshSession()) {
				// ...on failure try to login again
				endSession();
				if (!startSession()) {
					logger.severe("Could not start or refresh HTTP user session!");

					return null;
				}
			}

			ecocredits = APIClient.ecoCreditsPOST(ecoCreditParameters);
		} catch (ApiException e) {
			logger.severe(e.getMessage());

			return null;
		}

		return ecocredits;
	}

	/**
	 * USER_R_02 Get Eco-credits from evaluated recycled product.
	 * 
	 * @param productid   Product ID as in the unique identification properties of
	 *                    Product Master Data record.
	 * @param years       Years of product use thus far.
	 * @param extramonths Additional months of use added to years of use.
	 * @param eolstate    Product End of Life state, should be one of 'working',
	 *                    'repairable', 'broken'.
	 * @return the Eco-credits, or <code>null</code> in case of an inaccessible API.
	 */
	public Double EcoCreditsProductGET(String productid, Integer years, Integer extramonths,
			WEEEsEndOfLifeState eolstate) {
		Double ecocredits = null;

		try {
			// Try to refresh...
			if (!refreshSession()) {
				// ...on failure try to login again
				endSession();
				if (!startSession()) {
					logger.severe("Could not start or refresh HTTP user session!");

					return null;
				}
			}

			ecocredits = APIClient.ecoCreditsProductGET(productid, years, extramonths, eolstate.toString());
		} catch (ApiException e) {
			logger.severe(e.getMessage());

			return null;
		}

		return ecocredits;
	}

	/**
	 * Get WEEE product Eco-credits by supplying all parameters to path.
	 * 
	 * @param unukey      Electronic product UNU-KEY code.
	 * @param years       Years of product use thus far.
	 * @param extramonths Additional months of use added to years of use.
	 * @param eolstate    Product End of Life state, should be one of 'working',
	 *                    'repairable', 'broken'.
	 * @param A           Constant parameter weight A to be multiplied with product
	 *                    materials exergy.
	 * @param B           Constant parameter weight B to be multiplied with End of
	 *                    Life state factor.
	 * @param C           Constant parameter weight C to be multiplied by Lifetime
	 *                    Factor.
	 * @return the Eco-credits, or <code>null</code> in case of an inaccessible API.
	 */
	public Double EcoCreditsWEEEsGET(Integer unukey, Integer years, Integer extramonths, WEEEsEndOfLifeState eolstate,
			Double A, Double B, Double C) {
		Double ecocredits = null;

		try {
			// Try to refresh...
			if (!refreshSession()) {
				// ...on failure try to login again
				endSession();
				if (!startSession()) {
					logger.severe("Could not start or refresh HTTP user session!");

					return null;
				}
			}

			ecocredits = APIClient.ecoCreditsWEEEsGET(unukey, years, extramonths, eolstate.toString(), A, B,
					ecocredits);
		} catch (ApiException e) {
			logger.severe(e.getMessage());

			return null;
		}

		return ecocredits;
	}

	/**
	 * Get WEEE product Lifespan Weibull Distribution value.
	 * 
	 * @param unukey      Electronic product UNU-KEY code.
	 * @param years       Years of product use thus far.
	 * @param extramonths Additional months of use added to years of use.
	 * @return the Weibull Distribution Lifespan factor, or <code>null</code> in
	 *         case of an inaccessible API.
	 */
	public Double LifeSpanFactorWEEEsGET(Integer unukey, Integer years, Integer extramonths) {
		Double ecocredits = null;

		try {
			// Try to refresh...
			if (!refreshSession()) {
				// ...on failure try to login again
				endSession();
				if (!startSession()) {
					logger.severe("Could not start or refresh HTTP user session!");

					return null;
				}
			}

			ecocredits = APIClient.lifeSpanFactorWEEEsGET(unukey, years, extramonths);
		} catch (ApiException e) {
			logger.severe(e.getMessage());

			return null;
		}

		return ecocredits;
	}
}
