/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.openapi;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.validation.Valid;
import javax.validation.constraints.DecimalMin;
import javax.validation.constraints.Max;
import javax.validation.constraints.Min;
import javax.validation.constraints.Size;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.core.Response.Status;

import org.openapitools.api.EcocreditsApi;
import org.openapitools.model.EcoCreditParameters;
import org.openapitools.model.ProductType;

import gr.iccs.isense.circ4life.EcoCreditCalculator;
import gr.iccs.isense.circ4life.FoodParameters;
import gr.iccs.isense.circ4life.WEEEsLifetimeFactorCalculator;
import gr.iccs.isense.circ4life.WEEEsMaterialThermodynamicRarityCalculator;
import gr.iccs.isense.circ4life.WEEEsParameters;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO.WEEEsLifespanDistributionParameters;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO.WEEEsWeights;
import gr.iccs.isense.circ4life.datamodels.MaterialFactors;
import gr.iccs.isense.circ4life.rest.EcoCreditCalculatorModule;
import gr.iccs.isense.circ4life.rest.EcoCreditCalculatorModule.ErrorMessages;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;
import gr.iccs.isense.circ4life.rest.resources.validation.POSTDataValidator;

/**
 * /ecocredits end-point implementation.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class Ecocredits extends EcocreditsApi {
	/**
	 * Eco-credits calculator.
	 */
	private static final EcoCreditCalculator calculator = new EcoCreditCalculator();

	/**
	 * Eco-credits calculator Data Access Object for retrieving parameters and
	 * constants from persistent data layer.
	 */
	private static final EcoCreditCalculatorDAO calculatordao = new EcoCreditCalculatorDAO();

	/**
	 * Proxy of security context for session management.
	 */
	private final SecurityContext m_securityContext;

	/**
	 * Constructor.
	 * 
	 * @param securityContext proxy of security context for session management.
	 */
	public Ecocredits(@Context SecurityContext securityContext) {
		m_securityContext = securityContext;
	}

	@Override
	public Response ecoCreditsFoodGET(@DecimalMin("0") Double exergy, @DecimalMin("0") Double mass,
			@DecimalMin("0") Double D) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final EcoCreditParameters params = new EcoCreditParameters();
		final org.openapitools.model.FoodParameters foodparams = new org.openapitools.model.FoodParameters();
		foodparams.chemicalExergy(exergy).mass(mass).weightD(D);
		final org.openapitools.model.ProductType prodtype = new ProductType();
		prodtype.hskey(23);
		params.producttype(prodtype).foodparameters(foodparams);

		return ecoCreditsPOST(params);
	}

	@Override
	public Response ecoCreditsGET(@Min(1) @Max(999999) Integer hskey, @Min(0) @Max(100) Integer years,
			@Min(0) @Max(11) Integer extramonths, String eolstate) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		switch (calculatordao.getProductWasteType(hskey)) {
		case FOOD: {
			// In the case of food, there is no mass, exergy or lifetime parameters
			// involved. We need to return something strictly from the type of the product.
			// Can we? It depends on the database information that can be found for this
			// food HS key. We invoke the Food endpoint.
			final FoodParameters foodparams = calculatordao.getFoodParameters(hskey);
			if (null == foodparams) {
				EcoCreditCalculatorModule.logger
						.severe(ErrorMessages.FOODWASTE_PARAMS_NOT_FOUND + " for HS key: " + hskey);

				return Response.status(Status.BAD_REQUEST.getStatusCode(),
						ErrorMessages.FOODWASTE_PARAMS_NOT_FOUND + " for HS key: " + hskey).build();
			}

			return ecoCreditsFoodGET(foodparams.getChemicalExergy(), foodparams.getOrganicWasteMass(),
					foodparams.getWeightParameterD());
		}
		case WEEES: {
			// In the case of WEEEs we bring the parameters from the HS key and
			// invoke the WEEEs endpoint.
			final Integer unukey = calculatordao.getUNUKeyFromHSkey(hskey);
			if (null == unukey) {
				EcoCreditCalculatorModule.logger.severe(ErrorMessages.UNKNOWN_HSKEY + ": " + hskey);

				return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.UNKNOWN_HSKEY + ": " + hskey)
						.build();
			}
			final WEEEsWeights constants = calculatordao.getWEEEsWeights(unukey);
			if (null == constants) {
				EcoCreditCalculatorModule.logger
						.severe(ErrorMessages.CONSTANT_WEIGHTS_NOT_FOUND + " for UNU key: " + unukey);

				return Response.status(Status.BAD_REQUEST.getStatusCode(),
						ErrorMessages.CONSTANT_WEIGHTS_NOT_FOUND + " for UNU key: " + unukey).build();
			}

			return ecoCreditsWEEEsGET(unukey, years, extramonths, eolstate, constants.A, constants.B, constants.C);
		}
		case UNKNOWN:
		default: {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.UNKNOWN_HSKEY + ": " + hskey);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.UNKNOWN_HSKEY + ": " + hskey)
					.build();
		}
		} // switch
	}

	@Override
	public Response ecoCreditsPOST(@Valid EcoCreditParameters ecoCreditParameters) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		if (!POSTDataValidator.isValid(ecoCreditParameters)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_ECOCREDIT_PARAMS)
					.build();
		}

		final Integer hskey = ecoCreditParameters.getProducttype().getHskey();
		if (null == hskey) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.MISSING_HSKEY);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MISSING_HSKEY).build();
		}

		Double ecocredits = null;

		// Determine product waste type and use proper function
		switch (calculatordao.getProductWasteType(hskey)) {
		case FOOD: {
			try {
				ecocredits = calculateFoodEcoCredits(ecoCreditParameters.getFoodparameters());
			} catch (IllegalArgumentException e) {
				EcoCreditCalculatorModule.logger.severe(e.getMessage());

				return Response.status(Status.BAD_REQUEST.getStatusCode(), e.getMessage()).build();
			}
			if (null == ecocredits) {
				EcoCreditCalculatorModule.logger.severe(ErrorMessages.UNKNOWN_CALC_ERROR);

				return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.UNKNOWN_CALC_ERROR).build();
			}

			return Response.ok().entity(ecocredits).build();
		}
		case WEEES: {
			try {
				ecocredits = calculateWEEEsEcoCredits(ecoCreditParameters.getWeeesparameters());
			} catch (IllegalArgumentException e) {
				EcoCreditCalculatorModule.logger.severe(e.getMessage());

				return Response.status(Status.BAD_REQUEST.getStatusCode(), e.getMessage()).build();
			}
			if (null == ecocredits) {
				EcoCreditCalculatorModule.logger.severe(ErrorMessages.UNKNOWN_CALC_ERROR);

				return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.UNKNOWN_CALC_ERROR).build();
			}

			return Response.ok().entity(ecocredits).build();
		}
		case UNKNOWN:
		default: {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.UNKNOWN_HSKEY + ": " + hskey);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.UNKNOWN_HSKEY + ": " + hskey)
					.build();
		}
		} // switch
	}

	@Override
	public Response ecoCreditsProductGET(@Size(min = 14) String productid, @Min(0) @Max(100) Integer years,
			@Min(0) @Max(11) Integer extramonths, String eolstate) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final gr.iccs.isense.circ4life.EcoCreditParameters params = calculatordao
				.getProductEcoCreditParameters(productid, years, extramonths, DataConversion.convertEoLState(eolstate));
		if (null == params) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.PRODUCT_PARAMS_NOT_FOUND + ": " + productid);

			return Response.status(Status.BAD_REQUEST.getStatusCode(),
					ErrorMessages.PRODUCT_PARAMS_NOT_FOUND + ": " + productid).build();
		}

		Double ecocredits = null;
		try {
			ecocredits = calculator.calculate(params);
		} catch (IllegalArgumentException e) {
			EcoCreditCalculatorModule.logger.severe(e.getMessage());

			return Response.status(Status.BAD_REQUEST.getStatusCode(), e.getMessage()).build();
		}

		return Response.ok().entity(ecocredits).build();
	}

	@Override
	public Response ecoCreditsWEEEsGET(@Min(1) Integer unukey, @Min(0) @Max(100) Integer years,
			@Min(0) @Max(11) Integer extramonths, String eolstate, @DecimalMin("0") Double A, @DecimalMin("0") Double B,
			@DecimalMin("0") Double C) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final Integer hskey = calculatordao.getHSKeyFromUNUkey(unukey);
		if (null == hskey) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.UNKNOWN_UNUKEY + ": " + unukey);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.UNKNOWN_UNUKEY + ": " + unukey)
					.build();
		}

		final EcoCreditParameters params = new EcoCreditParameters();
		final org.openapitools.model.WEEEsParameters weeesparams = new org.openapitools.model.WEEEsParameters();
		final List<org.openapitools.model.MaterialFactors> materialfactors = new ArrayList<org.openapitools.model.MaterialFactors>();
		// Convert to OpenAPI model structure
		for (final MaterialFactors e : calculatordao.getMaterialFactors(unukey,
				DataConversion.convertEoLState(eolstate))) {
			final org.openapitools.model.MaterialFactors mf = new org.openapitools.model.MaterialFactors();
			mf.mass(e.getMass_in_grams()).rarity(e.getRarity()).recyclability(e.getRecyclability());
			materialfactors.add(mf);
		}
		weeesparams.weeestype(unukey).yearsofuse(years).monthsofuse(extramonths).eolstate(eolstate).weightA(A)
				.weightB(B).weightC(C).materialfactors(materialfactors);
		final org.openapitools.model.ProductType prodtype = new ProductType();
		prodtype.hskey(hskey).unukey(unukey);
		params.producttype(prodtype).weeesparameters(weeesparams);

		return ecoCreditsPOST(params);
	}

	@Override
	public Response lifeSpanFactorWEEEsGET(@Min(1) Integer unukey, @Min(0) @Max(100) Integer years,
			@Min(0) @Max(11) Integer extramonths) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		Double lifetimefactor = null;
		try {
			final WEEEsLifetimeFactorCalculator ltcalc = createWEEEsLifetimeFactorCalculator(unukey, years,
					extramonths);
			if (null == ltcalc) {
				EcoCreditCalculatorModule.logger.severe(ErrorMessages.CANNOT_CREATE_LIFETIMEFACTOR_CALC);

				return Response
						.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.CANNOT_CREATE_LIFETIMEFACTOR_CALC)
						.build();
			}

			lifetimefactor = ltcalc.getLifetimeFactor();
		} catch (IllegalArgumentException e) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.INVALID_LIFETIME_USAGE);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.INVALID_LIFETIME_USAGE).build();
		}

		return Response.ok().entity(lifetimefactor).build();
	}

	/**
	 * Calculate the Eco-credits for the Food waste case.
	 * 
	 * @param parameters OpenAPI model parameters.
	 * @return The Eco-credits.
	 * @throws IllegalArgumentException thrown if the specified parameters are not
	 *                                  valid.
	 */
	private Double calculateFoodEcoCredits(final org.openapitools.model.FoodParameters parameters)
			throws IllegalArgumentException {
		return calculateFoodEcoCredits(parameters.getChemicalExergy(), parameters.getMass(), parameters.getWeightD());
	}

	/**
	 * Calculate the Eco-credits for the Food waste case.
	 * 
	 * @param exergy the chemical exergy in case of Food waste.
	 * @param mass   the product mass in case of Food waste.
	 * @param D      the weight factor of the material exergy.
	 * @return The Eco-credits.
	 * @throws IllegalArgumentException thrown if the specified parameters are not
	 *                                  valid.
	 */
	private Double calculateFoodEcoCredits(final Double exergy, final Double mass, final Double D)
			throws IllegalArgumentException {
		final FoodParameters params = new FoodParameters(D, exergy, mass);

		return calculator.calculate(params);
	}

	/**
	 * Convert Material factors list from OpenAPI Data Model to internal class.
	 * 
	 * @param materialfactors OpenAPI model material factors.
	 * @return the converted list of material factors.
	 */
	private Collection<MaterialFactors> convertMaterialFactors(
			final Collection<org.openapitools.model.MaterialFactors> materialfactors) {
		final Collection<MaterialFactors> factors = new ArrayList<MaterialFactors>();

		for (final org.openapitools.model.MaterialFactors elem : materialfactors) {
			factors.add(new MaterialFactors(elem.getMass(), elem.getRarity(), elem.getRecyclability()));
		}

		return factors;
	}

	/**
	 * Create a WEEEs Lifespan Weibull distribution calculator from the specified
	 * parameters.
	 * 
	 * @param unukey      the UNU key.
	 * @param yearsofuse  years of product usage.
	 * @param monthsofuse extra months of usage added to yearsofuse.
	 * @return the lifetime factor parameters if a correspondence to the specified
	 *         parameters is found, <code>null</code> otherwise.
	 */
	private WEEEsLifetimeFactorCalculator createWEEEsLifetimeFactorCalculator(Integer unukey, Integer yearsofuse,
			Integer monthsofuse) {
		final WEEEsLifespanDistributionParameters lifetimeAB = calculatordao
				.getWEEEsLifespanDistributionParameters(unukey);
		if (null == lifetimeAB) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.LIFETIMEFACTOR_PARAMS_NOT_FOUND);

			return null;
		}

		return new WEEEsLifetimeFactorCalculator(yearsofuse, monthsofuse, lifetimeAB.A, lifetimeAB.B);
	}

	/**
	 * Calculate the Eco-credits for the Waste Electrical and Electronic Equipment
	 * case.
	 * 
	 * @param parameters OpenAPI model parameters.
	 * @return The Eco-credits.
	 * @throws IllegalArgumentException thrown if the specified parameters are not
	 *                                  valid.
	 */
	private Double calculateWEEEsEcoCredits(final org.openapitools.model.WEEEsParameters parameters)
			throws IllegalArgumentException {
		return calculateWEEEsEcoCredits(parameters.getWeeestype(), parameters.getEolstate(), parameters.getYearsofuse(),
				parameters.getMonthsofuse(), parameters.getWeightA(), parameters.getWeightB(), parameters.getWeightC(),
				parameters.getMaterialfactors());
	}

	/**
	 * Calculate the Eco-credits for the Waste Electrical and Electronic Equipment
	 * case.
	 * 
	 * @param unukey          the product type UNU key.
	 * @param eolstate        the End of Life state.
	 * @param yearsofuse      the product usage years.
	 * @param monthsofuse     extra months of product usage, added to the years.
	 * @param weightA         the weight factor of the material exergy.
	 * @param weightB         the weight factor of the End of Life state.
	 * @param weightC         the weight factor of the Lifetime factor.
	 * @param materialfactors the material factors list.
	 * @return The Eco-credits.
	 * @throws IllegalArgumentException thrown if the specified parameters are not
	 *                                  valid.
	 */
	private Double calculateWEEEsEcoCredits(Integer unukey, String eolstate, Integer yearsofuse, Integer monthsofuse,
			Double weightA, Double weightB, Double weightC,
			final Collection<org.openapitools.model.MaterialFactors> materialfactors) throws IllegalArgumentException {
		final WEEEsLifetimeFactorCalculator ltcalc = createWEEEsLifetimeFactorCalculator(unukey, yearsofuse,
				monthsofuse);
		if (null == ltcalc) {
			EcoCreditCalculatorModule.logger.severe(ErrorMessages.CANNOT_CREATE_LIFETIMEFACTOR_CALC);

			throw new IllegalArgumentException(ErrorMessages.CANNOT_CREATE_LIFETIMEFACTOR_CALC);
		}
		final WEEEsMaterialThermodynamicRarityCalculator materialcalc = new WEEEsMaterialThermodynamicRarityCalculator(
				convertMaterialFactors(materialfactors));
		final WEEEsParameters params = new WEEEsParameters(weightA, materialcalc, weightB,
				DataConversion.convertEoLState(eolstate), weightC, ltcalc);

		return calculator.calculate(params);
	}
}
