/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.validation;

import org.openapitools.model.EcoCreditParameters;
import org.openapitools.model.FoodParameters;
import org.openapitools.model.WEEEsParameters;

import gr.iccs.isense.circ4life.rest.EcoCreditCalculatorModule;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;

/**
 * POST data structures validator.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class POSTDataValidator {
	/**
	 * Validate POST Eco-credits parameters.
	 * 
	 * @param ecoCreditParameters the Eco-credits parameters.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final EcoCreditParameters ecoCreditParameters) {
		try {
			// The Product type HS key should be present
			ecoCreditParameters.getProducttype().getHskey().intValue();

			// Check specific tpye parameters
			final FoodParameters foodparams = ecoCreditParameters.getFoodparameters();
			final WEEEsParameters weeesparamas = ecoCreditParameters.getWeeesparameters();

			// Cannot miss both types of parameters
			if (null == foodparams && null == weeesparamas) {
				throw new IllegalArgumentException("Eco-credits parameters missing all parameter types!");
			}

			// Cannot set both either, we will handle it later
			if (null != foodparams && null != weeesparamas) {
				EcoCreditCalculatorModule.logger.warning("Eco-credits parameters contain both Food Waste and WEEEs!");
			}

			if (null != foodparams) {
				// all parameters should be present
				foodparams.getChemicalExergy().doubleValue();
				foodparams.getMass().doubleValue();
				foodparams.getWeightD().doubleValue();
			}

			if (null != weeesparamas) {
				DataConversion.convertEoLState(weeesparamas.getEolstate()).Value();
				// This can be empty or null
				// weeesparamas.getMaterialfactors();
				weeesparamas.getMonthsofuse().intValue();
				weeesparamas.getWeeestype().intValue();
				weeesparamas.getWeightA().doubleValue();
				weeesparamas.getWeightB().doubleValue();
				weeesparamas.getWeightC().doubleValue();
				weeesparamas.getYearsofuse().intValue();
			}

			// All OK
			return true;
		} catch (Exception e) {
			EcoCreditCalculatorModule.logger.severe("Eco-credits parameter validation error: " + e.getMessage());

			return false;
		}
	}
}
