/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest;

import java.io.File;
import java.util.logging.ConsoleHandler;
import java.util.logging.Logger;

import org.glassfish.jersey.jackson.JacksonFeature;
import org.glassfish.jersey.logging.LoggingFeature;
import org.glassfish.jersey.server.ResourceConfig;

import gr.iccs.isense.circ4life.logging.Log;
import gr.iccs.isense.circ4life.rest.config.ConfigProperties;
import gr.iccs.isense.circ4life.rest.resources.EcoAccountProfile;
import gr.iccs.isense.circ4life.rest.resources.EcoShoppingMedia;
import gr.iccs.isense.circ4life.rest.resources.Information;
import gr.iccs.isense.circ4life.rest.resources.openapi.EcoAccount;
import gr.iccs.isense.circ4life.rest.resources.openapi.EcoShopping;

/**
 * This class implements the Eco Account and Shopping module integration with
 * the ICT platform as a RESTful Web Service API.
 *
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EndUserModule extends ResourceConfig {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(EndUserModule.class);

	/**
	 * Uploaded profile image prefix. The filename suffix extension depends on the
	 * supported type.
	 */
	public final static String PROFILE_IMAGE_PREFIX = "ecoaccount";

	/**
	 * Uploaded product image prefix. The filename suffix extension depends on the
	 * supported type.
	 */
	public final static String PRODUCT_IMAGE_PREFIX = "product";

	/**
	 * Uploaded profile image size limit in bytes.
	 */
	public final static int PROFILE_IMAGE_SIZE_LIMIT = 256 * 1024;

	/**
	 * The file path where file uploads are saved under.
	 */
	private static File FileUploadsPath = null;

	/**
	 * REST service error messages.
	 */
	public static final class ErrorMessages {
		public final static String AUTHENTICATION_REQUIRED = "Authentication required";
		public final static String AUTHENTICATION_DENIED = "Authentication denied";
		public final static String AUTHENTICATION_ERROR = "Authentication error";
		public final static String AUTHENTICATION_NOT_VERIFIED_YET = "Eco-account email address is not verified yet";
		public final static String SESSIONLOGOUT_ERROR = "Cannot log out user from current session";
		public final static String SESSIONUSERID_FAILURE = "Cannot get user ID from session";
		public final static String ANONYMOUSID_FAILURE = "Cannot generate user anonymous ID";
		public final static String PRODUCT_NOTFOUND = "Product cannot be found";
		public final static String PRODUCT_MALFORMED_RECORD = "Product Master Data Record is malformed";
		public final static String ECOACCOUNT_NOTFOUND = "Eco-account cannot be found";
		public final static String ECOACCOUNT_EXISTS = "Eco-account already exists";
		public final static String MALFORMED_REGISTER_FORM = "Malformed user registration form";
		public final static String MALFORMED_REGISTER_FORM_IMAGE = "Malformed user registration form image upload";
		public final static String MALFORMED_RESIGN_FORM = "Malformed user resignation form";
		public final static String MALFORMED_RESET_PASSWORD_FORM = "Malformed user reset password form";
		public final static String MALFORMED_UPDATE_FORM = "Malformed profile update form";
		public final static String MALFORMED_LOGIN_FORM = "Malformed user login form";
		public final static String MALFORMED_LOGOUT_FORM = "Malformed user logout form";
		public final static String MALFORMED_PRODUCERS_FEEDBACK = "Malformed producers feedback form";
		public final static String PRODUCERS_FEEDBACK_FAILURE = "Cannot update producers feedback";
		public final static String MALFORMED_SYSTEM_FEEDBACK = "Malformed system feedback form";
		public final static String SYSTEM_FEEDBACK_FAILURE = "Cannot update system feedback";
		public final static String CONVERSION_ERROR = "Data conversion error";
		public final static String CREATE_ACCOUNT_ERROR = "Eco-account registration failed";
		public final static String REMOVE_ACCOUNT_ERROR = "Eco-account resignation failed";
		public final static String UPDATE_ACCOUNT_ERROR = "Eco-account update failed";
		public final static String EMAIL_SERVER_ERROR = "Eco-account E-mail service failed";
	}

	/**
	 * Constructor.
	 */
	public EndUserModule() {
		// REST resource classes
		register(Information.class);
		register(EcoAccountProfile.class);
		register(EcoShoppingMedia.class);
		register(EcoAccount.class);
		register(EcoShopping.class);

		// Jackson JSON serialization
		register(JacksonFeature.class);

		// Setup logging
		logger.setLevel(ConfigProperties.getLoggingLevel(EndUserModule.class));
		final ConsoleHandler loggerhandler = new ConsoleHandler();
		loggerhandler.setLevel(logger.getLevel());
		logger.addHandler(loggerhandler);
		logger.setUseParentHandlers(false);
		final LoggingFeature.Verbosity verbosity = ConfigProperties.getLoggingVerbosity(EndUserModule.class);
		register(new LoggingFeature(logger, logger.getLevel(), verbosity, LoggingFeature.DEFAULT_MAX_ENTITY_SIZE));

		logger.info("Logger level: " + logger.getLevel() + ", Verbosity: " + verbosity + ", Parent Logger Level: "
				+ logger.getParent().getLevel() + ", using parent logger handlers: " + logger.getUseParentHandlers());
	}

	/**
	 * Get the Web Service file uploads directory as a {@link File}. If the
	 * directory does not exists in the local filesystem, try to create it.
	 * 
	 * @return the file uploads directory as {@link File} on success,
	 *         <code>null</code> otherwise.
	 */
	public static final File getFileUploadsPath() {
		try {
			// Lazy initialization
			if (null == FileUploadsPath) {
				final String fileuploadsdircfg = ConfigProperties.getFileUploadsDir(EndUserModule.class);
				final File temp = new File(fileuploadsdircfg);
				final File fileuploadsdir = temp.getCanonicalFile();

				// If the directory does not exist try to create it
				if (!fileuploadsdir.exists() && !fileuploadsdir.mkdirs()) {
					logger.severe("Could not create non-existing file uploads directory [" + fileuploadsdircfg + "]!");

					return null;
				}

				// Check if the existing path is a directory
				if (!fileuploadsdir.isDirectory()) {
					logger.severe("File path [" + fileuploadsdircfg + "] exists but it is not a directory!");

					return null;
				}

				// Check if we can create files inside the directory
				if (!fileuploadsdir.canWrite()) {
					logger.severe("File path [" + fileuploadsdircfg + "] permissions does not allow write operations!");

					return null;
				}

				// All OK, directory exists or created successfully
				FileUploadsPath = fileuploadsdir;
			}

			return FileUploadsPath;
		} catch (Exception e) {
			logger.severe("Could not initialize File Uploads directory!");

			return null;
		}
	}

	/**
	 * Create if it does not exist the Web Service profile image directory of a
	 * specified user.
	 * 
	 * @param useruuid the user's Universally Unique ID.
	 * @return the profile image directory as {@link File} on success,
	 *         <code>null</code> otherwise.
	 */
	public static final File createProfileImagePath(final String useruuid) {
		final File profileimagedir = getProfileImagePath(useruuid);

		if (!profileimagedir.exists() && !profileimagedir.mkdirs()) {
			EndUserModule.logger.severe("Cannot create the target profile image directory!");

			return null;
		}

		return profileimagedir;
	}

	/**
	 * Get the Web Service profile image directory of a specified user.
	 * <strong>CAUTION:</strong> the directory path returned may not exist in the
	 * local filesystem.
	 * 
	 * @param useruuid the user's Universally Unique ID.
	 * @return the profile image directory as {@link File}.
	 */
	public static final File getProfileImagePath(final String useruuid) {
		return new File(getFileUploadsPath(), "/EndUserModuleWS/ecoaccounts/" + useruuid);
	}

	/**
	 * Get the Web Service product image directory of a specified product.
	 * <strong>CAUTION:</strong> the directory path returned may not exist in the
	 * local filesystem.
	 * 
	 * @param productid the GTIN14 unique product ID.
	 * @return the product image directory as {@link File}.
	 */
	public static final File getProductImagePath(final String productid) {
		return new File(getFileUploadsPath(), "/EndUserModuleWS/products/" + productid);
	}
}
