/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.file.Files;

import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.core.StreamingOutput;

import javax.ws.rs.core.Response.Status;

import gr.iccs.isense.circ4life.rest.EndUserModule;
import gr.iccs.isense.circ4life.rest.EndUserModule.ErrorMessages;

/**
 * REST end-point for Eco-account profile image access.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
@Path("/ecoaccount/profile")
public class EcoAccountProfile {
	/**
	 * Proxy of security context for session management.
	 */
	private final SecurityContext m_securityContext;

	/**
	 * Constructor.
	 * 
	 * @param securityContext proxy of security context for session management.
	 */
	public EcoAccountProfile(@Context SecurityContext securityContext) {
		m_securityContext = securityContext;
	}

	/**
	 * Get the uploaded Eco-account profile image.
	 * 
	 * @return the profile image of the logged in user's Eco-account.
	 */
	@GET
	@Path("/image")
	@Produces({ "image/png", "image/jpg" })
	public Response getProfileImage() {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		// Get the profile image directory and return the first JPG or PNG image found
		// in there
		final File profileimagedir = EndUserModule.getProfileImagePath(getPrincipalName());
		if (!profileimagedir.exists()) {
			EndUserModule.logger.warning("Uploaded profile image directory not found for user " + getPrincipalName());

			return Response.status(Status.NOT_FOUND).build();
		}

		final FilenameFilter filter = new FilenameFilter() {
			@Override
			public boolean accept(File dir, String name) {
				if (name.equalsIgnoreCase(EndUserModule.PROFILE_IMAGE_PREFIX + ".jpg")
						|| name.equalsIgnoreCase(EndUserModule.PROFILE_IMAGE_PREFIX + ".png")) {
					// matched our predefined filename
					return true;
				}

				return false;
			}
		};

		for (final String imagefile : profileimagedir.list(filter)) {
			EndUserModule.logger.fine("Profile image file found: " + imagefile);

			final File profileimage = new File(profileimagedir, imagefile);

			// Open the image and return a stream to the client
			return Response.ok(new StreamingOutput() {
				@Override
				public void write(OutputStream output) throws IOException, WebApplicationException {
					Files.copy(profileimage.toPath(), output);
				}
			}).build();
		}

		EndUserModule.logger.warning("Uploaded profile image file not found for user " + getPrincipalName());
		return Response.status(Status.NOT_FOUND).build();
	}

	/**
	 * @return the user name from an authenticated session, <code>null</code>
	 *         otherwise.
	 */
	private String getPrincipalName() {
		final String principalname = AccessControlManager.isEnabled()
				? AccessControlManager.getPrincipalName(m_securityContext)
				: null;

		EndUserModule.logger.finest("User session principal name: " + principalname);

		return principalname;
	}
}
