/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.openapi;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Base64;
import java.util.Date;
import java.util.List;
import java.util.concurrent.ThreadLocalRandom;
import java.util.regex.Pattern;

import javax.validation.Valid;
import javax.validation.constraints.Size;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.core.StreamingOutput;
import javax.ws.rs.core.UriInfo;

import org.openapitools.api.EcoaccountApi;
import org.openapitools.model.EcoAccountForgotPasswordForm;
import org.openapitools.model.EcoAccountLoginForm;
import org.openapitools.model.EcoAccountLogoutForm;
import org.openapitools.model.EcoAccountRegistrationForm;
import org.openapitools.model.EcoAccountResignationForm;
import org.openapitools.model.EcoAccountUpdateForm;
import org.openapitools.model.EcoBalance;
import org.openapitools.model.InlineResponse200;
import org.openapitools.model.InlineResponse2001;
import org.openapitools.model.ProducersFeedbackForm;
import org.openapitools.model.Product;
import org.openapitools.model.ProductType;
import org.openapitools.model.SystemFeedbackForm;

import gr.iccs.isense.circ4life.dao.CommonDAO.EcoAccountRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.IndividualProductData;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductID;
import gr.iccs.isense.circ4life.dao.CommonDAO.SystemFeedback;
import gr.iccs.isense.circ4life.dao.CommonDAO.UserProductFeedback;
import gr.iccs.isense.circ4life.dao.EcoAccountDAO;
import gr.iccs.isense.circ4life.rest.EndUserModule;
import gr.iccs.isense.circ4life.rest.EndUserModule.ErrorMessages;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager.AccessTokens;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager.LoginResult;
import gr.iccs.isense.circ4life.rest.resources.validation.POSTDataValidator;
import net.glxn.qrgen.core.image.ImageType;
import net.glxn.qrgen.javase.QRCode;

/**
 * /ecoaccount end-point implementation.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EcoAccount extends EcoaccountApi {
	/**
	 * Eco-account Data Access Object.
	 */
	private static final EcoAccountDAO ecoaccountdao = new EcoAccountDAO();

	/**
	 * The name of the field property that holds the profile image URL string.
	 */
	private static final String ECOACCOUNTPROFILEIMAGEPROPERTY = "EcoAccount.image";

	/**
	 * Proxy of security context for session management.
	 */
	private final SecurityContext m_securityContext;

	/**
	 * Proxy of HTTP request headers.
	 */
	private final HttpHeaders m_httpHeaders;

	/**
	 * Proxy of application and request URI information.
	 */
	private final UriInfo m_uriInfo;

	/**
	 * Constructor.
	 * 
	 * @param securityContext proxy of security context for session management.
	 * @param httpHeaders     proxy of HTTP request headers.
	 * @param uriInfo         proxy of application and request URI information.
	 */
	public EcoAccount(@Context SecurityContext securityContext, @Context HttpHeaders httpHeaders,
			@Context UriInfo uriInfo) {
		m_securityContext = securityContext;
		m_httpHeaders = httpHeaders;
		m_uriInfo = uriInfo;
	}

	@Override
	public Response ecoAccountAnonymousConsumerIdGET() {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final String anonymoususerid = ecoaccountdao.getEcoAccountAnonymousID(getPrincipalName());
		if (null == anonymoususerid) {
			EndUserModule.logger.severe(ErrorMessages.ANONYMOUSID_FAILURE);

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.ANONYMOUSID_FAILURE)
					.build();
		}

		return Response.ok().entity(anonymoususerid).build();
	}

	@Override
	public Response ecoAccountAnonymousConsumerQRCodeGET() {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final String anonymoususerid = ecoaccountdao.getEcoAccountAnonymousID(getPrincipalName());
		if (null == anonymoususerid) {
			EndUserModule.logger.severe(ErrorMessages.ANONYMOUSID_FAILURE);

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.ANONYMOUSID_FAILURE)
					.build();
		}

		return Response.ok(new StreamingOutput() {
			@Override
			public void write(OutputStream output) throws IOException, WebApplicationException {
				generateQRCodePNG(anonymoususerid).writeTo(output);
			}
		}, "image/png").build();
	}

	@Override
	public Response ecoAccountRecycleBinUserIdGET() {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final long recyclebinuserid = ecoaccountdao.getEcoAccountRecycleBinUserID(getPrincipalName());
		if (0 > recyclebinuserid) {
			EndUserModule.logger.severe(ErrorMessages.ANONYMOUSID_FAILURE);

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.ANONYMOUSID_FAILURE)
					.build();
		}

		// Use a conversion to String in order to always include all 11 digits.
		final String response = String.format("%1$011d", recyclebinuserid);

		return Response.ok().entity(response).build();
	}

	@Override
	public Response ecoAccountRecycleBinUserQRCodeGET() {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final long recyclebinuserid = ecoaccountdao.getEcoAccountRecycleBinUserID(getPrincipalName());
		if (0 > recyclebinuserid) {
			EndUserModule.logger.severe(ErrorMessages.ANONYMOUSID_FAILURE);

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.ANONYMOUSID_FAILURE)
					.build();
		}

		// Use a conversion to String in order to always include all 11 digits.
		final String response = String.format("%1$011d", recyclebinuserid);

		return Response.ok(new StreamingOutput() {
			@Override
			public void write(OutputStream output) throws IOException, WebApplicationException {
				generateQRCodePNG(response).writeTo(output);
			}
		}, "image/png").build();
	}

	@Override
	public Response ecoAccountEcoBalanceGET(@Size(min = 2, max = 2) String locale) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final EcoAccountRecord ecoaccount = ecoaccountdao.getEcoAccount(getPrincipalName());
		if (null == ecoaccount) {
			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.ECOACCOUNT_NOTFOUND).build();
		}

		final EcoBalance balance = new EcoBalance();
		try {
			balance.ecocredits(ecoaccount.getEcoBalance().getEcoCredit().getValue().doubleValue())
					.ecopoints(ecoaccount.getEcoBalance().getEcoDebit().getValue().doubleValue());
		} catch (Exception e) {
			EndUserModule.logger.severe(ErrorMessages.CONVERSION_ERROR + ", details: " + e.getMessage());

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.CONVERSION_ERROR)
					.build();
		}

		return Response.ok().entity(balance).build();
	}

	@Override
	public Response ecoAccountFeedbackProducersPOST(@Valid List<ProducersFeedbackForm> producersFeedbackForm) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		if (!POSTDataValidator.isValid(producersFeedbackForm)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_PRODUCERS_FEEDBACK)
					.build();
		}

		try {
			final ArrayList<UserProductFeedback> feedback = new ArrayList<UserProductFeedback>();
			for (final ProducersFeedbackForm form : producersFeedbackForm) {
				feedback.add(new UserProductFeedback().withFreetext(form.getComment())
						.withProductid(new ProductID(form.getProductgtin())).withUserscore(form.getUserscore())
						.withUseruuid(getPrincipalName()));
			}

			if (!ecoaccountdao.updateProductsFeedback(feedback)) {
				EndUserModule.logger.severe(ErrorMessages.PRODUCERS_FEEDBACK_FAILURE);

				return Response
						.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.PRODUCERS_FEEDBACK_FAILURE)
						.build();
			}

			// All OK
			return Response.ok().build();
		} catch (Exception e) {
			EndUserModule.logger.severe(ErrorMessages.PRODUCERS_FEEDBACK_FAILURE + ", details: " + e.getMessage());

			return Response
					.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.PRODUCERS_FEEDBACK_FAILURE)
					.build();
		}
	}

	@Override
	public Response ecoAccountFeedbackSystemPOST(@Valid SystemFeedbackForm systemFeedbackForm) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		if (!POSTDataValidator.isValid(systemFeedbackForm)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_SYSTEM_FEEDBACK).build();
		}

		try {
			final SystemFeedback feedback = new SystemFeedback().withFreetext(systemFeedbackForm.getComment())
					.withUserscore(systemFeedbackForm.getUserscore()).withUseruuid(getPrincipalName());
			if (!ecoaccountdao.updateSystemFeedback(feedback)) {
				EndUserModule.logger.severe(ErrorMessages.SYSTEM_FEEDBACK_FAILURE);

				return Response
						.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.SYSTEM_FEEDBACK_FAILURE)
						.build();
			}

			// All OK
			return Response.ok().build();
		} catch (Exception e) {
			EndUserModule.logger.severe(ErrorMessages.SYSTEM_FEEDBACK_FAILURE + ", details: " + e.getMessage());

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.SYSTEM_FEEDBACK_FAILURE)
					.build();
		}
	}

	@Override
	public Response ecoAccountForgotPasswordPOST(@Valid EcoAccountForgotPasswordForm ecoAccountForgotPasswordForm) {
		// Protect this endpoint by allowing only the pre-defined user to access it. If
		// left public, a malicious user can flood us with reset password requests and
		// eventually cause a DoS, even if the forms are not valid.
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		// Allow only the pre-defined user
		final String username = AccessControlManager.getUsername(getPrincipalName());
		if (null == username || 0 != username.compareToIgnoreCase("EndUserModule")) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_DENIED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_DENIED).build();
		}

		// Validate the registration form
		if (!POSTDataValidator.isValid(ecoAccountForgotPasswordForm)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_RESET_PASSWORD_FORM)
					.build();
		}

		// Check if email address given belongs to a registered user
		final String useruuid = AccessControlManager.getEmailUUID(ecoAccountForgotPasswordForm.getEmail());
		if (null == useruuid) {
			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.ECOACCOUNT_NOTFOUND).build();
		}

		// Send the reset email
		if (!AccessControlManager.sendResetPasswordEmail(useruuid)) {
			// Cannot contact email service to fulfill the request
			return Response.status(Status.BAD_GATEWAY.getStatusCode(), ErrorMessages.EMAIL_SERVER_ERROR).build();
		}

		// All OK, user is registered with this email and he has been sent an email with
		// password update instructions.
		return Response.ok().build();
	}

	@Override
	public Response ecoAccountUpdatePOST(@Valid EcoAccountUpdateForm ecoAccountUpdateForm) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}
		final String username = AccessControlManager.getUsername(getPrincipalName());
		// Validate the registration form
		if (!POSTDataValidator.isValid(username, ecoAccountUpdateForm)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_UPDATE_FORM).build();
		}

		try {
			final String useruuid = getPrincipalName();

			// Check if the updated email belongs to another user already, fail if it does
			if (null != ecoAccountUpdateForm.getEmail()) {
				final String useruuidtocheck = AccessControlManager.getEmailUUID(ecoAccountUpdateForm.getEmail());

				if (null != useruuidtocheck && !useruuid.equalsIgnoreCase(useruuidtocheck)) {
					EndUserModule.logger.severe(ErrorMessages.ECOACCOUNT_EXISTS + " with email "
							+ ecoAccountUpdateForm.getEmail() + " and UUID " + useruuidtocheck);

					return Response.status(Status.CONFLICT.getStatusCode(), ErrorMessages.ECOACCOUNT_EXISTS).build();
				}
			}

			// Check if the updated Recycle Bin ID belongs to another user already, fail if
			// it does
			if (null != ecoAccountUpdateForm.getRecyclebinuserid()) {
				final String useruuidtocheck = ecoaccountdao
						.getEcoAccountIDFromRecycleBinUserID(ecoAccountUpdateForm.getRecyclebinuserid());

				if (null != useruuidtocheck && !useruuid.equalsIgnoreCase(useruuidtocheck)) {
					EndUserModule.logger.severe(ErrorMessages.ECOACCOUNT_EXISTS + " with Recycle Bin ID "
							+ ecoAccountUpdateForm.getRecyclebinuserid() + " and UUID " + useruuidtocheck);

					return Response.status(Status.CONFLICT.getStatusCode(), ErrorMessages.ECOACCOUNT_EXISTS).build();
				}
			}

			// Try to update the account
			final Date birthday = (null == ecoAccountUpdateForm.getBirthday()) ? null
					: DataConversion.readDate(ecoAccountUpdateForm.getBirthday());
			final URL profileimgURL = (null == ecoAccountUpdateForm.getImageurl()) ? null
					: new URL(ecoAccountUpdateForm.getImageurl()),
					userURL = (null == ecoAccountUpdateForm.getUrl()) ? null : new URL(ecoAccountUpdateForm.getUrl()),
					orgURL = (null == ecoAccountUpdateForm.getOrganization()
							|| null == ecoAccountUpdateForm.getOrganization().getUrl()) ? null
									: new URL(ecoAccountUpdateForm.getOrganization().getUrl());
			final String country = (null == ecoAccountUpdateForm.getAddress()) ? null
					: ecoAccountUpdateForm.getAddress().getCountry(),
					postalcode = (null == ecoAccountUpdateForm.getAddress()) ? null
							: ecoAccountUpdateForm.getAddress().getPostalcode(),
					streetaddr = (null == ecoAccountUpdateForm.getAddress()) ? null
							: ecoAccountUpdateForm.getAddress().getStreet(),
					city = (null == ecoAccountUpdateForm.getAddress()) ? null
							: ecoAccountUpdateForm.getAddress().getCity(),
					region = (null == ecoAccountUpdateForm.getAddress()) ? null
							: ecoAccountUpdateForm.getAddress().getRegion(),
					orgName = (null == ecoAccountUpdateForm.getOrganization()) ? null
							: ecoAccountUpdateForm.getOrganization().getName(),
					locale = (null == ecoAccountUpdateForm.getLocale()) ? null : ecoAccountUpdateForm.getLocale();

			if (!ecoaccountdao.updateEcoAccount(useruuid, ecoAccountUpdateForm.getPassword(),
					ecoAccountUpdateForm.getEmail(), ecoAccountUpdateForm.getFirstname(),
					ecoAccountUpdateForm.getLastname(), birthday, country, postalcode, streetaddr, city, region,
					ecoAccountUpdateForm.getTelephone(), profileimgURL, userURL, ecoAccountUpdateForm.getGender(),
					orgName, orgURL, locale)) {
				EndUserModule.logger.severe("Cannot update Eco-account!");

				return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.UPDATE_ACCOUNT_ERROR)
						.build();
			}

			// Check if there is a Recycle Bin User ID and try to set it, do not fail the
			// update if this fails
			if (null != ecoAccountUpdateForm.getRecyclebinuserid()) {
				if (!mapRecycleBinUserID(useruuid, ecoAccountUpdateForm.getRecyclebinuserid())) {
					EndUserModule.logger.warning("Could not update Recycle Bin User ID! Eco-account ID: " + useruuid);
				}
			}

			// Check if the profile image requires upload and perform it if everything is OK
			// If anything goes wrong, do not fail the update, just set the default
			// profile picture
			if (null != ecoAccountUpdateForm.getImageupload() && null == ecoAccountUpdateForm.getImageurl()) {
				EndUserModule.logger
						.info("Eco-account update form contains uploaded profile image, trying to decode and upload.");

				final File profileimg = decodeAndSaveProfileImage(useruuid, ecoAccountUpdateForm.getImageupload());
				if (null == profileimg) {
					// Something went wrong with the image uploaded, use the default instead
					EndUserModule.logger.warning("Could not decode and save uploaded profile image!");

					// This account will not have set this property so it will be served with the
					// placeholder default.
					EndUserModule.logger.info("Setting default profile image...");
					if (!ecoaccountdao.updateEcoAccountTextProperty(useruuid, ECOACCOUNTPROFILEIMAGEPROPERTY, null)) {
						EndUserModule.logger
								.warning("Could not set Eco-account property " + ECOACCOUNTPROFILEIMAGEPROPERTY);
					}
				} else {
					// Update the Eco-account picture
					EndUserModule.logger.info("Saved uploaded profile image to " + profileimg.getPath()
							+ ", updating Eco-account record accordingly...");
					final URL uploadURL = new URL(m_uriInfo.getBaseUri().toString() + "ecoaccount/profile/image");
					if (!ecoaccountdao.updateEcoAccountTextProperty(useruuid, ECOACCOUNTPROFILEIMAGEPROPERTY,
							uploadURL.toString())) {
						EndUserModule.logger
								.warning("Could not set Eco-account property " + ECOACCOUNTPROFILEIMAGEPROPERTY);
					}
				}
			}

			// All OK, user profile updated properly
			return Response.ok().build();
		} catch (Exception e) {
			EndUserModule.logger.severe(ErrorMessages.UPDATE_ACCOUNT_ERROR + ", details: " + e.getMessage());

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.UPDATE_ACCOUNT_ERROR)
					.build();
		}
	}

	@Override
	public Response ecoAccountLoginPOST(@Valid EcoAccountLoginForm ecoAccountLoginForm) {
		if (!POSTDataValidator.isValid(ecoAccountLoginForm)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_LOGIN_FORM).build();
		}

		try {
			final LoginResult result = AccessControlManager.ecoAccountLogin(AccessControlManager.ENDUSERMODULE_CLIENTID,
					ecoAccountLoginForm.getUsername(), ecoAccountLoginForm.getPassword());

			switch (result.error_code) {
			case EMAIL_NOT_VERIFIED: {
				return Response.status(Status.CONFLICT.getStatusCode(), ErrorMessages.AUTHENTICATION_NOT_VERIFIED_YET)
						.build();
			}
			case WRONG_CREDENTIALS: {
				return Response.status(Status.UNAUTHORIZED.getStatusCode(), ErrorMessages.AUTHENTICATION_ERROR).build();
			}
			case OK:
			default: {
				final InlineResponse200 response = new InlineResponse200().accesstoken(result.tokens.getAccessToken())
						.refreshtoken(result.tokens.getRefreshToken());

				return Response.ok().entity(response).build();
			}
			}
		} catch (Exception e) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_ERROR + ", details: " + e.getMessage());

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.AUTHENTICATION_ERROR)
					.build();
		}
	}

	@Override
	public Response ecoAccountLogoutPOST(@Valid EcoAccountLogoutForm ecoAccountLogoutForm) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		if (!POSTDataValidator.isValid(ecoAccountLogoutForm)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_LOGOUT_FORM).build();
		}

		try {
			// Get current Access Token from Bearer header.
			final List<String> authheaders = m_httpHeaders.getRequestHeader("authorization");
			String bearer = null;
			final Pattern bearerregex = Pattern.compile("^bearer.*", Pattern.CASE_INSENSITIVE);
			for (final String authheader : authheaders) {
				if (bearerregex.matcher(authheader).matches()) {
					// We found the bearer authorization header
					bearer = authheader.substring(6, authheader.length()).trim();
					break;
				}
			}

			if (null == bearer) {
				EndUserModule.logger.severe("Could not extract Authorization Bearer Token from HTTP request headers!");

				return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.SESSIONLOGOUT_ERROR)
						.build();
			}

			final AccessTokens tokens = new AccessTokens(bearer, ecoAccountLogoutForm.getRefreshtoken());

			if (!AccessControlManager.endSession(AccessControlManager.ENDUSERMODULE_CLIENTID, tokens)) {
				EndUserModule.logger.severe(ErrorMessages.SESSIONLOGOUT_ERROR);

				return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.SESSIONLOGOUT_ERROR)
						.build();
			}

			return Response.ok().entity("Logout success").build();
		} catch (Exception e) {
			EndUserModule.logger.severe(ErrorMessages.SESSIONLOGOUT_ERROR + ", details: " + e.getMessage());

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.SESSIONLOGOUT_ERROR)
					.build();
		}
	}

	@Override
	public Response ecoAccountPurchaseRecycleHistoryGET(@Size(min = 2, max = 2) String locale) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final EcoAccountRecord ecoaccount = ecoaccountdao.getEcoAccount(getPrincipalName());
		if (null == ecoaccount) {
			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.ECOACCOUNT_NOTFOUND).build();
		}

		try {
			final InlineResponse2001 response = new InlineResponse2001();
			final ArrayList<Product> purchased = new ArrayList<Product>();
			final ArrayList<Product> recycled = new ArrayList<Product>();

			for (final IndividualProductData p : ecoaccountdao.getEcoAccountPurchasedList(getPrincipalName(), null,
					locale)) {
				purchased.add(convertProduct(p));
			}
			response.setPurchased(purchased);

			for (final IndividualProductData p : ecoaccountdao.getEcoAccountRecycledList(getPrincipalName(), locale)) {
				recycled.add(convertProduct(p));
			}
			response.setRecycled(recycled);

			return Response.ok().entity(response).build();
		} catch (Exception e) {
			EndUserModule.logger.severe(ErrorMessages.CONVERSION_ERROR + ", details: " + e.getMessage());

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.CONVERSION_ERROR)
					.build();
		}
	}

	@Override
	public Response ecoAccountRegistrationPOST(@Valid EcoAccountRegistrationForm ecoAccountRegistrationForm) {
		// Protect this endpoint by allowing only the pre-defined user to access it. If
		// left public, a malicious user can flood us with registration requests and
		// eventually cause a DoS, even if the registration forms are not valid.
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		// Allow only the pre-defined user
		final String username = AccessControlManager.getUsername(getPrincipalName());
		if (null == username || 0 != username.compareToIgnoreCase("EndUserModule")) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_DENIED);

			return Response.status(Status.UNAUTHORIZED.getStatusCode(), ErrorMessages.AUTHENTICATION_DENIED).build();
		}

		// Validate the registration form
		if (!POSTDataValidator.isValid(ecoAccountRegistrationForm)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_REGISTER_FORM).build();
		}

		// UUID of newly created user
		String useruuid = null;
		try {
			// Check if the current username or email already exists on a registered user
			if (null != AccessControlManager.getEmailUUID(ecoAccountRegistrationForm.getEmail())) {
				return Response.status(Status.CONFLICT.getStatusCode(), ErrorMessages.ECOACCOUNT_EXISTS).build();
			}

			if (null != AccessControlManager.getUsernameUUID(ecoAccountRegistrationForm.getUsername())) {
				return Response.status(Status.CONFLICT.getStatusCode(), ErrorMessages.ECOACCOUNT_EXISTS).build();
			}

			// Try to create the account
			final Date birthday = (null == ecoAccountRegistrationForm.getBirthday()) ? null
					: DataConversion.readDate(ecoAccountRegistrationForm.getBirthday());
			final URL profileimgURL = (null == ecoAccountRegistrationForm.getImageurl()) ? null
					: new URL(ecoAccountRegistrationForm.getImageurl()),
					userURL = (null == ecoAccountRegistrationForm.getUrl()) ? null
							: new URL(ecoAccountRegistrationForm.getUrl()),
					orgURL = (null == ecoAccountRegistrationForm.getOrganization()
							|| null == ecoAccountRegistrationForm.getOrganization().getUrl()) ? null
									: new URL(ecoAccountRegistrationForm.getOrganization().getUrl());
			final String country = (null == ecoAccountRegistrationForm.getAddress()) ? null
					: ecoAccountRegistrationForm.getAddress().getCountry(),
					postalcode = (null == ecoAccountRegistrationForm.getAddress()) ? null
							: ecoAccountRegistrationForm.getAddress().getPostalcode(),
					streetaddr = (null == ecoAccountRegistrationForm.getAddress()) ? null
							: ecoAccountRegistrationForm.getAddress().getStreet(),
					city = (null == ecoAccountRegistrationForm.getAddress()) ? null
							: ecoAccountRegistrationForm.getAddress().getCity(),
					region = (null == ecoAccountRegistrationForm.getAddress()) ? null
							: ecoAccountRegistrationForm.getAddress().getRegion(),
					orgName = (null == ecoAccountRegistrationForm.getOrganization()) ? null
							: ecoAccountRegistrationForm.getOrganization().getName(),
					locale = (null == ecoAccountRegistrationForm.getLocale()) ? null
							: ecoAccountRegistrationForm.getLocale();

			useruuid = null; // Just to be sure
			useruuid = ecoaccountdao.createEcoAccount(ecoAccountRegistrationForm.getUsername(),
					ecoAccountRegistrationForm.getPassword(), ecoAccountRegistrationForm.getEmail(),
					ecoAccountRegistrationForm.getFirstname(), ecoAccountRegistrationForm.getLastname(), birthday,
					country, postalcode, streetaddr, city, region, ecoAccountRegistrationForm.getTelephone(),
					profileimgURL, userURL, ecoAccountRegistrationForm.getGender(), orgName, orgURL, locale);

			// User account creation failed
			if (null == useruuid) {
				EndUserModule.logger.severe("Cannot create Eco-account!");

				return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.CREATE_ACCOUNT_ERROR)
						.build();
			}

			// Check if there is a Recycle Bin User ID and try to set it, do not fail the
			// registration if this fails
			if (!mapRecycleBinUserID(useruuid, ecoAccountRegistrationForm.getRecyclebinuserid())) {
				EndUserModule.logger.warning("Could not set Recycle Bin User ID! Eco-account ID: " + useruuid);
			}

			// Check if the profile image requires upload and perform it if everything is OK
			// If anything goes wrong, do not fail the registration, just set the default
			// profile picture
			if (null != ecoAccountRegistrationForm.getImageupload()
					&& null == ecoAccountRegistrationForm.getImageurl()) {
				EndUserModule.logger.info(
						"Eco-account registration form contains uploaded profile image, trying to decode and upload.");

				final File profileimg = decodeAndSaveProfileImage(useruuid,
						ecoAccountRegistrationForm.getImageupload());
				if (null == profileimg) {
					// Something went wrong with the image uploaded, use the default instead
					EndUserModule.logger.warning("Could not decode and save uploaded profile image!");

					// This account will not have set this property so it will be served with the
					// placeholder default.
					EndUserModule.logger.info("Setting default profile image...");
					if (!ecoaccountdao.updateEcoAccountTextProperty(useruuid, ECOACCOUNTPROFILEIMAGEPROPERTY, null)) {
						EndUserModule.logger
								.warning("Could not set Eco-account property " + ECOACCOUNTPROFILEIMAGEPROPERTY);
					}
				} else {
					// Update the Eco-account picture
					EndUserModule.logger.info("Saved uploaded profile image to " + profileimg.getPath()
							+ ", updating Eco-account record accordingly...");
					final URL uploadURL = new URL(m_uriInfo.getBaseUri().toString() + "ecoaccount/profile/image");
					if (!ecoaccountdao.updateEcoAccountTextProperty(useruuid, ECOACCOUNTPROFILEIMAGEPROPERTY,
							uploadURL.toString())) {
						EndUserModule.logger
								.warning("Could not set Eco-account property " + ECOACCOUNTPROFILEIMAGEPROPERTY);
					}
				}
			}

			// All OK, user is registered and should have received an email with
			// instructions on how to activate account
			return Response.ok().build();
		} catch (Exception e) {
			EndUserModule.logger.severe(ErrorMessages.CREATE_ACCOUNT_ERROR + ", details: " + e.getMessage());

			// Something unhandled and serious happened, we should remove the Eco-account if
			// already created.
			if (AccessControlManager.existsUserUUID(useruuid)) {
				if (!ecoaccountdao.removeEcoAccount(useruuid)) {
					// Well if this also fails, nothing we can do than log the error.
					EndUserModule.logger
							.severe("Could not remove user [" + useruuid + "] from ACM after registration error!");
				}
			}

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.CREATE_ACCOUNT_ERROR)
					.build();
		}
	}

	@Override
	public Response ecoAccountResignationPOST(@Valid EcoAccountResignationForm ecoAccountResignationForm) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		// Validate the registration form
		if (!POSTDataValidator.isValid(ecoAccountResignationForm)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_RESIGN_FORM).build();
		}

		// Perform the resignation IF AND ONLY IF, the provided username and email
		// belong to the currently logged in user.
		final String useruuid = getPrincipalName();
		try {
			// Get all information in ACM and Database and make sure everything is in sync
			// in order to remove the user ONCE AND FOR ALL.
			final EcoAccountRecord ecoaccount = ecoaccountdao.getEcoAccount(useruuid);
			final String email = ecoAccountResignationForm.getEmail(),
					acmemail = AccessControlManager.getEmail(useruuid),
					dbemail = (null == ecoaccount) ? null : ecoaccount.getEcoAccount().getEmail();
			if (null == acmemail || 0 != email.compareToIgnoreCase(acmemail) || null == dbemail
					|| !email.contentEquals(dbemail)) {
				// The logged in user principal UUID is not in sync with the ACM and Database.
				// Malicious attack?
				return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_RESIGN_FORM).build();
			}

			// Remove the account, bye bye dear user, you will be missed.
			if (!ecoaccountdao.removeEcoAccount(useruuid)) {
				EndUserModule.logger.severe(ErrorMessages.REMOVE_ACCOUNT_ERROR);

				return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.REMOVE_ACCOUNT_ERROR)
						.build();
			}

			// All OK, user is not registered any more and all his data are removed from the
			// system.
			return Response.ok().build();
		} catch (Exception e) {
			EndUserModule.logger.severe(ErrorMessages.REMOVE_ACCOUNT_ERROR + ", details: " + e.getMessage());

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.REMOVE_ACCOUNT_ERROR)
					.build();
		}
	}

	@Override
	public Response ecoAccountUserRecordGET(@Size(min = 2, max = 2) String locale) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final EcoAccountRecord ecoaccount = ecoaccountdao.getEcoAccount(getPrincipalName());
		if (null == ecoaccount) {
			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.ECOACCOUNT_NOTFOUND).build();
		}

		return Response.ok().entity(ecoaccount).build();
	}

	/**
	 * @return the user name from an authenticated session, <code>null</code>
	 *         otherwise.
	 */
	private String getPrincipalName() {
		final String principalname = AccessControlManager.isEnabled()
				? AccessControlManager.getPrincipalName(m_securityContext)
				: null;

		EndUserModule.logger.finest("User session principal name: " + principalname);

		return principalname;
	}

	/**
	 * Generate a QR Code PNG object from the input String using UTF-8 character
	 * set.
	 * 
	 * @param input the {@link String} input text to use for generating the QR code.
	 * @return the {@link QRCode} object with PNG image type.
	 */
	private QRCode generateQRCodePNG(final String input) {
		return QRCode.from(input).withCharset("UTF-8").to(ImageType.PNG);
	}

	/**
	 * Convert from internal Data Model of individual product to OpenAPI model.
	 * 
	 * @param individualproduct the object that contains the data.
	 * @return the converted object, or <code>null</code> in case of any error.
	 */
	private Product convertProduct(final IndividualProductData individualproduct) {
		final Product product = new Product();

		// If the IndividualProductData object comes from the Eco-account recycle list
		// it may contain a non-valid GTIN from a recycled product that the evaluator
		// knew the type but had no access to a GTIN number. So we will try our best to
		// return an object, even without a GTIN.
		Long gtin = -1L;
		try {
			gtin = Long.parseLong(individualproduct.getGtin14());
		} catch (Exception e) {
			gtin = -1L;
		}

		// Try also our best to get the Eco-credits if they are present
		double ecocredits = 0.0;
		try {
			ecocredits = individualproduct.getEcoCredits().getValue();
		} catch (Exception e) {
			ecocredits = 0.0;
		}

		try {
			final ProductType prodtype = new ProductType().hskey(individualproduct.getProductType().HSkey);
			final String serialnum = individualproduct.getSerialNumber(), name = individualproduct.getName(),
					manufacturer = null == individualproduct.getManufacturer() ? null
							: individualproduct.getManufacturer().getName(),
					model = individualproduct.getProductID(), purchasedate = individualproduct.getPurchaseDate(),
					recycledate = individualproduct.getRecycleDate(), price = individualproduct.getPrice();

			product.gtin(gtin).type(prodtype).serialnum(serialnum).manufacturer(manufacturer).model(model).name(name)
					.price(price).purchasedate(purchasedate).recycledate(recycledate).ecocredits(ecocredits);
		} catch (Exception e) {
			EndUserModule.logger.severe(ErrorMessages.CONVERSION_ERROR + " during conversion of " + individualproduct
					+ ", details: " + e.getMessage());

			return null;
		}

		return product;
	}

	/**
	 * Decode the specified Base64 ASCII {@link String}, check if the contents are a
	 * PNG or JPG image of 256KiB maximum size and save it locally. Return the saved
	 * file path.
	 * 
	 * @param useruuid    the username of the Eco-account as represented in the user
	 *                    session security context.
	 * @param base64Image {@link String} containing Base64 ASCII encoded PNG or JPG
	 *                    image data.
	 * @return the saved image {@link File} on success, <code>null</code> otherwise.
	 * @throws IOException if an I/O error occurs writing to or creating the file.
	 */
	private File decodeAndSaveProfileImage(final String useruuid, final String base64Image) throws IOException {
		if (null == useruuid || null == base64Image) {
			EndUserModule.logger.severe("Cannot decode user registration profile image upload, null parameters!");

			return null;
		}

		final File fileuploadsdir = EndUserModule.getFileUploadsPath();
		if (null == fileuploadsdir) {
			EndUserModule.logger.severe(
					"Cannot access the file uploads directory, user registration profile image upload cannot be saved!");

			return null;
		}

		// Decode Base64 ASCII to image data bytes, input encoded String is in UTF-8
		final byte[] decoded = Base64.getDecoder().decode(base64Image.getBytes("UTF-8"));
		// Check image size, max allowed 256KiB
		if (decoded.length > EndUserModule.PROFILE_IMAGE_SIZE_LIMIT) {
			EndUserModule.logger.severe("User registration profile image exceeds limit of "
					+ EndUserModule.PROFILE_IMAGE_SIZE_LIMIT + " bytes");

			return null;
		}

		// Find out which image type the bytes represent, JPG or PNG, by checking the
		// magic bytes at the start of the array:
		// JPEG: https://en.wikipedia.org/wiki/JPEG (magic number: ff d8 ff)
		// PNG: https://en.wikipedia.org/wiki/Portable_Network_Graphics (magic number:
		// 89 50 4e 47 0d 0a 1a 0a)
		String imageext = null;
		if ((byte) 0xff == decoded[0] && (byte) 0xd8 == decoded[1] && (byte) 0xff == decoded[2]) {
			imageext = "jpg";
		} else if ((byte) 0x89 == decoded[0] && (byte) 0x50 == decoded[1] && (byte) 0x4e == decoded[2]
				&& (byte) 0x47 == decoded[3] && (byte) 0x0d == decoded[4] && (byte) 0x0a == decoded[5]
				&& (byte) 0x1a == decoded[6] && (byte) 0x0a == decoded[7]) {
			imageext = "png";
		} else {
			EndUserModule.logger.severe("Unknown format in user registration profile image upload!");

			return null;
		}

		// Create a child directory for final save target
		final File profileimagedir = EndUserModule.createProfileImagePath(useruuid);
		// Save the image in the final target directory
		final File profileimage = new File(profileimagedir, EndUserModule.PROFILE_IMAGE_PREFIX + "." + imageext);
		final FileOutputStream fos = new FileOutputStream(profileimage);
		fos.write(decoded);
		fos.close();

		// All OK
		return profileimage;
	}

	private boolean mapRecycleBinUserID(final String useruuid, final Long RecycleBinUserID) {
		// Generate a new unique Recycle Bin User ID
		long recyclebinuuid = generateRecycleBinUserID();

		if (null != RecycleBinUserID) {
			// Get the provided Recycle Bin User ID if it is set
			recyclebinuuid = RecycleBinUserID.longValue();
		}

		// Check that it does not already exist and repeat generation until we find a
		// Recycle Bin User ID that is not already mapped.
		String acmuuid = ecoaccountdao.getEcoAccountIDFromRecycleBinUserID(recyclebinuuid);
		while (null != acmuuid) {
			recyclebinuuid = generateRecycleBinUserID();
			acmuuid = ecoaccountdao.getEcoAccountIDFromRecycleBinUserID(recyclebinuuid);
		}

		// Set the generated recycle bin user ID to the newly created Eco-account
		if (!ecoaccountdao.updateEcoAccountWithRecycleBinUserID(useruuid, recyclebinuuid)) {
			return false;
		}

		// All OK
		return true;
	}

	private long generateRecycleBinUserID() {
		return ThreadLocalRandom.current().nextLong(0L, 99999999999L + 1L);
	}
}
