/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.openapi;

import javax.validation.constraints.Size;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.core.Response.Status;

import org.openapitools.api.EcoshoppingApi;

import gr.iccs.isense.circ4life.dao.CommonDAO.ProductRecord;
import gr.iccs.isense.circ4life.dao.ProductMasterDataDAO;
import gr.iccs.isense.circ4life.json.JacksonJSONSerialization;
import gr.iccs.isense.circ4life.rest.EndUserModule;
import gr.iccs.isense.circ4life.rest.EndUserModule.ErrorMessages;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;

/**
 * /ecoshopping end-point implementation.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EcoShopping extends EcoshoppingApi {
	/**
	 * Product Master Data Access Object.
	 */
	private static final ProductMasterDataDAO productmasterdatadao = new ProductMasterDataDAO();

	/**
	 * Proxy of security context for session management.
	 */
	private final SecurityContext m_securityContext;

	/**
	 * Constructor.
	 * 
	 * @param securityContext proxy of security context for session management.
	 */
	public EcoShopping(@Context SecurityContext securityContext) {
		m_securityContext = securityContext;
	}

	@Override
	public Response ecoShoppingEcoPointsGET(@Size(min = 14) String productid) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final Double ecopoints = productmasterdatadao.getProductEcoPoints(productid);
		if (null == ecopoints) {
			EndUserModule.logger.severe(ErrorMessages.PRODUCT_NOTFOUND);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.PRODUCT_NOTFOUND).build();
		}

		return Response.ok().entity(ecopoints).build();
	}

	@Override
	public Response ecoShoppingProductGET(@Size(min = 14) String productid, @Size(min = 2, max = 2) String locale) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final ProductRecord productrec = productmasterdatadao.getProductRecord(productid, locale);
		if (null == productrec) {
			EndUserModule.logger.severe(ErrorMessages.PRODUCT_NOTFOUND);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.PRODUCT_NOTFOUND).build();
		}

		return Response.ok().entity(productrec).build();
	}

	@Override
	public Response ecoShoppingProductManufacturerGET(@Size(min = 14) String productid,
			@Size(min = 2, max = 2) String locale) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final ProductRecord productrec = productmasterdatadao.getProductRecord(productid, locale);
		if (null == productrec) {
			EndUserModule.logger.severe(ErrorMessages.PRODUCT_NOTFOUND);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.PRODUCT_NOTFOUND).build();
		}

		try {
			return Response.ok().entity(productrec.getMasterData().getProductData().getManufacturer()).build();
		} catch (Exception e) {
			// Something really wrong with the product master data record
			EndUserModule.logger.severe(ErrorMessages.PRODUCT_MALFORMED_RECORD + ", Details: "
					+ JacksonJSONSerialization.JSONtoString(productrec));

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.PRODUCT_MALFORMED_RECORD)
					.build();
		}
	}

	@Override
	public Response ecoShoppingProductReusabilityGET(@Size(min = 14) String productid,
			@Size(min = 2, max = 2) String locale) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			EndUserModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final ProductRecord productrec = productmasterdatadao.getProductRecord(productid, locale);
		if (null == productrec) {
			EndUserModule.logger.severe(ErrorMessages.PRODUCT_NOTFOUND);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.PRODUCT_NOTFOUND).build();
		}

		try {
			// TODO: Should we return something else apart from Recources?
			return Response.ok().entity(productrec.getMasterData().getLifeCycle().getResources()).build();
		} catch (Exception e) {
			// Something really wrong with the product master data record
			EndUserModule.logger.severe(ErrorMessages.PRODUCT_MALFORMED_RECORD + ", Details: "
					+ JacksonJSONSerialization.JSONtoString(productrec));

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.PRODUCT_MALFORMED_RECORD)
					.build();
		}
	}
}
