/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.validation;

import java.util.Date;
import java.util.List;

import org.openapitools.model.EcoAccountForgotPasswordForm;
import org.openapitools.model.EcoAccountLoginForm;
import org.openapitools.model.EcoAccountLogoutForm;
import org.openapitools.model.EcoAccountRegistrationForm;
import org.openapitools.model.EcoAccountResignationForm;
import org.openapitools.model.EcoAccountUpdateForm;
import org.openapitools.model.ProducersFeedbackForm;
import org.openapitools.model.SystemFeedbackForm;

import gr.iccs.isense.circ4life.rest.EndUserModule;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;
import gr.iccs.isense.circ4life.validation.EmailAddressValidator;
import gr.iccs.isense.circ4life.validation.EndUserRegistrationFormValidator;
import gr.iccs.isense.circ4life.validation.EndUserResignationFormValidator;
import gr.iccs.isense.circ4life.validation.EndUserSessionFormValidator;

/**
 * POST data structures validator.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class POSTDataValidator {
	/**
	 * Validate POST feedback forms for producers.
	 * 
	 * @param producersFeedbackForm the feedback forms for producers.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final List<ProducersFeedbackForm> producersFeedbackForm) {
		try {
			int count = 0;
			for (final ProducersFeedbackForm form : producersFeedbackForm) {
				form.getProductgtin().longValue();
				final int score = form.getUserscore().intValue();
				if (1 > score || 5 < score) {
					throw new IllegalArgumentException("Producers feedback score out of range [1,5]");
				}
				++count;
			}

			if (0 == count) {
				// Empty list, not allowed
				throw new IllegalArgumentException("Empty producers feedback form list");
			}

			// All OK
			return true;
		} catch (Exception e) {
			EndUserModule.logger.severe("User producers feedback form validation error: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Validate POST feedback form for system.
	 * 
	 * @param systemFeedbackForm the feedback form for system.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final SystemFeedbackForm systemFeedbackForm) {
		try {
			final int score = systemFeedbackForm.getUserscore().intValue();
			if (1 > score || 5 < score) {
				throw new IllegalArgumentException("System feedback score out of range [1,5]");
			}

			// All OK
			return true;
		} catch (Exception e) {
			EndUserModule.logger.severe("User system feedback form validation error: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Validate POST Eco-account login form.
	 * 
	 * @param ecoAccountLoginForm the Eco-account login form.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final EcoAccountLoginForm ecoAccountLoginForm) {
		try {
			return EndUserSessionFormValidator.isValid(ecoAccountLoginForm.getUsername(),
					ecoAccountLoginForm.getPassword());
		} catch (Exception e) {
			EndUserModule.logger.severe("User login form validation error: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Validate POST Eco-account logout form.
	 * 
	 * @param ecoAccountLogoutForm the Eco-account logout form.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final EcoAccountLogoutForm ecoAccountLogoutForm) {
		try {
			return EndUserSessionFormValidator.isValid(ecoAccountLogoutForm.getRefreshtoken());
		} catch (Exception e) {
			EndUserModule.logger.severe("User logout form validation error: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Validate POST Eco-account registration form.
	 * 
	 * @param ecoAccountRegistrationForm the Eco-account registration form.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final EcoAccountRegistrationForm ecoAccountRegistrationForm) {
		try {
			final Date birthday = (null == ecoAccountRegistrationForm.getBirthday()) ? null
					: DataConversion.readDate(ecoAccountRegistrationForm.getBirthday());
			final String country = (null == ecoAccountRegistrationForm.getAddress()) ? null
					: ecoAccountRegistrationForm.getAddress().getCountry(),
					postalcode = (null == ecoAccountRegistrationForm.getAddress()) ? null
							: ecoAccountRegistrationForm.getAddress().getPostalcode(),
					streetaddr = (null == ecoAccountRegistrationForm.getAddress()) ? null
							: ecoAccountRegistrationForm.getAddress().getStreet(),
					city = (null == ecoAccountRegistrationForm.getAddress()) ? null
							: ecoAccountRegistrationForm.getAddress().getCity(),
					region = (null == ecoAccountRegistrationForm.getAddress()) ? null
							: ecoAccountRegistrationForm.getAddress().getRegion(),
					orgname = (null == ecoAccountRegistrationForm.getOrganization()) ? null
							: ecoAccountRegistrationForm.getOrganization().getName(),
					orgurl = (null == ecoAccountRegistrationForm.getOrganization()) ? null
							: ecoAccountRegistrationForm.getOrganization().getUrl();

			// User profile image is not tested here
			return EndUserRegistrationFormValidator.isValid(ecoAccountRegistrationForm.getUsername(), // REQUIRED -> ACM
					ecoAccountRegistrationForm.getPassword(), // REQUIRED -> ACM
					ecoAccountRegistrationForm.getEmail(), // REQUIRED -> ACM,DB
					ecoAccountRegistrationForm.getFirstname(), // REQUIRED -> ACM,DB
					ecoAccountRegistrationForm.getLastname(), // REQUIRED -> ACM,DB
					birthday, // OPTIONAL -> DB
					country, // OPTIONAL -> DB
					postalcode, // OPTIONAL -> DB
					streetaddr, // OPTIONAL -> DB
					city, // OPTIONAL -> DB
					region, // OPTIONAL -> DB
					ecoAccountRegistrationForm.getTelephone(), // OPTIONAL -> DB
					ecoAccountRegistrationForm.getUrl(), // OPTIONAL -> DB
					ecoAccountRegistrationForm.getGender(), // OPTIONAL -> DB
					orgname, // OPTIONAL -> DB
					orgurl, // OPTIONAL -> DB
					ecoAccountRegistrationForm.getRecyclebinuserid() // OPTIONAL -> DB
			);
		} catch (Exception e) {
			EndUserModule.logger.severe("User registration form validation error: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Validate POST Eco-account resignation form.
	 * 
	 * @param ecoAccountResignationForm the Eco-account resignation form.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final EcoAccountResignationForm ecoAccountResignationForm) {
		try {
			return EndUserResignationFormValidator.isValid(ecoAccountResignationForm.getEmail());
		} catch (Exception e) {
			EndUserModule.logger.severe("User resignation form validation error: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Validate POST Eco-account reset password form.
	 * 
	 * @param ecoAccountForgotPasswordForm the Eco-account reset password form.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final EcoAccountForgotPasswordForm ecoAccountForgotPasswordForm) {
		try {
			return EmailAddressValidator.isValid(ecoAccountForgotPasswordForm.getEmail());
		} catch (Exception e) {
			EndUserModule.logger.severe("User reset password form validation error: " + e.getMessage());

			return false;
		}
	}

	/**
	 * Validate POST Eco-account profile update form.
	 * 
	 * @param username             The human friendly username as returned by the
	 *                             Access Control Manager.
	 * @param ecoAccountUpdateForm the Eco-account profile update form.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final String username, final EcoAccountUpdateForm ecoAccountUpdateForm) {
		try {
			// Flag that gets true, whenever a field is not null in the form. If all fields
			// are null, this flag will cause the validation to fail.
			boolean fieldfound = false;

			final String addressCountry = (null == ecoAccountUpdateForm.getAddress()) ? null
					: ecoAccountUpdateForm.getAddress().getCountry(),
					postalCode = (null == ecoAccountUpdateForm.getAddress()) ? null
							: ecoAccountUpdateForm.getAddress().getPostalcode(),
					streetAddress = (null == ecoAccountUpdateForm.getAddress()) ? null
							: ecoAccountUpdateForm.getAddress().getStreet(),
					addressLocality = (null == ecoAccountUpdateForm.getAddress()) ? null
							: ecoAccountUpdateForm.getAddress().getCity(),
					addressRegion = (null == ecoAccountUpdateForm.getAddress()) ? null
							: ecoAccountUpdateForm.getAddress().getRegion();
			if (null != addressCountry || null != postalCode || null != streetAddress || null != addressLocality
					|| null != addressRegion) {
				fieldfound = true;
			}

			final Date birthDate = (null == ecoAccountUpdateForm.getBirthday()) ? null
					: DataConversion.readDate(ecoAccountUpdateForm.getBirthday());
			if (null != birthDate) {
				fieldfound = true;
			}

			final String email = ecoAccountUpdateForm.getEmail();
			if (null != email) {
				fieldfound = true;
				EndUserRegistrationFormValidator.validateEmail(email);
			}

			final String givenName = ecoAccountUpdateForm.getFirstname();
			if (null != givenName) {
				fieldfound = true;
				EndUserRegistrationFormValidator.validateGivenName(givenName);
			}

			final String gender = ecoAccountUpdateForm.getGender();
			if (null != gender) {
				fieldfound = true;
			}

			if (null != ecoAccountUpdateForm.getImageupload()) {
				fieldfound = true;
			}

			final String imageURL = ecoAccountUpdateForm.getImageurl();
			if (null != imageURL) {
				fieldfound = true;
				EndUserRegistrationFormValidator.validateHTTPURL(imageURL);
			}

			final String familyName = ecoAccountUpdateForm.getLastname();
			if (null != familyName) {
				fieldfound = true;
				EndUserRegistrationFormValidator.validateFamilyName(familyName);
			}

			final String organization = (null == ecoAccountUpdateForm.getOrganization()) ? null
					: ecoAccountUpdateForm.getOrganization().getName(),
					organizationURL = (null == ecoAccountUpdateForm.getOrganization()) ? null
							: ecoAccountUpdateForm.getOrganization().getUrl();
			if (null != organization || null != organizationURL) {
				fieldfound = true;
			}

			final String password = ecoAccountUpdateForm.getPassword();
			if (null != password) {
				fieldfound = true;
				EndUserRegistrationFormValidator.validateCredentials(username, password);
			}

			final Long recyclebinuserid = ecoAccountUpdateForm.getRecyclebinuserid();
			if (null != recyclebinuserid) {
				fieldfound = true;
			}

			final String telephone = ecoAccountUpdateForm.getTelephone();
			if (null != telephone) {
				fieldfound = true;
			}

			final String userURL = ecoAccountUpdateForm.getUrl();
			if (null != userURL) {
				fieldfound = true;
			}

			// Start checking the fields one-by-one.
			// Did we manage to have at least one field set? if not short-circuit and skip
			// the validation.
			if (!fieldfound) {
				EndUserModule.logger.severe("Empty Eco-account profile update form!");

				return false;
			} else {
				// Use the registration validator methods, the fields are the same.
				// password, email, givenName, familyName checked in place above
				EndUserRegistrationFormValidator.validate(birthDate, addressCountry, postalCode, streetAddress,
						addressLocality, addressRegion, telephone, userURL, gender, organization, organizationURL,
						recyclebinuserid);

				// All OK
				return true;
			}
		} catch (Exception e) {
			EndUserModule.logger.severe("User profile update form validation error: " + e.getMessage());

			return false;
		}
	}
}
