/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.test;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import org.junit.Assert;
import org.junit.Test;
import org.openapitools.model.Address;
import org.openapitools.model.EcoAccountForgotPasswordForm;
import org.openapitools.model.EcoAccountLoginForm;
import org.openapitools.model.EcoAccountLogoutForm;
import org.openapitools.model.EcoAccountRegistrationForm;
import org.openapitools.model.EcoAccountUpdateForm;
import org.openapitools.model.Organization;
import org.openapitools.model.ProducersFeedbackForm;
import org.openapitools.model.Product;
import org.openapitools.model.ProductType;
import org.openapitools.model.SystemFeedbackForm;

import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import gr.iccs.isense.circ4life.json.JacksonJSONSerialization;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;
import gr.iccs.isense.circ4life.rest.resources.validation.POSTDataValidator;

/**
 * Tests for EndUserModule REST Web service.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class EndUserModuleWSTests {
	private final static String VALIDFORMS = "/EndUserRegistrationFormsValid.txt";
	private final static String INVALIDFORMS = "/EndUserRegistrationFormsInvalid.txt";

	private final Calendar calendar = Calendar.getInstance();
	private final String today = DataConversion.formatDate(Date.from(calendar.toInstant()));
	private final ProductType productType = new ProductType();
	private final Product product = new Product();
	private final ProducersFeedbackForm producersFeedbackForm = new ProducersFeedbackForm();
	private final SystemFeedbackForm systemFeedbackForm = new SystemFeedbackForm();
	private final EcoAccountLoginForm ecoAccountLoginForm = new EcoAccountLoginForm();
	private final EcoAccountLogoutForm ecoAccountLogoutForm = new EcoAccountLogoutForm();
	private final EcoAccountRegistrationForm ecoAccountRegistrationForm = new EcoAccountRegistrationForm();
	private final EcoAccountForgotPasswordForm ecoAccountForgotPasswordForm = new EcoAccountForgotPasswordForm();
	private final String validusername = "username";
	private final String validpassword = "123456789012345";
	private final String shortpassword = "1234567";
	private final String invalidemail = "user@127.0.0.1";
	private final String imageURL = "https://circ4life.iccs.gr/images/placeholders/ecoaccount.png";
	private final Long validrecyclebinuserid = 16004000001L;
	private final Long invalidrecyclebinuserid = -1L;
	private final int currentyearminus4 = Calendar.getInstance().get(Calendar.YEAR) - 4;
	private String tooLongFreeText = createStringWithLength(257);

	@Test
	public void testEcoAccountRegistrationForms() throws IOException {
		System.out.println("USER REGISTRATION FORM");
		System.out.println("---------------------------------------------");
		{
			// VALID
			System.out.println("VALID POST DATA");

			final BufferedReader reader = new BufferedReader(
					new InputStreamReader(EndUserModuleWSTests.class.getResourceAsStream(VALIDFORMS), "UTF-8"));

			String form;
			while (null != (form = reader.readLine())) {
				try {
					System.out.println("Form -> " + form);

					// deserialize and validate form
					Assert.assertTrue(POSTDataValidator
							.isValid(JacksonJSONSerialization.StringtoJSON(form, EcoAccountRegistrationForm.class)));
				} catch (Exception e) {
					System.err.println(e.getMessage());
					// Should NOT reach here
					Assert.assertTrue(false);
				}
			}

			reader.close();

			System.out.println("---------------------------------------------");
		}

		{
			System.out.println("INVALID POST DATA");

			// INVALID
			final BufferedReader reader = new BufferedReader(
					new InputStreamReader(EndUserModuleWSTests.class.getResourceAsStream(INVALIDFORMS), "UTF-8"));

			String form;
			while (null != (form = reader.readLine())) {
				try {
					System.out.println("Form -> " + form);

					// deserialize and validate form
					Assert.assertFalse(POSTDataValidator
							.isValid(JacksonJSONSerialization.StringtoJSON(form, EcoAccountRegistrationForm.class)));
				} catch (Exception e) {
					System.err.println(e.getMessage());
					// Should reach here
					Assert.assertTrue(true);
				}
			}

			reader.close();

			System.out.println("---------------------------------------------");
		}
	}

	@Test
	public void testPOSTDataValidator() throws IOException {
		// ---------------------------------------------------------------------
		{
			System.out.println("USER LOGIN FORM");
			System.out.println("---------------------------------------------");

			System.out.println("VALID POST DATA");
			// User login form
			final EcoAccountLoginForm[] valid = { // VALID
					new EcoAccountLoginForm().username(validusername).password(validpassword),
					new EcoAccountLoginForm().username("user@example.com").password(validpassword)
					// End of tests
			};
			for (final EcoAccountLoginForm form : valid) {
				System.out.println(JacksonJSONSerialization.JSONtoString(form));
				Assert.assertTrue(POSTDataValidator.isValid(form));
			}

			System.out.println("---------------------------------------------");

			System.out.println("INVALID POST DATA");
			final EcoAccountLoginForm[] invalid = { // INVALID
					// Invalid usernames
					new EcoAccountLoginForm().username("user").password(validpassword),
					new EcoAccountLoginForm().username("user{-/*\\\"=`").password(validpassword),
					new EcoAccountLoginForm().username(invalidemail).password(validpassword),
					// Invalid passwords
					new EcoAccountLoginForm().username(validusername).password(shortpassword),
					new EcoAccountLoginForm().username(validusername).password(validusername),
					// End of tests
			};
			for (final EcoAccountLoginForm form : invalid) {
				System.out.println(JacksonJSONSerialization.JSONtoString(form));
				Assert.assertFalse(POSTDataValidator.isValid(form));
			}

			System.out.println("---------------------------------------------");
		}
		// ---------------------------------------------------------------------

		// ---------------------------------------------------------------------
		{
			System.out.println("USER LOGOUT FORM");
			System.out.println("---------------------------------------------");

			System.out.println("VALID POST DATA");
			// User logout form
			final EcoAccountLogoutForm[] valid = { // VALID
					new EcoAccountLogoutForm().refreshtoken(
							"eyJhbGciOiJIUzI1NiIsInR5cCIgOiAiSldUIiwia2lkIiA6ICI1ZWI2NWZjOS1iOGU1LTQ3YzUtYjFjMS1hYjcyN2Y5ZGZiMDMifQ."
									+ "eyJqdGkiOiJkNGQxMjc2OS1iOTk5LTQ2YWYtYmI0ZS1hNzg4NWI1ZGIwZmEiLCJleHAiOjE1Njk1MTI5NzMsIm5iZiI6MCwiaWF0Ij"
									+ "oxNTY5NTExMTczLCJpc3MiOiJodHRwOi8vMTI3LjAuMC4xOjgwODAvYXV0aC9yZWFsbXMvQ0lSQzRMSUZFIiwiYXVkIjoiaHR0cDovL"
									+ "zEyNy4wLjAuMTo4MDgwL2F1dGgvcmVhbG1zL0NJUkM0TElGRSIsInN1YiI6ImM3MDgxODA4LTU5NTctNDY5My1hZjMyLTIzMTJhMDY5Y"
									+ "WRjNSIsInR5cCI6IlJlZnJlc2giLCJhenAiOiJhZG1pbi1jbGkiLCJhdXRoX3RpbWUiOjAsInNlc3Npb25fc3RhdGUiOiJlNTEwMWEzNi"
									+ "00YzBjLTRjNzAtYjM1NS05NGFhMDE1MWMyYWMiLCJzY29wZSI6ImVtYWlsIHByb2ZpbGUifQ.e21mNIDVJYNlt042xs3gs7h7bTwkJ8frLaCEp-v03jc"),
					// End of tests
			};
			for (final EcoAccountLogoutForm form : valid) {
				System.out.println(JacksonJSONSerialization.JSONtoString(form));
				Assert.assertTrue(POSTDataValidator.isValid(form));
			}

			System.out.println("---------------------------------------------");

			System.out.println("INVALID POST DATA");
			final EcoAccountLogoutForm[] invalid = { // INVALID
					// Invalid refresh token
					new EcoAccountLogoutForm().refreshtoken(""),
					// End of tests
			};
			for (final EcoAccountLogoutForm form : invalid) {
				System.out.println(JacksonJSONSerialization.JSONtoString(form));
				Assert.assertFalse(POSTDataValidator.isValid(form));
			}

			System.out.println("---------------------------------------------");
		}
		// ---------------------------------------------------------------------

		// ---------------------------------------------------------------------
		{
			System.out.println("PRODUCERS FEEDBACK FORM");
			System.out.println("---------------------------------------------");

			System.out.println("VALID POST DATA");
			// Producers feedback
			final List<ProducersFeedbackForm> valid = new ArrayList<ProducersFeedbackForm>();
			valid.add(new ProducersFeedbackForm().productgtin(12345678L).userscore(3).comment("Some free text here"));
			System.out.println(JacksonJSONSerialization.JSONtoString(valid));
			Assert.assertTrue(POSTDataValidator.isValid(valid));

			System.out.println("---------------------------------------------");

			System.out.println("INVALID POST DATA");
			final List<ProducersFeedbackForm> empty = new ArrayList<ProducersFeedbackForm>();
			System.out.println(JacksonJSONSerialization.JSONtoString(empty));
			Assert.assertFalse(POSTDataValidator.isValid(empty));

			final List<ProducersFeedbackForm> invalid = new ArrayList<ProducersFeedbackForm>();
			invalid.add(new ProducersFeedbackForm().comment("Free text but no GTIN or score"));
			System.out.println(JacksonJSONSerialization.JSONtoString(invalid));
			Assert.assertFalse(POSTDataValidator.isValid(invalid));

			System.out.println("---------------------------------------------");
		}
		// ---------------------------------------------------------------------

		// ---------------------------------------------------------------------
		{
			System.out.println("SYSTEM FEEDBACK FORM");
			System.out.println("---------------------------------------------");

			System.out.println("VALID POST DATA");
			// System feedback form
			final SystemFeedbackForm valid = new SystemFeedbackForm().userscore(5).comment("Some free text here");
			System.out.println(JacksonJSONSerialization.JSONtoString(valid));
			Assert.assertTrue(POSTDataValidator.isValid(valid));

			System.out.println("---------------------------------------------");

			System.out.println("INVALID POST DATA");
			final SystemFeedbackForm invalid = new SystemFeedbackForm().comment("Some free text here");
			System.out.println(JacksonJSONSerialization.JSONtoString(invalid));
			Assert.assertFalse(POSTDataValidator.isValid(invalid));

			System.out.println("---------------------------------------------");
		}
		// ---------------------------------------------------------------------

		// ---------------------------------------------------------------------
		{
			System.out.println("RESET PASSWORD FORM");
			System.out.println("---------------------------------------------");

			System.out.println("VALID POST DATA");
			final EcoAccountForgotPasswordForm valid = new EcoAccountForgotPasswordForm().email("user@example.com");
			System.out.println(JacksonJSONSerialization.JSONtoString(valid));
			Assert.assertTrue(POSTDataValidator.isValid(valid));

			System.out.println("---------------------------------------------");

			System.out.println("INVALID POST DATA");
			final EcoAccountForgotPasswordForm[] invalid = { new EcoAccountForgotPasswordForm().email(invalidemail),
					new EcoAccountForgotPasswordForm().email("foobar") };
			for (final EcoAccountForgotPasswordForm form : invalid) {
				System.out.println(JacksonJSONSerialization.JSONtoString(form));
				Assert.assertFalse(POSTDataValidator.isValid(form));
			}
		}
		// ---------------------------------------------------------------------

		// ---------------------------------------------------------------------
		{
			System.out.println("UPDATE PROFILE FORM");
			System.out.println("---------------------------------------------");

			System.out.println("VALID POST DATA");
			final EcoAccountUpdateForm[] valid = { // VALID
					// Address
					new EcoAccountUpdateForm().address(new Address().country("GR").postalcode("15771")
							.street("9, Iroon Politechniou Str. Zografou").city("Athens")),
					// Birthday
					new EcoAccountUpdateForm().birthday("1900-01-01"),
					// Email
					new EcoAccountUpdateForm().email("user@example.com"),
					// First name
					new EcoAccountUpdateForm().firstname("FirstName"),
					// Gender
					new EcoAccountUpdateForm().gender("Male"),
					// Image URL
					new EcoAccountUpdateForm().imageurl(imageURL),
					// Last name
					new EcoAccountUpdateForm().lastname("LastName"),
					// Organization
					new EcoAccountUpdateForm()
							.organization(new Organization().name("ICCS").url("https://www.iccs.gr/")),
					// Password
					new EcoAccountUpdateForm().password(validpassword),
					// Recycle bin User ID
					new EcoAccountUpdateForm().recyclebinuserid(validrecyclebinuserid),
					// Telephone
					new EcoAccountUpdateForm().telephone("+302107723865"),
					// User URL
					new EcoAccountUpdateForm().url("https://circ4life.iccs.gr"),
					// All fields
					new EcoAccountUpdateForm()
							.address(new Address().country("GR").postalcode("15771")
									.street("9, Iroon Politechniou Str. Zografou").city("Athens"))
							.birthday("1900-01-01").email("user@example.com").firstname("FirstName").gender("Male")
							.imageurl(imageURL).lastname("LastName")
							.organization(new Organization().name("ICCS").url("https://www.iccs.gr/"))
							.password(validpassword).recyclebinuserid(validrecyclebinuserid).telephone("+302107723865")
							.url("https://circ4life.iccs.gr") };
			for (final EcoAccountUpdateForm form : valid) {
				System.out.println(JacksonJSONSerialization.JSONtoString(form));
				Assert.assertTrue(POSTDataValidator.isValid(validusername, form));
			}

			System.out.println("---------------------------------------------");

			System.out.println("INVALID POST DATA");
			final EcoAccountUpdateForm[] invalid = { // INVALID
					// Empty
					new EcoAccountUpdateForm(),
					// Invalid country code
					new EcoAccountUpdateForm().address(new Address().country("Foobar").postalcode("15771")
							.street("9, Iroon Politechniou Str. Zografou").city("Athens")),
					// Empty Address fields
					new EcoAccountUpdateForm().address(new Address().country("GR").postalcode("")
							.street("9, Iroon Politechniou Str. Zografou").city("Athens")),
					new EcoAccountUpdateForm()
							.address(new Address().country("GR").postalcode("15771").street("").city("Athens")),
					new EcoAccountUpdateForm().address(new Address().country("GR").postalcode("15771")
							.street("9, Iroon Politechniou Str. Zografou").city("")),
					// Invalid birthday
					new EcoAccountUpdateForm().birthday(currentyearminus4 + "-01-01"),
					// Invalid email
					new EcoAccountUpdateForm().email(invalidemail),
					// Invalid first name
					new EcoAccountUpdateForm().firstname("a"), new EcoAccountUpdateForm().firstname("a23"),
					// Invalid gender
					new EcoAccountUpdateForm().gender(tooLongFreeText),
					// Invalid image URL
					new EcoAccountUpdateForm().imageurl("Foobar"),
					// Invalid last nameF
					new EcoAccountUpdateForm().lastname("a"), new EcoAccountUpdateForm().lastname("a23"),
					// Invalid Organization
					new EcoAccountUpdateForm().organization(new Organization().name("ICCS").url("foobar")),
					// Invalid password
					new EcoAccountUpdateForm().password(shortpassword),
					// Invalid Recycle Bin ID
					new EcoAccountUpdateForm().recyclebinuserid(invalidrecyclebinuserid),
					// Invalid telephone
					new EcoAccountUpdateForm().telephone("+210-1234567"),
					// Invalid User URL
					new EcoAccountUpdateForm().url("foobar") };
			for (final EcoAccountUpdateForm form : invalid) {
				System.out.println(JacksonJSONSerialization.JSONtoString(form));
				Assert.assertFalse(POSTDataValidator.isValid(validusername, form));
			}
		}
		// ---------------------------------------------------------------------
	}

	@Test
	public void testJSONSerialization() throws JsonGenerationException, JsonMappingException, IOException {
		fillJSONProperties();

		System.out.println("---- JSON representations ----");
		final JsonFactory messagePackFactory = new JsonFactory();
		messagePackFactory.configure(JsonGenerator.Feature.AUTO_CLOSE_TARGET, false);
		final ObjectMapper objmapper = new ObjectMapper(messagePackFactory);
		// TODO: Add assertions
		System.out.println("-- Compact");
		objmapper.writeValue(System.out, productType);
		System.out.println();
		objmapper.writeValue(System.out, product);
		System.out.println();
		objmapper.writeValue(System.out, producersFeedbackForm);
		System.out.println();
		objmapper.writeValue(System.out, systemFeedbackForm);
		System.out.println();
		objmapper.writeValue(System.out, ecoAccountLoginForm);
		System.out.println();
		objmapper.writeValue(System.out, ecoAccountLogoutForm);
		System.out.println();
		objmapper.writeValue(System.out, ecoAccountRegistrationForm);
		System.out.println();
		objmapper.writeValue(System.out, ecoAccountForgotPasswordForm);
		System.out.println();

		System.out.println();

		System.out.println("-- Pretty");
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, productType);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, product);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, producersFeedbackForm);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, systemFeedbackForm);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, ecoAccountLoginForm);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, ecoAccountLogoutForm);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, ecoAccountRegistrationForm);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, ecoAccountForgotPasswordForm);
		System.out.println();

		System.out.println();

		System.out.println("-- To String and Back");
		Object obj = null;
		final ByteArrayOutputStream byteos = new ByteArrayOutputStream();
		objmapper.writeValue(byteos, productType);
		obj = objmapper.readerFor(ProductType.class).readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, product);
		obj = objmapper.readerFor(Product.class).readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, producersFeedbackForm);
		obj = objmapper.readerFor(ProducersFeedbackForm.class)
				.readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, systemFeedbackForm);
		obj = objmapper.readerFor(SystemFeedbackForm.class)
				.readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, ecoAccountLoginForm);
		obj = objmapper.readerFor(EcoAccountLoginForm.class)
				.readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, ecoAccountLogoutForm);
		obj = objmapper.readerFor(EcoAccountLogoutForm.class)
				.readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, ecoAccountRegistrationForm);
		obj = objmapper.readerFor(EcoAccountRegistrationForm.class)
				.readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, ecoAccountForgotPasswordForm);
		obj = objmapper.readerFor(EcoAccountForgotPasswordForm.class)
				.readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, product);
		final String productString = byteos.toString(StandardCharsets.UTF_8.toString());
		obj = objmapper.readerFor(Product.class).readValue(productString);
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		System.out.println("-- Date format");

		objmapper.writeValue(byteos, ecoAccountRegistrationForm);
		final String EcoAccountRegistrationFormString = byteos.toString(StandardCharsets.UTF_8.toString());
		obj = objmapper.readerFor(EcoAccountRegistrationForm.class).readValue(EcoAccountRegistrationFormString);
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		// Check that dates are properly formatted
		final String formatteddate = String.format("%1$tY-%1$tm-%1$td", calendar);
		String pattern = "^.*\"purchasedate\":.*\"" + formatteddate + "\".*";
		Assert.assertTrue(productString.matches(pattern));
		pattern = "^.*\"birthday\":.*\"" + formatteddate + "\".*";
		Assert.assertTrue(EcoAccountRegistrationFormString.matches(pattern));

		System.out.println("------------------------------");
	}

	private void fillJSONProperties() {
		productType.hskey(847130).unukey(303).text("Some text");
		product.gtin(1234567891234L).manufacturer("Some manufacturer").model("Some model").name("Some name")
				.purchasedate(today).serialnum("Some serial number").type(productType);
		producersFeedbackForm.comment("Some comment").productgtin(1234567891234L).userscore(1);
		systemFeedbackForm.comment("Some comment").userscore(1);
		ecoAccountLoginForm.password("User password").username("Username or email");
		ecoAccountLogoutForm.refreshtoken("_RefreshToken_");

		final Address address = new Address().city("City").country("Country").postalcode("123456").region("Region")
				.street("Street address");
		final Organization organization = new Organization().name("Organization name")
				.url("http://example.organization.com");
		ecoAccountRegistrationForm.address(address).birthday(today).email("user@email.com").firstname("First name")
				.gender("Female").lastname("Last name").organization(organization).telephone("210-12345678")
				.url("http://example.com").username(validusername).password(validpassword).email("user@example.com");
		ecoAccountForgotPasswordForm.email("user@email.com");
	}

	private String createStringWithLength(final int length) {
		final StringBuffer strbuffer = new StringBuffer(length);

		for (int i = 0; i < length; ++i) {
			strbuffer.append('A');
		}

		return strbuffer.toString();
	}
}
