/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.logging.ConsoleHandler;
import java.util.logging.Logger;

import org.glassfish.jersey.jackson.JacksonFeature;
import org.glassfish.jersey.logging.LoggingFeature;
import org.glassfish.jersey.server.ResourceConfig;

import gr.iccs.isense.circ4life.client.EcoCreditCalculatorClient;
import gr.iccs.isense.circ4life.logging.Log;
import gr.iccs.isense.circ4life.rest.config.ConfigProperties;
import gr.iccs.isense.circ4life.rest.resources.Information;
import gr.iccs.isense.circ4life.rest.resources.openapi.Ecocredits;
import gr.iccs.isense.circ4life.rest.resources.openapi.ProductTypes;
import gr.iccs.isense.circ4life.rest.resources.openapi.PurchaseHistory;
import gr.iccs.isense.circ4life.rest.resources.openapi.RecycleEvaluation;

/**
 * This class implements the Recycle/Traceability module integration with the
 * ICT platform as a RESTful Web Service API.
 *
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class RecycleModule extends ResourceConfig {
	/**
	 * Logger.
	 */
	public final static Logger logger = Log.getConfiguredLogger(RecycleModule.class);

	/**
	 * REST service error messages.
	 */
	public static final class ErrorMessages {
		public final static String AUTHENTICATION_REQUIRED = "Authentication required";
		public final static String INIT_ECOCREDITSWS_FAILURE = "Cannot initialize Eco-credits service connection";
		public final static String ACCESS_ECOCREDITSWS_FAILURE = "Cannot access Eco-credits service";
		public final static String INVALID_EOLSTATE = "Invalid End of Life state";
		public final static String INVALID_CONSUMERID = "Invalid consumer identification code";
		public final static String UNKNOWN_CONSUMERID = "Unknown consumer identification code";
		public final static String CONVERSION_ERROR = "Data conversion error";
		public final static String PRODUCT_TYPES_EMPTY = "Product Types list is empty";
		public final static String MISSING_EVALUATION_REPORT = "Missing evaluation report";
		public final static String MISSING_EVALUATION_REPORT_ITEM = "Missing evaluation report's product item";
		public final static String MISSING_EVALUATION_REPORT_ITEM_WEIGHT = "Missing evaluation report's weight value";
		public final static String MISSING_EVALUATION_REPORT_LIFETIME = "Missing evaluation report's product lifetime";
		public final static String MISSING_EVALUATION_REPORT_DATE = "Missing evaluation report's product recycle date";
		public final static String INVALID_EVALUATION_REPORT_STATE = "Invalid evaluation report's product end of life state";
		public final static String INVALID_EVALUATION_REPORT_ITEM_WASTETYPE = "Invalid evaluation report's item waste type";
		public final static String ECOACCOUNT_UPDATE_FAILURE = "Could not update Eco-account";
		public final static String MALFORMED_RECYCLE_EVALUATION_REPORT = "Malformed recycle item evaluation report";
	}

	/**
	 * Property the holds the end-point of EcoCreditCalculator REST Web Service.
	 */
	private static final String PROPERTY_ECOCREDITSCALCULATORWS_URL = "ecocreditscalc.url";

	/**
	 * Property the holds the user name for accessing the EcoCreditCalculator REST
	 * Web Service.
	 */
	private static final String PROPERTY_ECOCREDITSCALCULATORWS_USER = "ecocreditscalc.username";

	/**
	 * Property the holds the password for accessing the EcoCreditCalculator REST
	 * Web Service.
	 */
	private static final String PROPERTY_ECOCREDITSCALCULATORWS_PASS = "ecocreditscalc.password";

	/**
	 * EcoCreditCalculator REST Web Service client.
	 */
	private static EcoCreditCalculatorClient ecocreditsclient = null;

	/**
	 * Constructor.
	 */
	public RecycleModule() {
		// REST resource classes
		register(Information.class);
		register(Ecocredits.class);
		register(ProductTypes.class);
		register(PurchaseHistory.class);
		register(RecycleEvaluation.class);

		// Jackson JSON serialization
		register(JacksonFeature.class);

		// Setup logging
		logger.setLevel(ConfigProperties.getLoggingLevel(RecycleModule.class));
		final ConsoleHandler loggerhandler = new ConsoleHandler();
		loggerhandler.setLevel(logger.getLevel());
		logger.addHandler(loggerhandler);
		logger.setUseParentHandlers(false);
		final LoggingFeature.Verbosity verbosity = ConfigProperties.getLoggingVerbosity(RecycleModule.class);
		register(new LoggingFeature(logger, logger.getLevel(), verbosity, LoggingFeature.DEFAULT_MAX_ENTITY_SIZE));

		logger.info("Logger level: " + logger.getLevel() + ", Verbosity: " + verbosity + ", Parent Logger Level: "
				+ logger.getParent().getLevel() + ", using parent logger handlers: " + logger.getUseParentHandlers());
	}

	/**
	 * Access the singleton EcoCreditCalculator REST Web Service client.
	 * 
	 * @return the singleton EcoCreditCalculator REST Web Service client on
	 *         successful initialization, <code>null</code> otherwise.
	 */
	public static EcoCreditCalculatorClient ecoCreditsClient() {
		// Lazy initialization on first access
		if (null == ecocreditsclient) {
			// Get Eco-credits web service details from application configuration
			final String ecocreditscalc_url = ConfigProperties.getConfigProperty(RecycleModule.class,
					PROPERTY_ECOCREDITSCALCULATORWS_URL);
			if (null == ecocreditscalc_url) {
				logger.severe(
						"Property " + PROPERTY_ECOCREDITSCALCULATORWS_URL + " not found in configuration properties!");
				return null;
			}
			logger.config("Eco-credits calculator service URL: " + ecocreditscalc_url);

			final String ecocreditscalc_user = ConfigProperties.getConfigProperty(RecycleModule.class,
					PROPERTY_ECOCREDITSCALCULATORWS_USER);
			if (null == ecocreditscalc_user) {
				logger.severe(
						"Property " + PROPERTY_ECOCREDITSCALCULATORWS_USER + " not found in configuration properties!");
				return null;
			}
			logger.config("Eco-credits calculator service user: " + ecocreditscalc_user);

			final String ecocreditscalc_pass = ConfigProperties.getConfigProperty(RecycleModule.class,
					PROPERTY_ECOCREDITSCALCULATORWS_PASS);
			if (null == ecocreditscalc_pass) {
				logger.severe(
						"Property " + PROPERTY_ECOCREDITSCALCULATORWS_PASS + " not found in configuration properties!");
				return null;
			}

			try {
				final URI basePath = new URI(ecocreditscalc_url);
				ecocreditsclient = new EcoCreditCalculatorClient(basePath, ecocreditscalc_user, ecocreditscalc_pass);
				ecocreditsclient.startSession();
			} catch (URISyntaxException e) {
				logger.severe("Malformed URI value of property " + PROPERTY_ECOCREDITSCALCULATORWS_URL);
				return null;
			}

		}
		return ecocreditsclient;
	}
}
