/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.openapi;

import javax.validation.constraints.Max;
import javax.validation.constraints.Min;
import javax.validation.constraints.Size;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.core.Response.Status;

import org.openapitools.api.EcocreditsApi;

import gr.iccs.isense.circ4life.client.EcoCreditCalculatorClient;
import gr.iccs.isense.circ4life.datamodels.WEEEsEndOfLifeState;
import gr.iccs.isense.circ4life.rest.RecycleModule;
import gr.iccs.isense.circ4life.rest.RecycleModule.ErrorMessages;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;

/**
 * /ecocredits end-point implementation.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class Ecocredits extends EcocreditsApi {
	/**
	 * Proxy of security context for session management.
	 */
	private final SecurityContext m_securityContext;

	/**
	 * Constructor.
	 * 
	 * @param securityContext proxy of security context for session management.
	 */
	public Ecocredits(@Context SecurityContext securityContext) {
		m_securityContext = securityContext;
	}

	@Override
	public Response ecoCreditsGET(@Size(min = 14) String productid, @Min(0) @Max(100) Integer years,
			@Min(0) @Max(11) Integer extramonths, String eolstate) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RecycleModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final EcoCreditCalculatorClient client = RecycleModule.ecoCreditsClient();
		if (null == client) {
			RecycleModule.logger.severe(ErrorMessages.INIT_ECOCREDITSWS_FAILURE);

			return Response
					.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.INIT_ECOCREDITSWS_FAILURE)
					.build();
		}

		final WEEEsEndOfLifeState eolenum = DataConversion.convertEoLState(eolstate);
		if (null == eolenum) {
			RecycleModule.logger.severe(ErrorMessages.INVALID_EOLSTATE);

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.INVALID_EOLSTATE)
					.build();
		}

		final Double ecocredits = client.EcoCreditsProductGET(productid, years, extramonths, eolenum);
		if (null == ecocredits) {
			RecycleModule.logger.severe(ErrorMessages.ACCESS_ECOCREDITSWS_FAILURE);

			return Response
					.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.ACCESS_ECOCREDITSWS_FAILURE)
					.build();
		}

		return Response.ok().entity(ecocredits).build();
	}

	@Override
	public Response ecoCreditsProductEstimationGET(@Size(min = 14) String productid, String eolstate) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RecycleModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		// TODO: Get the product's initial production in order to estimate the lifetime
		// with the help of the current date.
		final Integer years = 0, extramonths = 0;

		return ecoCreditsGET(productid, years, extramonths, eolstate);

	}

	@Override
	public Response ecoCreditsProductTypeEstimationGET(@Min(1) @Max(999999) Integer hskey, String eolstate) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RecycleModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final EcoCreditCalculatorClient client = RecycleModule.ecoCreditsClient();
		if (null == client) {
			RecycleModule.logger.severe(ErrorMessages.INIT_ECOCREDITSWS_FAILURE);

			return Response
					.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.INIT_ECOCREDITSWS_FAILURE)
					.build();
		}

		final WEEEsEndOfLifeState eolenum = DataConversion.convertEoLState(eolstate);
		if (null == eolenum) {
			RecycleModule.logger.severe(ErrorMessages.INVALID_EOLSTATE);

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.INVALID_EOLSTATE)
					.build();
		}

		// TODO: Get the product's initial production in order to estimate the lifetime
		// with the help of the current date.
		final Integer years = 0, extramonths = 0;

		final Double ecocredits = client.EcoCreditsGET(hskey, years, extramonths, eolenum);
		if (null == ecocredits) {
			RecycleModule.logger.severe(ErrorMessages.ACCESS_ECOCREDITSWS_FAILURE);

			return Response
					.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.ACCESS_ECOCREDITSWS_FAILURE)
					.build();
		}

		return Response.ok().entity(ecocredits).build();
	}
}
