/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.openapi;

import java.util.ArrayList;

import javax.validation.constraints.Max;
import javax.validation.constraints.Min;
import javax.validation.constraints.Size;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.core.Response.Status;

import org.openapitools.api.PurchaseHistoryApi;
import org.openapitools.model.Product;
import org.openapitools.model.ProductType;

import gr.iccs.isense.circ4life.dao.CommonDAO.IndividualProductData;
import gr.iccs.isense.circ4life.dao.EcoAccountDAO;
import gr.iccs.isense.circ4life.rest.RecycleModule;
import gr.iccs.isense.circ4life.rest.RecycleModule.ErrorMessages;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;

/**
 * /purchase_history end-point implementation.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class PurchaseHistory extends PurchaseHistoryApi {
	/**
	 * Eco-account Data Access Object.
	 */
	private static final EcoAccountDAO ecoaccountdao = new EcoAccountDAO();

	/**
	 * Proxy of security context for session management.
	 */
	private final SecurityContext m_securityContext;

	/**
	 * Constructor.
	 * 
	 * @param securityContext proxy of security context for session management.
	 */
	public PurchaseHistory(@Context SecurityContext securityContext) {
		m_securityContext = securityContext;
	}

	@Override
	public Response purchaseHistoryByTypeGET(@Min(0) @Max(99999999999L) Long consumerid,
			@Min(1) @Max(999999) Integer hskey, @Size(min = 2, max = 2) String locale) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RecycleModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final String ecoaccountid = ecoaccountdao.getEcoAccountIDFromRecycleBinUserID(consumerid);
		// Check that the user ID is not empty
		if (null == ecoaccountid) {
			RecycleModule.logger.severe(ErrorMessages.UNKNOWN_CONSUMERID + ": " + consumerid);

			return Response
					.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.UNKNOWN_CONSUMERID + ": " + consumerid)
					.build();
		}

		final ArrayList<Product> purchased_products = new ArrayList<Product>();
		Product p = null;
		try {
			for (final IndividualProductData indproddata : ecoaccountdao.getEcoAccountPurchasedList(ecoaccountid, hskey,
					locale)) {
				p = convertProduct(indproddata);
				if (null != p) {
					purchased_products.add(p);
				}
			}
		} catch (Exception e) {
			RecycleModule.logger.severe(ErrorMessages.CONVERSION_ERROR);

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.CONVERSION_ERROR)
					.build();
		}

		return Response.ok().entity(purchased_products).build();
	}

	@Override
	public Response purchaseHistoryGET(@Min(0) @Max(99999999999L) Long consumerid,
			@Size(min = 2, max = 2) String locale) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RecycleModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		return purchaseHistoryByTypeGET(consumerid, null, locale);
	}

	/**
	 * Convert from internal Data Model of individual product to OpenAPI model.
	 * 
	 * @param individualproduct the object that contains the data
	 * @return the converted object, or <code>null</code> in case of any error.
	 */
	private Product convertProduct(final IndividualProductData individualproduct) {
		final Product product = new Product();

		try {
			final Long gtin = Long.parseLong(individualproduct.getGtin14());
			final ProductType prodtype = new ProductType().hskey(individualproduct.getProductType().HSkey);
			final String serialnum = individualproduct.getSerialNumber(), name = individualproduct.getName(),
					manufacturer = null == individualproduct.getManufacturer() ? null
							: individualproduct.getManufacturer().getName(),
					model = individualproduct.getProductID(), purchasedate = individualproduct.getPurchaseDate(),
					recycledate = individualproduct.getRecycleDate(), price = individualproduct.getPrice();

			product.gtin(gtin).type(prodtype).serialnum(serialnum).manufacturer(manufacturer).model(model).name(name)
					.price(price).purchasedate(purchasedate).recycledate(recycledate);
		} catch (Exception e) {
			RecycleModule.logger.severe(ErrorMessages.CONVERSION_ERROR + " during conversion of " + individualproduct
					+ ", details: " + e.getMessage());

			return null;
		}

		return product;
	}
}
