/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.openapi;

import java.util.Date;

import javax.validation.Valid;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.core.Response.Status;

import org.openapitools.api.RecycleEvaluationApi;
import org.openapitools.model.Product;
import org.openapitools.model.RecycleEvaluationReport;
import org.openapitools.model.RecycleEvaluationReportReport;
import org.openapitools.model.RecycleEvaluationReportReportItem;
import org.openapitools.model.RecycleEvaluationReportReportLifetime;

import gr.iccs.isense.circ4life.dao.EcoAccountDAO;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO;
import gr.iccs.isense.circ4life.dao.CommonDAO.IndividualProductData;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductType;
import gr.iccs.isense.circ4life.dao.EcoCreditCalculatorDAO.WasteType;
import gr.iccs.isense.circ4life.datamodels.WEEEsEndOfLifeState;
import gr.iccs.isense.circ4life.rest.RecycleModule;
import gr.iccs.isense.circ4life.rest.RecycleModule.ErrorMessages;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;
import gr.iccs.isense.circ4life.rest.resources.validation.POSTDataValidator;

/**
 * /recycle_evaluation end-point implementation.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class RecycleEvaluation extends RecycleEvaluationApi {
	/**
	 * Eco-account Data Access Object.
	 */
	private static final EcoAccountDAO ecoaccountdao = new EcoAccountDAO();

	/**
	 * Eco-credits calculator Data Access Object.
	 */
	private static final EcoCreditCalculatorDAO calculatordao = new EcoCreditCalculatorDAO();

	/**
	 * Proxy of security context for session management.
	 */
	private final SecurityContext m_securityContext;

	/**
	 * Constructor.
	 * 
	 * @param securityContext proxy of security context for session management.
	 */
	public RecycleEvaluation(@Context SecurityContext securityContext) {
		m_securityContext = securityContext;
	}

	@Override
	public Response recycleEvaluationPOST(@Valid RecycleEvaluationReport recycleEvaluationReport) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RecycleModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		if (!POSTDataValidator.isValid(recycleEvaluationReport)) {
			return Response
					.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_RECYCLE_EVALUATION_REPORT)
					.build();
		}

		// Get the user's input and check every object
		final Long recycleBinUserID = recycleEvaluationReport.getConsumer();
		if (null == recycleBinUserID) {
			RecycleModule.logger.severe(ErrorMessages.INVALID_CONSUMERID);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.INVALID_CONSUMERID).build();
		}

		final String ecoaccountid = ecoaccountdao.getEcoAccountIDFromRecycleBinUserID(recycleBinUserID);
		// In this case the ecoaccountid is of the Eco-account we want to update, not
		// the current logged in user's session user id. So we do not need to check with
		// the Access Control Manager whether this needs to match the current user. This
		// endpoint is reached by an account that is given to a tool, not a consumer
		// end-user.
		if (null == ecoaccountid) {
			RecycleModule.logger.severe(ErrorMessages.UNKNOWN_CONSUMERID);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.UNKNOWN_CONSUMERID).build();
		}

		final RecycleEvaluationReportReport report = recycleEvaluationReport.getReport();
		if (null == report) {
			RecycleModule.logger.severe(ErrorMessages.MISSING_EVALUATION_REPORT);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MISSING_EVALUATION_REPORT).build();
		}

		final RecycleEvaluationReportReportItem item = report.getItem();
		if (null == item) {
			RecycleModule.logger.severe(ErrorMessages.MISSING_EVALUATION_REPORT_ITEM);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MISSING_EVALUATION_REPORT_ITEM)
					.build();
		}

		final String recycledatestr = report.getRecycledate();
		if (null == recycledatestr) {
			RecycleModule.logger.severe(ErrorMessages.MISSING_EVALUATION_REPORT_DATE);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MISSING_EVALUATION_REPORT_DATE)
					.build();
		}

		try {
			final Product product = item.getProduct();
			final ProductType prodtype = new ProductType(item.getWastetype());
			final Date recycledate = DataConversion.readDate(recycledatestr);

			// Decide your path: WEEE or Bio-waste?
			final WasteType wastetype = calculatordao.getProductWasteType(prodtype.HSkey);
			switch (wastetype) {
			case FOOD: {
				// Bio-waste require weight that is not negative
				if (null == item.getWeight() || item.getWeight().doubleValue() < 0.0) {
					RecycleModule.logger.severe(ErrorMessages.MISSING_EVALUATION_REPORT_ITEM_WEIGHT);

					return Response.status(Status.BAD_REQUEST.getStatusCode(),
							ErrorMessages.MISSING_EVALUATION_REPORT_ITEM_WEIGHT).build();
				}

				final Double bio_weight = item.getWeight();

				if (ecoaccountdao.updateEcoAccountWithRecycledBioWasteItem(ecoaccountid, recycledate, bio_weight,
						prodtype)) {
					// All OK
					return Response.ok().build();
				} else {
					RecycleModule.logger.severe(ErrorMessages.ECOACCOUNT_UPDATE_FAILURE);

					return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(),
							ErrorMessages.ECOACCOUNT_UPDATE_FAILURE).build();
				}
			}

			case WEEES: {
				// WEEEs require evaluation report
				final RecycleEvaluationReportReportLifetime lifetime = report.getLifetime();
				if (null == lifetime) {
					RecycleModule.logger.severe(ErrorMessages.MISSING_EVALUATION_REPORT_LIFETIME);

					return Response.status(Status.BAD_REQUEST.getStatusCode(),
							ErrorMessages.MISSING_EVALUATION_REPORT_LIFETIME).build();
				}

				final String statestr = report.getState();
				WEEEsEndOfLifeState state = null;
				// State can be null, but cannot be something random
				if (null != statestr) {
					if (statestr.equalsIgnoreCase(WEEEsEndOfLifeState.BROKEN.toString())) {
						state = WEEEsEndOfLifeState.BROKEN;
					} else if (statestr.equalsIgnoreCase(WEEEsEndOfLifeState.REUSABLE.toString())) {
						state = WEEEsEndOfLifeState.REUSABLE;
					} else if (statestr.equalsIgnoreCase(WEEEsEndOfLifeState.WORKING.toString())) {
						state = WEEEsEndOfLifeState.WORKING;
					} else {
						RecycleModule.logger.warning("Item state is unknown: " + statestr);
					}
				}

				if (null == product) {
					// The user did not provide the product details, we only can rely on product
					// type
					if (ecoaccountdao.updateEcoAccountWithRecycledWEEEItem(ecoaccountid, recycledate,
							lifetime.getYears(), lifetime.getExtramonths(), state, prodtype)) {
						// All OK
						return Response.ok().build();
					} else {
						RecycleModule.logger.severe(ErrorMessages.ECOACCOUNT_UPDATE_FAILURE);

						return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(),
								ErrorMessages.ECOACCOUNT_UPDATE_FAILURE).build();
					}
				} else {
					// The user provided a product with details, including GTIN and Serial Number.
					final Date purchaseDate = DataConversion.readDate(product.getPurchasedate());
					final String gtin14 = DataConversion.formatGTIN14(product.getGtin()), productID = null,
							serialNumber = product.getSerialnum(), name = product.getName();
					final IndividualProductData recycledproduct = new IndividualProductData(prodtype, gtin14, productID,
							serialNumber, name, purchaseDate);
					recycledproduct.setRecycleDate(DataConversion.formatDate(recycledate));

					if (ecoaccountdao.updateEcoAccountWithRecycledWEEEItem(ecoaccountid, recycledate,
							lifetime.getYears(), lifetime.getExtramonths(), state, recycledproduct)) {
						// All OK
						return Response.ok().build();
					} else {
						RecycleModule.logger.severe(ErrorMessages.ECOACCOUNT_UPDATE_FAILURE);

						return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(),
								ErrorMessages.ECOACCOUNT_UPDATE_FAILURE).build();
					}
				}
			}

			case UNKNOWN:
			default: {
				return Response.status(Status.BAD_REQUEST.getStatusCode(),
						ErrorMessages.INVALID_EVALUATION_REPORT_ITEM_WASTETYPE).build();
			}
			}
		} catch (Exception e) {
			RecycleModule.logger.severe(ErrorMessages.ECOACCOUNT_UPDATE_FAILURE + ", details: " + e.getMessage());

			return Response
					.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.ECOACCOUNT_UPDATE_FAILURE)
					.build();
		}
	}
}
