/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.test;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Calendar;
import java.util.Date;

import org.junit.Assert;
import org.junit.Test;
import org.openapitools.model.Product;
import org.openapitools.model.RecycleEvaluationReport;
import org.openapitools.model.RecycleEvaluationReportReport;
import org.openapitools.model.RecycleEvaluationReportReportItem;
import org.openapitools.model.RecycleEvaluationReportReportLifetime;

import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import gr.iccs.isense.circ4life.dao.CommonDAO.IndividualProductData;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductType;
import gr.iccs.isense.circ4life.json.JacksonJSONSerialization;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;
import gr.iccs.isense.circ4life.rest.resources.validation.POSTDataValidator;

/**
 * Tests for RecycleModule REST Web service.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class RecycleModuleWSTests {
	private final Calendar calendar = Calendar.getInstance();
	private final String today = DataConversion.formatDate(Date.from(calendar.toInstant()));
	private final ProductType prodtype = new ProductType(847130, 303, "Laptops/tablets");
	private final IndividualProductData indivproduct = new IndividualProductData(prodtype, "04047111123453",
			"Some Product ID", "Some Serial Number", "Laptop example", DataConversion.readDate(today));
	private final org.openapitools.model.ProductType productType = new org.openapitools.model.ProductType();
	private final Product product = new Product();
	private final RecycleEvaluationReportReport recycleEvaluationReportReport = new RecycleEvaluationReportReport();
	private final RecycleEvaluationReport recycleEvaluationReport = new RecycleEvaluationReport();

	@Test
	public void testPOSTDataValidator() {
		// ---------------------------------------------------------------------
		{
			System.out.println("RECYCLE EVALUATION FORM");
			System.out.println("---------------------------------------------");

			System.out.println("VALID POST DATA");
			// Recycle evaluation report
			final RecycleEvaluationReport[] valid = { // VALID
					// Minimum required parameters
					new RecycleEvaluationReport().consumer(1234567891234L)
							.report(new RecycleEvaluationReportReport().recycledate("2019-01-01")
									.lifetime(new RecycleEvaluationReportReportLifetime().extramonths(7).years(2))
									.item(new RecycleEvaluationReportReportItem().wastetype(303))),
					// Full parameters
					new RecycleEvaluationReport().consumer(1234567891234L)
							.report(new RecycleEvaluationReportReport().recycledate("2019-01-01")
									.lifetime(new RecycleEvaluationReportReportLifetime().extramonths(7).years(2))
									.item(new RecycleEvaluationReportReportItem().wastetype(303)
											.product(new Product().gtin(12345678L).serialnum("Some Serial Number")
													.name("Some name").purchasedate("2017-01-01")
													.manufacturer("Some manufacturer").model("Some model")
													.type(new org.openapitools.model.ProductType().hskey(847130)
															.unukey(303).text("Laptops/tablets"))))), };
			for (final RecycleEvaluationReport form : valid) {
				System.out.println(JacksonJSONSerialization.JSONtoString(form));
				Assert.assertTrue(POSTDataValidator.isValid(form));
			}

			System.out.println("---------------------------------------------");

			System.out.println("INVALID POST DATA");
			final RecycleEvaluationReport[] invalid = { // INVALID
					// Empty
					new RecycleEvaluationReport(),
					// No consumer
					new RecycleEvaluationReport().report(new RecycleEvaluationReportReport().recycledate("2019-01-01")
							.lifetime(new RecycleEvaluationReportReportLifetime().extramonths(7).years(2))
							.item(new RecycleEvaluationReportReportItem().wastetype(303))),
					// No report
					new RecycleEvaluationReport().consumer(1234567891234L),
					// No recycle date
					new RecycleEvaluationReport().consumer(1234567891234L)
							.report(new RecycleEvaluationReportReport()
									.lifetime(new RecycleEvaluationReportReportLifetime().extramonths(7).years(2))
									.item(new RecycleEvaluationReportReportItem().wastetype(303))),
					// No lifetime
					new RecycleEvaluationReport().consumer(1234567891234L)
							.report(new RecycleEvaluationReportReport().recycledate("2019-01-01")
									.item(new RecycleEvaluationReportReportItem().wastetype(303))),
					// No recycled item
					new RecycleEvaluationReport().consumer(1234567891234L)
							.report(new RecycleEvaluationReportReport().recycledate("2019-01-01")
									.lifetime(new RecycleEvaluationReportReportLifetime().extramonths(7).years(2))),
					// Malformed date
					new RecycleEvaluationReport().consumer(1234567891234L)
							.report(new RecycleEvaluationReportReport().recycledate("2019+acv")
									.lifetime(new RecycleEvaluationReportReportLifetime().extramonths(7).years(2))
									.item(new RecycleEvaluationReportReportItem().wastetype(303))) };
			for (final RecycleEvaluationReport form : invalid) {
				System.out.println(JacksonJSONSerialization.JSONtoString(form));
				Assert.assertFalse(POSTDataValidator.isValid(form));
			}

			System.out.println("---------------------------------------------");
		}
		// ---------------------------------------------------------------------
	}

	@Test
	public void testJSONSerialization() throws JsonGenerationException, JsonMappingException, IOException {
		fillJSONProperties();

		System.out.println("---- JSON representations ----");
		final JsonFactory messagePackFactory = new JsonFactory();
		messagePackFactory.configure(JsonGenerator.Feature.AUTO_CLOSE_TARGET, false);
		final ObjectMapper objmapper = new ObjectMapper(messagePackFactory);
		// TODO: Add assertions
		System.out.println("-- Compact");
		objmapper.writeValue(System.out, productType);
		System.out.println();
		objmapper.writeValue(System.out, product);
		System.out.println();
		objmapper.writeValue(System.out, recycleEvaluationReportReport);
		System.out.println();
		objmapper.writeValue(System.out, recycleEvaluationReport);
		System.out.println();

		System.out.println();

		System.out.println("-- Pretty");
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, productType);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, product);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, recycleEvaluationReportReport);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, recycleEvaluationReport);
		System.out.println();

		System.out.println();

		System.out.println("-- To String and Back");
		Object obj = null;
		final ByteArrayOutputStream byteos = new ByteArrayOutputStream();
		objmapper.writeValue(byteos, productType);
		obj = objmapper.readerFor(org.openapitools.model.ProductType.class)
				.readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, product);
		obj = objmapper.readerFor(Product.class).readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, recycleEvaluationReportReport);
		obj = objmapper.readerFor(RecycleEvaluationReportReport.class)
				.readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, recycleEvaluationReport);
		obj = objmapper.readerFor(RecycleEvaluationReport.class)
				.readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		System.out.println("-- Date format");

		objmapper.writeValue(byteos, recycleEvaluationReportReport);
		final String recycleEvaluationReportReportString = byteos.toString(StandardCharsets.UTF_8.toString());
		obj = objmapper.readerFor(RecycleEvaluationReportReport.class).readValue(recycleEvaluationReportReportString);
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, recycleEvaluationReport);
		final String recycleEvaluationReportString = byteos.toString(StandardCharsets.UTF_8.toString());
		obj = objmapper.readerFor(RecycleEvaluationReport.class).readValue(recycleEvaluationReportString);
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		// Check that dates are properly formatted
		final String formatteddate = String.format("%1$tY-%1$tm-%1$td", calendar);
		String pattern = "^.*\"recycledate\":.*\"" + formatteddate + "\".*";
		Assert.assertTrue(recycleEvaluationReportReportString.matches(pattern));
		Assert.assertTrue(recycleEvaluationReportString.matches(pattern));

		System.out.println("------------------------------");
	}

	@Test
	public void testDataConversions() throws Exception {
		final JsonFactory messagePackFactory = new JsonFactory();
		messagePackFactory.configure(JsonGenerator.Feature.AUTO_CLOSE_TARGET, false);
		final ObjectMapper objmapper = new ObjectMapper(messagePackFactory);

		final Long gtin = Long.parseLong(indivproduct.getGtin14());
		final org.openapitools.model.ProductType prodtype = new org.openapitools.model.ProductType()
				.hskey(indivproduct.getProductType().HSkey);

		// DAO IndividualProductData to OpenAPI Product
		System.out.println("-----------");
		System.out.println("-- Original");
		objmapper.writeValue(System.out, indivproduct);
		System.out.println();

		final Product openapiproduct = new Product();
		openapiproduct.gtin(gtin).serialnum(indivproduct.getSerialNumber()).name(indivproduct.getName())
				.manufacturer(indivproduct.getManufacturer().getName()).model(indivproduct.getProductID())
				.type(prodtype).purchasedate(DataConversion.formatDate(indivproduct.getPurchaseDateObject()));
		System.out.println("-- OpenAPI generated");
		objmapper.writeValue(System.out, openapiproduct);
		System.out.println();
		System.out.println("-----------");
	}

	private void fillJSONProperties() {
		productType.hskey(847130).unukey(303).text("Some text");
		product.gtin(1234567891234L).manufacturer("Some manufacturer").model("Some model").name("Some name")
				.purchasedate(today).serialnum("Some serial number").type(productType);

		final RecycleEvaluationReportReportItem item = new RecycleEvaluationReportReportItem();
		item.wastetype(847130).product(product);
		final RecycleEvaluationReportReportLifetime lifetime = new RecycleEvaluationReportReportLifetime();
		lifetime.extramonths(7).years(2);

		recycleEvaluationReportReport.item(item).lifetime(lifetime).recycledate(today).state("working");
		recycleEvaluationReport.consumer(1234567891234L).report(recycleEvaluationReportReport);
	}
}
