/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.openapi;

import java.util.ArrayList;
import java.util.List;

import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.core.Response.Status;

import org.openapitools.api.EcoaccountApi;
import org.openapitools.model.EcoAccountAddPurchasesForm;
import org.openapitools.model.EcoAccountBalance;
import org.openapitools.model.EcoAccountBalanceUpdateForm;
import org.openapitools.model.Product;

import gr.iccs.isense.circ4life.dao.CommonDAO.EcoAccountRecord;
import gr.iccs.isense.circ4life.dao.CommonDAO.IndividualProductData;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductType;
import gr.iccs.isense.circ4life.dao.EcoAccountDAO;
import gr.iccs.isense.circ4life.dao.ProductMasterDataDAO;
import gr.iccs.isense.circ4life.dao.schema.Manufacturer;
import gr.iccs.isense.circ4life.rest.RetailerModule;
import gr.iccs.isense.circ4life.rest.RetailerModule.ErrorMessages;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;
import gr.iccs.isense.circ4life.rest.resources.validation.POSTDataValidator;

/**
 * /ecoaccount end-point implementation.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class AnonymousEcoAccount extends EcoaccountApi {
	/**
	 * Eco-account Data Access Object.
	 */
	private static final EcoAccountDAO ecoaccountdao = new EcoAccountDAO();

	/**
	 * Product Master Data Access Object.
	 */
	private static final ProductMasterDataDAO productmasterdatadao = new ProductMasterDataDAO();

	/**
	 * Proxy of security context for session management.
	 */
	private final SecurityContext m_securityContext;

	/**
	 * Constructor.
	 * 
	 * @param securityContext proxy of security context for session management.
	 */
	public AnonymousEcoAccount(@Context SecurityContext securityContext) {
		m_securityContext = securityContext;
	}

	@Override
	public Response ecoAccountAddPurchasesPOST(@Valid EcoAccountAddPurchasesForm ecoAccountAddPurchasesForm) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RetailerModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		if (!POSTDataValidator.isValid(ecoAccountAddPurchasesForm)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_PURCHASE_LIST).build();
		}

		final String consumerid = ecoAccountAddPurchasesForm.getConsumerid();

		final String useruuid = ecoaccountdao.getEcoAccountIDFromAnonymousID(consumerid);
		if (null == useruuid) {
			RetailerModule.logger.severe(ErrorMessages.ECOACCOUNT_NOTFOUND + ", Anonymous Consumer ID: " + consumerid);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.ECOACCOUNT_NOTFOUND).build();
		}

		// Check if the Eco-credits are present, if not set them equal to 0.0
		double ecocredits = 0.0;
		if (null != ecoAccountAddPurchasesForm.getEcocredits()) {
			ecocredits = ecoAccountAddPurchasesForm.getEcocredits().doubleValue();
		}

		try {
			// Convert the purchased products to our internal representation
			final @NotNull List<Product> products = ecoAccountAddPurchasesForm.getProducts();
			final List<IndividualProductData> purchases = new ArrayList<IndividualProductData>();
			for (final Product product : products) {
				purchases.add(convertProduct(product));
			}

			if (!ecoaccountdao.updateEcoAccountWithPurchases(useruuid, purchases,
					ecoAccountAddPurchasesForm.getEcodebits(), ecocredits)) {
				RetailerModule.logger.severe(ErrorMessages.ECOACCOUNT_UPDATE_FAILURE);

				return Response
						.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.ECOACCOUNT_UPDATE_FAILURE)
						.build();
			}

			// All OK
			return Response.ok().build();
		} catch (Exception e) {
			RetailerModule.logger.severe(ErrorMessages.ECOACCOUNT_UPDATE_FAILURE + ", details: " + e.getMessage());

			return Response
					.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.ECOACCOUNT_UPDATE_FAILURE)
					.build();
		}
	}

	@Override
	public Response ecoAccountBalanceUpdate(@Valid EcoAccountBalanceUpdateForm ecoAccountBalanceUpdateForm) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RetailerModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		if (!POSTDataValidator.isValid(ecoAccountBalanceUpdateForm)) {
			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.MALFORMED_PURCHASE_LIST).build();
		}

		final String consumerid = ecoAccountBalanceUpdateForm.getConsumerid();

		final String useruuid = ecoaccountdao.getEcoAccountIDFromAnonymousID(consumerid);
		if (null == useruuid) {
			RetailerModule.logger.severe(ErrorMessages.ECOACCOUNT_NOTFOUND + ", Anonymous Consumer ID: " + consumerid);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.ECOACCOUNT_NOTFOUND).build();
		}

		try {
			// Update Eco-balance
			if (!ecoaccountdao.updateEcoAccountBalance(useruuid,
					ecoAccountBalanceUpdateForm.getEcodebits().doubleValue(),
					ecoAccountBalanceUpdateForm.getEcocredits().doubleValue())) {
				RetailerModule.logger.severe(ErrorMessages.ECOACCOUNT_UPDATE_FAILURE);

				return Response
						.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.ECOACCOUNT_UPDATE_FAILURE)
						.build();
			}

			// All OK
			return Response.ok().build();
		} catch (Exception e) {
			RetailerModule.logger.severe(ErrorMessages.ECOACCOUNT_UPDATE_FAILURE + ", details: " + e.getMessage());

			return Response
					.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.ECOACCOUNT_UPDATE_FAILURE)
					.build();
		}
	}

	@Override
	public Response ecoAccountAnonymousGET(String consumerid) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RetailerModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final String useruuid = ecoaccountdao.getEcoAccountIDFromAnonymousID(consumerid);
		if (null == useruuid) {
			RetailerModule.logger.severe(ErrorMessages.ECOACCOUNT_NOTFOUND + ", Anonymous Consumer ID: " + consumerid);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.ECOACCOUNT_NOTFOUND).build();
		}

		return Response.ok().entity("Anonymous Consumer ID is valid").build();
	}

	@Override
	public Response ecoAccountEcobalanceGET(String consumerid) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RetailerModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final String useruuid = ecoaccountdao.getEcoAccountIDFromAnonymousID(consumerid);
		if (null == useruuid) {
			RetailerModule.logger.severe(ErrorMessages.ECOACCOUNT_NOTFOUND + ", Anonymous Consumer ID: " + consumerid);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.ECOACCOUNT_NOTFOUND).build();
		}

		try {
			final EcoAccountRecord ecoaccount = ecoaccountdao.getEcoAccount(useruuid);
			if (null == ecoaccount) {
				throw new Exception("Cannot retrieve Eco-account record for user ID " + useruuid);
			}

			// Fill in the returned object
			final EcoAccountBalance ecobalance = new EcoAccountBalance()
					.firstname(ecoaccount.getEcoAccount().getGivenName())
					.lastname(ecoaccount.getEcoAccount().getFamilyName())
					.ecodebits(ecoaccount.getEcoBalance().getEcoDebit().getValue())
					.ecocredits(ecoaccount.getEcoBalance().getEcoCredit().getValue());

			// All OK
			return Response.ok().entity(ecobalance).build();
		} catch (Exception e) {
			RetailerModule.logger.severe(ErrorMessages.CONVERSION_ERROR + ", details: " + e.getMessage());

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.CONVERSION_ERROR)
					.build();
		}
	}

	/**
	 * Convert from OpenAPI Model of individual product to internal Data model.
	 * 
	 * @param product the product the object that contains the data.
	 * @return the converted object, or <code>null</code> in case of any error.
	 * @throws Exception thrown if any error occurs during data conversion.
	 */
	private IndividualProductData convertProduct(final Product product) throws Exception {
		final String gtin14 = DataConversion.formatGTIN14(product.getGtin());

		ProductType type = productmasterdatadao.getProductType(gtin14);
		if (null == type) {
			RetailerModule.logger.warning(
					"Unknown product type for GTIN: " + gtin14 + ", we will try to check the user's input form...");

			if (null == product.getType()) {
				RetailerModule.logger.severe("Cannot determine product type for GTIN: " + gtin14);
				throw new Exception("Cannot determine product type for GTIN: " + gtin14);
			}

			type = new ProductType(product.getType().getHskey(), product.getType().getUnukey(),
					product.getType().getText());
		}

		final IndividualProductData indivproduct = new IndividualProductData(type, gtin14, null, product.getSerialnum(),
				product.getName(), DataConversion.readDate(product.getPurchasedate()));
		indivproduct.withManufacturer(new Manufacturer().withName(product.getManufacturer()))
				.withPrice(product.getPrice());

		return indivproduct;
	}
}
