/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.openapi;

import javax.validation.constraints.Size;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.SecurityContext;
import javax.ws.rs.core.Response.Status;

import org.openapitools.api.ProductApi;

import gr.iccs.isense.circ4life.dao.CommonDAO.ProductID;
import gr.iccs.isense.circ4life.dao.CommonDAO.ProductRecord;
import gr.iccs.isense.circ4life.json.JacksonJSONSerialization;
import gr.iccs.isense.circ4life.dao.ProductMasterDataDAO;
import gr.iccs.isense.circ4life.rest.RetailerModule;
import gr.iccs.isense.circ4life.rest.RetailerModule.ErrorMessages;
import gr.iccs.isense.circ4life.rest.resources.AccessControlManager;

/**
 * /product end-point implementation.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class Product extends ProductApi {
	/**
	 * Product Master Data Access Object.
	 */
	private static final ProductMasterDataDAO productmasterdatadao = new ProductMasterDataDAO();

	/**
	 * Proxy of security context for session management.
	 */
	private final SecurityContext m_securityContext;

	/**
	 * Constructor.
	 * 
	 * @param securityContext proxy of security context for session management.
	 */
	public Product(@Context SecurityContext securityContext) {
		m_securityContext = securityContext;
	}

	@Override
	public Response productBarcodeGET(@Size(min = 18, max = 42) String barcode, @Size(min = 2, max = 2) String locale) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RetailerModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		// Get GTIN code from Barcode
		final ProductID productid = productmasterdatadao.getProductIDFromBarcode(barcode);
		if (null == productid) {
			RetailerModule.logger.severe(ErrorMessages.UNSUPPORTED_BARCODE + ", barcode = " + barcode);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.UNSUPPORTED_BARCODE).build();
		}

		final ProductRecord productrec = productmasterdatadao.getProductRecord(productid.toString(), locale);
		if (null == productrec) {
			RetailerModule.logger.severe(ErrorMessages.PRODUCT_NOTFOUND);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.PRODUCT_NOTFOUND).build();
		}

		return Response.ok().entity(productrec).build();
	}

	@Override
	public Response productEcoPointsBarcodeGET(@Size(min = 18, max = 42) String barcode) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RetailerModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		// Get GTIN code from Barcode
		final ProductID productid = productmasterdatadao.getProductIDFromBarcode(barcode);
		if (null == productid) {
			RetailerModule.logger.severe(ErrorMessages.UNSUPPORTED_BARCODE + ", barcode = " + barcode);

			return Response.status(Status.BAD_REQUEST.getStatusCode(), ErrorMessages.UNSUPPORTED_BARCODE).build();
		}

		final ProductRecord productrec = productmasterdatadao.getProductRecord(productid.toString(), "en");
		if (null == productrec) {
			RetailerModule.logger.severe(ErrorMessages.PRODUCT_NOTFOUND);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.PRODUCT_NOTFOUND).build();
		}

		try {
			return Response.ok().entity(productrec.getEcoAccounting().getEcoPoints().getValue()).build();
		} catch (Exception e) {
			// Something really wrong with the product master data record
			RetailerModule.logger.severe(ErrorMessages.PRODUCT_MALFORMED_RECORD + ", Details: "
					+ JacksonJSONSerialization.JSONtoString(productrec));

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.PRODUCT_MALFORMED_RECORD)
					.build();
		}
	}

	@Override
	public Response productEcoPointsGET(@Size(min = 14) String productid) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RetailerModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		// Use the default locale for the Eco-points
		final ProductRecord productrec = productmasterdatadao.getProductRecord(productid, "en");
		if (null == productrec) {
			RetailerModule.logger.severe(ErrorMessages.PRODUCT_NOTFOUND);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.PRODUCT_NOTFOUND).build();
		}

		try {
			return Response.ok().entity(productrec.getEcoAccounting().getEcoPoints().getValue()).build();
		} catch (Exception e) {
			// Something really wrong with the product master data record
			RetailerModule.logger.severe(ErrorMessages.PRODUCT_MALFORMED_RECORD + ", Details: "
					+ JacksonJSONSerialization.JSONtoString(productrec));

			return Response.status(Status.INTERNAL_SERVER_ERROR.getStatusCode(), ErrorMessages.PRODUCT_MALFORMED_RECORD)
					.build();
		}
	}

	@Override
	public Response productGET(@Size(min = 14) String productid, @Size(min = 2, max = 2) String locale) {
		if (AccessControlManager.isEnabled() && !AccessControlManager.isUserAuthenticated(m_securityContext)) {
			RetailerModule.logger.severe(ErrorMessages.AUTHENTICATION_REQUIRED);

			return Response.status(Status.FORBIDDEN.getStatusCode(), ErrorMessages.AUTHENTICATION_REQUIRED).build();
		}

		final ProductRecord productrec = productmasterdatadao.getProductRecord(productid, locale);
		if (null == productrec) {
			RetailerModule.logger.severe(ErrorMessages.PRODUCT_NOTFOUND);

			return Response.status(Status.NOT_FOUND.getStatusCode(), ErrorMessages.PRODUCT_NOTFOUND).build();
		}

		return Response.ok().entity(productrec).build();
	}
}
