/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.rest.resources.validation;

import org.openapitools.model.EcoAccountAddPurchasesForm;
import org.openapitools.model.EcoAccountBalanceUpdateForm;
import org.openapitools.model.Product;

import gr.iccs.isense.circ4life.rest.RetailerModule;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;

/**
 * POST data structures validator.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class POSTDataValidator {
	/**
	 * Validate the consumer's purchase list.
	 * 
	 * @param purchaselistform the object containing the consumer's purchase list.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final EcoAccountAddPurchasesForm purchaselistform) {
		try {
			if (0 == purchaselistform.getConsumerid().length()) {
				throw new IllegalArgumentException("Empty consumer ID in consumer's purchase list");
			}

			if (0 == purchaselistform.getProducts().size()) {
				throw new IllegalArgumentException("Empty consumer's purchase list");
			}

			for (final Product p : purchaselistform.getProducts()) {
				// GTIN required
				p.getGtin().longValue();

				// Serial number required
				if (null == p.getSerialnum() || 0 == p.getSerialnum().length()) {
					throw new IllegalArgumentException("Empty serial number found in product");
				}

				// Name required
				if (null == p.getName() || 0 == p.getName().length()) {
					throw new IllegalArgumentException("Empty name found in product item of recycle evaluation report");
				}

				// Purchase data required
				DataConversion.readDate(p.getPurchasedate());
			}

			// Eco-debits required
			purchaselistform.getEcodebits().doubleValue();

			// All OK
			return true;
		} catch (Exception e) {
			RetailerModule.logger.severe("Consumer's purchase list validation error:" + e.getMessage());

			return false;
		}
	}

	/**
	 * Validate the consumer's balance update.
	 * 
	 * @param balanceform the object containing the consumer's balance update.
	 * @return <code>true</code> if the object is valid, <code>false</code>
	 *         otherwise.
	 */
	public static boolean isValid(final EcoAccountBalanceUpdateForm balanceform) {
		try {
			if (0 == balanceform.getConsumerid().length()) {
				throw new IllegalArgumentException("Empty consumer ID in consumer's balance update form");
			}

			if (0.0 > balanceform.getEcodebits().doubleValue()) {
				// Negative values not allowed
				throw new IllegalArgumentException("Negative Eco-debits in consumer's balance update form");
			}

			if (0.0 > balanceform.getEcocredits().doubleValue()) {
				// Negative values not allowed
				throw new IllegalArgumentException("Negative Eco-credits in consumer's balance update form");
			}

			// All OK
			return true;
		} catch (Exception e) {
			RetailerModule.logger.severe("Consumer's balance update form validation error:" + e.getMessage());

			return false;
		}
	}
}
