/**
 * Institute of Communication and Computer Systems (ICCS)
 * i-sense Research Group
 */
package gr.iccs.isense.circ4life.test;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import org.junit.Assert;
import org.junit.Test;
import org.openapitools.model.EcoAccountAddPurchasesForm;
import org.openapitools.model.EcoAccountBalanceUpdateForm;
import org.openapitools.model.Product;
import org.openapitools.model.ProductType;

import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import gr.iccs.isense.circ4life.json.JacksonJSONSerialization;
import gr.iccs.isense.circ4life.rest.resources.openapi.DataConversion;
import gr.iccs.isense.circ4life.rest.resources.validation.POSTDataValidator;

/**
 * Tests for RetailerModule REST Web service.
 * 
 * @author Miltiadis Koutsokeras, miltos.koutsokeras@iccs.gr
 */
public class RetailerModuleWSTests {
	private final Calendar calendar = Calendar.getInstance();
	private final String today = DataConversion.formatDate(Date.from(calendar.toInstant()));
	private final ProductType productType = new ProductType();
	private final Product product = new Product();
	private final EcoAccountAddPurchasesForm purchaseList = new EcoAccountAddPurchasesForm().consumerid("1234567890")
			.ecocredits(0.0).ecodebits(0.0);

	@Test
	public void testPOSTDataValidator() {
		// ---------------------------------------------------------------------
		{
			System.out.println("PURCHASES FORM");
			System.out.println("---------------------------------------------");

			System.out.println("VALID POST DATA");

			// Valid purchase list
			final List<Product> validpurchases = new ArrayList<Product>();
			validpurchases.add(new Product().gtin(12345678L).serialnum("Some Serial Number").name("Some name")
					.purchasedate("2017-01-01"));
			validpurchases.add(new Product().gtin(4047111123453L).manufacturer("Manufacturer").model("Model")
					.name("Some name").purchasedate("2016-03-02").serialnum("Some Serial Number")
					.type(new ProductType().hskey(847130).text("Laptops/tablets").unukey(303)));

			final List<Product> invalidpurchases = new ArrayList<Product>();
			invalidpurchases
					.add(new Product().serialnum("Some Serial Number").name("Some name").purchasedate("2017-01-01"));

			{
				// Consumer's purchase list
				final EcoAccountAddPurchasesForm valid = new EcoAccountAddPurchasesForm().consumerid("1234567890")
						.ecodebits(12.0).products(validpurchases);
				System.out.println(JacksonJSONSerialization.JSONtoString(valid));
				Assert.assertTrue(POSTDataValidator.isValid(valid));
			}

			{
				// Consumer's balance update form
				final EcoAccountBalanceUpdateForm valid = new EcoAccountBalanceUpdateForm().consumerid("1234567890")
						.ecodebits(12.0).ecocredits(25.0);
				System.out.println(JacksonJSONSerialization.JSONtoString(valid));
				Assert.assertTrue(POSTDataValidator.isValid(valid));
			}

			System.out.println("---------------------------------------------");

			System.out.println("INVALID POST DATA");

			{
				// Consumer's purchase list
				final EcoAccountAddPurchasesForm empty = new EcoAccountAddPurchasesForm();
				System.out.println(JacksonJSONSerialization.JSONtoString(empty));
				Assert.assertFalse(POSTDataValidator.isValid(empty));

				final EcoAccountAddPurchasesForm noconsumer = new EcoAccountAddPurchasesForm().ecodebits(12.0)
						.products(validpurchases);
				System.out.println(JacksonJSONSerialization.JSONtoString(noconsumer));
				Assert.assertFalse(POSTDataValidator.isValid(noconsumer));

				final EcoAccountAddPurchasesForm noecodebits = new EcoAccountAddPurchasesForm().consumerid("1234567890")
						.products(validpurchases);
				System.out.println(JacksonJSONSerialization.JSONtoString(noecodebits));
				Assert.assertFalse(POSTDataValidator.isValid(noecodebits));

				final EcoAccountAddPurchasesForm noproducts = new EcoAccountAddPurchasesForm().consumerid("1234567890")
						.ecodebits(12.0);
				System.out.println(JacksonJSONSerialization.JSONtoString(noproducts));
				Assert.assertFalse(POSTDataValidator.isValid(noproducts));

				final EcoAccountAddPurchasesForm emptyproducts = new EcoAccountAddPurchasesForm()
						.consumerid("1234567890").ecodebits(12.0).products(new ArrayList<Product>());
				System.out.println(JacksonJSONSerialization.JSONtoString(emptyproducts));
				Assert.assertFalse(POSTDataValidator.isValid(emptyproducts));

				final EcoAccountAddPurchasesForm invalidproduct = new EcoAccountAddPurchasesForm()
						.consumerid("1234567890").ecodebits(12.0).products(invalidpurchases);
				System.out.println(JacksonJSONSerialization.JSONtoString(invalidproduct));
				Assert.assertFalse(POSTDataValidator.isValid(invalidproduct));
			}

			{
				// Consumer's balance update form
				final EcoAccountBalanceUpdateForm empty = new EcoAccountBalanceUpdateForm();
				System.out.println(JacksonJSONSerialization.JSONtoString(empty));
				Assert.assertFalse(POSTDataValidator.isValid(empty));

				final EcoAccountBalanceUpdateForm noconsumer = new EcoAccountBalanceUpdateForm().ecodebits(12.0)
						.ecocredits(25.0);
				System.out.println(JacksonJSONSerialization.JSONtoString(noconsumer));
				Assert.assertFalse(POSTDataValidator.isValid(noconsumer));

				final EcoAccountBalanceUpdateForm noecodebits = new EcoAccountBalanceUpdateForm()
						.consumerid("1234567890").ecocredits(25.0);
				System.out.println(JacksonJSONSerialization.JSONtoString(noecodebits));
				Assert.assertFalse(POSTDataValidator.isValid(noecodebits));

				final EcoAccountBalanceUpdateForm noecocredits = new EcoAccountBalanceUpdateForm()
						.consumerid("1234567890").ecodebits(25.0);
				System.out.println(JacksonJSONSerialization.JSONtoString(noecocredits));
				Assert.assertFalse(POSTDataValidator.isValid(noecocredits));
			}

			System.out.println("---------------------------------------------");
		}
		// ---------------------------------------------------------------------
	}

	@Test
	public void testJSONSerialization() throws JsonGenerationException, JsonMappingException, IOException {
		fillJSONProperties();

		System.out.println("---- JSON representations ----");
		final JsonFactory messagePackFactory = new JsonFactory();
		messagePackFactory.configure(JsonGenerator.Feature.AUTO_CLOSE_TARGET, false);
		final ObjectMapper objmapper = new ObjectMapper(messagePackFactory);
		// TODO: Add assertions
		System.out.println("-- Compact");
		objmapper.writeValue(System.out, productType);
		System.out.println();
		objmapper.writeValue(System.out, product);
		System.out.println();
		objmapper.writeValue(System.out, purchaseList);
		System.out.println();

		System.out.println();

		System.out.println("-- Pretty");
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, productType);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, product);
		System.out.println();
		objmapper.writerWithDefaultPrettyPrinter().writeValue(System.out, purchaseList);
		System.out.println();

		System.out.println();

		System.out.println("-- To String and Back");
		Object obj = null;
		final ByteArrayOutputStream byteos = new ByteArrayOutputStream();
		objmapper.writeValue(byteos, productType);
		obj = objmapper.readerFor(ProductType.class).readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, product);
		obj = objmapper.readerFor(Product.class).readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		objmapper.writeValue(byteos, purchaseList);
		obj = objmapper.readerFor(EcoAccountAddPurchasesForm.class)
				.readValue(byteos.toString(StandardCharsets.UTF_8.toString()));
		objmapper.writeValue(System.out, obj);
		System.out.println();
		byteos.reset();

		System.out.println("------------------------------");
	}

	private void fillJSONProperties() {
		productType.hskey(847130).unukey(303).text("Some text");
		product.gtin(1234567891234L).manufacturer("Some manufacturer").model("Some model").name("Some name")
				.purchasedate(today).serialnum("Some serial number").type(productType);
		final List<Product> products = new ArrayList<>();
		products.add(product);
		purchaseList.consumerid("Consumer ID").ecodebits(1.0).products(products);
	}
}
